/* mus_lib.c. Version 2.0
 * Library of routines to handle MuSTARD.
 * 4 August 1998. JCH. First version - set up low level routines.
 * Nov 98: modify routines; add tests   GFM
 * 5 February 1999. JCH. Tidy up - call version 1.1
 * 3 March 1999. PWP. Remove unused variables; add return codes. V1.1b
 * 28 July 1999. JCH. New routine mus_read_compress_event to copy every
 * 2 bytes of the raw data into 1 word. Version 1.2
 * 31 July 1999. JCH. New routine mus_clear_buffer to see if there are
 * events in the buffer and clear them if so.
 * 12 Oct 1999. PWP. Change nevents to type int.
 * 4 August 2000. PWP. Add mus_add_histogram.
 * 1 June 2001. PWP. Revise mus_read_event to release the event page
 * even when an error has occured.
 * 3 Sept. 2001 PWP. Additions for use with BIT3 interface
 * 24 May 2002 BJG/PWP. revisions for NIVXI under linux
 * 24 June 2002 JCH. Add ATLAS RCC interface (for SBCs).
 */

/* includes and defines */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <time.h>

#ifdef NIVXI
#include <nivxi.h>
#endif

#ifdef OS9
#include <process.h>
#include <signal.h>
#include <types.h>
#endif

#ifdef RAID
#define CLOCKS_PER_SEC 100000
#endif

#ifdef QB_linux
#define WND_SIZE 0x20000
#include "../QB/InitVME.h"
#endif

#ifdef RCC
#include <rcc_error.h>
#include <vme_rcc.h>
#endif

#include "mus_local.h"
#include "mus_lib.h"

/* Global variables */
short mus_total = 0;           /*Total number of MuSTARDs*/
struct mus_info_structure *mus_info;	/*structure to hold MuSTARD info*/

/* Functions */

/*mus_add_histogram. PWP 04.08.00
 * Add the contents of MuSTARD's histogram memory to the
 * contents of the histogram stored in the host's memory.
 * Various error returns signal problems:
 * -1 - identifier invalid.
 * -2 - stream number invalid.
 * -3 - bad status from reading control register.
 * -4 - MRDOEN not clear, so memory not accessible.
 * Only 768 locations (6 chips) are read starting from chip 0, strip 0 (hilo = 0)
 * or from chip 8, strip 0 (hilo = 1).
 */
int mus_add_histogram(short identifier, short hilo, short stream, short *length, short **buffer)
{
  volatile unsigned short *ip;
  short mdaqen,mrdoen;
  int i,status;

  if(identifier<0 || identifier>=mus_total) return -1;

  if(stream<0 || stream>MAX_STREAM) return -2;  /* bugfix PWP 03.06.99 */

  ip = (unsigned short *)
        (mus_info[identifier].address + (stream<<12) + MUS_HISTMEM);

  status = mus_read_readout_cntl(identifier,&mdaqen, &mrdoen);
  if(status != 0) {
    printf("mus_add_histogram: Bad status from mus_read_readout_cntl: %d\n",status);
    return -3;
  }

  if(mrdoen != 0) {
    printf("mus_add_histogram: MRDOEN bit in control register not clear\n");
    return -4;
  }

/*  read out histogram contents for high or low 768-strip page.
 *  Note: as SCT hybrids are not normally populated with chips at
 *  addresses 6,7,14 or 15 these are skipped in the readout. */
 
  if(hilo) ip+= 1024;

  *length = 768;
  for(i=0;i<(*length);i++) *(*buffer)++ += *ip++;

  return 0;
}
/* end of mus_add_histogram */

/*mus_clear_buffer.    
 * Clear the MuSTARD buffer by checking if there is an event present and
 * clearing it if there is. Error returns:
 * -1 - identifier invalid.
 * -2 - bad status from reading control register.
 * -3 - MRDOEN not set, so can't access buffer.`
 * 
 * The number of events found is returned in the second argument
 */
int mus_clear_buffer(short identifier, int *nevents)
{
  volatile unsigned short *ip, test;
  short mdaqen, mrdoen;
  int status,loop;

  if(identifier<0 || identifier>=mus_total) return -1;

  status = mus_read_readout_cntl(identifier, &mdaqen, &mrdoen);
  if(status != 0) {
    printf("mus_clear_buffer: Bad status from mus_read_readout_cntl: %d\n",status);
    return -2;
  }
  
  if(mrdoen != 1) {
    printf("mus_clear_buffer: MRDOEN bit in control register not set\n");
    return -3;
  }

  ip = (unsigned short *)(mus_info[identifier].address + MUS_EVTBUFF);

  test = (unsigned short)0x8000;
  *nevents = 0;
  while (((*ip)&test) == 0) {
    (*nevents)++;
    *ip = 0;	/*clear the event*/
    loop = 1000000;
    while(loop-->0);         /*delay to let next event appear*/
  }
  return 0;
}
/* end of mus_clear_buffer */

/*mus_clear_histogram.
 * Clear the histogram memory of the MuSTARD for 1 stream. Various error
 * returns signal problems:
 * -1 - identifier invalid.
 * -2 - stream number invalid.
 * -3 - bad status from reading control register.
 * -4 - MRDOEN not clear, so memory not accessible.
 * All 2k locations are cleared (even though not all are used).
 */
int mus_clear_histogram(short identifier,short  stream)
{
  volatile unsigned short *ip;
  short mdaqen,mrdoen;
  int i,status,length;

  if(identifier<0 || identifier>=mus_total) return -1;

  if(stream<0 || stream>MAX_STREAM) return -2;  /* bugfix PWP 03.06.99 */

  ip = (unsigned short *)
        (mus_info[identifier].address +( stream<<12) + MUS_HISTMEM);

  status = mus_read_readout_cntl(identifier,&mdaqen, &mrdoen);
  if(status != 0) {
    printf("mus_clear_histogram: Bad status from mus_read_readout_cntl: %d\n",status);
    return -3;
  }

  if(mrdoen != 0) {
    printf("mus_clear_histogram: MRDOEN bit in control register not clear\n");
    return -4;
  }


/*  clear histogram contents */

  length = 0x800;
#ifdef BIT3
  for(i=0;i<length;i++){
   *ip++ = 0;
    NI_pause();
  }
#else
  for(i=0;i<length;i++){
   *ip++ = 0;
  }
#endif

  return 0;
}
/* end of mus_clear_histogram */

/*mus_exit.
 *Exit routine for MuSTARD library. Should be called once only to tidy things
 *up. Unmaps the modules and deletes the malloced memory.
 */
int mus_exit()
{
  int i;
#ifdef RCC
  VME_ErrorCode_t status;
#else
  int status;
#endif

/* loop over the modules */
  for(i=0;i<mus_total;i++) {
#ifdef RAID
    status = RD13_VmeUnmap(mus_info[i].address);
#endif
#ifdef NIVXI
    status = UnMapVXIAddress(mus_info[i].window);
#endif
#ifdef QB_linux
    munmap((void *)mus_info[i].address,WND_SIZE);
#endif
#ifdef RCC
    status = VME_MasterUnmap(mus_info[i].window);
#endif
  }  /*end of loop*/

#ifdef NIVXI
    CloseVXIlibrary();
#endif
#ifdef RCC
    status = VME_Close();
#endif

/* release memory */
  free((void *)mus_info);
  return 0;
}
/*end of mus_exit */

/*mus_read_clock_cntl.
 * Read the MuSTARD clock control register (bit 0 only is sensible).
 * Return -1 if identifier invalid.
 */
int mus_read_clock_cntl(short identifier, short *clock)
{
  volatile unsigned short *ip;

/*check identifier is valid.*/
  if(identifier<0 || identifier>=mus_total) return -1;

/*set the address of the clock register*/
  ip = (unsigned short *)(mus_info[identifier].address + MUS_CLKCNTL);

  *clock = (*ip)&1;  /*get bit 0 of register*/

  return 0;
}
/* end of mus_read_clock_cntl*/

/*mus_read_compress_event.
 * Read an event from the output buffer of the MuSTARD. Then pack each 2 
 * bytes of data into a word. This only makes sense in raw mode, and means
 * we have to provide pointers to the start of each stream of data.
 * Various error returns signal problems:
 * -1 - identifier invalid.
 * -2 - bad status from reading control register.
 * -3 - MRDOEN not set, so buffer not accessible.
 * -4 - timeout waiting for event availability
 * -5 - error bit (14) set in 1st word in buffer. Buffer still filled assuming
 *      bits 0-13 are the event length.
 * After the event is readout, the buffer is cleared by writing to the 1st
 * location.
 */
int mus_read_compress_event(short identifier, short *length, short **buffer)
{
  unsigned char *cbuffer;
  volatile unsigned short *ip, *ipbase, *str_ip;
  unsigned short str_p, test;
  short mdaqen, mrdoen, strnum, old_strnum, len;
  int i, error, loop, status;

  if(identifier<0 || identifier>=mus_total) return -1;

  ipbase = (unsigned short *)(mus_info[identifier].address + MUS_EVTBUFF);
  ip = ipbase;
  status = mus_read_readout_cntl(identifier,&mdaqen, &mrdoen);
  if(status != 0) {
    printf("mus_read_compress_event: Bad status from mus_read_readout_cntl: %d\n",status);
    return -2;
  }

  if(mrdoen != 1) {
    printf("mus_read_compress_event: MRDOEN bit in control register not set\n");
    return -3;
  }

  loop = 1000000;
/* JCH. 17 Feb 1999. The next two lines were originally just
 * while (((*ipbase)&0x8000)!=0 && loop>0) loop--;
 * but this fails on OS9 with the Microware compiler.
 * If it offends you, feel free to return to the original
 * code!
 */

  test = (unsigned short)0x8000;
  while (((*ipbase)&test)!=0 && loop>0) loop--;

  if(loop == 0) {
    printf("mus_read_compress_event: Timeout waiting for data\n");
    return -4;
  }

/* got event, read it out */
  len = (*ipbase)&0x3FFF;          /*bits 0-13*/
  error = (*ipbase)&0x4000;           /*bit 14*/

  for(i=0;i<4;i++) *(*buffer)++ = *ip++;   /*First 4 words copied as-is*/

  str_ip = (unsigned short *)(*buffer);
  str_p = 36;	/*Start of data - in bytes*/
  for(i=0;i<=MAX_STREAM;i++) *(str_ip+i) = 0;	/*Clear the pointers*/

  *buffer += 14;
  cbuffer = (char *)(*buffer);       /*point to first location after pointers*/

  old_strnum = -1;
  for(i=4;i<len;i++) {
     strnum = ((*ip)&0x1e00)>>9;
     if(strnum != old_strnum) {   /*Next stream - set pointer */
       while(4*(str_p/4) != str_p) {
	  *cbuffer++ = 0x0;
	  str_p++;		/*Move to next longword */
       }
       *(str_ip+strnum) = str_p/4;      /*pointer in longwords */
       old_strnum = strnum;
     }
     *cbuffer++ = (*ip++)&0xff;
     str_p++;
  }

  while(4*(str_p/4) != str_p) {
    *cbuffer++ = 0x0;
    str_p++;
  }
  *(str_ip+12) = str_p/4;	/*points just after end of event */
  *(str_ip+13) = 0x0;		/* for cleanliness*/
/* modify the length returned to be the correct number of words and set
   the buffer pointer to the end of the event */

  (*length) = str_p/4;
  *buffer = (short *)cbuffer;

  if(error != 0) return -5;		/*error bit set*/

/*now release the event*/
  *ipbase = 0;			/*data not relevant*/
  return 0;
}
/* end of mus_read_compress_event */

/*mus_read_event.
 * Read an event from the output buffer of the MuSTARD. Various error
 * returns signal problems:
 * -1 - identifier invalid.
 * -2 - bad status from reading control register.
 * -3 - MRDOEN not set, so buffer not accessible.
 * -4 - timeout waiting for event availability
 * -5 - error bit (14) set in 1st word in buffer. Buffer still filled assuming
 *      bits 0-13 are the event length.
 * After the event is readout, the buffer is cleared by writing to the 1st
 * location.
 */
int mus_read_event(short identifier, short *length, short **buffer)
{
  volatile unsigned short *ip, *ipbase;
  unsigned short test;
  short mdaqen,mrdoen;
  int i,error,loop,status;

  if(identifier<0 || identifier>=mus_total) return -1;

  ipbase = (unsigned short *)(mus_info[identifier].address + MUS_EVTBUFF);
  ip = ipbase;
  status = mus_read_readout_cntl(identifier,&mdaqen, &mrdoen);
  if(status != 0) {
    printf("mus_read_event: Bad status from mus_read_readout_cntl: %d\n",status);
    return -2;
  }

  if(mrdoen != 1) {
    printf("mus_read_event: MRDOEN bit in control register not set\n");
    return -3;
  }

  loop = 1000000;
/* JCH. 17 Feb 1999. The next two lines were originally just
 * while (((*ipbase)&0x8000)!=0 && loop>0) loop--;
 * but this fails on OS9 with the Microware compiler.
 * If it offends you, feel free to return to the original
 * code!
 */

  test = (unsigned short)0x8000;
  while (((*ipbase)&test)!=0 && loop>0) loop--;

  if(loop == 0) {
    printf("mus_read_event: Timeout waiting for data\n");
    return -4;
  }

/* got event, read it out */
  *length = (*ipbase)&0x3FFF;          /*bits 0-13*/
  error = (*ipbase)&0x4000;           /*bit 14*/

  for(i=0;i<(*length);i++) *(*buffer)++ = *ip++;

/*now release the event*/
  *ipbase = 0;			/*data not relevant*/

  if(error != 0) return -5;		/*error bit set*/

  return 0;
}
/* end of mus_read_event */

/*mus_read_histogram.
 * Read the histogram memory of the MuSTARD. Various error
 * returns signal problems:
 * -1 - identifier invalid.
 * -2 - stream number invalid.
 * -3 - bad status from reading control register.
 * -4 - MRDOEN not clear, so memory not accessible.
 * Only 768 locations (6 chips) are read starting from chip 0, strip 0 (hilo = 0)
 * or from chip 8, strip 0 (hilo = 1).
 */
int mus_read_histogram(short identifier, short hilo, short stream, short *length, short **buffer)
{
  volatile unsigned short *ip;
  short mdaqen,mrdoen;
  int i,status;

  if(identifier<0 || identifier>=mus_total) return -1;

  if(stream<0 || stream>MAX_STREAM) return -2;  /* bugfix PWP 03.06.99 */

  ip = (unsigned short *)
        (mus_info[identifier].address + (stream<<12) + MUS_HISTMEM);

  status = mus_read_readout_cntl(identifier,&mdaqen, &mrdoen);
  if(status != 0) {
    printf("mus_read_histogram: Bad status from mus_read_readout_cntl: %d\n",status);
    return -3;
  }

  if(mrdoen != 0) {
    printf("mus_read_histogram: MRDOEN bit in control register not clear\n");
    return -4;
  }

/*  read out histogram contents for high or low 768-strip page.
 *  Note: as SCT hybrids are not normally populated with chips at
 *  addresses 6,7,14 or 15 these are skipped in the readout. */
 
  if(hilo) ip+= 1024;

  *length = 768;
  for(i=0;i<(*length);i++) *(*buffer)++ = *ip++;

  return 0;
}
/* end of mus_read_histogram */

/*mus_read_readout_cntl.
 * Read the MuSTARD readout control register (bits 0&1 only are sensible).
 * Return -1 if identifier invalid.
 */
int mus_read_readout_cntl(short identifier, short *mdaqen, short *mroen)
{
  volatile unsigned short *ip;
  short i;

/*check identifier is valid.*/
  if(identifier<0 || identifier>=mus_total) return -1;

/*set the address of the readout control register*/
  ip = (unsigned short *)(mus_info[identifier].address + MUS_ROCNTL);

  i = (*ip)&3;  /*get bits 0&1 of register*/
  *mdaqen = (i&2)>>1;
  *mroen = (i&1);

  return 0;
}
/* end of mus_read_readout_cntl*/

/*mus_read_readout_modes.
 * Read the MuSTARD readout mode register (bits 0 and 1 only are sensible).
 * Decode into raw_mode and sample_all (this is a bit at variance with the
 * approach elsewhere).
 * Return -1 if identifier invalid.
 */
int mus_read_readout_modes(short identifier, short *raw_mode, short *sample_all)
{
  volatile unsigned short *ip;
  short i;

/*check identifier is valid.*/
  if(identifier<0 || identifier>=mus_total) return -1;

/*set the address of the readout mode register*/
  ip = (unsigned short *)(mus_info[identifier].address + MUS_ROMODE);

  i = (*ip)&3;  /*get bits 0&1 of register*/
  *raw_mode = (i&2)>>1;
  *sample_all = (i&1);

  return 0;
}
/* end of mus_read_readout_modes*/

/*mus_read_status.
 * Read the MuSTARD status register. Interpretation is left to the caller,
 * except that d15 (always 1) is cleared to avoid a negative number.
 * Return -1 if identifier invalid.
 */
int mus_read_status(short identifier, short *status)
{
  volatile unsigned short *ip;

/*check identifier is valid.*/
  if(identifier<0 || identifier>=mus_total) return -1;

/*set the address of the status register*/
  ip = (unsigned short *)(mus_info[identifier].address + MUS_STATUS);

  *status = (*ip)&0x7FFF;  /*get lower 15 bits of register*/

  return 0;
}
/* end of mus_read_status*/

/*mus_read_stream_enable.
 * Read whether a single stream is enabled or disabled. Streams 0-MAX_STREAM
 * are valid. Outside this range, return -2. Return -1 if identifier invalid.
 */
int mus_read_stream_enable(short identifier, short stream, short *enable)
{
  volatile unsigned short *ip;

/*check identifier is valid.*/
  if(identifier<0 || identifier>=mus_total) return -1;

/*check stream number is valid.*/
  if(stream<0 || stream>MAX_STREAM) return -2;

/*set the address - mode register for this stream*/
  ip = (unsigned short *)
    (mus_info[identifier].address + MUS_BUS + (stream<<4) + MUS_STRMODE);

  *enable = ((*ip)>>(stream%4))&1;  /*get enable or disable bit for this stream*/
  return 0;
}
/* end of mus_read_stream_enable*/

/*mus_read_stream_paf.
 * Read PAF register for a stream. Streams 0-MAX_STREAM are valid. Outside
 * this range, return -2. Return -1 if identifier invalid. The value of paf
 * is assumed to be interpreted external to this routine. There is one PAF
 * register for 4 streams.
 */
int mus_read_stream_paf(short identifier, short stream, short *paf)
{
  volatile unsigned short *ip;

/*check identifier is valid.*/
  if(identifier<0 || identifier>=mus_total) return -1;

/*check stream number is valid.*/
  if(stream<0 || stream>MAX_STREAM) return -2;

/*set the address - paf register for this stream*/
  ip = (unsigned short *)
    (mus_info[identifier].address + MUS_BUS + (stream<<4) + MUS_STRPAF);

  *paf = (*ip)&3;  /*get paf bits for this stream*/

  return 0;
}
/* end of mus_read_stream_paf*/

/*mus_read_stream_status.
 * Read status register for a stream. Streams 0-MAX_STREAM are valid. Outside
 * this range, return -2. Return -1 if identifier invalid. The value of the
 * register is assumed to be interpreted external to this routine, however
 * bits 5-15 are masked off as undefined. There is one status register per
 * stream.
 */
int mus_read_stream_status(short identifier, short stream, short *status)
{
  volatile unsigned short *ip;

/*check identifier is valid.*/
  if(identifier<0 || identifier>=mus_total) return -1;

/*check stream number is valid.*/
  if(stream<0 || stream>MAX_STREAM) return -2;

/*set the address - status register for this stream*/
  ip = (unsigned short *)
    (mus_info[identifier].address + MUS_BUS + (stream<<4) + MUS_STRSTATUS);

  *status = (*ip)&0x1F;  /*get status bits for this stream*/

  return 0;
}
/* end of mus_read_stream_status*/

/*mus_read_stream_trlim.
 * Read the trigger limit for a stream. Returns -1 if identifier invalid, -2
 * if illegal stream. The same register is used for 4 streams, but the access
 * is as if each is separate. trlim is not interpreted inside this routine.
 */
int mus_read_stream_trlim(short identifier, short stream, short *trlim)
{
  volatile unsigned short *ip;

/*check identifier is valid.*/
  if(identifier<0 || identifier>=mus_total) return -1;

/*check stream is valid.*/
  if(stream<0 || stream>MAX_STREAM) return -2;

/*set the address*/
  ip = (unsigned short *)
   (mus_info[identifier].address + MUS_BUS + (stream<<4) + MUS_STRTRL);

  *trlim = (*ip)&3;		/*read the register*/

  return 0;
}
/* end of mus_read_stream_trlim */

/*mus_read_testmode.
 * Read the MuSTARD test mode register. Only d0/d1 are relevant.
 * Return -1 if identifier invalid.
 */
int mus_read_testmode(short identifier, short *mode)
{
  volatile unsigned short *ip;

/*check identifier is valid.*/
  if(identifier<0 || identifier>=mus_total) return -1;

/*set the address of the test mode register*/
  ip = (unsigned short *)(mus_info[identifier].address + MUS_TESTMODE);

  *mode = (*ip)&3;  /*get lower 2 bits of register*/

  return 0;
}
/* end of mus_read_testmode*/

/*mus_release_event.
 * Release an event from the output buffer of the MuSTARD. There are a
 * number of error returns:
 * -1 - identifier invalid.
 * -2 - bad status from reading control register.
 * -3 - MRDOEN not set, so buffer not accessible.
 * -4 - no event to release.
 * This routine should only be needed in cases where normal readout has failed
 * and the user wants to clear the current event regardless.
 */
int mus_release_event(short identifier)
{
  volatile unsigned short *ipbase;
  short mdaqen,mrdoen;
  int status;

  if(identifier<0 || identifier>=mus_total) return -1;

  ipbase = (unsigned short *)(mus_info[identifier].address + MUS_EVTBUFF);
  status = mus_read_readout_cntl(identifier,&mdaqen, &mrdoen);
  if(status != 0) {
    printf("mus_release_event: Bad status from mus_read_readout_cntl: %d\n",status);
    return -2;
  }

  if(mrdoen != 1) {
    printf("mus_release_event: MRDOEN bit in control register not set\n");
    return -3;
  }

  if(((*ipbase)&0x8000)!=0) {
    printf("mus_release_event: No event to release\n");
    return -4;
  }

/*now release the event*/
  *ipbase = 0;			/*data not relevant*/
  return 0;
}
/* end of mus_release_event */

/*mus_reset.
 * Reset the module. Write 1 followed by 0 into the reset register. Return
 * -1 if identifier invalid.
 */
int mus_reset(short identifier)
{
  volatile unsigned short *ip;

/*check identifier is valid.*/
  if(identifier<0 || identifier>=mus_total) return -1;

/*set the address*/
  ip = (unsigned short *)(mus_info[identifier].address + MUS_RESET);
  *ip = 1;
  *ip = 0; /*set and clear the reset*/
  return 0;
}
/* end of mus_reset */

/*mus_set_clock_cntl.
 * Set the MuSTARD clock control register (bit 0 only is sensible).
 * Return -1 if identifier invalid.
 */
int mus_set_clock_cntl(short identifier, short clock)
{
  volatile unsigned short *ip;

/*check identifier is valid.*/
  if(identifier<0 || identifier>=mus_total) return -1;

/*set the address of the clock register*/
  ip = (unsigned short *)(mus_info[identifier].address + MUS_CLKCNTL);

  *ip = clock&1;  /*set bit 0 of register*/

  return 0;
}
/* end of mus_set_clock_cntl*/

/*mus_set_readout_cntl.
 * Set the MuSTARD readout control register (bits 0&1 only are sensible).
 * Return -1 if identifier invalid.
 */
int mus_set_readout_cntl(short identifier, short mdaqen, short mroen)
{
  volatile unsigned short *ip;

/*check identifier is valid.*/
  if(identifier<0 || identifier>=mus_total) return -1;

/*set the address of the readout control register*/
  ip = (unsigned short *)(mus_info[identifier].address + MUS_ROCNTL);

  *ip = ((mdaqen&1)<<1)+(mroen&1);  /*get bits 0&1 of register*/

  return 0;
}
/* end of mus_set_readout_cntl*/

/*mus_set_readout_modes.
 * Set the MuSTARD readout mode register (bits 0 and 1 only are sensible).
 * Decode from raw_mode and sample_all (this is a bit at variance with the
 * approach elsewhere).
 * Return -1 if identifier invalid.
 */
int mus_set_readout_modes(short identifier, short raw_mode, short sample_all)
{
  volatile unsigned short *ip;

/*check identifier is valid.*/
  if(identifier<0 || identifier>=mus_total) return -1;

/*set the address of the readout mode register*/
  ip = (unsigned short *)(mus_info[identifier].address + MUS_ROMODE);

  *ip = ((raw_mode&1)<<1) + (sample_all&1);  /*set bits 0&1 of register*/

  return 0;
}
/* end of mus_set_readout_modes*/

/*mus_set_stream_delay.
 * Set the delay for a single stream. Streams 0-MAX_STREAM are valid. Outside
 * this range, return -2. Return -1 if identifier invalid. Return -3 if delay
 * requested is outside range 0-24 (nsec). Wait inside routine until transfer
 * has completed (d4 on readout control register goes to 0) - return -4 if
 * register "never" set.
 */
int mus_set_stream_delay(short identifier, short stream, short delay)
{
  volatile unsigned short *ip;
  short i,j;
#ifdef OS9
  unsigned int ldelay;
#else
  long ldelay,t_zero;
#endif

/*check identifier is valid.*/
  if(identifier<0 || identifier>=mus_total) return -1;

/*check stream number is valid.*/
  if(stream<0 || stream>MAX_STREAM) return -2;

/*check delay is valid.*/
  if(delay<0 || delay>24) return -3;

/*set the address - delay register for this stream*/
  ip = (unsigned short *)
    (mus_info[identifier].address + MUS_DELAY + ((stream/4)<<4) + ((stream%4)<<1));

  *ip = delay;	/*write the delay for this stream*/

/* now wait until transfer done - give up a time slice on OS9, wait 200
   microsec elsewhere, and check d4 on readout control register. If loop
   >10 times, exit with return code -4 */

  ip = (unsigned short *)(mus_info[identifier].address + MUS_ROCNTL);

  for(i=0;i<10;i++){

#if defined(OS9)
    ldelay = 2;		/*give up a clock cycle*/
    _os9_sleep(&ldelay);
#else 
    ldelay = (long)CLOCKS_PER_SEC/5000;		/*200microsec*/
    if(ldelay<=0) ldelay=1;
    for(t_zero=(long)clock();(long)clock()-t_zero<=ldelay;);
#endif // OS9
    j = (*ip)&0x10;
    if(j == 0) return 0;	/*d4=0 means transfer completed*/
  }
  printf("transfer to delay chip on MuSTARD %d, stream %d never completed\n",
     identifier,stream);
  return -4;
}
/* end of mus_set_stream_delay*/

/*mus_set_stream_enable.
 *Enable or disable a single stream. Streams 0-MAX_STREAM are valid. Outside
 * this range, return -2. Return -1 if identifier invalid.
 */
int mus_set_stream_enable(short identifier, short stream, short enable)
{
  volatile unsigned short *ip;

/*check identifier is valid.*/
  if(identifier<0 || identifier>=mus_total) return -1;

/*check stream number is valid.*/
  if(stream<0 || stream>MAX_STREAM) return -2;

/*set the address - mode register for this stream*/
  ip = (unsigned short *)
    (mus_info[identifier].address + MUS_BUS + (stream<<4) + MUS_STRMODE);

  *ip = enable&1;	/*enable or disable this stream*/

  return 0;
}
/* end of mus_set_stream_enable*/

/*mus_set_stream_paf.
 * Set PAF register for a stream. Streams 0-MAX_STREAM are valid. Outside
 * this range, return -2. Return -1 if identifier invalid. The value of paf
 * is assumed to be set up external to this routine. There is one PAF
 * register for 4 streams.
 */
int mus_set_stream_paf(short identifier, short stream, short paf)
{
  volatile unsigned short *ip;

/*check identifier is valid.*/
  if(identifier<0 || identifier>=mus_total) return -1;

/*check stream number is valid.*/
  if(stream<0 || stream>MAX_STREAM) return -2;

/*set the address - paf register for this stream*/
  ip = (unsigned short *)
    (mus_info[identifier].address + MUS_BUS + (stream<<4) + MUS_STRPAF);

  *ip = (paf&3);  /*set paf bits for this stream*/

  return 0;
}
/* end of mus_set_stream_paf*/

/*mus_set_stream_trlim.
 * Set the trigger limit for a stream. Returns -1 if identifier invalid, -2
 * if illegal stream. The same register is used for 4 streams, but the access
 * is as if each is separate. trlim is assumed to have been set up correctly
 * external to this routine.
 */
int mus_set_stream_trlim(short identifier, short stream, short trlim)
{
  volatile unsigned short *ip;

/*check identifier is valid.*/
  if(identifier<0 || identifier>=mus_total) return -1;

/*check stream is valid.*/
  if(stream<0 || stream>MAX_STREAM) return -2;

/*set the address*/
  ip = (unsigned short *)
   (mus_info[identifier].address + MUS_BUS + (stream<<4) + MUS_STRTRL);
  *ip = (trlim&3);		/*set the register*/
  return 0;
}
/* end of mus_set_stream_trlim */

/*mus_set_testfifo.
 * Set the MuSTARD test fifo register. The value to be written is assumed
 * to be set up externally to this routine.
 * Return -1 if identifier invalid.
 */
int mus_set_testfifo(short identifier, short fifo)
{
  volatile unsigned short *ip;

/*check identifier is valid.*/
  if(identifier<0 || identifier>=mus_total) return -1;

/*set the address of the test mode register*/
  ip = (unsigned short *)(mus_info[identifier].address + MUS_TESTFIFO);

  *ip = fifo&0x3FFF;  /*set lower 14 bits of register*/

  return 0;
}
/* end of mus_set_testfifo*/

/*mus_set_testmode.
 * Set the MuSTARD test mode register. Only d0/d1 are relevant.
 * Return -1 if identifier invalid.
 */
int mus_set_testmode(short identifier, short mode)
{
  volatile unsigned short *ip;

/*check identifier is valid.*/
  if(identifier<0 || identifier>=mus_total) return -1;

/*set the address of the test mode register*/
  ip = (unsigned short *)(mus_info[identifier].address + MUS_TESTMODE);

  *ip = mode&3;  /*set lower 2 bits of register*/

  return 0;
}
/* end of mus_set_testmode*/

/*mus_set_testtrig.
 * Set the MuSTARD test trigger register. Data written is irrelevant.
 * Return -1 if identifier invalid.
 */
int mus_set_testtrig(short identifier)
{
  volatile unsigned short *ip;

/*check identifier is valid.*/
  if(identifier<0 || identifier>=mus_total) return -1;

/*set the address of the test trigger register*/
  ip = (unsigned short *)(mus_info[identifier].address + MUS_TESTTRIG);

  *ip = 0;  /*Data doesn't matter*/

  return 0;
}
/* end of mus_set_testtrig*/


/* mus_test_buffer (GFM Nov99)
 * Check whether any data is available in the output buffer;
 * return empty = 1 if bit 15 of the first word of the dual port RAM is set,
 * otherwise return empty = 0.   This routine intended for burst-mode
 * readout where a number of readout modules (Mustards, IRAMs etc.) are
 * cyclically accessed and read only when data is available.
 */

int mus_test_buffer(short identifier, short *empty)
{
  volatile unsigned short *ip, *ipbase, test;
  short mdaqen,mrdoen;
  int rcode;

  if(identifier<0 || identifier>=mus_total) return -1;

  ipbase = (unsigned short *)(mus_info[identifier].address + MUS_EVTBUFF);
  ip = ipbase;
  rcode = mus_read_readout_cntl(identifier,&mdaqen, &mrdoen);
  if(rcode != 0) {
    printf("mus_test_buffer: Bad rcode from mus_read_readout_cntl: %d\n",rcode);
    return -2;
  }

  if(mrdoen != 1) {
    printf("mus_test_buffer: MRDOEN bit in control register not set\n");
    return -3;
  }

  test = (unsigned short)0x8000;
  *empty = (*ipbase)&test;

  return 0;
}

/* mus_test_fifos (PWP 03.08.99)
 * Check input FIFOs are empty - to be called at eob to ensure that the 
 * last event has been processed correctly.  Returns a bitmap where if
 * bit n = 1, stream n is empty.
 */
 
int mus_test_fifos(short identifier, short *empty)
{
  short stream, enabled, status;
  short empty2 = 0;

  /*check identifier is valid.*/
  if(identifier<0 || identifier>=mus_total) return -1;

  for(stream = 0; stream<12; stream++){

    mus_read_stream_enable(identifier,stream,&enabled);
    if(enabled){
      mus_read_stream_status(identifier,stream,&status);    
      if(status&0x0001)   /* fifo is empty */
        empty2 |= 0x0001<<stream;
    }
  }
  *empty = (short)empty2;
  return 0;
}

/* end mus_test_fifos */

/*mus_vmemap.
 * Map the MuSTARD to the CPU address space. The input address is the true
 * VME address for UNIX and WNT implementations, mapped address for OS9 (this
 * is because the mapping in OS9 is CPU-specific and so it is not possible to
 * provide a generalised mapping within this routine). The mapping in UNIX/WNT
 * is A32.
 */
int mus_vmemap(unsigned long address, short *identifier)
{
#if defined(NIVXI)
  NIVXI_STATUS ret;
  UINT32 w,wstart, wend;
#elif defined(RCC)
  VME_ErrorCode_t status;
  VME_MasterMap_t master_map;
  u_int virtual_address;
  int handle;
#else
  int status;
#endif
  int am;
  char *addr;

  am = 0x9;
  *identifier = mus_total++;		/*Simplest approach? */

/* Use malloc/realloc to get space for the structure */
  if(mus_total == 1)
    mus_info = (struct mus_info_structure *) malloc(mus_info_size);
  else
    mus_info = (struct mus_info_structure *) realloc((void *)mus_info,mus_total*mus_info_size);

  mus_info[*identifier].address = address;
#ifdef NIVXI
  mus_info[*identifier].window = 0;
#endif

/* Now do the mapping - CPU-dependent - allow a 0x20000 window */

#ifdef OS9
  status = _os_permit((void *)address, 0x20000,2,0);
  if(status != 0) {
     printf("VME mapping failed, status = %d\n",status);
     return status;
  }
#endif

#ifdef RAID
  am = 0x9;		/* A32 access */
  status = RD13_VmeMap(address,am,0x20000,&addr,0);
  if(status != 0) {
     printf("VME mapping failed, status = %d\n",status);
     return status;
  }
  mus_info[*identifier].address = (unsigned long) addr;
#endif

#ifdef NIVXI
  if(mus_total == 1) {
	InitVXIlibrary();		/*only call once*/
	MapVXIAddressSize(0x20000);	/*allow up to 128kB window*/
  }
  addr = (char *)MapVXIAddress(3,address,5000,&w,&ret);
  if(ret != 0) {
     printf("VME mapping failed, status = %d\n",ret);
     return (int)ret;
  }
  printf("mus_vmemap: MapVXIAddress ret %d win %d addr %8x for vme %8x\n", 
		ret, w, addr, address);
  ret = GetWindowRange(w, &wstart, &wend);
  printf("mus_vmemap: WindowRange start %lx end %lx\n", wstart, wend);
  mus_info[*identifier].address = (unsigned long) addr;
  mus_info[*identifier].window = w;
#endif

#ifdef QB_linux
  Initqb(&address ,WND_SIZE,VME_A32);
  mus_info[*identifier].address = address;
#endif

#ifdef RCC
  if(mus_total == 1) status = VME_Open();
/*define the structure for the map */
  master_map.vmebus_address = address;
  master_map.window_size = 0x20000;
  master_map.address_modifier = VME_A32;
  master_map.options = 0;
  status = VME_MasterMap(&master_map,&handle);
  if(status != VME_SUCCESS) {
    VME_ErrorPrint(status);
    return (int)status;
  }
  status = VME_MasterMapVirtualAddress(handle,&virtual_address);
  printf("mus_vmemap: VME address %x mapped to %x, handle %d\n",
    address, virtual_address,handle);
  mus_info[*identifier].address = virtual_address;
  mus_info[*identifier].window = handle;
#endif

  return 0;
}
/*end of mus_vmemap */


/* mus_test_event.    GFM Sep98
 * Read and decode event header, then if no errors are detected, 
 * clear the Mustard event buffer quickly and return.  
 * If a chip error or protocol violation is detected, the buffer
 * is read as in a normal mus_read_event(), the returned data being
 * then available for external diagnosis.
 * This routine is specifically for use when sending exactly one
 * trigger and waiting for returned data, but where histograms are
 * the required output and speed is advantageous.  Note that errors
 * are only detected (and histograms filled) in DECODED mode.
 * Return codes are as for mus_read_event():
 * -1 - identifier invalid.
 * -2 - bad status from reading control register.
 * -3 - MRDOEN not set, so buffer not accessible.
 * -4 - timeout waiting for event availability
 * -5 - error bit (14) set in 1st word in buffer. Buffer filled assuming
 *      bits 0-13 are the event length.
 * After the event is readout, the buffer is cleared by writing to the 1st
 * location.
 */
int mus_test_event(short id, short *length, short **buffer, int *nevents, short *nerrors, short dump)
{
  volatile unsigned short *ip, *ipbase;
  short mdaqen,mrdoen;
  int i,error,rcode;
  int loop = 500000;
  short prev_nerrors;

  prev_nerrors = *nerrors;

  if(id<0 || id>=mus_total) return -1;

  ipbase = (unsigned short *)(mus_info[id].address + MUS_EVTBUFF);
  ip = ipbase;
  rcode = mus_read_readout_cntl(id,&mdaqen, &mrdoen);
  if(rcode != 0) {
    printf("mus_test_event: Error from mus_read_readout_cntl: %d\n",rcode);
    return -2;
  }
  if(mrdoen != 1) {
    printf("mus_test_event: MRDOEN bit in control register not set\n");
    return -3;
  }

  while (((*ipbase)&0x8000)!=0 && loop>0) loop--;
  if(loop == 0) {
    printf("mus_test_event: Timeout waiting for data\n");
    *ipbase = 0;  /*clear output buffer in case of partial event */
    return -4;
  }

  *length = (*ipbase)&0x3FFF;  /*bits 0-13*/
  error = (*ipbase)&0x4000;    /*bit 14:  protocol violation this event*/
  *nerrors = *(ipbase+1);      /*Number of accumulated protocol violations*/ 
  *nevents = (*(ipbase+3) << 16) + (*(ipbase+2));

  if ((error==0) && (*nerrors > prev_nerrors)) error = 1;  /* Interim fix while bit 14 doesn't work on some Mustards */ 

  if (error||dump) for (i=0;i<(*length);i++) *(*buffer)++ = *ip++;

  *ipbase = 0;	/* clear output buffer (release event)*/
  if(error != 0) return -5;    /*error bit set*/
  return 0;
}
/* end of mus_test_event */


/* ### Test routines
 * Following routines should not normally be needed for running of a fully
 * debugged system. They are provided for the assistance of users trying
 * to get a MuSTARD working.
 */

/* mus_test_stren.
 * Check the stream enable bits for all 12 channels.
 */

int mus_test_stren(short id) 
{
  volatile unsigned short *ip;
  int n;

  printf("mus_test_stren: ");

    for (n=0; n<12; n++) {
    ip = (unsigned short *) (mus_info[id].address + MUS_BUS + (n * 0x10));
    printf(" %x",(*ip) & 0xf);
    }
  printf("\n");
  return 0;
}
/* end of mus_test_stren */

/* mus_loop_stren
 * read stream register many times and print result
 */

int mus_loop_stren(short id, short offset)
{
  volatile unsigned short *ip;
  int n;

  ip = (unsigned short *) (mus_info[id].address + MUS_BUS + offset);

  printf("mus_loop_stren: offset %x\n ", offset);

  for (n=0; n<10000; n++) {
    printf("%x\n",(*ip) & 0xf);
  }
  return 0;
}

/* end of mus_loop_stren */

/* mus_hmem.
 * Test the histogram memory - write a pattern into the memory and read it
 * back.
 */

int mus_test_hmem(short id)
{
  volatile unsigned short *ip, *nip;
  unsigned short val,n;

  printf("mus_test_hmem: ");

  ip = (unsigned short *) (mus_info[id].address + MUS_HISTMEM);
  val = *ip;
  printf(" %x ",val); 
#ifdef BIT3
  for (n=0; n<1000; n++) {
    nip = ip + n;
    *nip = n;
    NI_pause();
  }
#else
  for (n=0; n<1000; n++) {
    nip = ip + n;
    *nip = n;
  }
#endif
  for (n=0; n<10; n++) {
    nip = ip + n;
    val = *nip;
    printf(" %x ",val); 
  }
  printf("\n");
  return 0;
}

/* end of mus_test_hmem */

void mus_dump_event(short identifier, unsigned short *data){
  int i;
  int word_count, error_count, event_count;

  printf("mus_dump_event:  header words:   %4x %4x %4x %4x\n", data[0], data[1], data[2], data[3]);
  
  word_count = data[0] & 0x3fff;
  if (word_count < 4) printf("   word_count %d is smaller than header !!!\n", word_count);

  error_count = data[1];
  event_count = data[2] + (data[3] << 8);

  printf("   word_count %d   error_count %d   event_count %d\n",
	 word_count, error_count, event_count);

  if (word_count < 7) printf("   word_count too small for stream data !!!\n");

  for (i=4; i<word_count; i++){
     printf("   word %4d  value %4x ", i, (unsigned short)data[i]);
	 if (data[i] & 0x8000) {
		 printf("  STREAM HEADER  L1 %1x  BC %2x  Stream Number %4x\n", (data[i]&0x0f00)>>8, data[i]&0x00ff, data[i+1]);
         i++;
     } else if (data[i] & 0x2000)
		 printf("     NEXT HIT  Chip %1x  Strip %2x\n", (data[i]&0x0780)>>7, data[i]&0x007f);
     else 
		 printf("        DATA  %1x\n", data[i]&0x0007);
  }
}


void mus_debug_event(short identifier, short *data, short level){
  int i;
  int word_count, error_count, event_count;

  printf("mus_dump_event:  header words:   %4x %4x %4x %4x\n", data[0], data[1], data[2], data[3]);
  
  word_count = data[0] & 0x3fff;
  if (word_count < 4) printf("   word_count %d is smaller than header !!!\n", word_count);

  error_count = data[1];
  event_count = data[2] + (data[3] << 8);

  printf("   word_count %d   error_count %d   event_count %d\n",
	 word_count, error_count, event_count);

  if (word_count < 7) printf("   word_count too small for stream data !!!\n");

  for (i=4; i<word_count; i++){
     printf("   word %4d  value %4x ", i, (unsigned short)data[i]);
	 if (data[i] & 0x8000) {
		 printf("  STREAM HEADER  L1 %1x  BC %2x  Stream Number %4x\n", (data[i]&0x0f00)>>8, data[i]&0x00ff, data[i+1]);
         i++;
     } else if (data[i] & 0x2000)
		 printf("     NEXT HIT  Chip %1x  Strip %2x\n", (data[i]&0x0780)>>7, data[i]&0x007f);
     else 
		 printf("        DATA  %1x\n", data[i]&0x0007);
  }
}
 
