/* slog_lib.c. 
 *   Version 0.1  981026 GFM  
 *   Version 0.11 990719 PWP - adjustments for new firmware
 *   Version 0.12 990727 GFM - slog_set_clock_duty (switch only)
 *   Version 0.13            - adjustments for RAID
 *   Version 0.2  990703 PWP - additions for SLOGB
 *   Version 0.21 pwp 03.09.01 - additions for use with BIT3 interface
 *   Version 0.22 PWP 12.09.01 - add slog_test_mem routine
 *   Version 0.23 BJG/PWP 24.05.02 - support for NIVXI under linux
 *                                   and remove <0 tests on unsigned data
 *   Version 0.24 JCH 24.06.02 - support for ATLAS RCC library (for SBCs).
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <time.h>

#ifdef NIVXI
#include <nivxi.h>
#endif

#ifdef OS9
#include <process.h>
#include <signal.h>
#include <types.h>
#endif

#ifdef RAID
#define CLOCKS_PER_SEC 100000
#endif

#ifdef QB_linux
#define WND_SIZE 0x20000
#include "../QB/InitVME.h"
#endif

#ifdef RCC
#include <rcc_error.h>
#include <vme_rcc.h>
#endif

#include "slog_private.h"
#include "slog_lib.h"
#include "../stlib/sct_hardware.h"

short slog_total = 0;   /* Total number of SLOGs mapped as per RD13 usage */
struct slog_info_structure *slog_info;

/*************************************************************************
  Channel identifiers:
    For registers, eg, the enables:
     0-11: the 12 LVDS channels
     12,13: the 2 ECL outputs 0,1  (the NIM command outputs are always on)

    For RAM command channels:
     0-11: the 12 LVDS command channels
     12,13: the 2 NIM outputs 0,1
        (the ECL command outputs are selected by gmr_a, gmr_b from 0-11)
    

  Return codes:  unless otherwise specified,
     0 == no error
    -1 == id wrong (identifier negative or requested SLOG not mapped)
    -2 == ch wrong (channel negative or greater than allowed)
    -3 == other argument out of allowed range
    -4 == action possible with SLOGB only, and SLOGA is being used
*******************************************************************/


int slog_show_status(short id){
  unsigned short v;

  printf("slog_status: ");

  slog_get_runstate(id,&v);
  printf(" RUN %x",v); 

  slog_get_startaddr(id,&v);
  printf(" STADDR %x",v);

  slog_get_cycles(id,&v);
  printf(" CYCLES %d",v);

  slog_get_busy(id,&v);
  printf(" BENABLES %x  BSTATES %x ", v&7, ((v>>3)&7));
    
  if (v&0x40) printf("BUSY ON");
  else  printf("BUSY OFF");

  slog_get_clock(id,&v);
  printf("  CLOCKS %4x ",v);

  slog_get_internal(id,&v);
  printf("  INTCOM %4x ",v);  

  slog_get_external(id,&v);
  printf("  EXTCOM %4x ",v);  
  
  if(slog_get_clk_cntl(id,&v)==0)
    printf("  CLKCNTL %4x ",v);  

  printf("\n");
  return 0;
}


int slog_init(short id) {

  slog_stop(id);
  printf(".");
  slog_set_busy(id,0);
  printf(".");
  slog_set_startaddr(id,0);
  printf(".");
  slog_set_cycles(id,0);
  printf(".");
  slog_set_gmra(id,0);
  printf(".");
  slog_set_gmrb(id,1);
  printf(".");
  slog_test_revision(id);
  printf(".");
  slog_set_clk_cntl(id,0); /* PWP 11.10.00, set SLOG to defaults at power up */
  printf(".");
  slog_enable_clock(id);
  printf(".");
  slog_enable_external(id);
  printf(".");
  slog_enable_internal(id);
  printf("finished\n");

  return 0;
}


int slog_reset(short id){

  slog_stop(id);
  slog_set_startaddr(id,0);
  slog_set_cycles(id,0);
  slog_disable_busy(id,0);
  slog_disable_busy(id,1);
  slog_disable_busy(id,2);
  slog_enable_clock(id);
  slog_enable_external(id);
  slog_enable_internal(id);
  slog_set_gmra(id,0);
  slog_set_gmrb(id,1); 
  slog_test_revision(id);

  return 0;
}


int slog_clear(short id){
  volatile unsigned short *ip;
  unsigned int n,size;

  if ((id < 0) || (id >= slog_total)) return -1;

  ip = (unsigned short *)(slog_info[id].address + SLOG_BASE);
  size = SLOG_SIZE; 
         
#ifdef BIT3
  for (n=0; n<size; n++){
    *ip++ = 0x8000;
    NI_pause();
  }
#else
  for (n=0; n<size; n++){
    *ip++ = 0x8000;
  }
#endif
  
  return 0;
}


int slog_load(short id, unsigned short *cb, unsigned int size){
  volatile unsigned short *ip;
  unsigned int n;

  if ((id < 0) || (id >= slog_total)) return -1;
  if (size > SLOG_SIZE-2) return -3;

  slog_stop(id);
  
  ip = (unsigned short *)(slog_info[id].address + SLOG_BASE);

#ifdef BIT3
  for (n=0; n<size; n++) {
    *ip++ = *cb++;
    NI_pause();
  }
  *ip++ = 0x8000;
  NI_pause();
  *ip++ = 0x0000;
  NI_pause();
  *ip++ = 0x0000;
  NI_pause();
#else
  for (n=0; n<size; n++) {
    *ip++ = *cb++;
  }
  /* PWP bugfix 10.01.00 
   * write stop bit followed by two zeros - otherwise can get 
   * spurious output if location (n+1) not already empty */
  *ip++ = 0x8000;
  *ip++ = 0x0000;
  *ip++ = 0x0000;
#endif

  return 0;
}


int slog_fload(short id, char *cb_filespec){
  FILE *cbf;
  unsigned short buf[SLOG_SIZE];
  char inl[128];
  unsigned int rep, ci;
  unsigned short cs;  
  unsigned int inlcnt, bufcnt, n;

  if ((cbf = fopen(cb_filespec, "r")) == NULL) {
    printf("slog_fload: ERROR opening cb file %s\n", cb_filespec);
    return(-3);
  }
  inlcnt = 0;
  bufcnt = 0;
  while (fgets(inl, 128, cbf) != NULL) {
    if (inl[0] != '#') {
      sscanf(inl,"%d %x\n", &rep, &ci);  
      cs = (unsigned short) (ci & 0x0000FFFF);

      /*
       * I do not understand why the following line was ever here:
       * 
       * if (cs&0x0001) cs = 0x3fff; else cs = 0x0000;
       *
       * This means that the stop bit cannot be set from the file.
       * Hence the line is removed.  PWP 21.06.00
       */

#ifdef DEBUG
      printf("slog_fload: IN: line: %d  rep: %d  command short: %4x\n", 
         inlcnt, rep, cs);    
#endif
      for (n=0; n<rep; n++) {
        buf[bufcnt] = cs;
#ifdef DEBUG
        printf("slog_fload: OUT: bufcnt %d  buf %4x\n", bufcnt, buf[bufcnt]); 
#endif
        bufcnt++;  
        if (bufcnt > SLOG_SIZE) {
          printf("slog_fload: ERROR: cb too large, at input line %d\n",inlcnt);
          break;  
        }
      }   
    }
    inlcnt++;
  }
  slog_clear(id);
  slog_load(id, buf, bufcnt);
  fclose(cbf);

  return 0;
}


int slog_dump(short id){
  volatile unsigned short *ip;

  if ((id < 0) || (id >= slog_total)) return -1;

  printf("slog_dump: id %d\n\n", id);
  ip = (unsigned short *)(slog_info[id].address + SLOG_BASE);
  while ((*ip&0x8000)==0) {
    printf("%2x ", *ip++);
  }
  printf("\n"); 
  return 0;
}


int slog_dump_ram(short id, unsigned short start, unsigned short stop) {
  volatile unsigned short *ip;
  unsigned short p;
  FILE* fp;
  char fname[64];

  sprintf(fname,"%sslog_dump.txt",ST_DATA_PATH);

  fp = fopen(fname,"w");

  fprintf(fp, "slog_dump_ram: from %x to %x\n",start,stop);

  if ((id < 0) || (id >= slog_total)) return -1;
  if ((start > SLOG_SIZE) || (stop > SLOG_SIZE)) return -3;
  if (start > stop) return -4;

  ip = (unsigned short *)(slog_info[id].address + SLOG_BASE);

  for (p=start; p<stop; p++) {
    fprintf(fp, "%4x %4x\n", p, *(ip+p));
  }
  
  fclose(fp);

  return 0;
}


int slog_show_ram(short id, unsigned short start, unsigned short stop) {
  volatile unsigned short *ip;
  unsigned short p;
  unsigned short i = 0;

  printf("slog_show_ram: from %x to %x\n",start,stop);

  if ((id < 0) || (id >= slog_total)) return -1;
  if ((start > SLOG_SIZE) || (stop > SLOG_SIZE)) return -3;
  if (start > stop) return -4;

  ip = (unsigned short *)(slog_info[id].address + SLOG_BASE);  
  for (p=start; p<stop; p++) {
    printf("%4x ", *(ip+p));
    i++;
    if ((i%10)==0) printf("\n");
  }
  printf("\n");
  return 0;
}


int slog_start(short id){
  volatile unsigned short *ip;

  if ((id<0) || (id>=slog_total)) return -1;

  ip = (unsigned short *)(slog_info[id].address + SLOG_RUN);
  *ip = 0x1;

  return 0;
}


int slog_stop(short id){
  volatile unsigned short *ip;

  if ((id<0) || (id>=slog_total)) return -1;

  ip = (unsigned short *)(slog_info[id].address + SLOG_RUN);
  *ip = 0x0;

  return 0;
}


int slog_get_runstate(short id, unsigned short *runstate){
  volatile unsigned short *ip;
  
  if ((id<0) || (id>=slog_total)) return -1;

  ip = (unsigned short *)(slog_info[id].address + SLOG_RUN);
  *runstate = *ip;
  return 0;      
}


int slog_set_startaddr(short id, unsigned short startaddr){ 
  volatile unsigned short *ip;

  if ((id<0) || (id>=slog_total)) return -1;

  ip = (unsigned short *)(slog_info[id].address + SLOG_STARTADDR);
  *ip = startaddr;
  return 0; 
}


int slog_get_startaddr(short id, unsigned short *startaddr){
  volatile unsigned short *ip;

  if ((id<0) || (id>=slog_total)) return -1;

  ip = (unsigned short *)(slog_info[id].address + SLOG_STARTADDR);
  *startaddr = *ip;
  return 0; 
}


int slog_set_cycles(short id, unsigned short cycles){ 
  volatile unsigned short *ip;

  if ((id<0) || (id>=slog_total)) return -1;

  ip = (unsigned short *)(slog_info[id].address + SLOG_CYCLES);
  *ip = cycles;
  return 0; 
}


int slog_get_cycles(short id, unsigned short *cycles){
  volatile unsigned short *ip;

  if ((id<0) || (id>=slog_total)) return -1;

  ip = (unsigned short *)(slog_info[id].address + SLOG_CYCLES);
  *cycles = *ip;
  return 0; 
}


int slog_enable_busy(short id, unsigned short busy){
  unsigned short previous;
  volatile unsigned short *ip;

  if ((id<0) || (id>=slog_total)) return -1;
  if (busy>2) return -3;

  ip = (unsigned short *)(slog_info[id].address + SLOG_BUSY);
  previous = *ip&0x7;
  *ip = previous|(1<<busy);

  return 0;
}


int slog_disable_busy(short id, unsigned short busy){
  unsigned short previous;
  volatile unsigned short *ip;

  if ((id<0) || (id>=slog_total)) return -1;
  if (busy>2) return -3;

  ip = (unsigned short *)(slog_info[id].address + SLOG_BUSY);
  previous = *ip&0x7;
  if (busy==0) *ip = previous&0x6;
  else if (busy==1) *ip = previous&0x5;
  else if (busy==2) *ip = previous&0x3;

  return 0;
}


int slog_set_busy(short id, unsigned short bmap){ 
  volatile unsigned short *ip;

  if ((id<0) || (id>=slog_total)) return -1;
  if (bmap>7) return -3;

  ip = (unsigned short *)(slog_info[id].address + SLOG_BUSY);
  *ip = bmap;
  return 0; 
}


int slog_get_busy(short id, unsigned short *bmap){
  volatile unsigned short *ip;

  if ((id<0) || (id>=slog_total)) return -1;

  ip = (unsigned short *)(slog_info[id].address + SLOG_BUSY);
  *bmap = *ip&0x7;
  return 0; 
}


int slog_enable_clock(short id){     /* Enable all clock channels */
  volatile unsigned short *ip;

  if ((id<0) || (id>=slog_total)) return -1;

  ip = (unsigned short *)(slog_info[id].address + SLOG_CLOCK);
  *ip = 0x3fff;

  return 0;
}


int slog_disable_clock(short id){     /* Disable all clock channels */
  volatile unsigned short *ip;

  if ((id<0) || (id>=slog_total)) return -1;

  ip = (unsigned short *)(slog_info[id].address + SLOG_CLOCK);
  *ip = 0;

  return 0;
}


int slog_get_clock(short id, unsigned short *clock){     /* Return clock enables */
  volatile unsigned short *ip;

  if ((id<0) || (id>=slog_total)) return -1;

  ip = (unsigned short *)(slog_info[id].address + SLOG_CLOCK);
  *clock = *ip;
  
  return 0; 
}


void slog_test_regs2(short id){

  unsigned short c, i;
  volatile unsigned short *ipc, *ipi;

  ipc = (unsigned short *)(slog_info[id].address + SLOG_CLOCK);
  ipi = (unsigned short *)(slog_info[id].address + SLOG_INTERNAL);

  c = *ipc;  printf("clock:    %x\n", c);
  i = *ipi;  printf("internal: %x\n", i);
  c = *ipc;  printf("clock:    %x\n", c);
  i = *ipi;  printf("internal: %x\n", i);
  c = *ipc;  printf("clock:    %x\n", c);
  i = *ipi;  printf("internal: %x\n", i);

}


void slog_test_regs(short id){

  volatile unsigned short *ipc, *ipi;

  ipc = (unsigned short *)(slog_info[id].address + SLOG_CLOCK);
  ipi = (unsigned short *)(slog_info[id].address + SLOG_INTERNAL);

  printf("slog test registers:\n");
  printf("clock:    %x\n", *ipc);
  printf("internal: %x\n", *ipi);
  printf("clock:    %x\n", *ipc);
  printf("internal: %x\n", *ipi);
  printf("clock:    %x\n", *ipc);
  printf("internal: %x\n", *ipi);

}


int slog_set_clock(short id, unsigned short bmap){ 
  volatile unsigned short *ip;

  if ((id<0) || (id>=slog_total)) return -1;

  ip = (unsigned short *)(slog_info[id].address + SLOG_CLOCK);
  *ip = bmap;
  return 0; 
}


int slog_enable_external(short id){
  volatile unsigned short *ip;

  if ((id<0) || (id>=slog_total)) return -1;

  ip = (unsigned short *)(slog_info[id].address + SLOG_EXTERNAL);
  *ip = 0xffff;   /* new SLOG firmware, PWP 19.07.99 */

  return 0;
}


int slog_disable_external(short id){
  volatile unsigned short *ip;

  if ((id<0) || (id>=slog_total)) return -1;

  ip = (unsigned short *)(slog_info[id].address + SLOG_EXTERNAL);
  *ip = 0;

  return 0;
}


int slog_set_external(short id, unsigned short bmap){ 
  volatile unsigned short *ip;


  if ((id<0) || (id>=slog_total)) return -1;
  if (bmap > 0xffff) return -3; /* new SLOG firmware, PWP 19.07.99 */

  ip = (unsigned short *)(slog_info[id].address + SLOG_EXTERNAL);
  *ip = bmap;
  return 0; 
}


int slog_get_external(short id, unsigned short *external){
  volatile unsigned short *ip;

  if ((id<0) || (id>=slog_total)) return -1;

  ip = (unsigned short *)(slog_info[id].address + SLOG_EXTERNAL);
  *external = *ip;
  return 0; 
}


int slog_enable_internal(short id){
  volatile unsigned short *ip;

  if ((id<0) || (id>=slog_total)) return -1;

  ip = (unsigned short *)(slog_info[id].address + SLOG_INTERNAL);
  *ip = 0xffff;  /* new SLOG firmware, PWP 19.07.99 */
 
  return 0;
}


int slog_disable_internal(short id){
  volatile unsigned short *ip;

  if ((id<0) || (id>=slog_total)) return -1;

  ip = (unsigned short *)(slog_info[id].address + SLOG_INTERNAL);
  *ip = 0;

  return 0;
}


int slog_set_internal(short id, unsigned short bmap){ 
  volatile unsigned short *ip;

  if ((id<0) || (id>=slog_total)) return -1;
  if (bmap > 0xffff) return -3; /* new SLOG firmware, PWP 19.07.99 */

  ip = (unsigned short *)(slog_info[id].address + SLOG_INTERNAL);
  *ip = bmap;

  return 0; 
}


int slog_get_internal(short id, unsigned short *internal){
  volatile unsigned short *ip;

  if ((id<0) || (id>=slog_total)) return -1;

  ip = (unsigned short *)(slog_info[id].address + SLOG_INTERNAL);
  *internal = *ip;
  return 0; 
}


int slog_set_gmra(short id, unsigned short ch){
  volatile unsigned short *ip;

  if ((id<0) || (id>=slog_total)) return -1;
  if (ch>14) return -2;  /* nb: GMR includes upto d4 */

  ip = (unsigned short *)(slog_info[id].address + SLOG_GMRA);
  *ip = ch;
  return 0; 
}


int slog_set_gmrb(short id, unsigned short ch){ 
  volatile unsigned short *ip;
  if ((id<0) || (id>=slog_total)) return -1;
  if (ch>14) return -2;

  ip = (unsigned short *)(slog_info[id].address + SLOG_GMRB);
  *ip = ch;
  return 0; 
}


int slog_exit() {  /* Unmaps all SLOGs: call only once */
  int i;
#ifdef RCC
  VME_ErrorCode_t status;
#else
  int status;
#endif

  for(i=0;i<slog_total;i++) {
#ifdef RAID
    status = RD13_VmeUnmap(slog_info[i].address);
#endif
#ifdef NIVXI
    status = UnMapVXIAddress(slog_info[i].window);
#endif
#ifdef QB_Linux
    munmap(slog_info[i].address,WND_SIZE);
#endif
#ifdef RCC
    status = VME_MasterUnmap(slog_info[i].window);
#endif
  }  /*end of loop*/

#ifdef NIVXI
    CloseVXIlibrary();
#endif
#ifdef RCC
    status = VME_Close();
#endif

  free((void *)slog_info);
  return 0;
}


int slog_vmemap(unsigned long address, short *identifier) {

#if defined(NIVXI)
  short ret;
  UINT32 w,wstart,wend;
#elif defined(RCC)
  VME_ErrorCode_t status;
  VME_MasterMap_t master_map;
  u_int virtual_address;
  int handle;
#else
  int status;
#endif
  int am;
  char *addr;

  am = 0x9;
  *identifier = slog_total++; 

  if(slog_total == 1)
    slog_info = (struct slog_info_structure *) 
                malloc(slog_info_size);
  else
    slog_info = (struct slog_info_structure *) 
                realloc((void *)slog_info, slog_total*slog_info_size);

  slog_info[*identifier].address = address;
  slog_info[*identifier].revision = -1;
#ifdef NIVXI
  slog_info[*identifier].window = 0;
#endif

  /* SLOG decodes 0x20000 (128k) addresses */

#ifdef OS9
  status = _os_permit((void *)address, 0x20000,2,0);
  if(status != 0) {
     printf("VME mapping failed, status = %d\n",status);
     return status;
  }
#endif

#ifdef RAID
  am = 0x9;	/* A32;  Use am= 0x39 for A24 */
  status = RD13_VmeMap(address,am,0x20000,&addr,0);
  if(status != 0) {
     printf("slog_lib: VME mapping failed, status = %d\n",status);
     return status;
  }
  slog_info[*identifier].address = (unsigned long) addr;
#endif

#ifdef NIVXI
  if(slog_total == 1) {
    InitVXIlibrary();		
    MapVXIAddressSize(0x800000);	/* allow 8MB A32 window*/
    /* Workaround for four window limit: map all A32
     * devices in one window.  8MB is enough space for
     * 3 SLOGs and 6 MuSTARDs!!!
     * N.B. slog_vmemap must now be called first!!!
     * PWP 29.06.99 */
  }
  else {MapVXIAddressSize(0x18000);} /* allow 128k window */

  /* VXI-style access parameter, 2 for A24, 3 for A32 */
  addr = (char *)MapVXIAddress(3,address,5000,&w,&ret);
  if(ret != 0) {
     printf("slog_lib: NIVXI VME mapping failed, status = %d for address %x\n",ret,address);
     return (int)ret;
  }
  printf("slog_vmemap: MapVXIAddress ret %d win %d addr %8x for vme %8x\n", 
		ret, w, addr, address);
  ret = GetWindowRange(w, &wstart, &wend);
  printf("slog_vmemap: WindowRange start %lx end %lx\n", wstart, wend);
  slog_info[*identifier].address = (unsigned long) addr;
  slog_info[*identifier].window = w;
#endif

#ifdef QB_linux
  if(slog_total == 1) {
    InitVME();
  }		
  Initqb(&address ,WND_SIZE,VME_A24);
  slog_info[*identifier].address = address;
#endif

#ifdef RCC
  if(slog_total == 1) status = VME_Open();
/*Set up the structure for the map */
  master_map.vmebus_address = address;
  master_map.window_size = 0x20000;  /*128kB window*/
  master_map.address_modifier = VME_A32;
  master_map.options = 0;
  status = VME_MasterMap(&master_map, &handle);
  if(status != VME_SUCCESS) {
    VME_ErrorPrint(status);
    return (int)status;
  }
  status = VME_MasterMapVirtualAddress(handle, &virtual_address);
  printf("slog_vmemap: VME address %x mapped to %x, handle %d\n",
    address, virtual_address,handle);
  slog_info[*identifier].address = (unsigned long) virtual_address;
  slog_info[*identifier].window = handle;
#endif

  return 0;
}


int slog_all_abc( short id, 
  float vthr, float vcal, unsigned short delay, float febias, float feshaper) {

  int i,j;

  unsigned short pre1 = 0x0751;
  unsigned short pre2 = 0xcfc0; /* all chips 16bit value */
  unsigned short cmd[16];
  unsigned short cb[256];

  unsigned short bvthr;
  unsigned short bvcal;
  unsigned short bdelay;
  unsigned short bfebias;
  unsigned short bfeshaper;

  bvthr = (unsigned short) (vthr / 2.5);
  bvcal = (unsigned short) (vcal / 0.625);
  bdelay = delay;
  bfebias = (unsigned short) (febias / 9.2);
  bfeshaper = (unsigned short) (feshaper / 1.2); 

  cmd[0] = pre1;
  cmd[1] = pre2 | 0x10;  /* Strobe delay register */
  cmd[2] = bdelay&0x3f;
  cmd[3] = 0;

  cmd[4] = pre1;
  cmd[5] = pre2 | 0x18;  /* Threshold register */
  cmd[6] = ((bvthr&0xff)<<8) + (bvcal&0xff);
  cmd[7] = 0;

  cmd[8] = pre1;
  cmd[9] = pre2 | 0x38;  /* Bias DAC register */
  cmd[10] = ((bfebias&0x1f)<<8) + (bfeshaper&0x3f);
  cmd[11] = 0;

  cmd[12] = 0x0570;
  cmd[13] = 0xcfc0;  /* Re-enable data-taking */
  cmd[14] = 0;
  cmd[15] = 0;

  for (i=0; i<16; i++) {
    for (j=0; j<16; j++) {
      if ((cmd[i]>>(15-j)) & 0x1) cb[(i*16)+j] = 0x3fff;
      else cb[(i*16)+j] = 0x0000;
      printf("i %d  j %d  index %d  cb %x\n",
                i, j, (i*16)+j, cb[(i*16)+j]);
    }
  }

  slog_load(id, cb, 256);
  return 0;
}

/* new stuff for SLOGB, PWP 03.08.99 */

int slog_test_revision(short id){ 
  unsigned short bmap1, bmap2;
  volatile unsigned short *ip;
  unsigned short test = 0x1b00;

  /* cannot test clock duty bits in power up condition, since 
   * in the default state the switch setting is read back.  I have
   * deliberately not changed any of the enable bits CC3:0.  I think
   * we should avoid use of this test other than at startup to avoid
   * possible troubles due to clock glitches. PWP 03.08.99
   */

  if ((id<0) || (id>=slog_total)) return -1;

  ip = (unsigned short *)(slog_info[id].address + SLOG_CLK_CNTL);
  bmap1 = *ip;
  *ip = test;
  bmap2 = *ip;

  if ((bmap2&0xff00)==test){
    *ip = bmap1;                  /* restore original condition */
    slog_info[id].revision = 1;
  }

  return 0; 
}


int slog_set_clk_cntl(short id, unsigned short bmap){ 
  volatile unsigned short *ip;

  if ((id<0) || (id>=slog_total)) return -1;
  if (slog_info[id].revision<=0)  return -4;  /* not supported */

  ip = (unsigned short *)(slog_info[id].address + SLOG_CLK_CNTL);
  *ip = (bmap&0x1fff);

  return 0; 
}


int slog_get_clk_cntl(short id, unsigned short *clk_cntl){ 
  volatile unsigned short *ip;

  if ((id<0) || (id>=slog_total)) return -1;
  if (slog_info[id].revision<=0)  return -4;  /* not supported */

  ip = (unsigned short *)(slog_info[id].address + SLOG_CLK_CNTL);
  *clk_cntl = *ip&0x1fff;

  return 0; 
}



int slog_set_clock_duty(short id, unsigned short duty){ 
  char y[2];
  volatile unsigned short *ip;
  unsigned short bmap;

  if ((id<0) || (id>=slog_total)) return -1;

  if (duty > 15) duty = 15;

  if (slog_info[id].revision<=0){
#ifdef RAID
    printf("USER INTERVENTION REQUIRED !!!!\007\007\007\n");
#else
    printf("USER INTERVENTION REQUIRED !!!!\a\a\a\n");
#endif
    printf(" Set slog clock duty for slog # %d to %d\n", id, duty);
    printf("\n");
    printf(" ...rotary switch located on lower component side of Slog VME board\n");
    printf("\n");
    printf("HIT ANY KEY + RETURN WHEN READY TO CONTINUE\n");
    scanf("%s",y);

    return 0;
  }

  /* read clock control register */

  ip = (unsigned short *)(slog_info[id].address + SLOG_CLK_CNTL);
  bmap = *ip;

  /* clock duty is 4 bit register, CC7:4, 400pS steps.   Also set 
   * CC0 = 1 to use register value instead of default (switch) 
   */

  bmap = (bmap & 0x1f0e) | (duty<<4) | 0x0001;
  *ip = bmap;

  return 0; 
}


int slog_get_clock_duty(short id, unsigned short *duty){ 
  unsigned short bmap;
  volatile unsigned short *ip;

  *duty = 0;

  if ((id<0) || (id>=slog_total)) return -1;
  if (slog_info[id].revision<=0) return -4;  /* not supported */

  /* clock_duty is 4 bit register, CC7:4, 400pS steps */  

  ip = (unsigned short *)(slog_info[id].address + SLOG_CLK_CNTL);
  bmap = *ip;

  *duty = ((bmap & 0x00f0)>>4);

  return 0; 
}


int slog_set_clock_phase(short id, unsigned short phase){ 
  unsigned short bmap;
  volatile unsigned short *ip;

  if ((id<0) || (id>=slog_total)) return -1;

  if (phase > 31) phase = 31;

  if (slog_info[id].revision<=0) return -4;  /* not supported */

  /* read clock control register */

  ip = (unsigned short *)(slog_info[id].address + SLOG_CLK_CNTL);
  bmap = *ip;

  /* clock_phase is 5 bit register, CC12:8, 1nS steps */

  bmap = (bmap & 0x00ff) | (phase<<8) | 0x0002; /* PWP 22.08.00 vary LVDS clk/com phase */
  *ip = bmap;

  return 0; 
}


int slog_get_clock_phase(short id, unsigned short *phase){ 
  unsigned short bmap;
  volatile unsigned short *ip;

  *phase = 0;

  if ((id<0) || (id>=slog_total)) return -1;
  if (slog_info[id].revision<=0) return -4;  /* not supported */

  /* clock_phase is 5 bit register, CC12:8, 1nS steps */

  ip = (unsigned short *)(slog_info[id].address + SLOG_CLK_CNTL);
  bmap = *ip;

  *phase = ((bmap & 0x1f00)>>8);

  return 0; 
}

int slog_test_mem(short id){
  volatile unsigned short *ip;
  unsigned int n;

  int n_errs=0;

  if ((id < 0) || (id >= slog_total)) return -1;

  slog_stop(id);
  
  ip = (unsigned short *)(slog_info[id].address + SLOG_BASE);

  for (n=0; n<SLOG_SIZE; n++) {
    *ip++ = (unsigned short) n&0xffff;
  }

  ip = (unsigned short *)(slog_info[id].address + SLOG_BASE);

  for (n=0; n<SLOG_SIZE; n++) {
    if( ((unsigned short)n&0xffff) != *ip++){
      n_errs++;
    }
  }

  return n_errs;
}
