#define MAX_SCAN_DIMS 12

Int_t stan_fitraw(char* in_nt, char* out_nt, int run, int sno, int mid, int ascii, int postscript){

  // Routine to fit individual st scan histograms
  //
  // Cuts have been implemented as follows:
  // o Channels with <1% occupancy in any bin are marked as "DEAD"
  // o Channels with >98% occupancy in every bin are marked "STUCK"
  // o Channels which do not reach 100% efficiency are marked "UNDER"
  // o Channels which reach > 100% efficiency are marked "OVER"
  //   (Such a condition would imply decoder errors, inviting the
  //    user to check the digital integrity of their system)
  // o include only fits for which the mean value (e.g. vt50)
  //   falls within the scanned range
  // o include only fits for which the fitted sigma
  //   is positive and less than twice the mean sigma

  // Version 0.31 P.W.Phillips, 10.10.99
  // Version 0.32 BJG, 30.04.02   Fix some memory leaks

  char  f_name[64], h_name[32], h_title[64], scan_name[32];
  char  m_title[32],s_title[32],out_fname[128], s[256];
  char  stub[16], nt_members[64],nt_title[64]; 
  Int_t   i,j,k,x,y;
  Float_t err2, z;

  // Locality options

  int RAL = 1;                // path variations

  // Load other files needed by this one

  gROOT->LoadMacro("stan_util.cpp");
  gROOT->LoadMacro("stan_projection.cpp");

  // Global style variations

  stan_styles();

  // link to ntuple and form projection

  TNtuple * STNT = (TNtuple*) gDirectory->Get(in_nt);
  if(STNT==0){
    printf("stan_fitraw called for invalid NTuple name - exiting\n");
    return -1;
  }
  printf("st_fitraw called for NTuple %s sno %d mid %d\n",in_nt,sno,mid);


  Float_t l_nevt;
  Float_t l_var_settings[12][MAX_SCAN_DIMS];
  Int_t   l_var_type[MAX_SCAN_DIMS];

  Int_t n_dims = stan_projection(in_nt,sno,mid,l_var_type, l_var_settings, &l_nevt);
  if(n_dims<=0) return -2;
  TH2F * h_scan = (TH2F*) gDirectory->Get("h_scan");

// Choose appropriate fit function
// and form scan title

  switch (l_var_type[0]){
    case 1: // ST_VTHR
      sprintf(scan_name,"THRESHOLD (mV)");
      TF1 *f1 = new TF1 ("f1",erfc,  0, 1000, 3);
      f1->SetParNames("Constant","Mean","Sigma");
      break;
    case 2: // ST_VCAL
      sprintf(scan_name,"VCAL (mV)");
      TF1 *f1 = new TF1 ("f1", erf,  0, 1000, 3);
      f1->SetParNames("Constant","Mean","Sigma");
      break;
    case 3: // ST_STROBE_DELAY
      sprintf(scan_name,"Strobe Delay");
//      TF1 *f1 = new TF1 ("f1", tophat,  0, 1000, 3);
      TF1 *f1 = new TF1 ("f1", gaus,  0, 1000, 3);
      f1->SetParNames("Constant","Mean","Sigma");
      break;
    case 6: // ST_TRIM
      sprintf(scan_name,"TRIMDACS");
      TF1 *f1 = new TF1 ("f1",erfc,  0, 1000, 3);
      f1->SetParNames("Constant","Mean","Sigma");
      break;
    case 23: // ST_CLK_DUTY
      sprintf(scan_name,"Clock Duty");
      TF1 *f1 = new TF1 ("f1",gaus,  0, 1000, 3); // TODO better function
      f1->SetParNames("Constant","Mean","Sigma");
      break;
    case 25: // ST_TRIG_DELAY
      sprintf(scan_name,"Trig Delay");
      TF1 *f1 = new TF1 ("f1",gaus,  0, 1000, 3); // TODO better function
      f1->SetParNames("Constant","Mean","Sigma");
      break;
    case 101: // ST_VT50
      TF1 *f1 = new TF1 ("f1",pol2,  0, 1000, 3);
      f1->SetParNames("p0","p1","p2");
      break;
    case 102: // ST_VC50
      TF1 *f1 = new TF1 ("f1",pol2,  0, 1000, 3);
      f1->SetParNames("p0","p1","p2");
      break;
    default: // use gaussian for now 
      sprintf(scan_name,"Unknown");
      TF1 *f1 = new TF1 ("f1", gaus,  0, 1000, 3);
      f1->SetParNames("Constant","Mean","Sigma");
      break;
  }
  printf("scan_name is %s\n",scan_name);

  // titles for graphs - might as well do this now.

  sprintf(m_title,"Fitted Mean(%s)",scan_name);
  sprintf(s_title,"Fitted Sigma(%s)",scan_name);

  TPaveText *t1 = new TPaveText(0.0,0.0,1.0,1.0);
  t1->SetFillColor(10);
  t1->SetTextAlign(13);
  t1->SetTextSize(0.3);
  t1->SetTextColor(4);
  t1->AddText("ATLAS SCT Module Test");
  sprintf(s,"Scan %d   Module %d  Scan type: %s", sno, mid, scan_name);
  t1->AddText(s);

  // If we are processing raw data (and not fit results)
  // Calculate errors appropriate to binomial statistics 

  if(l_var_type[0]<100){
    for(x=0;x<h_scan->GetNbinsX();x++){
      for(y=0;y<h_scan->GetNbinsY();y++){
        z = h_scan->GetCellContent(x,y);
        err2 = (z+1) * (1+l_nevt-z) / l_nevt;
        h_scan->SetCellError(x,y,sqrt(err2));
      }
    }
  }

  // Before we draw the main canvas, open the postscript file
  // if the relevant option has been selected 

  if(postscript){
    if(RAL) sprintf(out_fname,"H:\\sctvar\\data\\stfit%d_%d_%d.ps",run,sno,mid);
    else    sprintf(out_fname,"C:\\sctvar\\data\\stfit%d_%d_%d.ps",run,sno,mid);

    TPostScript *ps = new TPostScript(out_fname,112);
    ps->NewPage();
  }

  // Open and draw the first canvas

  TCanvas* cfit=new TCanvas ("cfit","STAN_fitraw",1100,750);
  pad_t = new TPad("pad_t","Text",         0.01,0.91,0.99,0.99,33);
  pad_m = new TPad("pad_m","Fitted Means", 0.01,0.61,0.60,0.90,33);
  pad_s = new TPad("pad_s","Fitted Sigmas",0.01,0.31,0.60,0.60,33);
  pad_c = new TPad("pad_c","Fitted Chisq", 0.01,0.01,0.60,0.30,33);
  pad_p = new TPad("pad_p","Projections",  0.61,0.01,0.99,0.90,33);

  pad_t->Draw();
  pad_m->Draw();
  pad_s->Draw();
  pad_c->Draw();
  pad_p->Draw();
  pad_p->Divide(3,4);

  pad_m->cd();
  h_scan->Draw("COLZ");
  h_scan->SetXTitle("Channel Number");
  h_scan->SetYTitle(scan_name);

  cfit->Update();

  // Chip projections

  TH1 * h_dp[12];
  TH1 * h_ptr;
  Float_t chip_p0[12];
  Float_t chip_p1[12];
  Float_t chip_p2[12];

  for(i=0;i<12;i++){
    pad_p->cd(i+1);
    sprintf(h_name,"h_dp%d",i);
    sprintf(h_title,"Scan Projection for chip %d",i);

    // delete projection if it already exists
    if(h_ptr = (TH1F*) gDirectory->GetList()->FindObject(h_name)){
      delete h_ptr;
    }
    h_scan->ProjectionY(h_name,(i*128),((i+1)*128)-1);
    h_dp[i] = (TH1F*) gDirectory->GetList()->FindObject(h_name);

    h_dp[i]->SetTitle(h_title);
    h_dp[i]->SetXTitle(scan_name);
    h_dp[i]->SetLineColor(kRed);
    h_dp[i]->SetMarkerColor(kBlack);

    // Calculate errors appropriate to binomial statistics 
    Float_t p0_guess = h_dp[i]->GetMaximum();
    if (p0_guess==0) p0_guess = 1;          // protection against divide by zero
    for(x=1;x<=h_dp[i]->GetNbinsX();x++){
      z = h_dp[i]->GetBinContent(x);
      err2 = (z+1) * (1+p0_guess-z) / p0_guess;
      h_dp[i]->SetBinError(x,sqrt(err2));
    }

    // Fit the projections - unless the histograms are empty!!!

    if(p0_guess>1.0){

      if(l_var_type[0]==1){  // Threshold Scan
        Float_t p1_guess = find_level2(h_dp[i],0.50,1);          // last bin > 0%
        Float_t p2_guess = find_level2(h_dp[i],0.82,1)-p1_guess;  
        Float_t range_lo = find_bin2(h_dp[i],0.99,0);            // first bin > 99%
        Float_t range_hi = h_dp[i]->GetBinCenter(h_dp[i]->GetNbinsX());
        if(p2_guess<0) p2_guess = p2_guess*-1.0;
      } else if(l_var_type[0]==2){ // VCAL Scan
        Float_t p1_guess = find_level2(h_dp[i],0.50,0);          // first bin < 50%
        Float_t p2_guess = find_level2(h_dp[i],0.82,0)-p1_guess;
        Float_t range_lo = h_dp[i]->GetBinCenter(1);
        Float_t range_hi = find_bin2(h_dp[i],0.99,1);            // last bin > 99%
        if(p2_guess<0) p2_guess = p2_guess*-1.0;
      } else {
        Float_t p1_guess = h_dp[i]->GetMean(); 
        Float_t p2_guess = h_dp[i]->GetRMS(); 
        Float_t range_lo = h_dp[i]->GetBinCenter(1);
        Float_t range_hi = h_dp[i]->GetBinCenter(h_dp[i]->GetNbinsX());
      }

      f1->SetParameters(p0_guess,p1_guess,p2_guess);
      f1->SetParLimits(0,p0_guess,p0_guess);
      f1->SetRange(range_lo, range_hi);
      h_dp[i]->Fit("f1","RQ"); // quiet mode

      chip_p0[i] = f1->GetParameter(0); 
      chip_p1[i] = f1->GetParameter(1); 
      chip_p2[i] = f1->GetParameter(2);
    } else {
      chip_p0[i] = -1;
      chip_p1[i] = -1;
      chip_p2[i] = -1;
    }
    printf("Chip %d guesses p0=%3.2f p1=%3.2f p2=%3.2f rlo=%3.2f rhi=%3.2f\n",i,p0_guess,p1_guess,p2_guess, range_lo, range_hi); 
    printf("Chip %d fit     p0=%3.2f p1=%3.2f p2=%3.2f\n",i,chip_p0[i],chip_p1[i],chip_p2[i]); 
    h_dp[i]->Draw("e1p");
    cfit->Update();
  }

  int n_bins = h_dp[0]->GetNbinsX();
  float scan_lo = h_dp[0]->GetBinCenter(1);
  float scan_hi = h_dp[0]->GetBinCenter(n_bins);
  float half_range = (scan_hi - scan_lo)/4;

  Int_t i_code[1536];
  Float_t i_par0[1536];
  Float_t i_par1[1536];
  Float_t i_par2[1536];
  Float_t i_chisq[1536];

  // Fit channels one at a time, using "tweaked" start conditions where possible
  // This gives much more reliable results than "fitslices"

  p0_guess = l_nevt;
  if (p0_guess==0) p0_guess = 1;          // protection against divide by zero
  for(i=0;i<1536;i++){

    // delete projection if it already exists
    if(h_ptr = (TH1F*) gDirectory->GetList()->FindObject("h_tp")){
      delete h_ptr;
    }

    h_scan->ProjectionY("h_tp",i,i,"E");
    h_tp = (TH1F*) gDirectory->Get("h_tp");

    // error conditions: flag CH_UNDER and CH_OVER but still fit
    //                   do not attempt to fit CH_DEAD or CH_STUCK 

    Int_t code=0;

    if(h_tp->GetMaximum()<1.0){
      printf("channel %d is DEAD\n",i);
      code = CH_DEAD;
    } else if(h_tp->GetMinimum()>l_nevt*0.98){
      printf("channel %d is STUCK\n",i);
      code = CH_STUCK;
    } else if(h_tp->GetMaximum()<l_nevt){
      // printf("channel %d is UNDER THE TOP\n",i);
      code = CH_UNDER;
    } else if(h_tp->GetMaximum()>l_nevt){
      // printf("channel %d is OVER THE TOP\n",i);
      code = CH_OVER;
    }

    if(!(code&BAD_TEST)){ // attempt to fit CH_UNDER
                          // but skip CH_OVER, CH_DEAD and CH_STUCK

      Float_t p0_guess = l_nevt;
      if(l_var_type[0]==1){  // Threshold Scan
        Float_t p1_guess = find_level2(h_tp,0.50,1);          // last bin > 0%
        Float_t p2_guess = find_level2(h_tp,0.82,1)-p1_guess;  
        Float_t range_lo = find_bin2(h_tp,0.99,0);            // first bin > 99%
        Float_t range_hi = h_tp->GetBinCenter(h_tp->GetNbinsX());
        if(p2_guess<0) p2_guess = p2_guess*-1.0;
      } else if(l_var_type[0]==2){ // VCAL Scan
        Float_t p1_guess = find_level2(h_tp,0.50,0);          // first bin < 50%
        Float_t p2_guess = find_level2(h_tp,0.82,0)-p1_guess;
        Float_t range_lo = h_tp->GetBinCenter(1);
        Float_t range_hi = find_bin2(h_tp,0.99,1);            // last bin > 99%
        if(p2_guess<0) p2_guess = p2_guess*-1.0;
      } else {
        Float_t p1_guess = chip_p1[i/128];
        Float_t p2_guess = chip_p2[i/128];
        Float_t range_lo = h_tp->GetBinCenter(1);
        Float_t range_hi = h_tp->GetBinCenter(h_tp->GetNbinsX());
      }

      f1->SetParameters(p0_guess,p1_guess,p2_guess);
      f1->SetParLimits(0,p0_guess,p0_guess);
      f1->SetRange(range_lo, range_hi);
      h_tp->Fit("f1","0RQ"); // quiet mode

      // Validity checking on fit data:
      // mean value must be within scanned range
      // sigma must be greater than zero

      Float_t mean = f1->GetParameter(1);
      if(mean<scan_lo){
        // printf("fit for channel %d is UNDER THE BOTTOM\n",i);
        code += FIT_UNDER;
      } else if(mean>scan_hi){
        // printf("fit for channel %d is OVER THE TOP\n",i);
        code += FIT_OVER;
      }
      if(f1->GetParameter(2)<0) code+=SIG_UNDER;
    }

    if(!(code&BAD_TEST)){
      i_par0[i]=f1->GetParameter(0); 
      i_par1[i]=f1->GetParameter(1); 
      i_par2[i]=f1->GetParameter(2); 
      i_chisq[i]=f1->GetChisquare(); 
    } else {
      i_par0[i]=0; 
      i_par1[i]=0; 
      i_par2[i]=0; 
      i_chisq[i]=0; 
    }
    i_code[i]=code;
  }

  // projections of all sigma values - to facilitate range setting...

  printf("Projecting Sigma values for all channels....\n");
  TH1F* h_spall  =  new TH1F("h_spall", "Sigma, all channels", 50,0,100);
  for(i=0;i<1536;i++){
    if(!(i_code[i]&BAD_TEST)) h_spall->Fill(i_par2[i]);
  } 
  h_spall->Fit("gaus","0");
  f2=h_spall->GetFunction("gaus");

  Float_t sigma_m = f2->GetParameter(1);
  Float_t sigma_s = f2->GetParameter(2);
  if (sigma_m<0){  // fit failed - fall back to "sensible" defaults
    sigma_m = (scan_hi-scan_lo)/2;
    sigma_s = sigma_m/2;
    printf("Fit failed - reverting to cut at %3.2f\n", 2*sigma_m);
  } else {
    printf("Fitted mean of sigma values is %3.2f, therefore upper cut is %3.2f\n",
      sigma_m, 2*sigma_m);
  }
  Float_t sigma_lo = sigma_m - (6*sigma_s);
  Float_t sigma_hi = sigma_m + (6*sigma_s);

  // Create histograms to hold the (soon to be) fitted scan data.  

  TH1S * h_code  = new TH1S("h_code",  "Fit Code",     1536, 0.5, 1536.5);
  TH1F * h_par0  = new TH1F("h_par0",  "Fitted p0",    1536, 0.5, 1536.5);
  TH1F * h_par1  = new TH1F("h_par1",  "Fitted p1",    1536, 0.5, 1536.5);
  TH1F * h_par2  = new TH1F("h_par2",  "Fitted p2",    1536, 0.5, 1536.5);
  TH1F * h_chisq = new TH1F("h_chisq", "Fitted Chisq", 1536, 0.5, 1536.5);

  // Second pass at sigma data to mask excessive sigma values
  // also fill the histograms during this loop

  for(i=0;i<1536;i++){
    if(i_par2[i]>(2*sigma_m)){
      i_code[i]=i_code[i]+SIG_OVER;
    }
    if(!(i_code[i]&BAD_TEST)){
      h_par0->Fill(i,i_par0[i]); 
      h_par1->Fill(i,i_par1[i]); 
      h_par2->Fill(i,i_par2[i]); 
      h_chisq->Fill(i,i_chisq[i]); 
    }
    h_code->Fill(i,i_code[i]); 
  } 



// Draw the fit parameters using the pads of canvas 1

  pad_m->cd();
  h_par1->Draw("samep");
  pad_s->cd();
  h_par2->SetXTitle("Channel Number");
  h_par2->SetYTitle(s_title);
  h_par2->SetMaximum(2*sigma_m);
  h_par2->SetMinimum(0);
  h_par2->SetLineColor(kRed);
  h_par2->SetMarkerColor(kBlack);
  h_par2->Draw("ep");
  pad_c->cd();
  pad_c->SetLogy();
  h_chisq->SetXTitle("Channel Number");
  h_chisq->SetYTitle("Chisq");
  h_chisq->SetLineColor(kRed);
  h_chisq->SetMarkerColor(kBlack);
  h_chisq->Draw("ep");

  pad_t->cd();
  t1->Draw();
  cfit->Update();

  if(postscript) {
    ps->Close();
    delete ps;
  }


  // Tally total numbers of good and bad channels  

  short chip_bad_ch[13][10];
  short chip_good_ch[13];

  for(i=0;i<13;i++){
     chip_good_ch[i]=0;     // clear good channel count
    for(k=0;k<10;k++) {     // clear bad channel count
      chip_bad_ch[i][k]=0;
    } 
  }

  for(i=0;i<13;i++){
    for(j=0;j<128;j++){

      code=(Int_t)h_code->GetBinContent(j+(128*i));
      if(code){
        if (code&CH_DEAD)  chip_bad_ch[i][0]++;
        if (code&CH_STUCK) chip_bad_ch[i][1]++;
        if (code&CH_UNDER) chip_bad_ch[i][2]++;
        if (code&CH_OVER)  chip_bad_ch[i][3]++;
        if (code&FIT_UNDER) chip_bad_ch[i][4]++;
        if (code&FIT_OVER)  chip_bad_ch[i][5]++;
        if (code&SIG_UNDER) chip_bad_ch[i][6]++;
        if (code&SIG_OVER)  chip_bad_ch[i][7]++;

        if (code&DEAD_TEST) chip_bad_ch[i][8]++; // DEAD  channels
        if (code&DODGY_TEST) chip_bad_ch[i][9]++; // DODGY channels
      }
      if(!(code&BAD_TEST)) chip_good_ch[i]++;
    }
  }

  for(i=0;i<13;i++){
    chip_good_ch[12] += chip_good_ch[i];
    for(k=0;k<10;k++) chip_bad_ch[12][k] += chip_bad_ch[i][k];
  }

  // Print bad channel information to the screen

  printf("\nBad Channel Summary\n");
  printf("Chip\tDead\tStuck\tN_under\tN_over\tF_under\tF_over\tS_under\tS_over\n");
  for(i=0;i<12;i++){
    printf("%d\t%d\t%d\t%d\t%d\t%d\t%d\t%d\t%d\n", i,
      chip_bad_ch[i][0], chip_bad_ch[i][1],
      chip_bad_ch[i][2], chip_bad_ch[i][3],
      chip_bad_ch[i][4], chip_bad_ch[i][5],
      chip_bad_ch[i][6], chip_bad_ch[i][7]);
  }
  printf("total\t%d\t%d\t%d\t%d\t%d\t%d\t%d\t%d\n",
    chip_bad_ch[12][0], chip_bad_ch[12][1],
    chip_bad_ch[12][2], chip_bad_ch[12][3],
    chip_bad_ch[12][4], chip_bad_ch[12][5],
    chip_bad_ch[12][6], chip_bad_ch[12][7]);

  printf("\nChannel Totals\n");
  printf("Chip\tDead\tDodgy\tUseful\n");
  for(i=0;i<12;i++){
    printf("%d\t%d\t%d\t%d\n", i, chip_bad_ch[i][8], 
      chip_bad_ch[i][9], chip_good_ch[i]);
  }
  printf("total\t%d\t%d\t%d\n", chip_bad_ch[12][8], 
    chip_bad_ch[12][9], chip_good_ch[12]);

// Now write fit data to the NTuple STFITNTn

  TNtuple * STFITNT = (TNtuple*) gDirectory->Get(out_nt);
  if(STFITNT==0){
    printf("Creating new ROOT ntuple %s:\n",out_nt);
    sprintf(nt_title,"Fitted Data for run %d start scan %d",run,sno);
    printf("nt_title:%s\n",nt_title);
    sprintf(nt_members,"sno:mid");
    for(i=1; i<n_dims; i++){
      match_var(l_var_type[i],stub);
      sprintf(nt_members,"%s:%s",nt_members, stub);
    }
  switch (l_var_type[0]){
    case 1: // ST_VTHR
      sprintf(nt_members,"%s:ch:p0:vt50:sigma:chisq:code",nt_members);
      break;
    case 2: // ST_VCAL
      sprintf(nt_members,"%s:ch:p0:vc50:sigma:chisq:code",nt_members);
      break;
    case 3: // ST_STROBE_DELAY
      sprintf(nt_members,"%s:ch:p0:strb_m:sigma:chisq:code",nt_members);
      break;
    case 4: // ST_PREAMP
      sprintf(nt_members,"%s:ch:p0:preamp_m:sigma:chisq:code",nt_members);
      break;
    case 5: // ST_SHAPER
      sprintf(nt_members,"%s:ch:p0:shaper_m:sigma:chisq:code",nt_members);
      break;
    case 6: // ST_TRIM
      sprintf(nt_members,"%s:ch:p0:tr50:sigma:chisq:code",nt_members);
      break;
    default:
      sprintf(nt_members,"%s:ch:p0:p1:p2:chisq:code",nt_members);
      break;
  }
    
    printf("nt_members:%s\n",nt_members);
    TNtuple *STFITNT = new TNtuple(out_nt,nt_title,nt_members,32000);
  }
  printf("Writing fit data to NTuple %s\n",out_nt,sno,mid);

  for(i=0;i<1536;i++){
    Int_t chip = i/128;
    Float_t par0  = h_par0->GetBinContent(i);
    Float_t par1  = h_par1->GetBinContent(i);
    Float_t par2  = h_par2->GetBinContent(i);
    Float_t chisq = h_chisq->GetBinContent(i);
    Int_t code  = h_code->GetBinContent(i);

    switch(n_dims){
      case 4:
        STFITNT->Fill((Float_t)sno, (Float_t)mid, l_var_settings[chip][3], 
                       l_var_settings[chip][2], l_var_settings[chip][1], 
                       i, par0, par1, par2, chisq, code); 
        break;
      case 3:
        STFITNT->Fill((Float_t)sno, (Float_t)mid, l_var_settings[chip][2], 
                       l_var_settings[chip][1], i, par0, par1, par2, chisq, code); 
        break;
      case 2:
        STFITNT->Fill((Float_t)sno, (Float_t)mid, l_var_settings[chip][1], 
                       i, par0, par1, par2, chisq, code); 
        break;
      case 1:
        STFITNT->Fill((Float_t)sno, (Float_t)mid, i, par0, par1, par2, chisq, code); 
        break;
    }
  }

  if(ascii){ // write ascii output

    if(RAL) sprintf(f_name,"H:\\sctvar\\data\\stfit%d_%d_%d.txt",run,sno,mid);
    else    sprintf(f_name,"C:\\sctvar\\data\\stfit%d_%d_%d.txt",run,sno,mid);

    FILE* fitfile = fopen(f_name,"w");

    if(fitfile==0){
      printf("Could not open file for ascii output - returning");
      return -3;
    }

    if(n_dims>1){ // higher order scan - record conditions
    fprintf(fitfile,"#\n#Setup information:\n#\n");
    fprintf(fitfile,"Chip");
      for(i=1; i<n_dims; i++){
        match_var(l_var_type[i],stub);
        fprintf(fitfile,"\t%s", stub);
      }
      fprintf(fitfile,"\n");
      for(i=0;i<12;i++){
        fprintf(fitfile,"%d",i);
        for(j=1;j<n_dims;j++){
          fprintf(fitfile,"\t%3.2f",l_var_settings[i][j]);
        }
        fprintf(fitfile,"\n");
      }
    }

    fprintf(fitfile,"#\n#Summary information:\n#\n");
    fprintf(fitfile,"Chip\tp0\tp1\tp2\n");
    for(i=0;i<12;i++){
      fprintf(fitfile,"%d\t%3.2f\t%3.2f\t%3.2f\n",i,
        chip_p0[i],chip_p1[i],chip_p2[i]); 
    }
    fprintf(fitfile,"#\n#Bad Channel Information:\n#\n");
    fprintf(sumfile,"Chip\tDead\tStuck\tN_under\tN_over\tF_under\tF_over\tS_under\tS_over\n");
    for(i=0;i<12;i++){
      fprintf(sumfile,"%d\t%d\t%d\t%d\t%d\t%d\t%d\t%d\t%d\n", i,
        chip_bad_ch[i][0], chip_bad_ch[i][1],
        chip_bad_ch[i][2], chip_bad_ch[i][3],
        chip_bad_ch[i][4], chip_bad_ch[i][5],
        chip_bad_ch[i][6], chip_bad_ch[i][7]);
    }
    fprintf(sumfile,"total\t%d\t%d\t%d\t%d\t%d\t%d\t%d\t%d\n",
      chip_bad_ch[12][0], chip_bad_ch[12][1],
      chip_bad_ch[12][2], chip_bad_ch[12][3],
      chip_bad_ch[12][4], chip_bad_ch[12][5],
      chip_bad_ch[12][6], chip_bad_ch[12][7]);

    fprintf(sumfile,"#\n#Bad Channel Totals:\n");
    fprintf(sumfile,"Chip\tDead\tDodgy\tUseful\n");
    for(i=0;i<12;i++){
      fprintf(sumfile,"%d\t%d\t%d\t%d\n", i, chip_bad_ch[i][8], 
        chip_bad_ch[i][9],chip_good_ch[i]);
    }
    fprintf(sumfile,"total\t%d\t%d\t%d\n", chip_bad_ch[12][8], 
        chip_bad_ch[12][9],chip_good_ch[12]);

    fprintf(fitfile,"#\n#Fit information:\n#\n");
    fprintf(fitfile,"Ch\tpar0\tpar1\tpar2\tchisq\tcode\n");
    for(i=0;i<1536;i++){
      Float_t par0  = h_par0->GetBinContent(i);
      Float_t par1  = h_par1->GetBinContent(i);
      Float_t par2  = h_par2->GetBinContent(i);
      Float_t chisq = h_chisq->GetBinContent(i);
      Float_t code  = h_code->GetBinContent(i);
      fprintf(fitfile,"%3d\t %3.2f\t %3.2f\t %3.2f\t %3.2f\t %d\n",i , par0, par1, par2, chisq, code); 
    }
    fclose(fitfile);
  }
  
  return 0;
} 

Int_t stan_fitraw(char* in_nt, char* out_nt, Int_t run, Int_t ascii, Int_t postscript){

  TNtuple * STNT = (TNtuple*) gDirectory->Get(in_nt);
  if(STNT==0){
    printf("st_fit_nt called for invalid NTuple name - exiting\n");
    return -1;
  }
  printf("st_fit_nt called for NTuple %s\n",in_nt);

  Int_t min_mid = STNT->GetMinimum("mid");
  Int_t max_mid = STNT->GetMaximum("mid");
  Int_t max_sno = STNT->GetMaximum("sno");
  Int_t min_sno = STNT->GetMinimum("sno");

  printf("mid min %d max %d\n",min_mid,max_mid); 
  printf("sno min %d max %d\n",min_sno,max_sno); 
  for(Int_t mid=min_mid; mid<=max_mid; mid++){
    for(Int_t sno=min_sno; sno<=max_sno; sno++){
     printf("Fitting %s mid %d sno %d\n",in_nt,mid,sno);
      stan_fitraw(in_nt,out_nt,run,sno,mid,ascii,postscript);
    }
  }
  return sno;
}


Int_t stan_fitraw(Int_t run, Int_t ascii, Int_t postscript){
  char fname[64], in_nt[32], out_nt[32];

  // First, try to open ROOT data file for this run

  sprintf(fname,"H:\\sctvar\\data\\strun%d.root",run);
  TFile* out_file = new TFile(fname,"UPDATE");

  // If this fails, open and convert the ASCII Ntuple

  if(out_file->IsOpen()==kFALSE){
    printf("Failed to open root data file for run %d\n",run);
    printf("Attempting to process ASCII ntuple to produce root data file...\n");
    gROOT->LoadMacro("stan_ascii2root.cpp");
    stan_ascii2root(run);
  }

  Int_t sno = 1;

  while (sno){
    sprintf(out_nt,"STFIT%d",sno);
    TNtuple * STFIT = (TNtuple*) gDirectory->Get(out_nt);
    if(STFIT){
      printf("NTuple fit data for run %d sno %d already exists - exiting\n",run,sno);
      return -1;
    }

    sprintf(in_nt,"STNT%d",sno);
    TNtuple * STNT = (TNtuple*) gDirectory->Get(in_nt);
    if(!STNT){
      printf("Could not find NTuple object %s - exiting\n",in_nt);
      return -2;
    } else {
      sno = stan_fitraw(in_nt, out_nt, run, ascii, postscript);
      TNtuple * STFIT = (TNtuple*) gDirectory->Get(out_nt);
      STFIT->Write(out_nt);
    }
  }
  printf("stan_fitraw completed\n");
  return 0;
}


Int_t stan_fitraw(Int_t run, Int_t ascii){

  stan_fitraw(run,ascii,0);
}

Int_t stan_fitraw(Int_t run){

  stan_fitraw(run,0,0);
}

