#define KG_HI_MV    50
#define KG_HI_ENC 5000  // was 2000

void stan_kwikgain(Int_t run1, Int_t sno1, Float_t q1, Int_t run2, Int_t sno2, Float_t q2, 
                 Int_t moi, Int_t loi, Int_t ascii, Int_t postscript){

  // Routine to perform quick gain calculations
  //
  // o Uses "stan_fitscan.cpp" to fit scurves 
  // o Gain is calculated as (mean2-mean1) / (q2-q1)
  // o Offset is calculated as mean1 - (q1*gain)
  // o Noise is calculated as (noise1 + noise2)/2*gain
  // o q1 and q2 are user supplied parameters - in fC
  // o Calculations are not performed for any channel which is
  //   found to be "DEAD", "STUCK", "OVER", "FIT_UNDER",
  //   "FIT_OVER", "SIG_UNDER" or "SIG_OVER" in either scan

  // Version 0.2 P.W.Phillips, 11.11.99
  // Version 0.21 P.W.Phillips, 06.01.00
  //   Replace TObject->Copy by use of the "=" operator 
  //    - earlier code not compatible with 2.23/10
  //   Add time/place/module name stamps to plots
  // Version 0.22 P.W.Phillips, 08.03.00
  // Write summary in terms of chips 6-11 if link 1 selected
  // Version 0.3 PWP 04.08.00  - tuned for the system test
  //   show noise in mV at 2fC
  //   single summary file for all modules/links per run
  // Version 0.31 BJG, 30.04.02   Fix some memory leaks

  char  f_name[64], h_title[32];
  char  out_fname[128], s[256];
  int   bin,i,j,k;

  // Locality options now moved to stan_util.cpp

  // Options which change macro output
  int draw_fit_projections = 1;

  // Load other files needed by this one
  gROOT->LoadMacro("stan_fitscan.cpp");
  gROOT->LoadMacro("stan_util.cpp");

  // Global style variations
  stan_styles();
  gStyle->SetOptFit(1011);

  // fit first scan data 
  stan_fitscan(run1,sno1,moi,loi);
  TH1* h_mean1 = new TH1();
  TH1F* h_sigma1 = new TH1F();
  TH1S* h_code1 = new TH1S();
  h_mean1 = h_mean;
  h_sigma1 = h_sigma;
  h_code1 = h_code;

  sprintf(f_name,"ModuleName;%d",moi+1);
  TObjString *time  = (TObjString*)gDirectory->Get("Time");
  TObjString *place = (TObjString*)gDirectory->Get("Place");
  TObjString *mname = (TObjString*)gDirectory->Get(f_name);

  stan_fitscan(run2,sno2,moi,loi);
  TH1F* h_mean2 = new TH1F();
  TH1F* h_sigma2 = new TH1F();
  TH1S* h_code2 = new TH1S();
  h_mean2 = h_mean;
  h_sigma2 = h_sigma;
  h_code2 = h_code;

  // Create histograms to hold the analysed data  

  TH1S* h_codes  = new TH1S("h_codes", "Fit Code",     768,0.5,768.5);
  TH1F* h_gain   = new TH1F("h_gain",  "Gain (mV/fC)", 768,0.5,768.5);
  TH1F* h_offset = new TH1F("h_offset","Offset (mV)",  768,0.5,768.5);
  TH1F* h_mv     = new TH1F("h_mv",    "Noise (mV)",   768,0.5,768.5);
  TH1F* h_noise  = new TH1F("h_noise", "Noise (ENC)",  768,0.5,768.5);

  TH1F* h_gall  =  new TH1F("h_gall", "Gain, all channels",   40,40,80);
  TH1F* h_oall  =  new TH1F("h_oall", "Offset, all channels", 40,0,100);
  TH1F* h_mvall =  new TH1F("h_mvall","Noise (mV), all channels",  40,0,KG_HI_MV);
  TH1F* h_nall  =  new TH1F("h_nall", "Noise (ENC), all channels", 40,0,KG_HI_ENC);


  Float_t q = (q2+q1) /2.0;
  for(i=0;i<6;i++){
    for(j=0;j<128;j++){
      bin = j+(128*i);

      Int_t code=(Int_t)h_code1->GetBinContent(bin);
      if((code&BAD_TEST)==0) code=(Int_t)h_code2->GetBinContent(bin);

      if((code&BAD_TEST)==0){
        Float_t gain   = (h_mean2->GetBinContent(bin)-h_mean1->GetBinContent(bin))/(q2-q1);
        if(gain==0) gain = 0.1;  // protection against divide by zero errors, PWP 04.08.00
        Float_t offset = h_mean1->GetBinContent(bin) - (q1*gain);
        Float_t mv     = h_sigma1->GetBinContent(bin);
        Float_t noise  = (3125 * (h_sigma1->GetBinContent(bin)+h_sigma2->GetBinContent(bin))) / gain;

        h_gain->Fill(bin,gain);
        h_offset->Fill(bin,offset);
        h_mv->Fill(bin,mv);
        h_noise->Fill(bin,noise);

        h_gall->Fill(gain);
        h_oall->Fill(offset);
        h_mvall->Fill(mv);
        h_nall->Fill(noise);
      }
      h_codes->Fill(bin,code);
    }
  }


  // projections - to facilitate range setting...

  printf("Fitting Projections....\n");

  h_gall->Fit("gaus","0");
  f2=h_gall->GetFunction("gaus");
  Float_t gain_m = f2->GetParameter(1);
  Float_t gain_s = f2->GetParameter(2);
  Float_t gain_lo = gain_m - (4*gain_s);
  Float_t gain_hi = gain_m + (4*gain_s);

  h_oall->Fit("gaus","0");
  f3=h_oall->GetFunction("gaus");
  Float_t offset_m = f3->GetParameter(1);
  Float_t offset_s = f3->GetParameter(2);
  Float_t offset_lo = offset_m - (4*offset_s);
  Float_t offset_hi = offset_m + (4*offset_s);

  h_mvall->Fit("gaus","0");
  f4=h_mvall->GetFunction("gaus");
  Float_t mv_m = f4->GetParameter(1);
  Float_t mv_s = f4->GetParameter(2);
  Float_t mv_lo = mv_m - (4*mv_s);
  Float_t mv_hi = mv_m + (4*mv_s);

  h_nall->Fit("gaus","0");
  f5=h_nall->GetFunction("gaus");
  Float_t noise_m = f5->GetParameter(1);
  Float_t noise_s = f5->GetParameter(2);
  Float_t noise_lo = noise_m - (4*noise_s);
  Float_t noise_hi = noise_m + (4*noise_s);

  // Noise cannot be negative.  
  // Cut for bad channels and projection limits set by #define at top of this file

  if (mv_lo<0) mv_lo =0;
  if (mv_hi>KG_HI_ENC) mv_hi = KG_HI_MV;
  if (noise_lo<0) noise_lo =0;
  if (noise_hi>KG_HI_ENC) noise_hi = KG_HI_ENC;


// titles for graphs - might as well do this now.

  TPaveText *t1 = new TPaveText(0.0,0.0,1.0,1.0);
  t1->SetFillColor(10);
  t1->SetTextAlign(13);
  t1->SetTextSize(0.3);
  t1->SetTextColor(4);

  sprintf(s,"ATLAS SCT Quick evaluation of Gain/Offset/Noise");
  if(time)  sprintf(s,"%s  -  %s",s,time->GetName());
  if(place) sprintf(s,"%s  -  %s",s,place->GetName());
  if(mname) sprintf(s,"%s  -  Module %s",s,mname->GetName());
  t1->AddText(s);

//  sprintf(s,"ATLAS SCT quick evaluation of gain/offset/noise for moi, loi);
  sprintf(s,"module %d link %d  (Run %d  Scan %d  Charge %3.2ffC) - (Run %d  Scan %d Charge %3.2ffC)",
    moi,loi,run2, sno2, q2, run1, sno1, q1);
  t1->AddText(s);


// Before we draw the main canvas, open the postscript file
// if the relevant option has been selected 

  if(postscript){
    if(RAL) sprintf(out_fname,"M:\\sctvar\\data\\stgain%d_%d_%d_%d.ps",run1,sno1,moi,loi);
    else    sprintf(out_fname,"D:\\sctvar\\data\\stgain%d_%d_%d_%d.ps",run1,sno1,moi,loi);

    TPostScript *ps = new TPostScript(out_fname,112);
    ps->NewPage();
  }

// Open and draw the first canvas

  TCanvas* c1=new TCanvas ("c1","stan_kwikgain",1100,750);
  pad_t = new TPad("pad_t","Text",        0.01,0.91,0.99,0.99,33);
  pad_g = new TPad("pad_m","Gain",        0.01,0.61,0.80,0.90,33);
  pad_o = new TPad("pad_s","Offset",      0.01,0.31,0.80,0.60,33);
  pad_n = new TPad("pad_n","Noise (ENC)", 0.01,0.01,0.80,0.30,33);
  pad_p = new TPad("pad_p","Projections", 0.81,0.01,0.99,0.90,33);

  pad_t->Draw();
  pad_g->Draw();

  pad_o->Draw();

  pad_n->Draw();
  pad_p->Draw();
  pad_p->Divide(1,3);

  c1->Update();

  pad_t->cd();
  t1->Draw();

  pad_g->cd();
  sprintf(h_title,"Gain at %3.2ffC (mV/fC)",q);
  h_gain->SetXTitle("Channel Number");
  h_gain->SetYTitle(h_title);
  h_gain->SetMaximum(gain_hi);
  h_gain->SetMinimum(gain_lo);
  h_gain->Draw();

  pad_p->cd(1);
  h_gall->SetXTitle(h_title);
  h_gall->SetYTitle("n(Gain)");
  h_gall->Draw();

  pad_o->cd();
  sprintf(h_title,"Offset at %3.2ffC (mV)",q);
  h_offset->SetXTitle("Channel Number");
  h_offset->SetYTitle(h_title);
  h_offset->SetMaximum(offset_hi);
  h_offset->SetMinimum(offset_lo);
  h_offset->Draw();

  pad_p->cd(2);
  h_oall->SetXTitle(h_title);
  h_oall->SetYTitle("n(Offset)");
  h_oall->Draw();

  pad_n->cd();
  sprintf(h_title,"Input Noise at %3.2ffC (ENC)",q);
  h_noise->SetXTitle("Channel Number");
  h_noise->SetYTitle(h_title);
  h_noise->SetMaximum(noise_hi);
  h_noise->SetMinimum(noise_lo);
  h_noise->Draw();

  pad_p->cd(3);
  h_nall->SetXTitle(h_title);
  h_nall->SetYTitle("n(Noise)");
  h_nall->Draw();


  c1->Update();


  //  Create 18 histograms for projections of Gain, Offset and Noise

  TH1F* h_gp[6];
  TH1F* h_op[6];
  TH1F* h_mvp[6];
  TH1F* h_np[6];

  TH1F* h_gp[0]  =  new TH1F("h_gp0", "Gain, chip 0",  50,gain_lo,gain_hi);
  TH1F* h_gp[1]  =  new TH1F("h_gp1", "Gain, chip 1",  50,gain_lo,gain_hi);
  TH1F* h_gp[2]  =  new TH1F("h_gp2", "Gain, chip 2",  50,gain_lo,gain_hi);
  TH1F* h_gp[3]  =  new TH1F("h_gp3", "Gain, chip 3",  50,gain_lo,gain_hi);
  TH1F* h_gp[4]  =  new TH1F("h_gp4", "Gain, chip 4",  50,gain_lo,gain_hi);
  TH1F* h_gp[5]  =  new TH1F("h_gp5", "Gain, chip 5",  50,gain_lo,gain_hi);

  TH1F* h_op[0]  =  new TH1F("h_op0", "Offset, chip 0",  50,offset_lo,offset_hi);
  TH1F* h_op[1]  =  new TH1F("h_op1", "Offset, chip 1",  50,offset_lo,offset_hi);
  TH1F* h_op[2]  =  new TH1F("h_op2", "Offset, chip 2",  50,offset_lo,offset_hi);
  TH1F* h_op[3]  =  new TH1F("h_op3", "Offset, chip 3",  50,offset_lo,offset_hi);
  TH1F* h_op[4]  =  new TH1F("h_op4", "Offset, chip 4",  50,offset_lo,offset_hi);
  TH1F* h_op[5]  =  new TH1F("h_op5", "Offset, chip 5",  50,offset_lo,offset_hi);
  
  TH1F* h_mvp[0] =  new TH1F("h_mvp0","Output Noise, chip 0",  50,mv_lo,mv_hi);
  TH1F* h_mvp[1] =  new TH1F("h_mvp1","Output Noise, chip 1",  50,mv_lo,mv_hi);
  TH1F* h_mvp[2] =  new TH1F("h_mvp2","Output Noise, chip 2",  50,mv_lo,mv_hi);
  TH1F* h_mvp[3] =  new TH1F("h_mvp3","Output Noise, chip 3",  50,mv_lo,mv_hi);
  TH1F* h_mvp[4] =  new TH1F("h_mvp4","Output Noise, chip 4",  50,mv_lo,mv_hi);
  TH1F* h_mvp[5] =  new TH1F("h_mvp5","Output Noise, chip 5",  50,mv_lo,mv_hi);

  TH1F* h_np[0]  =  new TH1F("h_np0", "Input Noise, chip 0",  50,noise_lo,noise_hi);
  TH1F* h_np[1]  =  new TH1F("h_np1", "Input Noise, chip 1",  50,noise_lo,noise_hi);
  TH1F* h_np[2]  =  new TH1F("h_np2", "Input Noise, chip 2",  50,noise_lo,noise_hi);
  TH1F* h_np[3]  =  new TH1F("h_np3", "Input Noise, chip 3",  50,noise_lo,noise_hi);
  TH1F* h_np[4]  =  new TH1F("h_np4", "Input Noise, chip 4",  50,noise_lo,noise_hi);
  TH1F* h_np[5]  =  new TH1F("h_np5", "Input Noise, chip 5",  50,noise_lo,noise_hi);

  // Fill these Projections, marking additional bad channels in the process

  for (i=0;i<6;i++){
    for(j=0;j<128;j++){
      bin = j + (128*i);

      code=h_codes->GetBinContent(bin);
      gain=h_gain->GetBinContent(bin);
      offset=h_offset->GetBinContent(bin);
      mv=h_mv->GetBinContent(bin);
      noise=h_noise->GetBinContent(bin);

      if(gain<gain_lo) code += GAIN_UNDER;
      else if(gain>gain_hi) code += GAIN_OVER;
      if(offset<offset_lo) code += OFFSET_UNDER;
      else if(offset>offset_hi) code += OFFSET_OVER;
      if(noise<noise_lo) code += NOISE_UNDER;
      else if(noise>noise_hi) code += NOISE_OVER;


      if((code&BAD_TEST)==0){ // fill only good channels
        h_gp[i]->Fill(gain);
        h_op[i]->Fill(offset);
        h_mvp[i]->Fill(mv);
        h_np[i]->Fill(noise);
      } else {
        h_codes->SetBinContent(bin,code);
      }
    }
  }

  // ...and fit gaussians to them, storing the results in the arrays

  float chip_fit_par[4][6];
  float chip_fit_err[4][6];

  for(i=0;i<6;i++){
    if(h_gp[i]->GetMaximum()>0.0){
      h_gp[i]->Fit("gaus","0Q"); // quiet mode
      f2=h_gp[i]->GetFunction("gaus");
      chip_fit_par[0][i]=f2->GetParameter(1);
      chip_fit_err[0][i]=f2->GetParameter(2);
    } else {
      chip_fit_par[0][i]=-1;
      chip_fit_err[0][i]=-1;
    }
    if(h_op[i]->GetMaximum()>0.0){
      h_op[i]->Fit("gaus","0Q"); // quiet mode 
      f3=h_op[i]->GetFunction("gaus");
      chip_fit_par[1][i]=f3->GetParameter(1);
      chip_fit_err[1][i]=f3->GetParameter(2);
    } else {
      chip_fit_par[1][i]=-1;
      chip_fit_err[1][i]=-1;
    }
    if(h_mvp[i]->GetMaximum()>0.0){
      h_mvp[i]->Fit("gaus","0Q"); // quiet mode 
      f4=h_mvp[i]->GetFunction("gaus");
      chip_fit_par[2][i]=f4->GetParameter(1);
      chip_fit_err[2][i]=f4->GetParameter(2);
    } else {
      chip_fit_par[2][i]=-1;
      chip_fit_err[2][i]=-1;
    }
    if(h_np[i]->GetMaximum()>0.0){
      h_np[i]->Fit("gaus","0Q"); // quiet mode 
      f5=h_np[i]->GetFunction("gaus");
      chip_fit_par[3][i]=f5->GetParameter(1);
      chip_fit_err[3][i]=f5->GetParameter(2);
    } else {
      chip_fit_par[3][i]=-1;
      chip_fit_err[3][i]=-1;
    }
  }

  // Optionally, two canvases are created to show the projections
  // of fit data on a chip by chip basis.  This may be prevented by
  // setting draw_fit_projections = 0.

  if(draw_fit_projections){

  // Open and draw new canvas to show the Mean Projections

    if(postscript) ps->NewPage();
    TCanvas* c2=new TCanvas ("c2","stan_kwikgain_by_chip",1100,750);
    pad_mpt = new TPad("pad_mpt", "Text",                       0.00,0.92,1.00,1.00,33);
    pad_mp0 = new TPad("pad_mp0","Projection of Gain values",   0.00,0.69,1.00,0.92,33);
    pad_mp1 = new TPad("pad_mp1","Projection of Offset values", 0.00,0.46,1.00,0.69,33);
    pad_mp2 = new TPad("pad_mp2","Projection of Noise (mV)",    0.00,0.23,1.00,0.46,33);
    pad_mp3 = new TPad("pad_mp3","Projection of Noise (ENC)",   0.00,0.00,1.00,0.23,33);

    pad_mpt->Draw();
    pad_mp0->Draw();
    pad_mp1->Draw();
    pad_mp2->Draw();
    pad_mp3->Draw();
    pad_mp0->Divide(6,1,0,0);
    pad_mp1->Divide(6,1,0,0);
    pad_mp2->Divide(6,1,0,0);
    pad_mp3->Divide(6,1,0,0);
    c2->Update();

    for(i=0; i<6; i++){
      pad_mp0->cd(i+1);
      h_gp[i]->SetXTitle("Gain(mV/fC)");
      h_gp[i]->SetYTitle("N(Gain)");
      h_gp[i]->SetLineColor(kRed);
      h_gp[i]->SetMarkerColor(kBlack);
      h_gp[i]->Draw("ep");
    }

    for(i=0; i<6; i++){
      pad_mp1->cd(i+1);
      h_op[i]->SetXTitle("Offset(mV)");
      h_op[i]->SetYTitle("N(Mean)");
      h_op[i]->SetLineColor(kRed);
      h_op[i]->SetMarkerColor(kBlack);
      h_op[i]->Draw("ep");
    }

    for(i=0; i<6; i++){
      pad_mp2->cd(i+1);
      h_mvp[i]->SetXTitle("Noise(mV)");
      h_mvp[i]->SetYTitle("N(Mean)");
      h_mvp[i]->SetLineColor(kRed);
      h_mvp[i]->SetMarkerColor(kBlack);
      h_mvp[i]->Draw("ep");
    }

    for(i=0; i<6; i++){
      pad_mp3->cd(i+1);
      h_np[i]->SetXTitle("Noise(ENC)");
      h_np[i]->SetYTitle("N(Mean)");
      h_np[i]->SetLineColor(kRed);
      h_np[i]->SetMarkerColor(kBlack);
      h_np[i]->Draw("ep");
    }

    pad_mpt->cd();
    t1->Draw();
    c2->Update();
  }


  // Tally total numbers of good and bad channels  

  short chip_bad_ch[7][16];
  short chip_good_ch[7];

  for(i=0;i<7;i++){
     chip_good_ch[i]=0;     // clear good channel count
    for(k=0;k<16;k++) {     // clear bad channel count
      chip_bad_ch[i][k]=0;
    } 
  }

  for(i=0;i<6;i++){
    for(j=0;j<128;j++){

      code=(Int_t)h_codes->GetBinContent(j+(128*i));
      if(code){
        if (code&CH_DEAD)      chip_bad_ch[i][0]++;
        if (code&CH_STUCK)     chip_bad_ch[i][1]++;
        if (code&CH_UNDER)     chip_bad_ch[i][2]++;
        if (code&CH_OVER)      chip_bad_ch[i][3]++;
        if (code&FIT_UNDER)    chip_bad_ch[i][4]++;
        if (code&FIT_OVER)     chip_bad_ch[i][5]++;
        if (code&SIG_UNDER)    chip_bad_ch[i][6]++;
        if (code&SIG_OVER)     chip_bad_ch[i][7]++;
        if (code&GAIN_UNDER)   chip_bad_ch[i][8]++;
        if (code&GAIN_OVER)    chip_bad_ch[i][9]++;
        if (code&OFFSET_UNDER) chip_bad_ch[i][10]++;
        if (code&OFFSET_OVER)  chip_bad_ch[i][11]++;
        if (code&NOISE_UNDER)  chip_bad_ch[i][12]++;
        if (code&NOISE_OVER)   chip_bad_ch[i][13]++;

        if (code&DEAD_TEST) chip_bad_ch[i][14]++; // DEAD  channels
        if (code&DODGY_TEST) chip_bad_ch[i][15]++; // DODGY channels
      }
      if((code&BAD_TEST)==0) chip_good_ch[i]++;
    }
  }

  for(i=0;i<6;i++){
    chip_good_ch[6] += chip_good_ch[i];
    for(k=0;k<16;k++) chip_bad_ch[6][k] += chip_bad_ch[i][k];
  }

  // Summary data to screen

  // 08.01.00 PWP add code to use chip numbers 6 to 11 in case where link=1.
  //              (in response to a request from MM)

  Int_t   chip_offset=0;
  if(loi) chip_offset=6;

  if(time)  sprintf(s,"%s",time->GetName());
  if(place) sprintf(s,"%s  -  %s",s,place->GetName());
  if(mname) sprintf(s,"%s  -  Module %s",s,mname->GetName());

  if(time) printf("\n%s\n",s);
  printf("Scan 1: run %d scan %d charge %2.2f\n", run1, sno1, q1);
  printf("Scan 2: run %d scan %d charge %2.2f\n", run2, sno2, q2);
  printf("\nSummary for link %d\n",loi);

  printf("  Gain:         mean %3.2f mV/fC, sigma %3.2f mV/fC\n",gain_m,gain_s);
  printf("  Offset:       mean %3.2f mV,    sigma %3.2f mV\n",offset_m,offset_s);
  printf("  Output Noise: mean %3.2f mV,    sigma %3.2f mV (first scan, ie. 2fC)\n",mv_m,mv_s);
  printf("  Input Noise:  mean  %4.0f ENC,  sigma  %4.0f ENC (mean of two scans)\n",noise_m,noise_s);

  //  Print Summary Data (1)
  //  Fitted Gain/Offset/Noise for each chip

  printf("\nChip\tGain\tg_err\tOffset\to_err\tNoise\tn_err\tNoise\tn_err\t\n");
  for(i=0;i<6;i++){
    printf("%d\t%3.2f\t%3.2f\t%3.2f\t%3.2f\t%2.2f\t%2.2f\t%4.0f\t%4.0f\n",
      i+chip_offset,      chip_fit_par[0][i], chip_fit_err[0][i], 
      chip_fit_par[1][i], chip_fit_err[1][i], chip_fit_par[2][i],
      chip_fit_err[2][i], chip_fit_par[3][i], chip_fit_par[3][i]);
  }
  printf("\tmV/fC\t\tmV\t\tmV\t\tENC\n");

  //  Print Summary Data (2)
  //  Bad Channel Summary

  printf("\nBad Channel Summary (1)\n");
  printf("Chip\tDead\tStuck\tN_under\tN_over\tF_under\tF_over\tS_under\tS_over\n");
  for(i=0;i<6;i++){
    printf("%d\t%d\t%d\t%d\t%d\t%d\t%d\t%d\t%d\n", i+chip_offset,
      chip_bad_ch[i][0], chip_bad_ch[i][1],
      chip_bad_ch[i][2], chip_bad_ch[i][3],
      chip_bad_ch[i][4], chip_bad_ch[i][5],
      chip_bad_ch[i][6], chip_bad_ch[i][7]);
  }
  printf("total\t%d\t%d\t%d\t%d\t%d\t%d\t%d\t%d\n",
    chip_bad_ch[6][0], chip_bad_ch[6][1],
    chip_bad_ch[6][2], chip_bad_ch[6][3],
    chip_bad_ch[6][4], chip_bad_ch[6][5],
    chip_bad_ch[6][6], chip_bad_ch[6][7]);

  printf("\nBad Channel Summary (2)\n");
  printf("Chip\tG_under\tG_over\tO_under\tO_over\tNse_und\tNse_over\n");
  for(i=0;i<6;i++){
    printf("%d\t%d\t%d\t%d\t%d\t%d\t%d\n", i+chip_offset,
      chip_bad_ch[i][8], chip_bad_ch[i][9],
      chip_bad_ch[i][10], chip_bad_ch[i][11],
      chip_bad_ch[i][12], chip_bad_ch[i][13]);
  }
  printf("total\t%d\t%d\t%d\t%d\t%d\t%d\n",
    chip_bad_ch[6][8], chip_bad_ch[6][9],
    chip_bad_ch[6][10], chip_bad_ch[6][11],
    chip_bad_ch[6][12], chip_bad_ch[6][12]);


  printf("\nChannel Totals\n");
  printf("Chip\tDead\tDodgy\tUseful\n");
  for(i=0;i<6;i++){
    printf("%d\t%d\t%d\t%d\n", i+chip_offset, chip_bad_ch[i][14], 
      chip_bad_ch[i][15], chip_good_ch[i]);
  }
  printf("total\t%d\t%d\t%d\n", chip_bad_ch[6][14], 
    chip_bad_ch[6][15], chip_good_ch[6]);


  // Ascii file output - if relevant option(s) selected

  // Output Gain/Offset/Noise results to text file

  if(ascii){
    if(RAL) sprintf(f_name,"M:\\sctvar\\data\\stgain%d_%d_%d_%d.fit",run1,sno1,moi,loi);
    else    sprintf(f_name,"D:\\sctvar\\data\\stgain%d_%d_%d_%d.fit",run1,sno1,moi,loi);

    FILE* fitfile = fopen(f_name,"w");
    if(fitfile==NULL){
      printf("Could not open file for fit data ascii output\n");
    } else {
      fprintf(fitfile,"Stan_kwikgain Data - q1 = %3.2ffC, q2 = 3.2ffC\n",q1, q2);
      fprintf(fitfile,"ch   code  gain  offset noise\n");
      for(j=0;j<768;j++){
        fprintf(fitfile,"%3d 0x%04x %3.2f %3.2f %4.0f\n", j,
          h_codes->GetBinContent(j), h_gain->GetBinContent(j), 
          h_offset->GetBinContent(j), h_noise->GetBinContent(j));
        }
      fclose(fitfile);
    }
  }

  // Output summary data to text file

  if(ascii){

    if(RAL) sprintf(f_name,"M:\\sctvar\\data\\stgain%d_%d_%d_%d.sum",run1,sno1,moi,loi);
    else    sprintf(f_name,"D:\\sctvar\\data\\stgain%d_%d_%d_%d.sum",run1,sno1,moi,loi);

    FILE* sumfile = fopen(f_name,"w");
    if(sumfile==NULL){
      printf("Could not open file for summary data ascii output\n");
    } else {

      //  Append summary for each link to the summary file

      if((moi==0)&&(loi==0)){
        fprintf(sumfile,"Scan 1: run %d scan %d charge %2.2f\n", run1, sno1, q1);
        fprintf(sumfile,"Scan 2: run %d scan %d charge %2.2f\n", run2, sno2, q2);
      }

      if(loi==0){
        if(time)  sprintf(s,"%s",time->GetName());
        if(place) sprintf(s,"%s  -  %s",s,place->GetName());
        if(time)  fprintf(sumfile,"\n%s\n",s);
      }

      if(mname) fprintf(sumfile,"\nSummary for module %d link %d (%s)\n",moi, loi,mname->GetName());
      else      fprintf(sumfile,"\nSummary for module %d link %d\n",moi, loi);

      fprintf(sumfile,"  Gain:         mean %3.2f mV/fC, sigma %3.2f mV/fC\n",gain_m,gain_s);
      fprintf(sumfile,"  Offset:       mean %3.2f mV,    sigma %3.2f mV\n",offset_m,offset_s);
      fprintf(sumfile,"  Output Noise: mean %3.2f mV,    sigma %3.2f mV (first scan, at %2.2ffC)\n",mv_m,mv_s,q1);
      fprintf(sumfile,"  Input Noise:  mean  %4.0f ENC,   sigma %4.0f ENC (mean of two scans)\n",noise_m,noise_s);

      fprintf(sumfile,"\nChip\tGain\tg_err\tOffset\to_err\tNoise\tn_err\tNoise\tn_err\n");
      for(i=0;i<6;i++){
        fprintf(sumfile,"%d\t%3.2f\t%3.2f\t%3.2f\t%3.2f\t%2.2f\t%2.2f\t%4.0f\t%4.0f\n",
          i+chip_offset,      chip_fit_par[0][i], chip_fit_err[0][i],
          chip_fit_par[1][i], chip_fit_err[1][i], chip_fit_par[2][i],
          chip_fit_err[2][i], chip_fit_par[3][i], chip_fit_err[3][i]);
      }
      fprintf(sumfile,"\tmV/fC\t\tmV\t\tmV\t\tENC\n");

    //  Print Summary Data (2)
    //  Bad Channel Summary

      fprintf(sumfile,"\nBad Channel Summary (1)\n");
      fprintf(sumfile,"Chip\tDead\tStuck\tN_under\tN_over\tF_under\tF_over\tS_under\tS_over\n");
      for(i=0;i<6;i++){
        fprintf(sumfile,"%d\t%d\t%d\t%d\t%d\t%d\t%d\t%d\t%d\n", i+chip_offset,
          chip_bad_ch[i][0], chip_bad_ch[i][1],
          chip_bad_ch[i][2], chip_bad_ch[i][3],
          chip_bad_ch[i][4], chip_bad_ch[i][5],
          chip_bad_ch[i][6], chip_bad_ch[i][7]);
      }
      fprintf(sumfile,"total\t%d\t%d\t%d\t%d\t%d\t%d\t%d\t%d\n",
        chip_bad_ch[6][0], chip_bad_ch[6][1],
        chip_bad_ch[6][2], chip_bad_ch[6][3],
        chip_bad_ch[6][4], chip_bad_ch[6][5],
        chip_bad_ch[6][6], chip_bad_ch[6][7]);

      fprintf(sumfile,"\nBad Channel Summary (2)\n");
      fprintf(sumfile,"Chip\tG_under\tG_over\tO_under\tO_over\tNse_und\tNse_over\n");
      for(i=0;i<6;i++){
        fprintf(sumfile,"%d\t%d\t%d\t%d\t%d\t%d\t%d\n", i+chip_offset,
          chip_bad_ch[i][8], chip_bad_ch[i][9],
          chip_bad_ch[i][10], chip_bad_ch[i][11],
          chip_bad_ch[i][12], chip_bad_ch[i][13]);
      }
      fprintf(sumfile,"total\t%d\t%d\t%d\t%d\t%d\t%d\n",
        chip_bad_ch[6][8], chip_bad_ch[6][9],
        chip_bad_ch[6][10], chip_bad_ch[6][11],
        chip_bad_ch[6][12], chip_bad_ch[6][12]);


      fprintf(sumfile,"\nChannel Totals\n");
      fprintf(sumfile,"Chip\tDead\tDodgy\tUseful\n");
      for(i=0;i<6;i++){
        fprintf(sumfile,"%d\t%d\t%d\t%d\n", i+chip_offset, chip_bad_ch[i][14], 
          chip_bad_ch[i][15], chip_good_ch[i]);
      }
      fprintf(sumfile,"total\t%d\t%d\t%d\n", chip_bad_ch[6][14], 
        chip_bad_ch[6][15], chip_good_ch[6]);


    // List anomalous channels
  
      fprintf(sumfile,"\nBad Channel Data\n");
      fprintf(sumfile,"ch   code  gain  offset noise\n");
      for(j=0;j<768;j++){
        code = h_codes->GetBinContent(j);
        if(code){
          fprintf(sumfile,"%3d 0x%04x %3.2f %3.2f %4.0f\n", j, code,
            h_gain->GetBinContent(j), h_offset->GetBinContent(j), h_noise->GetBinContent(j));
        }
      }
      fclose(sumfile);
    }
  }


  if(time) delete time;
  if(place) delete place;
  if(mname) delete mname;
}

// Wrappers...

void stan_kwikgain(Int_t run, Int_t sno1, Float_t q1, Int_t sno2, Float_t q2, Int_t moi, Int_t loi){

  // same run, any scans, any module/link

  stan_kwikgain(run,sno1,q1,run,sno2,q2,moi,loi,1,1); 
}

void stan_kwikgain(Int_t run, Int_t sno, Float_t q1, Float_t q2, Int_t moi, Int_t loi){

  // same run, any scans, any module/link

  stan_kwikgain(run,sno,q1,run,(sno+1),q2,moi,loi,1,1); 
}

void stan_kwikgain(Int_t run, Int_t sno1, Float_t q1, Int_t sno2, Float_t q2){

  // same run, any scans, module 0, link 0

  stan_kwikgain(run,sno1,q1,run,sno2,q2,0,0,1,1); 
}

void stan_kwikgain(Int_t run, Int_t sno, Float_t q1, Float_t q2){

  // same run, consecutive scans, module 0, link 0

  stan_kwikgain(run,sno,q1,run,(sno+1),q2,0,0,1,1); 
}

void stan_kwikgain(){

  printf("Possible Uses of stan_kwikgain:\n\n");
  printf(".x stan_kwikgain.cpp(run,scan1,q1,q2)\n");
  printf(".x stan_kwikgain.cpp(run,scan1,q1,scan2,q2)\n");
  printf(".x stan_kwikgain.cpp(run,scan1,q1,q2,moi,loi)\n");
  printf(".x stan_kwikgain.cpp(run,scan1,q1,scan2,q2,moi,loi)\n");
  printf(".x stan_kwikgain.cpp(run1,scan1,q1,run2,scan2,q2,moi,loi,ascii,ps)\n");

}



