void FullBypassAnalysis(Int_t *run_array, Int_t *scan_array, Int_t moi, Float_t *vdd_array, Int_t n_scans, Int_t n_configs){
  
  // analysis of FullBypassTest
  // version 0.1 PWP 14.11.00
  // version 0.2 LE  19.02.00 - including BYPASS 13 and 14
  //                          - stricter definition of a good chip
  // version 0.3 LE  27.02.01 - uses full token/data bypass scheme
  //                          - loops over Vdd
  // version 0.4 PWP 25.05.01 - use parameter.h and Toolbox.cpp    
  // version 0.41 PWP 21.08.01 - database compliance including defect list
  // version 0.42 PWP 12.09.01 - further database compliance
  // version 0.43 PWP 02.11.01 - fixes to run with ROOT 3
  // version 0.44 BJG 30.04.02 - fix some memory leaks
  //

  #ifdef DATA_DIRECTORY
    // do not reload parameters.h
  #else
    // must load parameters.h
  #include <parameters.h>
  #endif

#define MAX_CONFIGS 62
#define N_TRIGS    100
#define MAX_VDD     20
  
  // Debug flag
  const Int_t debug = 0;
  
  // Various variables
  Int_t module, link, chip, chan, conf, ivdd;
  char f_name[128], h_name[128], s_direct[64], s_bypass[64];
  char mname[MAX_MODULES][32];
  Int_t conf_sum[MAX_LINKS];
  Int_t start, end;
  
  //
  // Status variables
  //
  
  Short_t chip_dead[MAX_LINKS][MAX_CHIPS];     // Chip returns zero data
  Short_t chip_good[MAX_LINKS][MAX_CHIPS];     // Chip returns full occupancy
  Short_t chip_error[MAX_LINKS][MAX_CHIPS];    // Chip returns crap data
  
  // Gives the status of a chip in a certain configuration.
  // '-1' unkown status
  // '0'  chip returns zero
  // '1'  chip returns full occpancy
  // '2'  chip returns crap data
  Short_t chip_status[MAX_CONFIGS][MAX_LINKS][MAX_CHIPS];
  
  // Gives the status of a particular direct/bypass
  // configuration. 
  // '-2' not tested
  // '-1' not applicable
  // '0'  configuration failing
  // '1'  configuration OK
  Short_t conf_status[MAX_CONFIGS][MAX_LINKS];
  
  // States if a chip is accessible or not in the read
  // out chain. 
  // '-1' chip not accessible.
  // 'n'  chip accesible via configuration #n.
  Short_t chip_accessible[MAX_LINKS][MAX_CHIPS];
  
  // Gives status of the data/token link, numbered according
  // to the numbers of the receiving chip;
  // '-2' token/data link not tested
  // '-1' token/data link not accessible
  // '0'  token/data link failing
  // '1'  token/data link OK
  Short_t direct_token[MAX_VDD][MAX_MODULES][MAX_LINKS][MAX_CHIPS];
  Short_t bypass_token[MAX_VDD][MAX_MODULES][MAX_LINKS][MAX_CHIPS];
  
  // Status of the whole module at a certain Vdd
  // '-2' no data
  // '-1' not analysed
  // '0'  failed
  // '1'  passed
  Short_t module_status[MAX_VDD][MAX_MODULES];
  Short_t module_passed[MAX_MODULES];

  // Summary data for the whole scan. Gives minimum Vdd
  // for correct operation of the link. The status array
  // states if:
  // '-3' token/data link does not work at any voltage (Vdd=0)
  // '-2' token/data link not tested (Vdd=0)
  // '-1' value states minimal accessible Vdd.
  // '0'  value states minimal functional Vdd.
  // '1'  token/data link works for all tested Vdd 
  Float_t vdd_direct[MAX_MODULES][MAX_LINKS][MAX_CHIPS];
  Float_t vdd_bypass[MAX_MODULES][MAX_LINKS][MAX_CHIPS];
  Short_t status_direct[MAX_MODULES][MAX_LINKS][MAX_CHIPS];
  Short_t status_bypass[MAX_MODULES][MAX_LINKS][MAX_CHIPS];
  
  // Load other files needed by this one
  gROOT->LoadMacro("stan_util.cpp");
  gROOT->LoadMacro("ToolBox.cpp");
  
  // 
  // Ensure that the module name array is cleared to start
  //
  
  for (module=0; module<MAX_MODULES; module++){
    sprintf(mname[module]," ");
  }
  
  //
  // sanity checks
  //

  if(n_configs>62){
    printf("*** Maximum value for ST_TOKEN is 62, using this instead\n");
    n_configs=62;
  }
  
  //
  // Loop over scans (different Vdds)
  //
  
  TFile *strun = 0;
  
  for(ivdd=0;ivdd<n_scans;ivdd++){

    // Load file
    sprintf(f_name,"%sstrun%d_%d.root",DATA_DIRECTORY,run_array[ivdd],scan_array[ivdd]);
    strun = new TFile(f_name);
    
    if(strun==0){
      printf("Failed to open file %s - EXITING!!\n",f_name);
      return;
    }

    //
    //  loop over modules
    //
    if(moi==-1){
      start = 0;
      end   = MAX_MODULES;
    } else {
      start = moi;
      end   = moi+1;
    }
    
    for (module=start; module<end; module++){
      module_status[ivdd][module] = -2;

      for(link=0;link<MAX_LINKS;link++){
	for(chip=0;chip<MAX_CHIPS;chip++){
	  chip_dead[link][chip]=0;
	  chip_good[link][chip]=0;
	  for(conf=0;conf<MAX_CONFIGS;conf++){
	    chip_status[conf][link][chip]=-1;
	  }
	}
      }
      
      for(link=0;link<MAX_LINKS;link++){

	// find histograms
	
	sprintf(h_name,"ModuleName;%d",module+1);
	TObjString* name = (TObjString*)strun->Get(h_name);
	if(name) {
	  sprintf(mname[module],"%s",name->GetName());
	  delete name;
	}
	
 	sprintf(h_name,"h_scan%d;%d",link,module+1);
	TH2F *h_scan = (TH2F*)strun->Get(h_name);
	
	if(h_scan){
	  for(conf=0;conf<n_configs;conf++){
	    for(chip=0;chip<MAX_CHIPS;chip++){
	      chip_status[conf][link][chip]=0;
	    }
	  }
	  if(h_scan->GetMean()>0){
	  module_status[ivdd][module] = -1;

	    printf("Analysing data for module %d link %d\n",
		   module, link);
	    
	    // Loop over each tested configurations
	    for(conf=0;conf<n_configs;conf++){
	      
	      for(chip=0;chip<MAX_CHIPS;chip++){
		chip_dead[link][chip]=0;
		chip_good[link][chip]=0;
		chip_error[link][chip]=0;
	      }
	            
	      // Categorise each channel
	      for(chan=0;chan<MAX_CHANS;chan++){
		chip = chan/128;
		
		Float_t occ = h_scan->GetBinContent((770*(conf+1))+chan+1);
		
		if( (occ>-3)&&(occ<3) ) {
		  // allow +/-2 hits failing per channel
		  chip_dead[link][chip]++;
		}else if( (occ>N_TRIGS-3)&&(occ<N_TRIGS+3) ) {
		  // allow +/-2 hits failing per channel
		  chip_good[link][chip]++;
		}else{
		  chip_error[link][chip]++;
		}
	      } // end for(chan)

	      if(debug){
		for(chip=0;chip<MAX_CHIPS;chip++){
		  printf("link%d chip %d: n_good %d, n_dead %d, n_error %d\n",
			 link,(chip+8*link),chip_good[link][chip],chip_dead[link][chip],chip_error[link][chip]);
		}
	      }
	      // How many chips gave data?
	      
	      if(debug) printf(" Conf %d, link%d, chip status: ",conf,link);
	      for(chip=0;chip<MAX_CHIPS;chip++){
		if(chip_good[link][chip]>=120){
		  chip_status[conf][link][chip]=1; // CHIP GAVE CORRECT DATA
		  if(debug) printf("%d ",chip_status[conf][link][chip]);
		}else if(chip_dead[link][chip]>=120){
		  chip_status[conf][link][chip]=0; // CHIP DEAD
		  if(debug) printf("%d ",chip_status[conf][link][chip]);
		}else{
		  chip_status[conf][link][chip]=2; // CHIP STATUS UNDEFINED
		  if(debug) printf("%d ",chip_status[conf][link][chip]);
		}
	      }
	      if(debug) printf("\n");
	    } // end for(conf)
	  } // end if(mean>0) [histogram is not empty]
	} // end if(h_scan) [histogram exists]
      } // end for(link)
      
    
      
      if(module_status[ivdd][module]>-2) {   // Only analyse data for modules
	                                     // with data.
	

	/* Loop over configuations to see which ones that work
	 * Each particular direct/byass configuration is checked and
	 * given the status in the array conf_status[conf][link]: 
	 * '-2' not tested
	 * '-1' not applicable
	 * '0'  configuration failing
	 * '1'  configuration OK   
	 */
	
	for(conf=0;conf<MAX_CONFIGS;conf++){
	  for(link=0;link<MAX_LINKS;link++){
	    for(chip=0;chip<MAX_CHIPS;chip++){
	      conf_status[conf][link] = -2;  // Not tested
	    }
	  }
	}
	

	for(conf=0;conf<n_configs;conf++){

	  //
	  // Generate the sum from the read pattern of chips
	  //

	  for(link=0;link<MAX_LINKS;link++){
	    conf_status[conf][link] = 0;
	    conf_sum[link] = 0;
	    for(chip=0;chip<MAX_CHIPS;chip++){
	      if(chip_status[conf][link][chip]==-1) {
		break;
	      }else if(chip_status[conf][link][chip]==0){
		// do nothing
	      }else if(chip_status[conf][link][chip]==1){
		conf_sum[link]+=pow(2,chip);
	      }else if(chip_status[conf][link][chip]==2){
		conf_sum[link] = -1;
	 	break;
	      } else {
		conf_sum[link] = -1;
		break;
	      }
	    }
	    if (debug) printf("link%d Vdd=%3.1f, conf=%d: conf_sum=%d\n",link,vdd_array[ivdd],conf,conf_sum[link]);
	  }
	   
	  //
	  // Check if pattern matches the predicted
	  //
	  if(conf<20) { // reading out the two links in parallel.
	    for(link=0;link<MAX_LINKS;link++){
	      if(conf_sum[link]==GetConf(conf)){
		conf_status[conf][link] = 1;
	      }else if(conf_sum[link]==-1){
		conf_status[conf][link] = 0;
	      }else{
		conf_status[conf][link] = 0;
	      }
	    }
	  } else if(conf<36) {
	    if((conf%2)==0){ // reading out via link0
	      if ( (conf_sum[0]==GetConf((conf-20)/2+12)) && (conf_sum[1]==2) ){
		conf_status[conf][0] = 1;
		conf_status[conf][1] = -1;
	      }else if((conf_sum[0]==-1)||(conf_sum[1]==-1)){
		conf_status[conf][0] = 0;
		conf_status[conf][1] = -1;
	      }else{
		conf_status[conf][0] = 0;
		conf_status[conf][1] = -1;
	      }
	    }else{ // reading out via link1
	      if ( (conf_sum[1]==GetConf((conf-21)/2+12)) && (conf_sum[0]==2) ){
		conf_status[conf][1] = 1;
		conf_status[conf][0] = -1;
	      }else if((conf_sum[0]==-1)||(conf_sum[1]==-1)){
		conf_status[conf][1] = 0;
		conf_status[conf][0] = -1;
	      }else{
		conf_status[conf][1] = 0;
		conf_status[conf][0] = -1;
	      }
	    }
	  }else if(conf<62){
	    if((conf%2)==0){ // reading out via link0
	      if ( (conf_sum[0]==GetConf((conf-36)/2+7)) && (conf_sum[1]==1) ){
		conf_status[conf][0] = 1;
		conf_status[conf][1] = -1;
	      }else if( (conf_sum[0]==-1) || (conf_sum[1]==-1) ){
		conf_status[conf][0] = 0;
		conf_status[conf][1] = -1;
	      }else{
		conf_status[conf][0] = 0;
		conf_status[conf][1] = -1;
	      }
	    }else{ // reading out via link1
	      if ( (conf_sum[1]==GetConf((conf-37)/2+7)) && (conf_sum[0]==1) ){
		conf_status[conf][1] = 1;
		conf_status[conf][0] = -1;
	      }else if( (conf_sum[0]==-1) || (conf_sum[1]==-1) ){
		conf_status[conf][1] = 0;
		conf_status[conf][0] = -1;
	      }else{
		conf_status[conf][1] = 0;
		conf_status[conf][0] = -1;
	      }
	    }
	  }else{
	    printf("ERROR: configuration variable out of range (>61)\n");
	  }
	  
	  if(debug) {
	    for(link=0;link<MAX_LINKS;link++){
	      printf("link%d configuration %d  status: %d\n",link,conf,conf_status[conf][link]);
	    }
	  }
	}
      
	/* Determine if a chip is accessible or not in the read out 
	 * chain. Stauts stored in array chip_accessible[link][chip]:
	 * '-1' chip not accessible.
	 * 'n'  chip accesible via configuration #n.
	 */
	for(link=0;link<MAX_LINKS;link++){
	  for(chip=0;chip<MAX_CHIPS;chip++){
	    chip_accessible[link][chip]=-1;
	    for(conf=SumFib(chip);conf<SumFib(chip+1);conf++){
	      if(conf_status[conf][link]==1){
		chip_accessible[link][chip]=conf;
		if(debug) printf("link%d chip %d accessible via configuration %d (value: %d)\n",
				 link,(chip+8*link),conf,GetConf(conf));
	      }else if(conf_status[conf][link]==0){
		if(debug) printf("link%d chip %d NOT accessible (conf: %d)\n",
				 link,(chip+8*link),conf);
	      }else if(conf_status[conf][link]==-2){
		if(debug) printf("link%d chip %d NOT tested (conf: %d)\n",
				 link,(chip+8*link),conf);
	      }
	    }
	  }
	}
      
	
	/* Analyse data to determine which token/data links that work
	 * 
	 * The token/data links are numbered 0-5 for each link.
	 * The results are stored for each supply voltage Vdd in 
	 * arrays called 'direct_token and 'bypass_token' according 
	 * to the following:
	 * '-2' token/data link not tested
	 * '-1' token/data link not accessible
	 * '0'  token/data link doesn't work properly
	 * '1' token/data link works properly.
	 */
	
	
	// Loop over token/data links to determine
	// if they are working, accessible or faulty.
	for(link=0;link<MAX_LINKS;link++){
	  for(chip=0;chip<MAX_CHIPS;chip++){
	    
	    if(chip_accessible[link][chip]!=-1){	  // token link accessible?

              switch(chip){
                case 4:
                  direct_token[ivdd][module][link][chip] = 
                    conf_status[ GetConfNo(GetConf(chip_accessible[link][chip]) + (int)pow(2,(chip+1))) ][link];
                  bypass_token[ivdd][module][link][chip] = 
                    conf_status[ 36 + 2*(chip_accessible[link][chip] - 7) + link ][link];
                  break;
		
                case 5:
                  direct_token[ivdd][module][link][chip] = 
                    conf_status[ 36 + 2*(chip_accessible[link][chip] - 7) + link ][link];
                  bypass_token[ivdd][module][link][chip] = 
                    conf_status[ 20 + 2*(chip_accessible[link][chip] - 12) + link ][link];
                  break;

                default:
                  direct_token[ivdd][module][link][chip] = 
                    conf_status[ GetConfNo(GetConf(chip_accessible[link][chip]) + (int)pow(2,(chip+1))) ][link];
                  bypass_token[ivdd][module][link][chip] = 
                    conf_status[ GetConfNo(GetConf(chip_accessible[link][chip]) + (int)pow(2,(chip+2))) ][link];
                  break;
              }

	    }else{	  // Chip not accessible, but do we really test that
	                  // configuration?

              switch(chip){
                case 4:
                  if(n_configs>=20){
                    direct_token[ivdd][module][link][chip] = -1;
                  }else{
                    direct_token[ivdd][module][link][chip] = -2;
                  }
                  if(n_configs>=(45+link)){
                    bypass_token[ivdd][module][link][chip] = -1;
                  }else{
                    bypass_token[ivdd][module][link][chip] = -2;
                  }
                  break;

                case 5:
                  if(n_configs>=(61+link)){
                    direct_token[ivdd][module][link][chip] = -1;
                  }else{
                    direct_token[ivdd][module][link][chip] = -2;
                  }
                  if(n_configs>=(35+link)){
                    bypass_token[ivdd][module][link][chip] = -1;
                  }else{
                    bypass_token[ivdd][module][link][chip] = -2;
                  }
                  break;

                default:
                  if(SumFib(chip+2)+link<=n_configs+1){
                    direct_token[ivdd][module][link][chip] = -1; // Not accessible
                  }else{
                    direct_token[ivdd][module][link][chip] = -2; // Not tested
                  }
                  if(SumFib(chip+3)+link<=n_configs+1){
                    bypass_token[ivdd][module][link][chip] = -1;
                  }else{
                    bypass_token[ivdd][module][link][chip] = -2;
                  }
                  break;
	      }
	    }
	  }
	}
	
	// And finally, has the module passed?
	
	module_status[ivdd][module]=1;
	for(link=0;link<MAX_LINKS;link++){
	  for(chip=0;chip<MAX_CHIPS;chip++){
	    if( (direct_token[ivdd][module][link][chip]==-1) || (bypass_token[ivdd][module][link][chip]==-1)
		|| (direct_token[ivdd][module][link][chip]==0) || (bypass_token[ivdd][module][link][chip]==0) ){
	      module_status[ivdd][module] = 0;
	    }
	  }
	}
      
	
	//
	// Results to screen
	//
	
        if(module_status[ivdd][module]==1){
	  printf("\n  Module %d (%s) has PASSED for Vdd = %3.1f\n\n",module,mname[module],vdd_array[ivdd]);
	}else{
	  printf("\n  Module %d (%s) has FAILED for Vdd = %3.1f\n\n",module,mname[module],vdd_array[ivdd]);
	}

	for(link=0;link<MAX_LINKS;link++){
	  for(chip=0;chip<MAX_CHIPS;chip++){

            switch(direct_token[ivdd][module][link][chip]){
              case -2:
                sprintf(s_direct,"Not Tested");
                break;
              case -1:
                sprintf(s_direct,"Not Accessible");
                break;
              case 0:
                sprintf(s_direct,"Failing");
                break;
              case 1:
                sprintf(s_direct,"OK");
                break;
              default:  
                sprintf(s_direct,"UNKNOWN");
                break;
	    }
	    
            switch(bypass_token[ivdd][module][link][chip]){
              case -2:
                sprintf(s_bypass,"Not Tested");
                break;
              case -1:
                sprintf(s_bypass,"Not Accessible");
                break;
              case 0:
                sprintf(s_bypass,"Failing");
                break;
              case 1:
                sprintf(s_bypass,"OK");
                break;
              default:  
                sprintf(s_bypass,"UNKNOWN");
                break;
	    }
	    
	    
            switch(chip){
              case 0:
                printf("M%d->S%d: \t%s\n",(chip+8*link),(chip+1+8*link),s_direct);
                printf("M%d->S%d: \t%s\n",(chip+8*link),(chip+2+8*link),s_bypass);
                break;
              case 3:
                printf("S%d->S%d: \t%s\n",(chip+8*link),(chip+1+8*link),s_direct);
                printf("S%d->E%d: \t%s\n",(chip+8*link),(chip+2+8*link),s_bypass);
                break;
              case 4:
                printf("S%d->E%d: \t%s\n",(chip+8*link),(chip+1+8*link),s_direct);
                printf("S%d->M%d: \t%s\n",(chip+8*link),(8*(link+1))%16,s_bypass);
                break;
              case 5:
                printf("E%d->M%d: \t%s\n",(chip+8*link),(8*(link+1))%16,s_direct);
                printf("E%d->S%d: \t%s\n",(chip+8*link),((8*(link+1))%16+1),s_bypass);
                break;
              default:  
                printf("S%d->S%d: \t%s\n",(chip+8*link),(chip+1+8*link),s_direct);
                printf("S%d->S%d: \t%s\n",(chip+8*link),(chip+2+8*link),s_bypass);
                break;
	    }
	  }
	}
	printf("\n");
      } // end if(module_staus>-2) 
    } // end for(module)

    // Close root file
    if(ivdd<(n_scans-1)){
      strun->Close();
      delete strun;
    }
  } // end for(ivdd)
  
  
  //
  // Loop over Vdd values to determine minimal operational values.
  //
  
  for(module=start;module<end;module++){

    // First of all, has the module passed all voltages?
    module_passed[module] = 1;
    for(ivdd=0;ivdd<n_scans;ivdd++){
      if(module_status[ivdd][module]!=1){
	module_passed[module]=0;
      }
    }

    if(module_status[0][module]>-2){
      printf("Data exists for Module %d\n",module);

      
      /* Summary data for the whole scan. Gives minimum Vdd
       * for correct operation of the link. The status array
       * states if:
       * '-3' token/data link does not work at any voltage (Vdd=0)
       * '-2' token/data link not tested (Vdd=0)
       * '-1' value states minimal accessible Vdd.
       * '0'  value states minimal functional Vdd.
       * '1'  token/data link works for all tested Vdd */

      Int_t failures = 0;
      for(link=0;link<MAX_LINKS;link++){
	for(chip=0;chip<MAX_CHIPS;chip++){

	  status_direct[module][link][chip]=-3; // Default: Doesn't work at all
	  vdd_direct[module][link][chip]=10.0;

	  for(ivdd=0;ivdd<n_scans;ivdd++){
	    if(direct_token[ivdd][module][link][chip]==-2){  // If it's not tested
	      vdd_direct[module][link][chip]=0.0;
	      status_direct[module][link][chip]=-2;
	      break;
	    }else if(direct_token[ivdd][module][link][chip]==1){  // If OK for this voltage
	      vdd_direct[module][link][chip]=vdd_array[ivdd];
	      status_direct[module][link][chip]=1;
	    }else if(direct_token[ivdd][module][link][chip]==0){  // If failure is true malfunctioning
	      status_direct[module][link][chip]=0;
	      break;
	    }else if(direct_token[ivdd][module][link][chip]==-1){  // If failure is link not accessible
	      status_direct[module][link][chip]=-1;
	      break;
	    }
	  }

	  
	  // Check if it doesn't work for the maximum tested voltage.
	  if(vdd_array[0]>vdd_array[n_scans-1]){
	    if(direct_token[0][module][link][chip]==0) status_direct[module][link][chip]=-3;
	  }else{
	    if(direct_token[n_scans-1][module][link][chip]==0) status_direct[module][link][chip]=-3;
	  }

	  status_bypass[module][link][chip]=-3;  // Default: Doesn't work at all
	  vdd_bypass[module][link][chip]=10.0;

	  for(ivdd=0;ivdd<n_scans;ivdd++){
	    if(bypass_token[ivdd][module][link][chip]==-2){  // If it's not tested
	      vdd_bypass[module][link][chip]=0.0;
	      status_bypass[module][link][chip]=-2;
	      break;
	    }else if(bypass_token[ivdd][module][link][chip]==1){  // If OK for this voltage
	      vdd_bypass[module][link][chip]=vdd_array[ivdd];
	      status_bypass[module][link][chip]=1;
	    }else if(bypass_token[ivdd][module][link][chip]==0){  // If failure is true malfunctioning
	      status_bypass[module][link][chip]=0;
	      break;
	    }else if(bypass_token[ivdd][module][link][chip]==-1){  // If failure is link not accessible
	      status_bypass[module][link][chip]=-1;
	      break;
	    }
	  }

	  // Check if it doesn't work for the maximum tested voltage.
	  if(vdd_array[0]>vdd_array[n_scans-1]){
	    if(bypass_token[0][module][link][chip]==0) status_bypass[module][link][chip]=-3;
	  }else{
	    if(bypass_token[n_scans-1][module][link][chip]==0) status_bypass[module][link][chip]=-3;
	  }
	}
      }

      // Output general PASS/FAIL to screen
      if(module_passed[module]==1){
	printf("\n  Module %d (%s) has PASSED for Vdd = %3.1f - %3.1f\n\n",module,mname[module],vdd_array[0],vdd_array[n_scans-1]);
      }else{
	printf("\n  Module %d (%s) has FAILED for Vdd = %3.1f - %3.1f\n\n",module,mname[module],vdd_array[0],vdd_array[n_scans-1]);
      }


      // Ouput header to screen
      if(vdd_array[0]>vdd_array[n_scans-1]){
	printf("\n  Minimal Vdd for correct operation of token/data link.\n");
	printf("  Module %d (%s)\n\n",module,mname[module]);
      }else{
	printf("\n  Maximal Vdd for correct operation of token/data link.\n");
	printf("  Module %d (%s)\n\n",module,mname[module]);
      }


      // Open file for output
      FILE* outfile = OpenResults("FullBypassTest", strun, module, 
                                  run_array[0], scan_array[0], module_passed[module],0); 

      if(outfile){
        fprintf(outfile,"%%FullBypassTest\n");
        fprintf(outfile,"#\n");

        fprintf(outfile,"#NConfigs  %d\n",n_configs);
        fprintf(outfile,"#Vdd = %3.1f - %3.1f\n",vdd_array[0],vdd_array[n_scans-1]);
        fprintf(outfile,"#\n");

        fprintf(outfile,"#Vmin\n");
        fprintf(outfile,"#token\tr_tkn\ttoken\tr_tkn\n");

        for(link=0;link<MAX_LINKS;link++){
          for(chip=0;chip<MAX_CHIPS;chip+=2){
            fprintf(outfile,"#%s \t\t%s\n",MseLabel(link,chip),MseLabel(link,chip+1));
            fprintf(outfile,"%1.2f \t %1.2f \t",vdd_direct[module][link][chip],vdd_bypass[module][link][chip]);
            fprintf(outfile,"%1.2f \t %1.2f\n", vdd_direct[module][link][chip+1],vdd_bypass[module][link][chip+1]);
            fprintf(outfile,"#\n");
          }
        }
        fprintf(outfile,"#Comment\n");
      }

      for(link=0;link<MAX_LINKS;link++){
	for(chip=0;chip<MAX_CHIPS;chip++){

          // Putting labels on the different failure modes.

          switch(status_direct[module][link][chip]){
            case -3:
              sprintf(s_direct,"Failing maximal tested Vdd");
              break;
            case -2:
              sprintf(s_direct,"Not Tested");
              break;
            case -1:
              if(vdd_array[0]>vdd_array[n_scans-1]){
                sprintf(s_direct,"Minimal accessible");
              }else{
                sprintf(s_direct,"Maximal accessible");
              }
              break;
            case 0:
              if(vdd_array[0]>vdd_array[n_scans-1]){
                sprintf(s_direct,"Minimal functional");
              }else{
                sprintf(s_direct,"Maximal functional");
              }
              break;
            case 1:
              if(vdd_array[0]>vdd_array[n_scans-1]){
                sprintf(s_direct,"Minimal tested");
              }else{
                sprintf(s_direct,"Maximal tested");
              }
              break;
            default:  
              sprintf(s_direct,"UNKNOWN");
              break;
          }

          switch(status_bypass[module][link][chip]){
            case -3:
              sprintf(s_bypass,"Failing maximal tested Vdd");
              break;
            case -2:
              sprintf(s_bypass,"Not Tested");
              break;
            case -1:
              if(vdd_array[0]>vdd_array[n_scans-1]){
                sprintf(s_bypass,"Minimal accessible");
              }else{
                sprintf(s_bypass,"Maximal accessible");
              }
              break;
            case 0:
              if(vdd_array[0]>vdd_array[n_scans-1]){
                sprintf(s_bypass,"Minimal functional");
              }else{
                sprintf(s_bypass,"Maximal functional");
              }
              break;
            case 1:
              if(vdd_array[0]>vdd_array[n_scans-1]){
                sprintf(s_bypass,"Minimal tested");
              }else{
                sprintf(s_bypass,"Maximal tested");
              }
              break;
            default:  
              sprintf(s_bypass,"UNKNOWN");
              break;
          }
	    

	  //
	  // Output data to screen
	  //

          switch(chip){
            case 0:
              printf("M%d->S%d: \t%3.1f V\t(%s)\n",(chip+8*link),(chip+1+8*link),vdd_direct[module][link][chip],s_direct);
              printf("M%d->S%d: \t%3.1f V\t(%s)\n",(chip+8*link),(chip+2+8*link),vdd_bypass[module][link][chip],s_bypass);
              break;
            case 3:
              printf("S%d->S%d: \t%3.1f V\t(%s)\n",(chip+8*link),(chip+1+8*link),vdd_direct[module][link][chip],s_direct);
              printf("S%d->E%d: \t%3.1f V\t(%s)\n",(chip+8*link),(chip+2+8*link),vdd_bypass[module][link][chip],s_bypass);
              break;
            case 4:
              printf("S%d->E%d: \t%3.1f V\t(%s)\n",(chip+8*link),(chip+1+8*link),vdd_direct[module][link][chip],s_direct);
              printf("S%d->M%d: \t%3.1f V\t(%s)\n",(chip+8*link),(8*(link+1))%16,vdd_bypass[module][link][chip],s_bypass);
              break;
            case 5:
              printf("E%d->M%d: \t%3.1f V\t(%s)\n",(chip+8*link),(8*(link+1))%16,vdd_direct[module][link][chip],s_direct);
              printf("E%d->S%d: \t%3.1f V\t(%s)\n",(chip+8*link),((8*(link+1))%16+1),vdd_bypass[module][link][chip],s_bypass);
              break;
            default:
              printf("S%d->S%d: \t%3.1f V\t(%s)\n",(chip+8*link),(chip+1+8*link),vdd_direct[module][link][chip],s_direct);
              printf("S%d->S%d: \t%3.1f V\t(%s)\n",(chip+8*link),(chip+2+8*link),vdd_bypass[module][link][chip],s_bypass);
              break;
	  }

	  //
	  // Output data to file
	  //

	  if(outfile!=NULL){

            if(chip%2==0){
              fprintf(outfile,"#%s \t\t%s\n",MseLabel(link,chip),MseLabel(link,chip+1));
              fprintf(outfile,"\"%s\" \t\"%s\" \t",s_direct,s_bypass);
            }else{
              fprintf(outfile,"\"%s\" \t\"%s\"\n",s_direct,s_bypass);
            }
	  }

	}
      }

      if(outfile!=NULL){

        Int_t n_defects=0;
        for(link=0;link<MAX_LINKS;link++){
          for(chip=0;chip<MAX_CHIPS;chip++){

            // Any link which does not work correctly at the maximum 
            // tested Vdd (normally 4.0V) should be recorded as a defect
	  
            if(status_direct[module][link][chip]==-3){
              //fprintf(outfile,"BYPASS\t%d\t%3.1f\n",((chip*2)+(12*link)),vdd_direct[module][link][chip]);
              ListDefect(outfile,"TOKEN", (((link*MAX_CHIPS)+chip)*128),(((link*MAX_CHIPS)+chip+1)*128)-1);
              n_defects++;
            }
            if(status_bypass[module][link][chip]==-3){
              //fprintf(outfile,"BYPASS\t%d\t%3.1f\n",(1+(chip*2)+(12*link)),vdd_bypass[module][link][chip]);
              ListDefect(outfile,"RTOKEN", (((link*MAX_CHIPS)+chip)*128),(((link*MAX_CHIPS)+chip+1)*128)-1);
              n_defects++;
            }
          }
        }
        if(n_defects==0){
          fprintf(outfile,"#No defects found!\n");
        }else{
          fprintf(outfile,"#%d defects found\n",n_defects);
        }

        fclose(outfile);
      }
    }
  }
  if(strun){
    strun->Close();
    delete strun;
  }
}


Int_t GetConfNo(Int_t token) {

  /* Returns the value of ST_TOKEN  that corresponds to a certain 
   * read out pattern. Chips included in the read out are considered 
   * to represent 1's and chips not included are considered to be 
   * 0's. This binary bit-pattern is converted into a value 'token', 
   * and this function returns the value of ST_TOKEN that corresopnds 
   * to that read out pattern.  
   *
   * This code is implemented as a look-up table to gain speed
   * perfomance. But I have keept the algorithm coded but commented 
   * out as a sort of documentation.
   *
   * LE 15/3 2001
   *
   * Revised PWP 01.11.01 to avoid possibility of out of range array access
   */

  #define MAX_CONF 60

  int conf[MAX_CONF] = {  1,   3,   5,   7,  11,  13,  15,  21,  23,  27,
		 	 29,  31,  43,  45,  47,  53,  55,  59,  61,  63,
			 85,  87,  91,  93,  95, 107, 109, 111, 117, 119,
			123, 125, 127, 171, 173, 175, 181, 183, 187, 189,
			191, 213, 215, 219, 221, 223, 235, 237, 239, 245, 
			247, 251, 253, 255,   0,   0,   0,   0,   0,   0}; 

  int confno = -1;

  for(int i=0;i<MAX_CONF;i++){
    if(conf[i]==token){
      confno  = i;
      break; 
    }  
  }
  return confno;

  /*
  unsigned conf[MAX_CONF];
  int confno = -1;

  for(int i=0;i<MAX_CONF;i++){
    conf[i] = 0;
  }

  for(int ichip=1;ichip<=8;ichip++){
    if(ichip==1){
      conf[0] = 0x1;
    } else if(ichip==2){
      conf[1] = 0x3;
    } else{
      for(int i=0;i<Fib(ichip-2);i++){
	conf[(int)SumFib(ichip-1)+i] = conf[(int)SumFib(ichip-3)+i] + (0x1 << (ichip-1));
      }
      for(int i=0;i<Fib(ichip-1);i++){
	conf[(int)SumFib(ichip-1)+(int)Fib(ichip-2)+i] = conf[(int)SumFib(ichip-2)+i] + (0x1 << (ichip-1));
      }
    }
  }
  */

}

Int_t GetConf(Int_t confno){

  /* Returns the value that corresponds to the read out pattern 
   * given for variable ST_TOKEN == confno. Chips included in the read 
   * out are considered to be 1's and chips not included 
   * are considered to be 0's. This binary bit-pattern is converted
   * into a value 'token'. This function returns that value corresponding
   * to the read out configuration where ST_TOKEN == confno.  
   *
   * This code is implemented as a look-up table to gain speed
   * perfomance. But I have keept the algorithm coded but commented 
   * out as a sort of documentation.
   *
   * LE 15/3 2001
   *
   * Revised PWP 01.11.01 to avoid possibility of out of range array access
   */

  #define MAX_CONF 60
 
  int conf[MAX_CONF] = {  1,   3,   5,   7,  11,  13,  15,  21,  23,  27,
		 	 29,  31,  43,  45,  47,  53,  55,  59,  61,  63,
			 85,  87,  91,  93,  95, 107, 109, 111, 117, 119,
			123, 125, 127, 171, 173, 175, 181, 183, 187, 189,
			191, 213, 215, 219, 221, 223, 235, 237, 239, 245, 
			247, 251, 253, 255,   0,   0,   0,   0,   0,   0}; 


  if(confno<0) return -1;
  else if(confno>=MAX_CONF) return -2;

  return conf[confno];

  /*
  unsigned int conf[MAX_CONF];

  for(int i=0;i<MAX_CONF;i++){
    conf[i] = 0;
  }

  for(int ichip=1;ichip<=8;ichip++){
    if(ichip==1){
      conf[0] = 0x1;
    } else if(ichip==2){
      conf[1] = 0x3;
    } else{
      for(int i=0;i<Fib(ichip-2);i++){
	conf[(int)SumFib(ichip-1)+i] = conf[(int)SumFib(ichip-3)+i] + (0x1 << (ichip-1));
      }
      for(int i=0;i<Fib(ichip-1);i++){
	conf[(int)SumFib(ichip-1)+(int)Fib(ichip-2)+i] = conf[(int)SumFib(ichip-2)+i] + (0x1 << (ichip-1));
      }
    }
  }
  */
}


Int_t Fib(Int_t n){
  // Return the Fibonacci number n
  // new version by PWP 01.11.01

  Int_t i = 1;
  Int_t m = 1;
  Int_t m1 = 1;
  Int_t m2 = 0;

  while(i<n){
    m = m1 + m2;
    m2 = m1;
    m1 = m;
    i++;
  }

  return m;
}


Int_t SumFib(Int_t n){
  // Return the sum of all Fibonacci numbers up to n
  // new version by PWP 01.11.01

  Int_t i = 1;
  Int_t m = 1;
  Int_t m1 = 1;
  Int_t m2 = 0;
  Int_t s  = 1;

  if(n<1) return 0;

  while(i<n){
    m = m1 + m2;
    m2 = m1;
    m1 = m;
    s+=m;
    i++;
  }

  return s;
}


void FullBypassAnalysis(Int_t run, Int_t scan, Int_t moi){

  Int_t n_scans = 6;
  Int_t n_configs = 36;
  Int_t run_array[20];
  Int_t scan_array[20];
  Float_t vdd_array[20];

  for(int i=0;i<n_scans;i++){
    vdd_array[i] = 4.0-i*0.1;
    scan_array[i] = scan+i;
    run_array[i] = run;
  }

  FullBypassAnalysis(run_array, scan_array, moi, vdd_array, n_scans, n_configs);
}


