void IVCurve(Float_t Vstart, Float_t Vstop, Float_t Vstep, Float_t Vend, 
             Float_t Imax, Float_t Twait, Int_t graphics,Int_t postscript){

  // Ramp voltage from Vstart to Vstop volts with
  // step size Vstep. At each step:
  //  1) wait Twait ms
  //  2) record current
  //
  // If at any point idet exceeds Imax, the voltage 
  // will not be raised any further.  Note that this 
  // limit is enforced purely in software: there is
  // also a hardware limit, read from the detector 
  // conifguration file: it is envisaged that this 
  // would usually have been set somewhat higher.
  //
  // At the end of the measurement, the voltage is 
  // set to Vend, excepting those modules for 
  // which the current Imax had been exceeded at
  // a voltage lower than Vend.
  //
  // Got all that? Feels like an episode of "Soap"...
  //
  // version 0.10 PWP 05.07.01
  // version 0.11 PWP 05.07.01 - tested and working with one module
  // version 0.12 PWP 21.08.01 - minor revisions
  // version 0.13 PWP 23.08.01 - database compliance
  // version 0.13 PWP 24.08.01 - write Vcc,Icc,Vdd,Idd to results file
  // version 0.14 PWP 12.09.01 - further database compliance
  // version 0.15 PWP 04.10.01 - user to choose between 150V, 350V and 500V
  // version 0.20 PWP 15.11.01 - new DB format
  // version 0.21 PWP 16.01.02 - placeholder for TIME_POWERED field
  // version 0.22 PWP 19.02.02 - upgrade default voltage to 200V
  // version 0.22 PWP 06.03.02 - extra option to ramp to 500V fast, then 
  //                             ramp down and switch off. Also reduce 
  //                             wait time on fast ramps from 10s to 5s
  // version 0.23 PWP 07.03.02 - name iv.txt file with time as well as date
  // version 0.24 PWP 14.03.02 - output I at 500V, but prefixed with # character
  //                             so as not to interfere with DB upload
  // version 0.25 PWP 21.03.02 - add web link to postscript plot if WEB_KEY defined
  //                             place leading zero in PS file timestamp
  // version 0.26 PWP 27.03.02 - extend f_name char array to 128 in length
  //                             to accomodate long web link filenames
  // version 0.27 BJG 30.04.02 - test failure on IV_TRIP defect
  // version 0.28 PWP 16.07.02 - add stopwatch timer and revise test options
  // version 0.29 PWP 21.08.02 - change "%ModIVScan" tag to "%DetModIV"
  // version 0.30 PWP 19.09.02 - add extra options for fast IV curves to 200 and 350 volts;
  //                             show hybrid temperatures and device type on plots;
  //                             add use of st stop facility - about turns off HV,
  //                             stop proceeds directly to requested end point
  // version 0.31 PWP 23.09.02 - bugfixes; make axis titles appear (finally!)
  // version 0.32 PWP 14.10.02 - no need to list each defect 12 times;
  //                             generate DB comments in the case that a module
  //                             trips or current limit is reached;
  //                             record current as 99999999 if current limit or trip reached

  #ifdef DATA_DIRECTORY
    // do not reload parameters.h
  #else
    // must load parameters.h
  #include <parameters.h>
  #endif

  #define VSAFETY 500
  #define ISAFETY 100
  #define TSAFETY 100

  #define MAX_VPOINTS 64

  Int_t link, chip;
  char devtype[MAX_MODULES][32];


  // Load other files needed by this one
  gROOT->LoadMacro("ToolBox.cpp");

  //
  // Sanity Checks
  //

  if((Vstart<0)||(Vstop<0)||(Vend<0)){
    printf("Detector bias cannot be negative - exiting\n");
    return;
  }

  if((Vstart>VSAFETY)||(Vstop>VSAFETY)||(Vend>VSAFETY)){
    printf("Detector bias cannot be greater than the safety limit (%d) - exiting\n");
    return;
  }

  if(Imax<0){
    printf("Imax (%d) cannot be a negative number - exiting\n",Imax);
    return;
  }

  if(Imax>ISAFETY){
    printf("Imax (%d) cannot be greater than the safety limit (%d) - exiting\n",
      Imax,ISAFETY);
    return;
  }

  if(Twait<TSAFETY){
    printf("Twait (%d) cannot be less than the safety limit (%d) - exiting\n",
      Twait,TSAFETY);
    return;
  }


  // Start the Stopwatch
  TStopwatch* patek_phillipe = new TStopwatch();

  //
  // Clear Arrays
  //

  Short_t module, point, n_points;

  Short_t keypoint[3];
  Short_t ilimpoint[MAX_MODULES];
  Short_t itrippoint[MAX_MODULES];

  Float_t Vset[MAX_VPOINTS];
  Float_t Vmon[MAX_MODULES][MAX_VPOINTS];
  Float_t Imon[MAX_MODULES][MAX_VPOINTS];

  Short_t status[MAX_MODULES];

  Short_t module_passed[MAX_MODULES];

  for(module=0;module<MAX_MODULES;module++){
    for(point=0;point<MAX_VPOINTS;point++){
      Imon[module][point]=-1;
      Vmon[module][point]=-1;
    }

    status[module]=0;
    ilimpoint[module]=-1;
    itrippoint[module]=-1;
    module_passed[module] = 1;

    e->GetDeviceType(module,devtype[module]);
  }

  for(point=0;point<3;point++){
    keypoint[point]=-1;
  }

  //
  // Calculation of set points
  //

  if(Vstep<0) Vstep = (float) -1.0 * Vstep;

  if(Vstart>Vstop){
    n_points = 1 + abs((int)(((Vstart-Vstop)+(Vstep/2))/Vstep));
    if(n_points>MAX_VPOINTS) n_points = MAX_VPOINTS;
  }else{
    n_points = 1 + abs((int)(((Vstop-Vstart)+(Vstep/2))/Vstep));
    if(n_points>MAX_VPOINTS) n_points = MAX_VPOINTS;
  }

  for(point=0;point<n_points;point++){
    if(Vstart>Vstop){
      Vset[point] = Vstart - (Vstep*point);
    }else{
      Vset[point] = Vstart + (Vstep*point);
    }
    if(Vset[point] == 150){
      keypoint[0]=point;
    }
    if(Vset[point] == 350){
      keypoint[1]=point;
    }
    if(Vset[point] == 500){
      keypoint[2]=point;
    }
  }


  //
  // Record IV curve
  //

  point=0;
  Short_t running=1;
  Int_t stop_loop=0;

  st_start_stopper(); // PWP 19.09.2002

  while(running){

    printf("IVCurve: point %d (%3.0fV)\n",point,Vset[point]);

    // Set desired Voltage
    for(module=0;module<MAX_MODULES;module++){
      if(status[module]==0){
        e->ConfigureVariable(module,30,Vset[point]);
      }
    }

    // Send Request
    e->HVRamp(0);

    // Wait Twait ms
    printf("Waiting %dmS for currents to settle\n",Twait);
    gSystem->Sleep(Twait);


    // Record Currents
    e->DCSQuery();

    for(module=0;module<MAX_MODULES;module++){
      if(module<e->GetMaxModules()){
        Vmon[module][point]=e->m[module]->DCS->vdet;
        Imon[module][point]=e->m[module]->DCS->idet;

        // check for trips
        if( (itrippoint[module]==-1)&&
            ((int)e->m[module]->DCS->HVstatus&0xC0) ){
          itrippoint[module]=point;
          module_passed[module]=0;
        }
      }

      if((Imon[module][point]>Imax)&&(status[module]==0)){
        printf("Module %d has exceeded Imax(%4.3fmicroA): V %3.0fV, I %4.3fmicroA\n",
          module, Imax, Vmon[module][point],Imon[module][point]);
        status[module]=1;
        ilimpoint[module]=point;
      }
    }

    point++;

    if((point>=n_points)||(point>=MAX_VPOINTS)){
      running=0;
    }

    stop_loop=st_do_stop();
    if(stop_loop){
      running=0;
    }
  }

  st_stop_stopper(); // PWP 19.09.02

  // 
  // Actions at end of IVCurve test
  //

  if(stop_loop){
    if(e->abort==1){
      // ABORT: ramp down to zero and exit 
      printf("\nABORT SIGNAL RECEIVED - ramping to zero\n\n");
      Vend = 0;
    }else{
      // STOP: set to requested end point
      printf("\nSTOP SIGNAL RECEIVED - ramping to requested endpoint, %3.0fV\n\n",Vend);
    }
  }

  //
  // Set desired endpoint voltage
  //

  for(module=0;module<MAX_MODULES;module++){
    if(status[module]==0){
      e->ConfigureVariable(module,30,Vend);
    }else if(Vmon[module][point-1]>Vend){
      e->ConfigureVariable(module,30,Vend);
    }
  }

  if(Vend==0){
    // Ramp down to zero and switch off
    e->HVRamp(1);
  }else{
    // Ramp as requested
    e->HVRamp(0);
  }


  // Save results to the database file
  // NB. Cannot use Toolbox since there is no scan file!!

  // Find today's date - needed for outfile output
  TDatime *dt = new TDatime;
  Int_t date = dt->GetDate();
  Int_t time = dt->GetTime();

  Char_t f_name[128];
  Char_t mname[32];

  FILE* outfile;

  for(module=0;module<MAX_MODULES;module++){
    if(module<e->GetMaxModules()){
      if(e->m[module]->present){

        // Open File

        e->GetModuleName(module,mname);
        sprintf(f_name,"%s%s_%d.txt",RESULTS_DIRECTORY,mname,date);
        outfile = fopen(f_name,"a+");
  
        if(outfile==NULL){
          printf("Could not open result outfile %s\n",f_name);
        } else {

          printf("File %s opened to append results\n",f_name);
          printf("     for IVCurve test\n\n");

          // Write Header

          fprintf(outfile,"#\n");
          fprintf(outfile,"%%NewTest\n");
          fprintf(outfile,"#\n");

          fprintf(outfile,"SERIAL NUMBER  : %s\n",mname);
          fprintf(outfile,"TEST MADE BY   : %s\n",gSystem->Getenv("SCTDB_USER"));
          fprintf(outfile,"LOCATION NAME  : %s\n",LOCATION_NAME);
          fprintf(outfile,"Run number     : %d\n",e->runnum); // no scan number!!

          Int_t date2,mon,year,hour,min,sec;

          // ROOT date format is YYYYMMDD
          year = date/10000;
          mon  = (date/100)%100;
          date2 = date%100;

          // ROOT time format is HHMMSS
          hour = time/10000;
          min  = (time/100)%100;
          sec  = time%100;

          fprintf(outfile,"TEST_DATE      : %02d/%02d/%04d\n",date2,mon,year);

          fprintf(outfile,"PASSED         : ");
          if(module_passed[module])
            fprintf(outfile,"YES\n");
          else
            fprintf(outfile,"NO\n");

          fprintf(outfile,"PROBLEM        : ");
          fprintf(outfile,"NO\n");


          fprintf(outfile,"#\n");
          fprintf(outfile,"%%DAQ_INFO\n");
          fprintf(outfile,"#\n");

          fprintf(outfile,"#HOST\n");
          fprintf(outfile,"\"%s\"\n",gSystem->Getenv("COMPUTERNAME"));

          fprintf(outfile,"#VERSION\n");
          fprintf(outfile,"\"%s\"\n",ANALYSIS_VERSION);

          fprintf(outfile,"#DUT\n");
          if(strlen(devtype[module]) > 0){ 
            fprintf(outfile,"\"%s\"\n",devtype[module]);
          }else{
            fprintf(outfile,".\n");
          }

          fprintf(outfile,"#TIME\n");
          fprintf(outfile,"\"%02d:%02d:%02d\"\n",hour,min,sec);


          fprintf(outfile,"#\n");
          fprintf(outfile,"%%DCS_INFO\n");
          fprintf(outfile,"#\n");

          fprintf(outfile,"#T0\tT1\n");
          fprintf(outfile,"%2.1f \t%2.1f\n",e->m[module]->DCS->T1,  e->m[module]->DCS->T2);

          fprintf(outfile,"#VDET\tIDET\n");
          fprintf(outfile,"%3.0f \t%1.3f\n",e->m[module]->DCS->vdet,e->m[module]->DCS->idet);

          fprintf(outfile,"#VCC\tICC\n");
          fprintf(outfile,"%1.2f \t%4.0f\n",e->m[module]->DCS->vcc, e->m[module]->DCS->icc);

          fprintf(outfile,"#VDD\tIDD\n");
          fprintf(outfile,"%1.2f \t%4.0f\n",e->m[module]->DCS->vdd, e->m[module]->DCS->idd);

          fprintf(outfile,"#TIME_POWERED\n");
          fprintf(outfile,".\n");


          fprintf(outfile,"#\n");
          fprintf(outfile,"%%DetModIV\n"); // PWP 21.08.02
          fprintf(outfile,"#\n");
          fprintf(outfile,"TEMPERATURE    : %2.1f\n",e->m[module]->DCS->T1);


          // Currents at 150V and 350V are mandatory
          // - record 99999999 microA if HV has tripped at this point
          //   or limited at an earlier point

          if((keypoint[0]>=0) && 
             ((ilimpoint[module]==-1)||(ilimpoint[module]>=keypoint[0])) &&
             ((itrippoint[module]==-1)||(itrippoint[module]>keypoint[0])) ){
            fprintf(outfile,"I LEAK 150     : %4.3f\n",Imon[module][keypoint[0]]);
          }else{
            fprintf(outfile,"I LEAK 150     : 99999999\n");
          }

          if((keypoint[1]>=0) && 
             ((ilimpoint[module]==-1)||(ilimpoint[module]>=keypoint[1])) &&
             ((itrippoint[module]==-1)||(itrippoint[module]>keypoint[1])) ){
            fprintf(outfile,"I LEAK 350     : %4.3f\n",Imon[module][keypoint[1]]);
          }else{
            fprintf(outfile,"I LEAK 350     : 99999999\n");
          }

          if((keypoint[2]>=0) && 
             ((ilimpoint[module]==-1)||(ilimpoint[module]>=keypoint[2])) &&
             ((itrippoint[module]==-1)||(itrippoint[module]>keypoint[2])) ){
            fprintf(outfile,"#I LEAK 500     : %4.3f\n",Imon[module][keypoint[2]]);
          }else{
            fprintf(outfile,"#I LEAK 500     : 99999999\n");
          }


          // Minor bugfix to next section PWP 14.10.02
          // We do not need to list each defect 12 times!
          Int_t n_defects=0;
          if(ilimpoint[module]>=0){
            ListDefect(outfile,"IV_LIMIT",0,1535);
            n_defects++;
          }
          if(itrippoint[module]>=0){
            ListDefect(outfile,"IV_TRIP",0,1535);
            n_defects++;
          }

          if(n_defects==0){
            fprintf(outfile,"#No defects found!\n");
          }else{
            fprintf(outfile,"#%d defects found\n",n_defects);
          }

          sprintf(f_name,"%s%s_iv_%d_%d.txt",RESULTS_DIRECTORY,mname,date,time);

          fprintf(outfile,"#\n");
          fprintf(outfile,"%%TEST Rawdata\n");
          fprintf(outfile,"FILENAME       : %s\n",f_name);


          // Add DB comments if current limit reached or trip level exceeded

          if(ilimpoint[module]>=0){
            sprintf(line, "HV Current Limit at %3.0fV %4.3fmicroA\n",
              Vmon[module][ilimpoint[module]], Imon[module][ilimpoint[module]]);
            ListComment(outfile,line);
          }
          if(itrippoint[module]>=0){
            sprintf(line, "HV Trip at %3.0fV %4.3fmicroA\n",
              Vmon[module][itrippoint[module]], Imon[module][itrippoint[module]]);
            ListComment(outfile,line);
          }


          // Add a URL to the postscript file

  #ifdef WEB_KEY        
          if(postscript){
            sprintf(f_name,"%s%s_IVCurve_%d_%06d.ps",WEB_KEY,mname,date,time);
            ListWebLink(outfile,"Plots (postscript)",f_name);
          }
  #endif

          fclose(outfile);

          // Record IV info
          sprintf(f_name,"%s%s_iv_%d_%d.txt",RESULTS_DIRECTORY,mname,date,time);
          outfile = fopen(f_name,"a+");

          fprintf(outfile,"TEST_DATE      : %02d/%02d/%04d\n",date2,mon,year);
          fprintf(outfile,"TEST_TIME      : %02d:%02d:%02d\n",hour,min,sec);
          fprintf(outfile,"TEMPERATURE    : %2.1f\n",e->m[module]->DCS->T1);
          fprintf(outfile,"#\n");

          fprintf(outfile,"#V\tI\n");
          for(point=0;point<n_points;point++){
            fprintf(outfile,"%3.0f\t%4.3f\n",
              Vmon[module][point],Imon[module][point]);
          }
          fprintf(outfile,"#V\tmicroA\n");
          fprintf(outfile,"#\n");

          fclose(outfile);
        } // end if(outfile)

      } // end if(present)
    }
  }

  //
  // Graphics and Postscript
  //

  Int_t dowait=0;

  TCanvas* civ =0;
  TGraph* gr = 0;
  TPostScript* ps =0;
  Char_t line[256];

  if(graphics!=0){
    dowait=1;
  }

  if((graphics!=0)||(postscript!=0)){
    graphics = 1;
    civ= new TCanvas ("civ","IV",1100,750);
  }

    // Open the postscript file if canvas is open
    // and valid data exists for the module

  if(civ){
    for(module=0;module<MAX_MODULES;module++){
      if(module<e->GetMaxModules()){
        if(e->m[module]->present){

          if(gr) gr->Delete();

          e->GetModuleName(module,mname);
          if(postscript){
          
            sprintf(f_name,"%s%s_IVCurve_%d_%06d.ps",PS_DIRECTORY,mname,date,time);
            ps = new TPostScript(f_name,112);

            if(ps){
              printf("File %s opened for PS output\n",f_name);
            } else {
              printf("ERROR: failed to open PS file %s\n",f_name);
            }
          }

          civ->cd();
          civ->Clear();

          TPad *pad_text   = new TPad("pad_text", "Text",        0.00,0.90,1.00,1.00,33);
          TPad *pad_plot   = new TPad("pad_plot", "Plots",       0.00,0.00,1.00,0.90,33);

          pad_text->Draw();
          pad_plot->Draw();
          pad_text->cd();

          TPaveText *t1 = new TPaveText(0.0,0.0,1.0,1.0);
          t1->SetFillColor(10);
          t1->SetTextAlign(13);
          t1->SetTextSize(0.3);
          t1->SetTextColor(4);
          sprintf(line,"ATLAS SCT Module Test:  IV Characteristic");
          sprintf(line,"%s  -  %s",line,dt->AsString());
          //if(place) sprintf(line,"%s  -  %s",line,place->GetName());
          t1->AddText(line);
          sprintf(line,"Run %d   Module %d   %s", e->runnum, module, mname);
          if(strlen(devtype[module]) > 0){ 
            sprintf(line, "%s   Type %s", line, devtype[module]);
          }
          sprintf(line, "%s   T0 %2.1fC   T1 %2.1fC", line, e->m[module]->DCS->T1,  e->m[module]->DCS->T2);
          t1->AddText(line);
          t1->Draw();

          pad_plot->cd();

          gr = new TGraph(n_points,Vmon[module],Imon[module]);

          if(gr){
            gr->SetTitle("IV Curve");
            gr->SetMarkerColor(kBlue);
            gr->SetMarkerSize(1.0);
            gr->SetMarkerStyle(20);
            gr->Draw("ALP");
            gr->GetXaxis()->SetTitle("Bias Voltage (V)");
            gr->GetYaxis()->SetTitle("Current (microA)");
            gPad->Modified(); // How to make the axis titles appear!
          }

          civ->Update();

          if(ps){
            ps->Close();
            delete ps;
          }

          if(dowait){
            printf("Enter a string to continue\n");
            scanf("%s",line);
          }

        } // end if(present)
      }
    } // end for(modules)
  } // end if(canvas)

  printf("IVCurve completed\n");
  patek_phillipe->Print();
  delete patek_phillipe;
}

void IVCurve(Float_t Vstop, Float_t Vend){

  // IV curve from 10 to Vstop in 10V steps
  // postscript output enabled
  // set volts to Vend when done

  IVCurve(10,Vstop,10,Vend,10,10000,0,1);
}

void IVCurve(Float_t Vstop){

  // IV curve from 10 to Vstop in 10V steps
  // postscript output enabled
  // leave volts at 200V when done

  IVCurve(10,Vstop,10,200,10,10000,0,1);
}

void IVCurve(){

  // IV curve to 200, 350 or 500V 
  // as selected by the user
  // leave volts at 200V when done
  // PWP 04.10.01

  Short_t x = -1;

  while((x<0)||(x>8)){
    printf("Record IVCurve to which upper voltage:\n");
    printf("  0 - don't bother\n");
    printf("  1 - 200V in 10V steps, 10s wait, end at 200V\n");
    printf("  2 - 350V in 10V steps, 10s wait, end to 200V\n");
    printf("  3 - 500V in 10V steps, 10s wait, end to 200V (estimate 13min 40sec)\n");
    printf("  4 - 500V in 25V steps, 08s wait, end at 200V (estimate 04min 25sec)\n");
    printf("  5 - 500V in 25V steps, 04s wait, end at 200V (estimate 03min 05sec)\n");
    printf("  6 - 500V in 25V steps, 04s wait, ramp down and SWITCH OFF HV\n");
    printf("  7 - 200V in 25V steps, 04s wait, end at 200V\n");
    printf("  8 - 350V in 25V steps, 04s wait, end to 200V\n");
    printf("      (Time estimates assume that ramp rate 2 is selected.)\n");

    scanf("%d",&x);
  }

  // Float_t Vstart, 
  // Float_t Vstop, 
  // Float_t Vstep, 
  // Float_t Vend, 
  // Float_t Imax,  microA
  // Float_t Twait,
  // Int_t graphics
  // Int_t postscript

  switch(x){
    case 1:
      printf("IVCurve to 200V in 10V steps\n");
      IVCurve(10,200,10,200,10,10000,0,1);
      break;
    case 2:
      printf("IVCurve to 350V in 10V steps\n");
      IVCurve(10,350,10,200,10,10000,0,1);
      break;
    case 3:
      printf("IVCurve to 500V in 10V steps\n");
      IVCurve(10,500,10,200,10,10000,0,1);
      break;
    case 4:
      printf("IVCurve to 500V in 25V steps, 8sec wait\n");
      IVCurve(25,500,25,200,10,8000,0,1);
      break;
    case 5:
      printf("IVCurve to 500V in 25V steps, 4sec wait\n");
      IVCurve(25,500,25,200,10,4000,0,1);
      break;
    case 6:
      printf("IVCurve to 500V in 25V steps, then ramp down and switch off HV\n");
      IVCurve(25,500,25,0,10,4000,0,1);
      break;
    case 7:
      printf("IVCurve to 200V in 25V steps\n");
      IVCurve(25,200,25,200,10,4000,0,1);
      break;
    case 8:
      printf("IVCurve to 350V in 25V steps\n");
      IVCurve(25,350,25,200,10,4000,0,1);
      break;
    default:
      printf("IVCurve: exit without doing anything!!\n");
      break;
  }
}
