void KwikFit(int run, int scan, int moi, int graphics, int dowait, int postscript){

  //
  //  A routine to fit st scan histograms - FAST!!!
  //  The fit routine used is based upon the integrated fitting within
  //  stdll, but compiled into a standalone DLL (KwikFit.dll)
  //
  // The fitted data is saved back into the root file to be compatible
  // with the on line fitting.  If online fit data is present in the
  // file, the exisiting fit data is used instead!!
  //
  //  A reduced set of cuts have been implemented as follows:
  //  o Channels with <1% occupancy in any bin are marked as "DEAD"
  //  o Channels with >98% occupancy in every bin are marked "STUCK"
  //  o Channels which do not reach 100% efficiency are marked "UNDER"
  //  o Channels which reach > 100% efficiency are marked "OVER"
  //    (Such a condition would imply decoder errors, inviting the
  //     user to check the digital integrity of their system)
  //
  //  Version 0.1, PWP 26.10.00
  //  Version 0.2, LE  23.05.01
  //  Version 0.3, BJG 30.04.02 - Fix some memory leaks
  //


  #ifdef DATA_DIRECTORY
    // do not reload parameters.h
  #else
    // must load parameters.h
  #include <parameters.h>
  #endif

  char  f_name[128], out_fname[128];
  char  h_name[64], h_title[64];
  char  scan_name[64];
  char  s[256], unit[32];
  short scan_type, fit_type;


  printf("KwikFit called for run %d scan %d module %d\n",run,scan,moi);

  //
  // Load other files needed by this one
  //

  gSystem->Load("KwikFit.dll");
  gROOT->LoadMacro("stan_util.cpp");

  // Global style variations
  stan_styles();
  gStyle->SetOptStat(0);


  //
  // Open the root data file and obtain pointers to the event histogram
  //

  sprintf(f_name,"%sstrun%d_%d.root",DATA_DIRECTORY,run,scan);

  TFile *strun = new TFile(f_name,"UPDATE");
  TH1F *h_nevent= (TH1F*)strun->Get("h_scan_evcnt");

  //
  // Open the output canvas (and postscript file if requested)
  //

  Int_t ps_page=0;
  Int_t screen_page=0;

  TCanvas* c1 =0;
  TPostScript* ps =0;
  TPad *pad_t = 0;
  TPad *pad_p = 0;

  if((graphics!=0)||(postscript!=0)){
    c1=new TCanvas ("c1","KwikFit",1100,750);
  }

  if(postscript){
    ps_page=0;

    sprintf(out_fname,"%sstrun%d_%d_%d.ps",PS_DIRECTORY,run,scan,moi);

    ps = new TPostScript(out_fname,112);
  }

  TObjString *time  = 0;
  TObjString *place = 0;
  TObjString *mname = 0;

//
//  loop over modules here
//

  Int_t start = 0;
  Int_t end   = 0;

  if(moi==-1){
    start = 0;
    end   = MAX_MODULES;

  } else {
    start = moi;
    end   = moi+1;
  }


  Int_t last_good = -1;

  for(moi=start;moi<end;moi++){

    if((screen_page>0)&&(dowait)){
      printf("Enter a string to continue\n");
      scanf("%s",s);
    }

    ps_page++;
    screen_page++;
    if(ps){
      ps->NewPage();
    }

    sprintf(h_name,"ModuleName;%d",moi+1);
    if(time) delete time;
    if(place) delete place;
    if(mname) delete mname;
    time  = (TObjString*)strun->Get("Time");
    place = (TObjString*)strun->Get("Place");
    mname = (TObjString*)strun->Get(h_name);

    sprintf(h_name,"h_scan0;%d",moi+1);
    TH2F *h_scan0= (TH2F*)strun->Get(h_name);
    sprintf(h_name,"h_scan1;%d",moi+1);
    TH2F *h_scan1= (TH2F*)strun->Get(h_name);

    sprintf(h_name,"h_mean0;%d",moi+1);
    TH2F *h_mean0= (TH2F*)strun->Get(h_name);
    sprintf(h_name,"h_mean1;%d",moi+1);
    TH2F *h_mean1= (TH2F*)strun->Get(h_name);

    sprintf(h_name,"h_sigma0;%d",moi+1);
    TH2F *h_sigma0= (TH2F*)strun->Get(h_name);
    sprintf(h_name,"h_sigma1;%d",moi+1);
    TH2F *h_sigma1= (TH2F*)strun->Get(h_name);

    sprintf(h_name,"h_code0;%d",moi+1);
    TH2F *h_code0= (TH2F*)strun->Get(h_name);
    sprintf(h_name,"h_code1;%d",moi+1);
    TH2F *h_code1= (TH2F*)strun->Get(h_name);

    TKwikFit *kf = 0;

    //
    // only attempt plotting or fitting if data exists for the module
    //

    if((h_scan0!=0)||(h_scan1!=0)){

      last_good = moi;

      sprintf(scan_name,"Unknown");
      sprintf(unit,"");
      TString title = h_scan0->GetTitle();
      title.ToUpper();
      if (title.Contains("THRESHOLD")){
        fit_type = 1;
        scan_type = 1;
        sprintf(scan_name,"THRESHOLD (mV)");
        sprintf(unit,"mV");
      }
      if (title.Contains("VCAL")){
        fit_type = 2;
        scan_type = 2;
        sprintf(scan_name,"VCAL (mV)");
        sprintf(unit,"mV");
      }
      if (title.Contains("STROBE")){
        fit_type = 1;
        scan_type = 3;
        sprintf(scan_name,"Strobe Delay");
        sprintf(unit,"");
      }
      if (title.Contains("QTHR")){
        fit_type = 1;
        scan_type = 41;
        sprintf(scan_name,"QTHR (fC)");
        sprintf(unit,"fC");
      }
      if (title.Contains("QCAL")){
        fit_type = 2;
        scan_type = 42;
        sprintf(scan_name,"QCAL (fC)");
        sprintf(unit,"fC");
      }


      if(c1){
        c1->cd();
        c1->Clear();
        pad_t = new TPad("pad_t","Text",   0.00,0.90,1.00,1.00,33);
        pad_p = new TPad("pad_p","Plots",  0.00,0.00,1.00,0.90,33);

        pad_t->Draw();
        pad_p->Draw();
        pad_p->Divide(2,3,0,0);

        pad_t->cd();

        TPaveText *t1 = new TPaveText(0.0,0.0,1.0,1.0);
        t1->SetFillColor(10);
        t1->SetTextAlign(13);
        t1->SetTextSize(0.3);
        t1->SetTextColor(4);
  
        sprintf(s,"ATLAS SCT Module Test");
        if(time)  sprintf(s,"%s  -  %s",s,time->GetName());
        if(place) sprintf(s,"%s  -  %s",s,place->GetName());
        if(mname) sprintf(s,"%s  -  Module %s",s,mname->GetName());
        t1->AddText(s);
        sprintf(s,"Run %d   Scan %d   Module %d   Scan type: %s", run, scan, moi, scan_name);
        t1->AddText(s);

        t1->Draw();

        pad_p->cd(1);
        h_scan0->Draw("COLZ");
        h_scan0->SetXTitle("Channel Number");
        pad_p->cd(2);
        h_scan1->Draw("COLZ");
        h_scan1->SetXTitle("Channel Number");

        c1->Update();
      }

      //
      // Fit the data if it wasn't already done
      //

      if(h_mean0==0){
        kf = new TKwikFit(moi);
        if(mname) kf->SetModuleName(mname->GetName());

        kf->LoadScan(h_scan0,h_scan1,h_nevent);
        kf->PrintScan();

        printf("\nFitting Data for module %d...\n",moi);
        kf->FitScan(fit_type,-1);
        kf->PrintScan();

        sprintf(h_title,"Sigma (%s)",scan_name);
        kf->h_sigma0->SetMarkerColor(kBlue);
        kf->h_sigma0->SetXTitle("Channel Number");
        kf->h_sigma0->SetYTitle(h_title);
        kf->h_sigma1->SetMarkerColor(kBlue);
        kf->h_sigma1->SetXTitle("Channel Number");
        kf->h_sigma1->SetYTitle(h_title);

        sprintf(h_title,"Fit Code");
        kf->h_code0->SetMarkerColor(kBlue);
        kf->h_code0->SetXTitle("Channel Number");
        kf->h_code0->SetYTitle(h_title);
        kf->h_code1->SetMarkerColor(kBlue);
        kf->h_code1->SetXTitle("Channel Number");
        kf->h_code1->SetYTitle(h_title);

        sprintf(h_title,"Chisq/NDF");
        kf->h_chisq0->SetMarkerColor(kBlue);
        kf->h_chisq0->SetXTitle("Channel Number");
        kf->h_chisq0->SetYTitle(h_title);
        kf->h_chisq1->SetMarkerColor(kBlue);
        kf->h_chisq1->SetXTitle("Channel Number");
        kf->h_chisq1->SetYTitle(h_title);

      } else {
        printf("\nFitted Data exists for module %d...\n",moi);
      }


      if(c1){
        if(h_mean0==0){

          pad_p->cd(1);
          kf->h_mean0->Draw("same");
          pad_p->cd(2);
          kf->h_mean1->Draw("same");

          pad_p->cd(3);
          kf->h_sigma0->Draw("P");
          pad_p->cd(4);
          kf->h_sigma1->Draw("P");

          pad_p->cd(5);
          kf->h_code0->Draw("P");
          pad_p->cd(6);
          kf->h_code1->Draw("P");

        } else {

          pad_p->cd(1);
          h_mean0->Draw("same");
          pad_p->cd(2);
          h_mean1->Draw("same");

          pad_p->cd(3);
          h_sigma0->Draw("P");
          pad_p->cd(4);
          h_sigma1->Draw("P");

          pad_p->cd(5);
          h_code0->Draw("P");
          pad_p->cd(6);
          h_code1->Draw("P");

        }
        c1->Update();
      }

      //
      // Save the histograms in the root file
      //     - if they were not already there
      //

      if(h_mean0==0){
        kf->h_mean0->Write();
        kf->h_mean1->Write();
        kf->h_sigma0->Write();
        kf->h_sigma1->Write();
        kf->h_code0->Write();
        kf->h_code1->Write();
        // kf->h_chisq0->Write();
        // kf->h_chisq1->Write();

        kf->Delete();
      }
 

    } else {
      printf("\nNo data for module %d\n",moi);
    }
  }

  //
  // Note that the action of closing the root file causes all plots to disappear!!
  //

  if(time) delete time;
  if(place) delete place;
  if(mname) delete mname;

  if(ps) { 
    ps->Close();
    delete ps;
  }
  strun->Close();
  delete strun;
}


//
// Wrappers
//

void KwikFit(int run, int scan){

  //
  // Fit the data for all modules 
  //
  // Output to root file only
  // 

  KwikFit(run,scan,-1,0,0,0);
}

void KwikFit(int run, int scan, int moi){

  //
  // Fit the data for the specified module 
  //
  // Output also to screen and PS file
  // 

  KwikFit(run,scan,moi,1,0,1);
}


void KwikFit(){

  printf("Possible Uses of KwikFit:\n\n");
  printf(".x KwikFit.cpp(run,scan)\n");
  printf(".x KwikFit.cpp(run,scan,module)\n");
  printf(".x KwikFit.cpp(run,scan,module,graphics,dowait,postscript)\n\n");

}

