//
// TLTT.cpp   Class implementation of Long-Term 
//            Tests of Hybrids or Modules
//
// Version 0.1  DGC  17/10/01  First version for release
// Version 0.2  PWP  20/12/01  Fix memory leaks in UpdateDisplay
//                             add code specific to modules
//                             numerous other small changes
// Version 0.21  PWP  18/01/02  More small changes to support module testing
// Version 0.22  PWP  06/02/02  Addition of ALARM in case of HV trip
// Version 0.23  BJG  27/03/02  Additions
// Version 0.24  BJG/PWP 19/04/02  Revise calculation of temp graph bounds to handle -ve numbers
//                                 do not switch off power at end if testing modules
// Version 0.25  BJG  30/04/02  Some memory leak fixes
//                              Fix Idet results
// Version 0.26  BJG/PWP 30/05/02 Revise use of DCSTree to avoid creation of multiple trees
//                                track noise occupancy at 1fC as the test progresses
//                                NB this requires a valid response curve to be held in memory
//                                - so the first point is skipped
// Version 0.27  BJG/PWP 08.07.02 Add IVLTT for use to monitor detector currents as a 
//                                function of time without other electrical measurements
// Version 0.28  PWP 10.07.02  bugfixes
// Version 0.29  DGC 12/07/02  Addition for hybrid cold test
// Version 0.30  BJG 13.08.02  Append data to log file if file already exists
//                             (file is no longer overwritten in the event of a restart)
//                             VDET is now recorded in the log file;
//                             MAX_CURRENT raised to 4.5microA
// Version 0.31  PWP 21.08.02  bugfix re. "%TEST Rawdata" tag
// Version 0.32  BJG 23/09/02  Put alarm recovery in new function
//
// To do: Checking of results of confirmation tests to track failures
//        Correct database output etc
//

#ifdef DATA_DIRECTORY
  // do not reload parameters.h
#else
  // must load parameters.h
#include <parameters.h>
#endif

#include "TLTT.h"

const Float_t max_temp = 39.0 ;
const Float_t max_current = 4.5 ; //microA

// public functions first

// Constructor
// Small is time between DCS updates
// Big is time between confirmation tests
TLTT::TLTT(Int_t small, Int_t big, Int_t total){

  printf("TLTT constructed with args %d %d %d\n",small,big,total) ;

  // Default setup is for normal Hybrid LTT
  testing_modules = 0;
  dcs_only = 0;
  cold_test = 0;

  graph0 = 0;
  graphi0 = 0;
  graphid0 = 0;

  small_interval = small ;
  big_interval = big ;
  total_length = total ;

  number_of_big_loops = total_length / big_interval ;

  iloop_readings=0 ;

  printf("Will run for %d big loops............\n",number_of_big_loops) ;

  // 
  // Ensure that the arrays are cleared to start
  //

  for(Int_t module=0;module<MAX_MODULES;module++){
    sprintf(mname[module]," ");
    e->GetModuleName(module,mname[module]) ;
  }  

  tmax_actual =-999. ;
  tmin_actual = 999. ;
  imax_actual =-999999. ;
  imin_actual = 999999. ;
  idmax_actual =-999999. ;
  idmin_actual = 999999. ;
  nomax_actual =-999999. ;
  nomin_actual = 999999. ;

  start_run = e->runnum ;
  start_scan = e->scannum+1 ;

  failComment[0] = '\0';
}

// testing_modules turns on recording of idet
// dcs_only        turns off conf/char test
// cold_test       on for cold hybrid test

void TLTT::HybridLTT(){
  // PWP 20.12.01
  testing_modules = 0;
  cold_test = 0 ;
  dcs_only = 0;
  DoTest();
}

void TLTT::ModuleLTT(){
  // PWP 20.12.01
  testing_modules = 1;
  cold_test = 0 ;
  dcs_only = 0;
  DoTest();
}

// Record IV Characteristics over time.
void TLTT::IVLTT() {
  testing_modules = 1;
  cold_test = 0 ;
  dcs_only = 1;
  DoTest();
}

void TLTT::HybridColdTest(){
  // DGC 10.07.02
  testing_modules = 0;
  cold_test = 1 ;
  dcs_only = 0 ;
  DoTest();
}

// now private functions

void TLTT::InitialiseTiming() {

  start_dt = new TDatime ;
  
  // Start a stopwatch
  
  if(timer) delete timer;
  timer = new TStopwatch();
  start_time = timer->GetRealTime();

  Int_t start_secs = start_dt->GetTime()%100 ;
  Int_t start_mins = int(start_dt->GetTime()/100)%100 ;
  Int_t start_hour = int(start_dt->GetTime()/10000) ;

  Int_t start_day = start_dt->GetDate()%100 ;
  Int_t start_mon = int(start_dt->GetDate()/100)%100 ;
  Int_t start_year = int(start_dt->GetDate()/10000) ;

  sprintf(start_str,"%02d/%02d/%04d at %02d:%02d:%02d",start_day,start_mon,start_year,start_hour,start_mins,start_secs) ;

  Int_t end_secs = start_secs + total_length ;
  Int_t end_mins = start_mins + int(end_secs/60) ;
  end_secs = end_secs % 60 ;
  Int_t end_hour = start_hour + int(end_mins/60) ;
  end_mins = end_mins % 60 ;
  Int_t end_day = start_day + int(end_hour/24) ;
  end_hour = end_hour % 24 ;
  Int_t end_mon = start_mon ;

  switch(end_mon){
    case 2: // February
      if(end_day>28){
        end_day-=28;
        end_mon++;
      }
      break;

    case 4:  //April
    case 6:  // June
    case 9:  // September
    case 11: // November
      if(end_day>30){
        end_day-=30;
        end_mon++;
      }
      break;

    default: // some boring month!!
      if(end_day>31){
        end_day-=31;
        end_mon++;
      }
      break;
  }

  Int_t end_year = start_year ;
  if (end_mon>12) {
    end_mon-=12 ;
    end_year++ ;
  }

  sprintf(end_str,"%02d/%02d/%04d at %02d:%02d:%02d",end_day,end_mon,end_year,end_hour,end_mins,end_secs) ;

  //
  // Allow 10 LV trips for each hybrid/module
  //

  for(Int_t n=0;n<e->GetMaxModules();n++){
    if(e->m[n]->present){
      e->m[n]->DCS->maxLVtrips =10;
    }
  }
  e->DCSAction();
}

void TLTT::ActOnAlarm() {
  switch(alarm){
  case 1:
    printf("*****************************\n") ;
    printf("*** OVERTEMPERATURE ALARM ***\n") ;
    printf("*** STOPPING RUN NOW!!!!! ***\n") ;
    printf("*****************************\n") ;
    strcpy(failComment, "Test failed due to over temperature");
    break;

  case 2:
    printf("*****************************\n") ;
    printf("*** HVTRIP ALARM          ***\n") ;
    printf("*** STOPPING RUN NOW!!!!! ***\n") ;
    printf("*****************************\n") ;
    strcpy(failComment, "Test failed due to High Voltage Trip");
    break;

  case 3:
    printf("*****************************\n") ;
    printf("*** IV ALARM              ***\n") ;
    printf("*** STOPPING RUN NOW!!!!! ***\n") ;
    printf("*****************************\n") ;
    strcpy(failComment, "Test failed due to High Detector Bias Current");
    break;

  default:
    printf("*****************************\n") ;
    printf("*** UNKNOWN  ALARM        ***\n") ;
    printf("*** STOPPING RUN NOW!!!!! ***\n") ;
    printf("*****************************\n") ;
    sprintf(failComment, "Test failed due to unknown alarm (%d)", alarm);
    break;
  }

  e->LVOff();   // switch off LV
  e->HVRamp(1); // switch off HV
  dcs_outfile->Close();

  WriteSummaryFile();
}

void TLTT::DoTest() {
  // revised PWP 20.12.01
  // revised BJG/PWP 30.05.02

  InitialiseTiming() ;
  OpenLogFiles() ;
  OpenDisplay() ;

  e->abort=0;

  if(dcs_only == 1) {
    e->scan_number++;
  }

  // Main loop

  Int_t big_loop = 0 ;
  Int_t stop_loop = 0;
  Int_t small_loop = 0;

  if(dcs_only==0) {
    // Load the file once only.
    gROOT->ProcessLine(".L ConfirmationTest.cpp");
  }

  while(e->abort==0 && big_loop<number_of_big_loops){
  
    MakeMonitoringMeasurements();

    if(alarm){
      ActOnAlarm();
      return;
    }

    if((e->abort==0) && (dcs_only==0)) {
      TDatime* dt = new TDatime;
      printf("%8.8d %8.8d Time since start: %ld seconds\n",dt->GetDate(),dt->GetTime(),timer->GetRealTime()-start_time);
      printf("Loop %d: Entering Measurement phase\n",big_loop);
      delete dt;
      
      // Make Measurements Here....
      // (Free running triggers will already have been stopped)

      DoPeriodicTests();
    }

    // Monitor currents for up to "big_loop_end" seconds

    if (e->abort==0) {
      
      small_loop = 0;
      stop_loop=0;

      st_start_stopper();

      Double_t big_loop_end = big_interval + (int((timer->GetRealTime()-start_time)/big_interval))*big_interval + start_time;

      while((timer->GetRealTime()<big_loop_end)&&(stop_loop==0)){

        printf("Time since start: %ld seconds\n",timer->GetRealTime()-start_time);
 
        MakeMonitoringMeasurements();

        if(alarm){
          ActOnAlarm();
          return;
        }


        if (e->abort==0) {
          printf("Loop %d_%d: Entering trigger phase\n",big_loop,small_loop);

          e->DCSAction();        // Recover LV trips
          e->ExecuteConfigs();   // reinitialise modules
          gSystem->Sleep(1000);

          if(dcs_only==0) {
            e->RunFree();          // start triggers
            gSystem->Sleep(1000);
          }
          // wait small interval
    
          Double_t small_loop_end = small_interval + (int((timer->GetRealTime()-start_time)/small_interval))*small_interval + start_time;

          stop_loop=st_do_stop();
          while(stop_loop==0 && timer->GetRealTime()<small_loop_end){
            // DO NOWT!! in a low-CPU way...
            gSystem->Sleep(1000) ;
            stop_loop=st_do_stop();
          }

          small_loop++ ;

          if(dcs_only==0) {
            // Stop free running triggers
            e->RunStop();
            gSystem->Sleep(1000);
          }
        }

      }  // End of big loop
    }  
    st_stop_stopper();
  
    big_loop++;

  } // end of main loop

  if(e->abort==0){

    printf("Final Measurements\n");

    //
    // Make Measurements Here....
    //

    DoFinalTests() ;
  }

  // Record monitored data once more before shutdown
  MakeMonitoringMeasurements();

  if(alarm) {
    ActOnAlarm();
    return;
  }

  // Save file and Power off

  e->HVRamp(1); // switch off HV
  if(testing_modules != 1 && cold_test != 1){
    // do not switch off LV if testing modules, or hybrids cold -
    // we don't want to freeze them unduly!!  BJG/PWP 19.04.02
    //                                 (Cold hybrids: DGC 10.07.02)
    e->LVOff();   // switch off LV
  }
  dcs_outfile->Close();

  WriteSummaryFile() ;
}
  
void TLTT::OpenLogFiles() {

  // Open log files, text (1 file/module) and for DCS root objects
  // Also switch on external temp monitoring

  for(Int_t n=0;n<e->GetMaxModules();n++){
    if(e->m[n]->present){
      sprintf(txt_fname[n],"%s%s_LDT_%d.txt",RESULTS_DIRECTORY,mname[n],start_dt->GetDate());

      FILE *txtfile = fopen(txt_fname[n],"a+");
      fprintf(txtfile, "# Starting new LTT run: %d scan %d\n", e->runnum, e->scannum);
      fclose(txtfile);
    }
  }

  sprintf(dcs_fname,"%sDCSntuple_%d_%d.root",DATA_DIRECTORY,e->runnum,e->scannum);
  sprintf(psfilename,"%slongterm_%d_%d.ps",PS_DIRECTORY,e->runnum,e->scannum) ;

  // File for DCS information (temperatures, currents)

  dcs_outfile = new TFile(dcs_fname,"CREATE");

  // Create a TTree for the DCS information

  DCSTree = new TTree("DCSTree","DCSTree");
  DCSTree->Branch("DCSBranch","TModuleDCS",&dcs,8000,1);
  DCSTree->Branch("NoiseOcc", no_val, "no/F");
  DCSTree->SetAutoSave(1000);  
}
  
void TLTT::OpenDisplay() {

  // Create a canvas to display the monitored DCS values

  Char_t ltttitle[128];

  gStyle->SetHistFillColor(46);
  gStyle->SetTitleW(0.6);
  gStyle->SetTitleH(0.08);

  sprintf(ltttitle,"Long Term Test, Run %d",e->runnum);
  LTTCanvas = new TCanvas("LTTCanvas",ltttitle,1100,750);
  LTTCanvas->SetFillColor(18);

  LTTInfo = new TPad("LTTInfo","LTT Information",0.0,0.9,1.0,1.0) ;
  LTTInfo->Draw() ;
  tpaveinfo = new TPaveText(0.0,0.0,0.5,1.0) ;
  tpaveinfo->SetTextFont(102) ;
  tpaveinfo2 = new TPaveText(0.5,0.0,1.0,1.0) ;
  tpaveinfo2->SetTextFont(102) ;
  LTTCanvas->Draw() ;

  if(testing_modules){
    LTTPadTGraph = new TPad("LTTPadTGraph","LTT Temperature History",0.0,0.675,0.8,0.900);
    LTTPadTGraph->Draw();
    LTTPadTVal = new TPad("LTTPadTVal","LTT Temperatures",0.8,0.675,1.0,0.900);
    LTTPadTVal->Draw();
    tpavet = new TPaveText(0.0,0.0,1.0,1.0) ;
    tpavet->SetTextFont(102) ;

    LTTPadIGraph = new TPad("LTTPadIGraph","LTT Current History",0.0,0.450,0.8,0.675) ;
    LTTPadIGraph->Draw() ;
    LTTPadIVal = new TPad("LTTPadIVal","LTT Currents",0.8,0.450,1.0,0.675) ;
    LTTPadIVal->Draw() ;
    tpavei = new TPaveText(0.0,0.0,1.0,1.0) ;
    tpavei->SetTextFont(102) ;

    LTTPadIdetGraph = new TPad("LTTPadIdetGraph","LTT Detector Current History",0.0,0.225,0.8,0.450) ;
    LTTPadIdetGraph->Draw() ;
    LTTPadIdetVal = new TPad("LTTPadIdetVal","LTT Detector Currents",0.8,0.225,1.0,0.450) ;
    LTTPadIdetVal->Draw() ;
    tpaveid = new TPaveText(0.0,0.0,1.0,1.0) ;
    tpaveid->SetTextFont(102) ;

    LTTPadNOGraph = new TPad("LTTPadNOGraph","LTT Noise Occupancy History",0.0,0.000,0.8,0.225) ;
    LTTPadNOGraph->SetLogy() ;
    LTTPadNOGraph->Draw() ;
    LTTPadNOVal = new TPad("LTTPadNOVal","LTT Noise Occupancy",0.8,0.000,1.0,0.225) ;
    LTTPadNOVal->Draw() ;
    tpaveno = new TPaveText(0.0,0.0,1.0,1.0) ;
    tpaveno->SetTextFont(102) ;
  }else{
    LTTPadTGraph = new TPad("LTTPadTGraph","LTT Temperature History",0.0,0.6,0.8,0.9);
    LTTPadTGraph->Draw();
    LTTPadTVal = new TPad("LTTPadTVal","LTT Temperatures",0.8,0.6,1.0,0.9);
    LTTPadTVal->Draw();
    tpavet = new TPaveText(0.0,0.0,1.0,1.0) ;
    tpavet->SetTextFont(102) ;

    LTTPadIGraph = new TPad("LTTPadIGraph","LTT Current History",0.0,0.3,0.8,0.6) ;
    LTTPadIGraph->Draw() ;
    LTTPadIVal = new TPad("LTTPadIVal","LTT Currents",0.8,0.3,1.0,0.6) ;
    LTTPadIVal->Draw() ;
    tpavei = new TPaveText(0.0,0.0,1.0,1.0) ;
    tpavei->SetTextFont(102) ;

    LTTPadNOGraph = new TPad("LTTPadNOGraph","LTT Noise Occupancy History",0.0,0.000,0.8,0.3) ;
    LTTPadNOGraph->SetLogy() ;
    LTTPadNOGraph->Draw() ;
    LTTPadNOVal = new TPad("LTTPadNOVal","LTT Noise Occupancy",0.8,0.000,1.0,0.3) ;
    LTTPadNOVal->Draw() ;
    tpaveno = new TPaveText(0.0,0.0,1.0,1.0) ;
    tpaveno->SetTextFont(102) ;
  }

  trange[0]=0 ;
  trange[1]=total_length / 3600. ;
}

void TLTT::DoPeriodicTests() {
  // Run the test 
  ConfirmationTest(1);
}

void TLTT::DoFinalTests() {
  // Only do this test if we are testing modules
  // PWP 13.02.02
  // or hybrids cold             DGC 10.07.02
  if(((testing_modules == 1) || (cold_test == 1)) && (dcs_only == 0)){
    gROOT->ProcessLine(".L CharacterisationTest.cpp");
    CharacterisationTest(1);
  }
}

void TLTT::MakeMonitoringMeasurements() {
  // Revised PWP 20.12.01

  Int_t n ;
  FILE* txtfile ; 
  
  TDatime *update_dt = new TDatime;

  Int_t update_secs = update_dt->GetTime()%100 ;
  Int_t update_mins = int(update_dt->GetTime()/100)%100 ;
  Int_t update_hour = int(update_dt->GetTime()/10000) ;
  Int_t update_day  = update_dt->GetDate()%100 ;
  Int_t update_mon  = int(update_dt->GetDate()/100)%100 ;
  Int_t update_year = int(update_dt->GetDate()/10000) ;

  sprintf(update_str,"%02d/%02d/%04d at %02d:%02d:%02d",
          update_day,update_mon,update_year,update_hour,update_mins,update_secs) ;

  alarm = 0;

  if(dcs_only==0) { // Start triggers
    e->RunFree();
    gSystem->Sleep(1000);
  }

  // Update contents of each module's DCS structure
  e->DCSQuery();
  e->HVStatus();

  if(dcs_only==0) { // Stop triggers
    e->RunStop();
    gSystem->Sleep(1000);
  }

  // Process DCS data to spot error conditions
  for(n=0;n<e->GetMaxModules();n++){  
    if(e->m[n]->present){
      time[n][iloop_readings]=((timer->GetRealTime()-start_time)/3600.);

      t1_val[n][iloop_readings]=e->m[n]->DCS->T1 ;
      if (t1_val[n][iloop_readings]>tmax_actual) {
        tmax_actual=t1_val[n][iloop_readings] ;
      }
      if (t1_val[n][iloop_readings]<tmin_actual) {
        tmin_actual=t1_val[n][iloop_readings] ;
      }
      t2_val[n][iloop_readings]=e->m[n]->DCS->T2 ;
      if (t2_val[n][iloop_readings]>tmax_actual) {
        tmax_actual=t2_val[n][iloop_readings] ;
      }
      if (t2_val[n][iloop_readings]<tmin_actual) {
        tmin_actual=t2_val[n][iloop_readings] ;
      }

      icc_val[n][iloop_readings]=e->m[n]->DCS->icc ;
      if (icc_val[n][iloop_readings]>imax_actual) {
        imax_actual=icc_val[n][iloop_readings] ;
      }
      if (icc_val[n][iloop_readings]<imin_actual) {
        imin_actual=icc_val[n][iloop_readings] ;
      }
      idd_val[n][iloop_readings]=e->m[n]->DCS->idd ;
      if (idd_val[n][iloop_readings]>imax_actual) {
        imax_actual=idd_val[n][iloop_readings] ;
      }
      if (idd_val[n][iloop_readings]<imin_actual) {
        imin_actual=idd_val[n][iloop_readings] ;
      }

      if(testing_modules){
        idet_val[n][iloop_readings]=e->m[n]->DCS->idet ;
        if (idet_val[n][iloop_readings]>idmax_actual) {
          idmax_actual=idet_val[n][iloop_readings] ;
        }
        if (idet_val[n][iloop_readings]<idmin_actual) {
          idmin_actual=idet_val[n][iloop_readings] ;
        }

        if (e->m[n]->DCS->HVtrip) {
          printf("Module %d HV trip: V %f I %f\n Ilim %f",n,e->m[n]->DCS->vdet,e->m[n]->DCS->idet,e->m[n]->DCS->idet_lim) ;
          alarm=2 ;
        }

        if ( (e->m[n]->DCS->idet > max_current) ) {
          printf("Module %d Over current: Idet %f IMAX %f\n",n,e->m[n]->DCS->idet,max_current) ;
          alarm=3 ;
        }
      }// end if(testing_modules)

      if ( (e->m[n]->DCS->T1>max_temp) || (e->m[n]->DCS->T2>max_temp)) {
        printf("Module %d Overtemp:T1 %f T2 %f TMAX %f\n",n,e->m[n]->DCS->T1,e->m[n]->DCS->T2,max_temp) ;
        alarm=1 ;
      }
    }
  }

  // Record Noise Occupancy if there are no error conditions
  if((iloop_readings>0) && (alarm==0) && (dcs_only==0)){

    e->DCSAction();        // Recover LV trips
    e->ExecuteConfigs();   // reinitialise modules

    e->ConfigureVariable(41,1.0); // set 1.0fC threshold
    e->burst_trtype=0;     // L1A only
    e->burst_ntrigs=32000; // maximum whilst avoiding overflow
    e->burst_type = 11;    // TriggerBurst2

    e->ExecuteConfigs();   // reinitialise modules
    e->ExecuteBurst();     // execute burst
    st_display_burst();    // display results

    for(n=0;n<e->GetMaxModules();n++){  
      if(e->m[n]->present){
        no_val[n][iloop_readings]=e->m[n]->FractionalOccupancy();
        if (no_val[n][iloop_readings]>nomax_actual) {
          nomax_actual=no_val[n][iloop_readings] ;
        }
        if (no_val[n][iloop_readings]<nomin_actual) {
          nomin_actual=no_val[n][iloop_readings] ;
        }
      }
    }
  }else{ // no occupancy measurement
    for(n=0;n<e->GetMaxModules();n++){  
      if(e->m[n]->present){
        no_val[n][iloop_readings]=0;
      }
    }
  }

  // Save results in ASCII form
  for(n=0;n<e->GetMaxModules();n++){  
    if(e->m[n]->present){
      txtfile = fopen(txt_fname[n],"a+");

      if(txtfile==NULL){
        printf("Could not open result file %s\n",txt_fname);
      } else {
        //               dat time  tim idet  vdet  icc   idd   T1    T2    NO
        fprintf(txtfile,"%8d %6.6d %6d %4.3f %3.1f %7.1f %7.1f %7.3f %7.3f %3.2e\n",
                e->m[n]->DCS->date,
                e->m[n]->DCS->time,
                timer->GetRealTime()-start_time,
                e->m[n]->DCS->idet, // added PWP 20.12.01
                e->m[n]->DCS->vdet, // added BJG 13.08.02
                e->m[n]->DCS->icc,
                e->m[n]->DCS->idd,
                e->m[n]->DCS->T1,
                e->m[n]->DCS->T2,
                no_val[n][iloop_readings]);
        printf("%02d %8d %6.6d %6d %4.3f %3.1f %7.1f %7.1f %7.3f %7.3f %3.2e\n",
               n, e->m[n]->DCS->date,
               e->m[n]->DCS->time,
               timer->GetRealTime()-start_time,
               e->m[n]->DCS->idet, // added PWP 20.12.01
               e->m[n]->DCS->vdet, // added BJG 13.08.02
               e->m[n]->DCS->icc,
               e->m[n]->DCS->idd,
               e->m[n]->DCS->T1,
               e->m[n]->DCS->T2,
               no_val[n][iloop_readings]);

        fclose(txtfile) ;
      }
    }
  }

  // Save results in ROOT form
  for(n=0;n<e->GetMaxModules();n++){
    if(e->m[n]->present){
      DCSTree->SetBranchAddress("DCSBranch",&e->m[n]->DCS) ;
      DCSTree->Fill();
      DCSTree->SetBranchAddress("NoiseOcc",&no_val[n][iloop_readings]);
      DCSTree->Fill();
    }
  }

  // Tidy up!
  delete update_dt ;
  iloop_readings++ ;
  UpdateDisplay();
}
  
void TLTT::UpdateDisplay() {

  Int_t n ;
  Float_t trangemin,trangemax ;
  Float_t irangemin,irangemax ;
  Float_t idrangemin,idrangemax ;
  Float_t norangemin,norangemax ;
  Char_t tstr[64] ;
  Float_t zero2[2] = {0,0};
  Float_t notzero2[2];
  notzero2[0]=1E-7;
  notzero2[1]=1E-3;

  trangemin=(int(tmin_actual/5.0)-1.0)*5.0 ; // previous form did not work for negative numbers BJG/PWP 19.04.02
  trangemax=(int(tmax_actual/5.0)+1.0)*5.0 ;

  irangemin=int((imin_actual-10.0)/100.)*100.0 ;
  irangemax=(int(imax_actual/100.0)+1.0)*100.0 ;

  idrangemin=idmin_actual*0.9 ;
  idrangemax=idmax_actual*1.1 ;

  // TODO - revise, recall we have log scale
  norangemin=nomin_actual*0.9 ;
  norangemax=nomax_actual*1.1 ;

  LTTInfo->cd() ;
  gPad->Clear();
    
  tpaveinfo->Clear() ;
  tpaveinfo->SetFillColor(10) ;
  tpaveinfo->Draw() ;
  tpaveinfo->SetTextAlign(13) ;
  tpaveinfo->SetTextSize(0.25) ;
  tpaveinfo->SetTextColor(1) ;
  sprintf(tstr,"Run number %d at %s",e->runnum, LOCATION_NAME) ;
  tpaveinfo->AddText(tstr) ;
  sprintf(tstr,"Test started at %s",start_str) ;
  tpaveinfo->AddText(tstr) ;
  sprintf(tstr,"last update was %s",update_str) ;
  tpaveinfo->AddText(tstr) ;
  tpaveinfo->AddText(" ") ;
  sprintf(tstr,"expected end at %s",end_str) ;
  tpaveinfo->AddText(tstr) ;
  tpaveinfo->AddText(" ") ;

  tpaveinfo2->Clear() ;
  tpaveinfo2->SetFillColor(10) ;
  tpaveinfo2->Draw() ;
  tpaveinfo2->SetTextAlign(13) ;
  tpaveinfo2->SetTextSize(0.25) ;
  tpaveinfo2->SetTextColor(1) ;
  sprintf(tstr,"%d Modules under test", e->num_modules) ;
  tpaveinfo2->AddText(tstr) ;

  // Do two columns
  for(int i=0; i<(e->num_modules+1)/2; i++) {
    if(i*2+1 == e->num_modules) { 
      sprintf(tstr,"%2d: %s", i*2, e->m[i*2]->name) ;
    } else {
      sprintf(tstr,"%2d: %s    %2d: %s", i*2, e->m[i*2]->name, i*2+1, e->m[i*2+1]->name) ;
    }
    tpaveinfo2->AddText(tstr) ;
  }

  LTTPadTVal->cd() ;
  gPad->Clear();

  tpavet->Clear() ;
  tpavet->SetFillColor(10) ;
  tpavet->Draw() ;
  tpavet->SetTextAlign(13) ;
  tpavet->SetTextSize(0.08) ;
  tpavet->SetTextColor(1) ;
  tpavet->AddText("Present values:") ;
  tpavet->AddText(" ") ;
  tpavet->AddText("Module   T1/C  T2/C") ;
  for(n=0;n<e->GetMaxModules();n++){
    if(e->m[n]->present){
      sprintf(tstr,"%4d%9.1f%6.1f",n,t1_val[n][iloop_readings-1],t2_val[n][iloop_readings-1]) ;
      TText *ttdrawn=tpavet->AddText(tstr) ;
      ttdrawn->SetTextColor(n+2) ;
    }
    else {
      tpavet->AddText(" ") ;
    }
  }

  LTTPadTGraph->cd() ;
  LTTPadTGraph->Clear() ;

  if (iloop_readings>1) {
    delete markert1 ;
    delete markert2 ;
    delete keyt1 ;
    delete keyt2 ;
  }
  if(graph0==0){
    graph0 = new TGraph(2, trange, zero2) ;
    graph0->SetTitle("Temperature histories") ;
  }
  graph0->Draw("AP") ;
  graph0->GetXaxis()->SetTitle("hours") ;
  graph0->GetXaxis()->SetTitleSize(0.05) ;
  graph0->GetXaxis()->CenterTitle() ;
  graph0->GetYaxis()->SetTitle("deg C") ;
  graph0->GetYaxis()->SetTitleSize(0.05) ;
  graph0->GetYaxis()->CenterTitle() ;
  graph0->SetMinimum(trangemin) ;
  graph0->SetMaximum(trangemax) ;

  Float_t xkey1 = (graph0->GetXaxis()->GetXmax() - graph0->GetXaxis()->GetXmin())*0.8 +
    graph0->GetXaxis()->GetXmin() ;
  Float_t xkey2 = (graph0->GetXaxis()->GetXmax() - graph0->GetXaxis()->GetXmin())*0.9 +
    graph0->GetXaxis()->GetXmin() ;
  Float_t xeps = (graph0->GetXaxis()->GetXmax() - graph0->GetXaxis()->GetXmin())*0.01 ;
  Float_t ykey = (graph0->GetYaxis()->GetXmax() - graph0->GetYaxis()->GetXmin())*1.05 +
    graph0->GetYaxis()->GetXmin() ;

  for(n=0;n<e->GetMaxModules();n++){    
    if(e->m[n]->present){
      if (iloop_readings>1) {
        grapht1[n]->Delete();
        grapht2[n]->Delete();
      }
      grapht1[n] = new TGraph(iloop_readings, time[n], t1_val[n]) ;
      grapht1[n]->SetLineColor(n+2) ;
      grapht1[n]->SetMarkerStyle(24) ;
      grapht1[n]->SetMarkerColor(n+2) ;
      grapht1[n]->Draw("LP") ;
      grapht2[n] = new TGraph(iloop_readings, time[n], t2_val[n]) ;
      grapht2[n]->SetLineColor(n+2) ;
      grapht2[n]->SetMarkerColor(n+2) ;
      grapht2[n]->SetMarkerStyle(25) ;
      grapht2[n]->Draw("LP") ;
    }
  }
  markert1 = new TMarker(xkey1,ykey,24) ;
  markert1->Draw() ;
  keyt1 = new TText(xkey1+xeps,ykey,"T1") ;
  keyt1->SetTextAlign(12) ;
  keyt1->Draw() ;
  markert2 = new TMarker(xkey2,ykey,25) ;
  markert2->Draw() ;
  keyt2 = new TText(xkey2+xeps,ykey,"T2") ;
  keyt2->SetTextAlign(12) ;
  keyt2->Draw() ;

  LTTPadIVal->cd() ;
  gPad->Clear();

  tpavei->Clear() ;
  tpavei->SetFillColor(10) ;
  tpavei->Draw() ;
  tpavei->SetTextAlign(13) ;
  tpavei->SetTextSize(0.08) ;
  tpavei->SetTextColor(1) ;
  tpavei->AddText("Present values:") ;    
  tpavei->AddText(" ") ;
  tpavei->AddText("Module Icc(mA) Idd(mA)") ;
  for(n=0;n<e->GetMaxModules();n++){
    if(e->m[n]->present){
      sprintf(tstr,"%4d%8.0f%7.0f",n,icc_val[n][iloop_readings-1],idd_val[n][iloop_readings-1]) ;
      TText *ttdrawn=tpavei->AddText(tstr) ;
      ttdrawn->SetTextColor(n+2) ;
    }
    else {
      tpavei->AddText(" ") ;
    }
  }
    
  LTTPadIGraph->cd() ;
  LTTPadIGraph->Clear() ;
  if (iloop_readings>1) {
    delete markeri1 ;
    delete markeri2 ;
    delete keyi1 ;
    delete keyi2 ;
  }
  if(graphi0==0){
    graphi0 = new TGraph(2, trange, zero2);
    graphi0->SetTitle("Current histories");
  }
  graphi0->Draw("AP");
  graphi0->SetMinimum(irangemin) ;
  graphi0->SetMaximum(irangemax) ;
  graphi0->GetXaxis()->SetTitle("hours") ;
  graphi0->GetXaxis()->SetTitleSize(0.05) ;
  graphi0->GetXaxis()->CenterTitle() ;
  graphi0->GetYaxis()->SetTitle("mA") ;
  graphi0->GetYaxis()->SetTitleSize(0.05) ;
  graphi0->GetYaxis()->CenterTitle() ;
  xkey1 = (graphi0->GetXaxis()->GetXmax() - graphi0->GetXaxis()->GetXmin())*0.8 +
      graphi0->GetXaxis()->GetXmin() ;
  xkey2 = (graphi0->GetXaxis()->GetXmax() - graphi0->GetXaxis()->GetXmin())*0.9 +
      graphi0->GetXaxis()->GetXmin() ;
  xeps = (graphi0->GetXaxis()->GetXmax() - graphi0->GetXaxis()->GetXmin())*0.01 ;
  ykey = (graphi0->GetYaxis()->GetXmax() - graphi0->GetYaxis()->GetXmin())*1.05 +
      graphi0->GetYaxis()->GetXmin() ;
  for(n=0;n<e->GetMaxModules();n++){
      if(e->m[n]->present){
      if (iloop_readings>1) {
        graphi1[n]->Delete();
        graphi2[n]->Delete();
      }
      graphi1[n] = new TGraph(iloop_readings, time[n], icc_val[n]) ;
      graphi1[n]->SetLineColor(n+2) ;
      graphi1[n]->SetMarkerColor(n+2) ;
      graphi1[n]->SetMarkerStyle(24) ;
      graphi1[n]->Draw("LP") ;
      graphi2[n] = new TGraph(iloop_readings, time[n], idd_val[n]) ;
      graphi2[n]->SetLineColor(n+2) ;
      graphi2[n]->SetMarkerColor(n+2) ;
      graphi2[n]->SetMarkerStyle(25) ;
      graphi2[n]->Draw("LP") ;
    }
  }
  markeri1 = new TMarker(xkey1,ykey,24) ;
  markeri1->Draw() ;
  keyi1 = new TText(xkey1+xeps,ykey,"Icc") ;
  keyi1->SetTextAlign(12) ;
  keyi1->Draw() ;
  markeri2 = new TMarker(xkey2,ykey,25) ;
  markeri2->Draw() ;
  keyi2 = new TText(xkey2+xeps,ykey,"Idd") ;
  keyi2->SetTextAlign(12) ;
  keyi2->Draw() ;

  if(testing_modules){

    // Detector Currents

    LTTPadIdetVal->cd() ;
    gPad->Clear();

    tpaveid->Clear() ;
    tpaveid->SetFillColor(10) ;
    tpaveid->Draw() ;
    tpaveid->SetTextAlign(13) ;
    tpaveid->SetTextSize(0.08) ;
    tpaveid->SetTextColor(1) ;
    tpaveid->AddText("Present values:") ;    
    tpaveid->AddText(" ") ;
    tpaveid->AddText("Module Vdet(V) Idet(#muA)") ;
    for(n=0;n<e->GetMaxModules();n++){
      if(e->m[n]->present){
        sprintf(tstr,"%4d%8.0f%7.2f",n,e->m[n]->DCS->vdet,idet_val[n][iloop_readings-1]) ;
        TText *ttdrawn=tpaveid->AddText(tstr) ;
        ttdrawn->SetTextColor(n+2) ;
      }
      else {
        tpaveid->AddText(" ") ;
      }
    }

    LTTPadIdetGraph->cd() ;

    LTTPadIdetGraph->Clear() ;
    if (iloop_readings>1) {
      delete markerid;
      delete keyid;
    }
    if(graphid0==0){
      graphid0 = new TGraph(2, trange, zero2);
      graphid0->SetTitle("Detector Current history");
    }
    graphid0->Draw("AP");
    graphid0->SetMinimum(idrangemin) ;
    graphid0->SetMaximum(idrangemax) ;
    graphid0->GetXaxis()->SetTitle("hours") ;
    graphid0->GetXaxis()->SetTitleSize(0.05) ;
    graphid0->GetXaxis()->CenterTitle() ;
    graphid0->GetYaxis()->SetTitle("#muA") ;
    graphid0->GetYaxis()->SetTitleSize(0.05) ;
    graphid0->GetYaxis()->CenterTitle() ;
    Float_t xkey1 = (graphid0->GetXaxis()->GetXmax() - graphid0->GetXaxis()->GetXmin())*0.8 +
        graphid0->GetXaxis()->GetXmin() ;
    Float_t xkey2 = (graphid0->GetXaxis()->GetXmax() - graphid0->GetXaxis()->GetXmin())*0.9 +
        graphid0->GetXaxis()->GetXmin() ;
    Float_t xeps = (graphid0->GetXaxis()->GetXmax() - graphid0->GetXaxis()->GetXmin())*0.01 ;
    Float_t ykey = (graphid0->GetYaxis()->GetXmax() - graphid0->GetYaxis()->GetXmin())*1.05 +
        graphid0->GetYaxis()->GetXmin() ;
    for(n=0;n<e->GetMaxModules();n++){
      if(e->m[n]->present){
        if (iloop_readings>1) {
          graphid1[n]->Delete();
        }
        graphid1[n] = new TGraph(iloop_readings, time[n], idet_val[n]) ;
        graphid1[n]->SetLineColor(n+2) ;
        graphid1[n]->SetMarkerColor(n+2) ;
        graphid1[n]->SetMarkerStyle(24) ;
        graphid1[n]->Draw("LP") ;
      }
    }
    markerid = new TMarker(xkey1,ykey,24) ;
    markerid->Draw() ;
    keyid = new TText(xkey1+xeps,ykey,"Idet") ;
    keyid->SetTextAlign(12) ;
    keyid->Draw() ;

  }

  // Noise Occupancy
  {
    LTTPadNOVal->cd() ;
    gPad->Clear();

    tpaveno->Clear() ;
    tpaveno->SetFillColor(10) ;
    tpaveno->Draw() ;
    tpaveno->SetTextAlign(13) ;
    tpaveno->SetTextSize(0.08) ;
    tpaveno->SetTextColor(1) ;
    tpaveno->AddText("Present values:") ;    
    tpaveno->AddText(" ") ;
    tpaveno->AddText("Module   NO") ;
    for(n=0;n<e->GetMaxModules();n++){
      if(e->m[n]->present){
        sprintf(tstr,"%4d    %-6.2e",n,no_val[n][iloop_readings-1]) ;
        TText *ttdrawn=tpaveno->AddText(tstr) ;
        ttdrawn->SetTextColor(n+2) ;
      }
      else {
        tpaveno->AddText(" ") ;
      }
    }

    LTTPadNOGraph->cd() ;
    LTTPadNOGraph->Clear() ;
    if (iloop_readings>1) {
      delete markerno;
      delete keyno;
    }
    if(graphno0==0){
      graphno0 = new TGraph(2, trange, notzero2);
      graphno0->SetTitle("Detector Noise Occupancy");
    }
    graphno0->Draw("AP");
    graphno0->GetXaxis()->SetTitle("hours") ;
    graphno0->GetXaxis()->SetTitleSize(0.05) ;
    graphno0->GetXaxis()->CenterTitle() ;
    graphno0->GetYaxis()->SetTitle("Occupancy") ;
    graphno0->GetYaxis()->SetTitleSize(0.05) ;
    graphno0->GetYaxis()->CenterTitle() ;

    xkey1 = (graphno0->GetXaxis()->GetXmax() - graphno0->GetXaxis()->GetXmin())*0.8 +
        graphno0->GetXaxis()->GetXmin() ;
    xkey2 = (graphno0->GetXaxis()->GetXmax() - graphno0->GetXaxis()->GetXmin())*0.9 +
        graphno0->GetXaxis()->GetXmin() ;
    xeps = (graphno0->GetXaxis()->GetXmax() - graphno0->GetXaxis()->GetXmin())*0.01 ;
    for(n=0;n<e->GetMaxModules();n++){
      if(e->m[n]->present){
        if (iloop_readings>1) {
          graphno1[n]->Delete();
        }
        graphno1[n] = new TGraph(iloop_readings, time[n], no_val[n]) ;
        graphno1[n]->SetLineColor(n+2) ;
        graphno1[n]->SetMarkerColor(n+2) ;
        graphno1[n]->SetMarkerStyle(24) ;
        graphno1[n]->Draw("LP") ;
      }
    }
    markerno = new TMarker(xkey1,1,24) ;
    markerno->Draw() ;
    keyno = new TText(xkey1+xeps,1,"NO") ;
    keyno->SetTextAlign(12) ;
    keyno->Draw() ;
  }

  LTTCanvas->Update() ;

  LTTCanvas->Print(psfilename) ;

}

void TLTT::WriteSummaryFile() {

  Char_t summ_fname[128] ;
  Char_t rawd_fname[128] ;
  Char_t f_scan1name[128] ;
  FILE* fwr ;
  FILE* fraw ;

  Int_t date = start_dt->GetDate();

  gROOT->LoadMacro("ToolBox.cpp");

  for(Short_t n=0;n<e->GetMaxModules();n++){
    if(e->m[n]->present){

      Int_t start_day = start_dt->GetDate()%100 ;
      Int_t start_mon = int(start_dt->GetDate()/100)%100 ;
      Int_t start_year = int(start_dt->GetDate()/10000) ;

      Double_t duration = timer->GetRealTime()-start_time ;
      duration /= 3600. ;
      Int_t duration_hours = int(duration) ;

      Float_t t1max=-999999. ;
      Float_t t1min= 999999. ;
      Float_t t2max=-999999. ;
      Float_t t2min= 999999. ;
      for (Int_t i=0;i<iloop_readings;i++) {
        if (t1_val[n][i]>t1max) {
          t1max=t1_val[n][i] ;
        }
        if (t1_val[n][i]<t1min) {
          t1min=t1_val[n][i] ;
        }
        if (t2_val[n][i]>t2max) {
          t2max=t2_val[n][i] ;
        }
        if (t2_val[n][i]<t2min) {
          t2min=t2_val[n][i] ;
        }
      }
      Float_t iccmax=-999999. ;
      Float_t iccmin= 999999. ;
      Float_t iddmax=-999999. ;
      Float_t iddmin= 999999. ;
      Float_t idetmax=-999999. ;
      Float_t idetmin= 999999. ;
      for (Int_t i=0;i<iloop_readings;i++) {
        if (idet_val[n][i]<idetmin) {
          idetmin=idet_val[n][i] ;
        }
        if (idet_val[n][i]>idetmax) {
          idetmax=idet_val[n][i] ;
        }
        if (icc_val[n][i]>iccmax) {
          iccmax=icc_val[n][i] ;
        }
        if (icc_val[n][i]<iccmin) {
          iccmin=icc_val[n][i] ;
        }
        if (idd_val[n][i]>iddmax) {
          iddmax=idd_val[n][i] ;
        }
        if (idd_val[n][i]<iddmin) {
          iddmin=idd_val[n][i] ;
        }
      }
      sprintf(f_scan1name,"%sstrun%d_%d.root",DATA_DIRECTORY,start_run,start_scan);
      TFile *strun = new TFile(f_scan1name);
      if(strun==0){
        printf("Failed to open file %s - Skipping writing LTT summary file!!\n",f_scan1name);
      } else {
        // Fail with explanatory comment
        Int_t pass = 1;
        if(strlen(failComment)) pass = 0;

        FILE* outfile = OpenResults("LongTermTest",strun,n,start_run,start_scan,pass,0) ;
        if (outfile) {
          fprintf(outfile,"#\n");
          fprintf(outfile,"%%LongTermTest\n");

          fprintf(outfile,"#\n");
          fprintf(outfile,"#Duration\n") ;
          fprintf(outfile,"%.1f\n",duration) ;
          fprintf(outfile,"#Time of first failure\n");
          // Need to add this when it is available... -1 for no failure?
          fprintf(outfile,"-1.0\n") ;
          fprintf(outfile,"#T0 min max\n") ;
          fprintf(outfile,"%.1f \t%.1f\n",t1min,t1max) ;
          fprintf(outfile,"#T1 min max\n") ;
          fprintf(outfile,"%.1f \t%.1f\n",t2min,t2max) ;
          fprintf(outfile,"#Icc min max\n") ;
          fprintf(outfile,"%d \t%d\n",int(iccmin),int(iccmax)) ;
          fprintf(outfile,"#Idd min max\n") ;
          fprintf(outfile,"%d \t%d\n",int(iddmin),int(iddmax)) ;
          fprintf(outfile,"#Idet min max\n") ;
          if(testing_modules){
            fprintf(outfile,"%1.3f \t%1.3f\n",idetmin,idetmax) ;
          }else{
            fprintf(outfile,". \t.\n");
          }
          fprintf(outfile,"#\n") ;
          // Here need to add code for defects
          // Why there was a failure
          if(pass == 0)
            ListComment(outfile, failComment);
	
          fprintf(outfile,"%%TEST Rawdata\n") ; // PWP bugfix 21.08.02
          sprintf(rawd_fname,"%s%s_ltt_%d-%d.txt",RESULTS_DIRECTORY,mname[n],start_run,start_scan);
          fprintf(outfile,"FILENAME       : %s\n",rawd_fname) ;
          fclose(outfile) ;
        }
        strun->Close() ;
        delete strun;
      }

      fraw=fopen(rawd_fname,"a+") ;
      if(fraw==NULL){
        printf("Could not open raw data output file %s\n",rawd_fname);
      } else {
        for (Int_t i=0;i<iloop_readings;i++) {
          if(testing_modules){
            fprintf(fraw,"%16.3f%16.3f%16.3f%16.3f%16.3f%16.3f%16.3e\n",
                    time[n][i], t1_val[n][i],t2_val[n][i],
                    icc_val[n][i],idd_val[n][i], idet_val[n][i], no_val[n][i]) ;
          }else{
            fprintf(fraw,"%16.3f%16.3f%16.3f%16.3f%16.3f%16.3e\n",
                    time[n][i],t1_val[n][i],t2_val[n][i],
                    icc_val[n][i],idd_val[n][i], no_val[n][i]) ;
          }
        }
        fclose(fraw) ;
      }

      sprintf(summ_fname,"%sLONGTERM_%s_%d.txt",RESULTS_DIRECTORY,mname[n],date);
      fwr=fopen(summ_fname,"a+") ;            
      if(fwr==NULL){
        printf("Could not open results summary file %s\n",summ_fname);
      } else {
        fprintf(fwr,"atlasPartsId   : %s\n",mname[n]) ;
        fprintf(fwr,"dateStart      : %2d/%2d/%4d\n",start_day,start_mon,start_year) ;   
        fprintf(fwr,"testLength     : %f\n",duration) ;
        fprintf(fwr,"testLength     : %d\n",duration_hours) ;
        fprintf(fwr,"t1Max          : %f\n",t1max) ;
        fprintf(fwr,"t1Min          : %f\n",t1min) ;
        fprintf(fwr,"t2Max          : %f\n",t2max) ;
        fprintf(fwr,"t2Min          : %f\n",t2min) ;
        fprintf(fwr,"iccMax         : %f\n",iccmax) ;
        fprintf(fwr,"iccMin         : %f\n",iccmin) ;
        fprintf(fwr,"iddMax         : %f\n",iddmax) ;
        fprintf(fwr,"iddMin         : %f\n",iddmin) ;
        if(testing_modules){
          fprintf(fwr,"idetMax        : %f\n",idetmax) ;
          fprintf(fwr,"idetMin        : %f\n",idetmin) ;
        }
        fprintf(fwr,"RawData follows\n") ;
        for (Int_t i=0;i<iloop_readings;i++) {
          if(testing_modules){
            //            time  t1    t2    icc   idd   idet  no
            fprintf(fwr,"%16.3f%16.3f%16.3f%16.3f%16.3f%16.3f%16.3e\n",
                    time[n][i], t1_val[n][i],t2_val[n][i],
                    icc_val[n][i],idd_val[n][i], idet_val[n][i], no_val[n][i]) ;
          }else{      //  time  t1    t2    icc   idd   no 
            fprintf(fwr,"%16.3f%16.3f%16.3f%16.3f%16.3f%16.3e\n",
                    time[n][i], t1_val[n][i],t2_val[n][i],
                    icc_val[n][i],idd_val[n][i], no_val[n][i]) ;
          }
        }

        fclose(fwr) ;
      }
    }
  }
}
