// Macro for beta source tests - Analyses
// Created by Pavel Reznicek (reznicek@ipnp.troja.mff.cuni.cz)
// Last update 29/07/2002
// Search for string "EXAMPLES:" in this file to see important (most usefull) examples of use

Float_t b_cut=3.29;  // maximal allowed deviation from mean value <x>
                     // |x-<x>|<=b_cut*sigma where sigma=sqrt(<x^2>-<x>^2)

B_pos2 B_RunCrossPosition(Int_t runnum=-1,Int_t scannum=-1,Char_t *filename=NULL){
// finds mean beam position (mean channel number) and it's width on both links
// this position corresponds to beam position across strips
/* EXAMPLES:
   B_RunCrossPosition(); - calculate beam position for the last measured burst
   B_RunCrossPosition(14,3); - calculate beam position for burst from data file: sctvar/data/btrun14_3.root
   B_RunCrossPosition(-1,-1,"d:\\btrun56_7.root"); - calculate beam position for burst from data file: d:/btrun56_7.root
*/
  Int_t i,j;
  B_pos2 ret;

  if(!B_IsOpenedDataFile(runnum,scannum,filename))  // open requested data file
    if(!B_DataFile(runnum,scannum,filename)){
      printf("B_RunCrossPosition: File can not be opened !\n");
      ret.mean[0]=ret.sigma[0]=ret.width[0]=ret.mean[1]=ret.sigma[1]=ret.width[1]=-1;
      ret.n[0]=ret.nmax[0]=ret.n[1]=ret.nmax[1]=0;
      return ret;
    }
  
  Int_t nentries=b_events->GetEntries();
  Float_t pos0[768/2],pos1[768/2];
  Int_t counts0,counts1;
  b_events->SetBranchAddress("pos0",&pos0);
  b_events->SetBranchAddress("pos1",&pos1);
  b_events->SetBranchAddress("counts0",&counts0);
  b_events->SetBranchAddress("counts1",&counts1);

  Float_t beam_mean[2]={768/2,768/2};
  Float_t beam_sigma[2]={770,770};
  Float_t last_mean[2],last_sigma[2];
  Int_t naccept[2]={0,0},lastaccept[2]={0,0},maxaccept[2]={0,0};
  Short_t stop[2]={0,0};
  
  do{
    for(i=0;i<2;i++) if(!stop[i]){  // initial values
      last_mean[i]=beam_mean[i];
      last_sigma[i]=beam_sigma[i];
      lastaccept[i]=naccept[i];
      beam_sigma[i]=beam_mean[i]=naccept[i]=maxaccept[i]=0;
    }

    for(i=0;i<nentries;i++){  // calculate <n*x> and <n*x^2>
      b_events->GetEntry(i);
      if(!(i%10) || i==nentries-1) printf("%cReading data: %.1f%%",13,100.*(Float_t)(i+1)/(Float_t)nentries);
      if(counts0 && !stop[0]){
        maxaccept[0]+=counts0;
        for(j=0;j<counts0;j++) if(fabs(pos0[j]-last_mean[0])<=last_sigma[0]*b_cut){
          beam_mean[0]+=pos0[j];
          beam_sigma[0]+=pos0[j]*pos0[j];
          naccept[0]++;
        }
      }
      if(counts1 && !stop[1]){
        maxaccept[1]+=counts1;
        for(j=0;j<counts1;j++) if(fabs(pos1[j]-last_mean[1])<=last_sigma[1]*b_cut){
          beam_mean[1]+=pos1[j];
          beam_sigma[1]+=pos1[j]*pos1[j];
          naccept[1]++;
        }
      }
    }
    
    for(i=0;i<2;i++)  // too few hits
      if(naccept[i]==0){
        printf("%c- Link0: No hits left !                    \n",13);
        stop[i]=1;
        beam_mean[i]=768/2;
        beam_sigma[i]=769/2;
        naccept[i]=maxaccept[i]=0;
      }else{
        if(!stop[i]){  // calculate <x> and sqrt(<x^2>-<x>^2)
          beam_mean[i]/=(Float_t)naccept[i];
          beam_sigma[i]=sqrt(beam_sigma[i]/(Float_t)naccept[i]-beam_mean[i]*beam_mean[i]);
          stop[i]=(lastaccept[i]==naccept[i]);
        }
        printf("%c- Link%d: Beam cross strip position: channel %.1f\n",13,i,beam_mean[i]);
        printf("  Link%d: Beam width (sigma around mean): %.1f channels\n",i,beam_sigma[i]);
        printf("  Link%d: Number of accepted hits: %d of %d\n",i,naccept[i],maxaccept[i]);
      }
  }while(!stop[0] || !stop[1]);

  for(i=0;i<2;i++){
    ret.mean[i]=beam_mean[i];          // mean position: <x>
    ret.sigma[i]=beam_sigma[i];        // sigma: sqrt(<x^2>-<x>^2)
    ret.width[i]=beam_sigma[i]*b_cut;  // width: b_cut*sigma
    ret.n[i]=naccept[i];               // number of accepted hits: |x-<x>|<=width
    ret.nmax[i]=maxaccept[i];          // number of all hits
  }
  B_SetEventBranchAddress();  // recover addresses of 'b_events' branches
  return ret;
}

B_pos B_RunAlongPosition(Int_t runnum=-1,Int_t scannum=-1,Char_t *filename=NULL){
// finds mean beam position (mean value of hit_position_link0+hit_position_link1-767) and it's width
// this position corresponds to beam position along strips
/* EXAMPLES:
   B_RunAlongPosition(); - calculate beam position for the last measured burst
   B_RunAlongPosition(14,3); - calculate beam position for burst from data file: sctvar/data/btrun14_3.root
   B_RunAlongPosition(-1,-1,"d:\\btrun56_7.root"); - calculate beam position for burst from data file: d:/btrun56_7.root
*/
  Int_t i;
  B_pos ret;
  
  if(!B_IsOpenedDataFile(runnum,scannum,filename))  // open requested data file
    if(!B_DataFile(runnum,scannum,filename)){
      printf("B_RunAlongPosition: File can not be opened !\n");
      ret.mean=ret.sigma=ret.width=-1;
      ret.n=ret.nmax=0;
      return ret;
    }
  
  Int_t nentries=b_events->GetEntries();
  Float_t pos0[768/2],pos1[768/2];
  Int_t counts0,counts1;
  b_events->SetBranchAddress("pos0",&pos0);
  b_events->SetBranchAddress("pos1",&pos1);
  b_events->SetBranchAddress("counts0",&counts0);
  b_events->SetBranchAddress("counts1",&counts1);

  Float_t beam_mean=0,beam_sigma=770*2;
  Float_t last_mean,last_sigma;
  Int_t naccept,lastaccept,maxaccept;
  
  do{
    last_mean=beam_mean;  // initial values
    last_sigma=beam_sigma;
    lastaccept=naccept;
    beam_sigma=beam_mean=naccept=maxaccept=0;

    for(i=0;i<nentries;i++){  // calculate <n*x> and <n*x^2>
      b_events->GetEntry(i);
      if(!(i%10) || i==nentries-1) printf("%cReading data: %.1f%%",13,100.*(Float_t)(i+1)/(Float_t)nentries);
      if(counts0==1 && counts1==1){
        maxaccept++;
        if(fabs(pos0[0]+pos1[0]-767-last_mean)<=last_sigma*b_cut){
          beam_mean+=pos0[0]+pos1[0]-767;
          beam_sigma+=(pos0[0]+pos1[0]-767)*(pos0[0]+pos1[0]-767);
          naccept++;
        }
      }
    }

    if(naccept==0){  // too few hits
      printf("%c- No hits left !                    \n",13);
      ret.mean=0;
      ret.sigma=769;
      ret.n=ret.nmax=0;
      ret.width=ret.sigma*b_cut;
      B_SetEventBranchAddress();  // recover addresses of 'b_events' branches
      return ret;
    }
    beam_mean/=(Float_t)naccept;                                       // calculate <x>
    beam_sigma=sqrt(beam_sigma/(Float_t)naccept-beam_mean*beam_mean);  // calculate sqrt(<x^2>-<x>^2)
    printf("%c- Beam along strip position: difference in channel %.1f\n  Beam width (sigma around mean): %.1f channels\n  Number of accepted hits: %d of %d\n",
           13,beam_mean,beam_sigma,naccept,maxaccept);
  }while(lastaccept!=naccept);
  
  ret.mean=beam_mean;         // mean position: <x>
  ret.sigma=beam_sigma;       // sigma: sqrt(<x^2>-<x>^2)
  ret.width=ret.sigma*b_cut;  // width: b_cut*sigma
  ret.n=naccept;              // number of accepted hits: |x-<x>|<=width
  ret.nmax=maxaccept;         // number of all hits
  B_SetEventBranchAddress();  // recover addresses of 'b_events' branches
  return ret;
}

Short_t b_ignore_zero_mult=1;  // if 1 then don't consider events with no active channels to multiplicity calculations
                               // usefull for noisy trigger system (a lot of non-beta-particle triggers)

B_mult B_RunMultiplicity(Int_t runnum=-1,Int_t scannum=-1,Char_t *filename=NULL){
// finds mean multiplicity
/* EXAMPLES:
   B_RunMultiplicity(); - calculate multiplicity for the last measured burst
   B_RunMultiplicity(14,3); - calculate multiplicity for burst from data file: sctvar/data/btrun14_3.root
   B_RunMultiplicity(-1,-1,"d:\\btrun56_7.root"); - calculate multiplicity for burst from data file: d:/btrun56_7.root
*/
  Int_t i;
  B_mult ret;
  
  if(!B_IsOpenedDataFile(runnum,scannum,filename))  // open requested data file
    if(!B_DataFile(runnum,scannum,filename)){
      printf("B_RunMultiplicity: File can not be opened !\n");
      ret.mult[0]=ret.mult[1]=ret.sigma[0]=ret.sigma[1]=-1;
      ret.n[0]=ret.n[1]=0;
      return ret;
    }
  
  Int_t nentries=b_events->GetEntries();
  Int_t mults0,mults1;
  b_events->SetBranchAddress("mult0",&mults0);
  b_events->SetBranchAddress("mult1",&mults1);

  Int_t n0=0,n1=0,nz0=0,nz1=0;
  Float_t mult0=0,mult1=0;
  Float_t sigma0=0,sigma1=1;
  Float_t multmp;

  for(i=0;i<nentries;i++){  // calculate <n*x> and <n*x^2>
    b_events->GetEntry(i);
    if(!(i%10) || i==nentries-1) printf("%cReading data: %.1f%%",13,100.*(Float_t)(i+1)/(Float_t)nentries);
    sigma0+=mults0*mults0;
    mult0+=mults0;
    sigma1+=mults1*mults1;
    mult1+=mults1;
    if(!b_ignore_zero_mult || mults0) n0++;
    if(!mults0) nz0++;
    if(!b_ignore_zero_mult || mults1) n1++;
    if(!mults1) nz1++;
  }
  
  ret.nz[0]=nz0;  // number of events with zero multiplicity on link 0
  ret.nz[1]=nz1;  // number of events with zero multiplicity on link 1
  ret.n[0]=n0;    // number of considered events on link 0 (is equal to number of all events if b_ignore_zero_mult=0)
  ret.n[1]=n1;    // number of considered events on link 1 (is equal to number of all events if b_ignore_zero_mult=0)
  if(n0){
    ret.mult[0]=mult0/n0;                                     // link 0 mean multiplicity: <x>
    ret.sigma[0]=sqrt((sigma0/n0-(mult0/n0)*(mult0/n0))/n0);  // link 0 sigma: sqrt((<x^2>-<x>^2)/n)
  }else{  // no hits
    ret.mult[0]=0;
    ret.sigma[0]=0;
  }
  if(n1){
    ret.mult[1]=mult1/n1;                                     // link 1 mean multiplicity: <x>
    ret.sigma[1]=sqrt((sigma1/n1-(mult1/n1)*(mult1/n1))/n1);  // link 1 sigma: sqrt((<x^2>-<x>^2)/n)
  }else{  // no hits
    ret.mult[1]=0;
    ret.sigma[1]=0;
  }
  printf("%cMultiplicity (link 1): %.5f  , sigma = %.5f\n",13,ret.mult[0],ret.sigma[0]);
  printf(  "Multiplicity (link 0): %.5f  , sigma = %.5f\n"   ,ret.mult[1],ret.sigma[1]);
  B_SetEventBranchAddress();  // recover addresses of 'b_events' branches
  return ret;
}

B_clust B_RunClusterSize(B_pos2 &posC,B_pos &posA,Int_t runnum=-1,Int_t scannum=-1,Char_t *filename=NULL){
// finds mean cluster size (consider hits in beam region only - see above B_RunCrossPosition and B_RunAlongPosition)
//                                                            - determined by mean and width
/* EXAMPLES:
   B_RunClusterSize(pC,pA); - calculate cluster size for the last measured burst
   B_RunClusterSize(pC,pA,14,3); - calculate cluster size for burst from data file: sctvar/data/btrun14_3.root
   B_RunClusterSize(pC,pA,-1,-1,"d:\\btrun56_7.root"); - calculate cluster size for burst from data file: d:/btrun56_7.root
*/
  Int_t i,j;
  B_clust ret;
  
  if(!B_IsOpenedDataFile(runnum,scannum,filename))  // open requested data file
    if(!B_DataFile(runnum,scannum,filename)){
      printf("B_RunClusterSize: File can not be opened !\n");
      ret.clust[0]=ret.clust[1]=ret.sigma[0]=ret.sigma[1]=-1;
      ret.n[0]=ret.n[1]=0;
      return ret;
    }
  
  Int_t nentries=b_events->GetEntries();
  Int_t clusts0[768/2],clusts1[768/2];
  Float_t pos0[768/2],pos1[768/2];
  Int_t counts0,counts1;
  b_events->SetBranchAddress("clust0",&clusts0);
  b_events->SetBranchAddress("clust1",&clusts1);
  b_events->SetBranchAddress("pos0",&pos0);
  b_events->SetBranchAddress("pos1",&pos1);
  b_events->SetBranchAddress("counts0",&counts0);
  b_events->SetBranchAddress("counts1",&counts1);
  
  Int_t n0=0,n1=0;
  Float_t clust0=0,clust1=0;
  Float_t sigma0=0,sigma1=1;

  for(i=0;i<nentries;i++){  // calculate <n*x> and <n*x^2>
    b_events->GetEntry(i);
    if(!(i%10) || i==nentries-1) printf("%cReading data: %.1f%%",13,100.*(Float_t)(i+1)/(Float_t)nentries);
    for(j=0;j<counts0;j++)
      if(fabs(pos0[j]-posC.mean[0])<=posC.width[0]){
        n0++;
        clust0+=clusts0[j];
        sigma0+=clusts0[j]*clusts0[j];
      }
    for(j=0;j<counts1;j++)
      if(fabs(pos1[j]-posC.mean[1])<=posC.width[1]){
        n1++;
        clust1+=clusts1[j];
        sigma1+=clusts1[j]*clusts1[j];
      }
  }
  
  ret.n[0]=n0;  // number of accepted (in beam region) hits on link 0
  ret.n[1]=n1;  // number of accepted (in beam region) hits on link 1
  if(n0){
    ret.clust[0]=clust0/n0;                                     // link 0 mean cluster size: <x>
    ret.sigma[0]=sqrt((sigma0/n0-(clust0/n0)*(clust0/n0))/n0);  // link 0 sigma: sqrt((<x^2>-<x>^2)/n)
  }else{
    ret.clust[0]=1;
    ret.sigma[0]=0;
  }
  if(n1){
    ret.clust[1]=clust1/n1;                                     // link 1 mean cluster size: <x>
    ret.sigma[1]=sqrt((sigma1/n1-(clust1/n1)*(clust1/n1))/n1);  // link 1 sigma: sqrt((<x^2>-<x>^2)/n)
  }else{
    ret.clust[1]=1;
    ret.sigma[1]=0;
  }
  printf("%cCluster size (link 1): %.5f  , sigma = %.5f\n",13,ret.clust[0],ret.sigma[0]);
  printf(  "Cluster size (link 0): %.5f  , sigma = %.5f\n"   ,ret.clust[1],ret.sigma[1]);
  B_SetEventBranchAddress();  // recover addresses of 'b_events' branches
  return ret;
}

Float_t b_near=10;  // minimal difference in channel number on link 0 and link 1: hit_position_link0+hit_position_link1-767
                    // defines minimal along strip width of beam - see above B_RunAlongPosition

#define B_HIT_ON_LINK0     1   // found hit on link 0 in beam region (see above B_RunClusterSize)
#define B_HIT_ON_LINK1     2   // found hit on link 1 in beam region (see above B_RunClusterSize)
#define B_HIT_ON_BOTH      4   // found neighbouring hits on link 0 and 1 in beam region
#define B_NOMASK_ON_LINK0  8   // correction to masked channels on link 0
#define B_NOMASK_ON_LINK1  16  // correction to masked channels on link 1

B_eff B_RunEfficiency(B_pos2 &posC,B_pos &posA,Int_t runnum=-1,Int_t scannum=-1,Char_t *filename=NULL){
// calculates efficiency (consider hits in beam region only - see above B_RunClusterSize)
/* EXAMPLES:
   B_RunEfficiency(pC,pA); - calculate efficiency for the last measured burst
   B_RunEfficiency(pC,pA,14,3); - calculate efficiency for burst from data file: sctvar/data/btrun14_3.root
   B_RunEfficiency(pC,pA,-1,-1,"d:\\btrun56_7.root"); - calculate efficiency for burst from data file: d:/btrun56_7.root
*/
  Int_t i,j,k,l;
  B_eff ret;

  if(!B_IsOpenedDataFile(runnum,scannum,filename))  // open requested data file
    if(!B_DataFile(runnum,scannum,filename)){
      printf("B_RunEfficiency: File can not be opened !\n");
      ret.eff01=ret.eff10=ret.eff=ret.n0=ret.n1=ret.n01=0;
      ret.err01=ret.err10=ret.err=-1;
      return ret;
    }

  Int_t nentries=b_events->GetEntries();
  Int_t clusts0[768/2],clusts1[768/2];
  Float_t pos0[768/2],pos1[768/2];
  Int_t counts0,counts1;
  b_events->SetBranchAddress("clust0",&clusts0);
  b_events->SetBranchAddress("clust1",&clusts1);
  b_events->SetBranchAddress("pos0",&pos0);
  b_events->SetBranchAddress("pos1",&pos1);
  b_events->SetBranchAddress("counts0",&counts0);
  b_events->SetBranchAddress("counts1",&counts1);

  Int_t n0=0,n1=0,n01=0,found,chn;
  Float_t near=posA.width;
  if(posA.width<b_near) near=b_near;  // use calculated along strips beam positions and width only if this width is greater
                                      // than value of variable 'b_near' - see above

  for(i=0;i<nentries;i++){  // loop over all in-beam-region hits to calculate efficiency
    b_events->GetEntry(i);
    if(!(i%10) || i==nentries-1) printf("%cReading data: %.1f%%",13,100.*(Float_t)(i+1)/(Float_t)nentries);
    found=B_NOMASK_ON_LINK0+B_NOMASK_ON_LINK1;  // there are no masked channels on link 0 and link 1
    for(j=0;j<counts0;j++){   // is there hit on link 0 in beam region ?
      for(k=0;k<counts1;k++)  // are there neighbouring hits on link 0 and 1 in beam region ?
        if(fabs(pos0[j]-posC.mean[0])<posC.width[0] && fabs(pos1[j]-posC.mean[1])<posC.width[1] &&
           pos0[j]-(clusts0[j]-1)/2 <= 767+posA.mean-pos1[k]+near+(clusts1[k]-1)/2 &&
           pos0[j]+(clusts0[j]-1)/2 >= 767+posA.mean-pos1[k]-near-(clusts1[k]-1)/2){
          found|=B_HIT_ON_BOTH;
          break;
        }
      if(fabs(pos0[j]-posC.mean[0])<posC.width[0]){
        found|=B_HIT_ON_LINK0;
        //for(l=-near-(clusts0[j]-1)/2;l<=near+(clusts0[j]-1)/2;l++){  // find masked channels in beam region on link 0
        //  chn=(Int_t)(767+l+posA.mean-pos0[j]);
        //  if(chn<0 || chn>767 || l<0 || l>767) continue;
        //  if(b_masks[chn+768]){
        //    found&=0xff-B_NOMASK_ON_LINK0;
        //    break;
        //  }
        //}
      }
    }
    for(k=0;k<counts1;k++)  // is there hit on link 1 in beam region ?
      if(fabs(pos1[k]-posC.mean[1])<posC.width[1]){
        found|=B_HIT_ON_LINK1;
        //for(l=-near-(clusts1[k]-1)/2;l<=near+(clusts1[k]-1)/2;l++){  // find masked channels in beam region on link 1
        //  chn=(Int_t)(767+l+posA.mean-pos1[k]);
        //  if(chn<0 || chn>767 || l<0 || l>767) continue;
        //  if(b_masks[chn]){
        //    found&=0xff-B_NOMASK_ON_LINK1;
        //    break;
        //  }
        //}
      }
    if(found&B_HIT_ON_BOTH) n01++;
    if(found&B_HIT_ON_LINK0 && (found&B_NOMASK_ON_LINK0 || found&B_HIT_ON_BOTH)) n0++;
    if(found&B_HIT_ON_LINK1 && (found&B_NOMASK_ON_LINK1 || found&B_HIT_ON_BOTH)) n1++;
  }
    
  ret.n0=n0;    // number of events containing hit in beam region on link 0
  ret.n1=n1;    // number of events containing hit in beam region on link 1
  ret.n01=n01;  // number of events containing neighbouring hits in beam region on link 0 and 1
  if(n0!=0){
    ret.eff01=100.*(Float_t)n01/(Float_t)n0;  // efficiency of link 1 (link 0 is used as reference)
    ret.err01=sqrt(n01+n0);                   // error on efficiency of link 1
  }else{
    ret.eff01=0;
    ret.err01=0;
  }
  if(n1!=0){
    ret.eff10=100.*(Float_t)n01/(Float_t)n1;  // efficiency of link 0 (link 1 is used as reference)
    ret.err10=sqrt(n01+n0);                   // error on efficiency of link 0
  }else{
    ret.eff10=0;
    ret.err10=0;
  }
  printf("%cEfficiency (link 1 [<-0]): %.2f%% = %d / %d , sigma = %.2f %%\n"  ,13,ret.eff01,n01,n0,ret.err01);
  printf(  "Efficiency (link 0 [<-1]): %.2f%% = %d / %d , sigma = %.2f %%\n\n",   ret.eff10,n01,n1,ret.err10);
  ret.eff=(ret.eff01+ret.eff10)/2;                        // mean efficiency over link 0 and 1
  ret.err=sqrt(ret.err01*ret.err01+ret.err10*ret.err10);  // error on mean efficiency over link 0 and 1
  B_SetEventBranchAddress();  // recover addresses of 'b_events' branches
  return ret;
}