/* optif_lib.c. 
 * Version 1.4. J.C.Hill. 24 June 2002.
 * V1.0. 10 August 2000. JCH.
 * Only supported on RAID and NI-VXI at present - other platforms if needed.
 * However, only optif_vmemap should be affected (apart from the includes).
 * V1.1. 16 August 2000. JCH.
 * Bug fixes in optif_read_info (stupid errors!).
 * Also add routines optif_read_laser and optif_read_threshold.
 * Due to a misunderstanding, all BPM read routines need changing to
 * return the information from the data structure, as the registers are
 * writeonly. Also several other bugs fixed.
 * V1.2 09 October 2000. PWP
 * change default values for mark space registers in optif_init
 * V1.3 24 May 2002 BJG/PWP
 * trivial changes to support NIVXI under linux
 * V1.4 24 June 2002 JCH. Support for ATLAS RCC library (for SBCs).
 */

/* includes and defines */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#ifdef NIVXI
#include <nivxi.h>
#endif

#ifdef RAID
#define CLOCKS_PER_SEC 100000
#endif

#ifdef QB_linux
#define WND_SIZE 0x1000
#include "../QB/InitVME.h"
#endif

#ifdef RCC
#include <rcc_error.h>
#include <vme_rcc.h>
#endif

#include "optif_local.h"
#include "optif_lib.h"

/* Global variables */
short optif_total = 0;           /*Total number of OPTIF boards*/
struct optif_info_structure *optif_info;      /*structure to hold OPTIF info*/


/* What follows are the set of routines to communicate with the
 * BPM4B chips on the Optif board. They have been named 
 * optif_bpm_xxxxx to make clear that they talk to the BPM 
 * registers, as opposed to the OPTIF stuff. For this reason,
 * optif_reset_bpm is not in this section, as no BPM register is
 * involved in that reset mechanism. So sue me...
 * The channelmask, pulsewidth and testdata registers in fact point to
 * a single BPM4B register, but stream-specific values are held in the
 * internal structure. This makes it easier to change things to deal with
 * a channel split of either 3:3 or 4:2 with the transmit stream mapping
 * from BPM4B to Optif.
 */

/* optif_bpm_read_channelmask
 * This routine reads the channelmask on the BPM4B.
 * The input arguments are the board identifier and the stream number
 * and the mask set is returned. There is one register address per channel,
 * but this maps to a single register. Hence the bit pattern is the
 * same for all streams on that BPM. The mask read back is compared to
 * that stored for the channel. Note that 0 means the channel is ON,
 * 1 means OFF.
 * return: -1 means invalid identifier 
 *         -2 means invalid stream number.
 *         -3 means structure still has the initialisation value.
 */

int optif_bpm_read_channelmask (identifier, stream, mask)
int identifier;
short stream;
short *mask;
{

  if ( identifier<0 || identifier>=optif_total ) return -1;

  if ( stream<0 || stream>MAX_TRANSMIT_STREAM ) return -2;

/* If the value is unset, return -3 */
  if(optif_info[identifier].bpm_channelmask[stream] < 0 ) return -3;

/* only read the lowest 4 bits */
  *mask = (optif_info[identifier].bpm_channelmask[stream])&0xF;

  return 0;

}
/* end of optif_bpm_read_channelmask */


/* optif_bpm_read_coarsedelay
 * This routine reads the coarse delay register on the BPM4B.
 * The input arguments are the board identifier and the stream number
 * The delay is returned. There is one register address per channel.
 * Only the lowest 5 bits are used. The value read is compared to the
 * internal structure and corrected if necessary.
 * return: -1 means invalid identifier 
 *         -2 means invalid stream number.
 *         -3 means the structure still has the initialisation value.
 */

int optif_bpm_read_coarsedelay (identifier, stream, delay)
int identifier;
short stream;
short *delay;
{

  if ( identifier<0 || identifier>=optif_total ) return -1;

  if ( stream<0 || stream>MAX_TRANSMIT_STREAM ) return -2;

/* If the value is unset, return -3 */
  if(optif_info[identifier].bpm_coarsedelay[stream] < 0 ) return -3;

/* only read the lowest 5 bits */
  *delay = (optif_info[identifier].bpm_coarsedelay[stream])&0x1F;

  return 0;

}
/* end of optif_bpm_read_coarsedelay */


/* optif_bpm_read_finedelay
 * This routine reads the fine delay register on the BPM4B.
 * The input arguments are the board identifier and the stream number
 * The delay is returned. There is one register address per channel.
 * Only the lowest 7 bits are used. The value read is compared to the
 * internal structure and corrected if necessary.
 * return: -1 means invalid identifier 
 *         -2 means invalid stream number.
 *         -3 means structure still has initialsation value.
 */

int optif_bpm_read_finedelay (identifier, stream, delay)
int identifier;
short stream;
short *delay;
{

  if ( identifier<0 || identifier>=optif_total ) return -1;

  if ( stream<0 || stream>MAX_TRANSMIT_STREAM ) return -2;

/* If the value is unset, return -3 */
  if(optif_info[identifier].bpm_finedelay[stream] < 0 ) return -3;

/* only read the lowest 7 bits */
  *delay = (optif_info[identifier].bpm_finedelay[stream])&0x7F;

  return 0;

}
/* end of optif_bpm_read_finedelay */


/* optif_bpm_read_mkspneg
 * This routine reads the mark-space delay for the negative edge on
 * the BPM4B.
 * The input arguments are the board identifier and the stream number
 * The delay read is output. There is one register address per channel.
 * Only the lowest 6 bits are used. The value read is compared to the
 * internal structure and this is corrected if necessary.
 * return: -1 means invalid identifier 
 *         -2 means invalid stream number.
 *         -3 means the structure still has the initialisation value.
 */

int optif_bpm_read_mkspneg (identifier, stream, mksp)
int identifier;
short stream;
short *mksp;
{

  if ( identifier<0 || identifier>=optif_total ) return -1;

  if ( stream<0 || stream>MAX_TRANSMIT_STREAM ) return -2;

/* If the value is unset, return -3 */
  if(optif_info[identifier].bpm_mkspneg[stream] < 0 ) return -3;

/* only read the lowest 6 bits */
  *mksp=(optif_info[identifier].bpm_mkspneg[stream])&0x3F;



  return 0;

}
/* end of optif_bpm_read_mkspneg */


/* optif_bpm_read_mksppos
 * This routine reads the mark-space delay for the positive edge
 * on the BPM4B.
 * The input arguments are the board identifier and the stream number
 * The delay is output. There is one register address per channel.
 * Only the lowest 6 bits are used. The value read is compared to the
 * internal structure and this is corrected if necessary.
 * return: -1 means invalid identifier.
 *         -2 means invalid stream number.
 *         -3 means the structure still has the initialisation value.
 */

int optif_bpm_read_mksppos (identifier, stream, mksp)
int identifier;
short stream;
short *mksp;
{
  
  if ( identifier<0 || identifier>=optif_total ) return -1;

  if ( stream<0 || stream>MAX_TRANSMIT_STREAM ) return -2;

/* If the value is unset, return -3 */
  if(optif_info[identifier].bpm_mksppos[stream] < 0 ) return -3;

/* only set the lowest 6 bits */
  *mksp = (optif_info[identifier].bpm_mksppos[stream])&0x3F;

  return 0;

}
/* end of optif_bpm_read_mksppos */


/* optif_bpm_read_mode
 * This routine reads the mode bit in the syncreset register on
 * the BPM4B. The other bits are ignored.
 * The input arguments are the board identifier and the stream number
 * The mode is returned. There is one register address per channel.
 * The value read is copied to the internal structure if the two don't
 * match.
 * return: -1 means invalid identifier 
 *         -2 means invalid stream number.
 *         -3 means the structure still has the initialsation value.
 */

int optif_bpm_read_mode (identifier, stream, mode)
int identifier;
short stream;
short *mode;
{
  
  if ( identifier<0 || identifier>=optif_total ) return -1;

  if ( stream<0 || stream>MAX_TRANSMIT_STREAM ) return -2;

/* If the value is unset, return -3 */
  if(optif_info[identifier].bpm_syncreset[stream] < 0 ) return -3;

/* Read the register and get the mode bit*/
  *mode = ((optif_info[identifier].bpm_syncreset[stream])>>1)&0x1;


  return 0;

}
/* end of optif_bpm_read_mode */


/* optif_bpm_read_pulsewidth
 * This routine reads the pulse width register on the BPM4B.
 * The input arguments are the board identifier and the stream number.
 * The width is output. There is one register address per channel, but
 * there is only a single register per BPM.
 * Only the lowest 2 bits are used. The value read is compared to the
 * internal structure and this is corrected if necessary.
 * return: -1 means invalid identifier.
 *         -2 means invalid stream number.
 *         -3 means the structure still has the initialisation value.
 */

int optif_bpm_read_pulsewidth (identifier, stream, width)
int identifier;
short stream;
short *width;
{
  
  if ( identifier<0 || identifier>=optif_total ) return -1;

  if ( stream<0 || stream>MAX_TRANSMIT_STREAM ) return -2;
 
/* If the value is unset, return -3 */
  if(optif_info[identifier].bpm_pulsewidth[stream] < 0 ) return -3;

/* only read the lowest 2 bits */
  *width = (optif_info[identifier].bpm_pulsewidth[stream])&0x3;

  return 0;

}
/* end of optif_bpm_read_pulsewidth */


/* optif_bpm_read_random
 * This routine reads the random/fixed bit in the syncreset register
 * on the BPM4B.
 * The input arguments are the board identifier and the stream number.
 * The setting is output. There is one register address per channel.
 * The value read is compared to the appropriate bit in the internal
 * structure and corrected if necessary.
 * return: -1 means invalid identifier 
 *         -2 means invalid stream number.
 *         -3 means structure still has initialisation value.
 */

int optif_bpm_read_random (identifier, stream, random)
int identifier;
short stream;
short *random;
{
  
  if ( identifier<0 || identifier>=optif_total ) return -1;

  if ( stream<0 || stream>MAX_TRANSMIT_STREAM ) return -2;

/* If the value is unset, return -3 */
  if(optif_info[identifier].bpm_syncreset[stream] < 0 ) return -3;

/* Read the register */
  *random = ((optif_info[identifier].bpm_syncreset[stream])>>2)&0x1;

  return 0;

}
/* end of optif_bpm_read_random */


/* optif_bpm_read_testdata
 * This routine reads the test data register on the BPM4B.
 * The input arguments are the board identifier and the stream number
 * The value is output. There is one register address per channel, but
 * only a single real register per BPM.
 * Only the lowest 4 bits are used. The value read is compared to the
 * internal structure and this is corrected if necessary.
 * return: -1 means invalid identifier 
 *         -2 means invalid stream number.
 *         -3 means the structure still has the initialisation value.
 */

int optif_bpm_read_testdata (identifier, stream, value)
int identifier;
short stream;
short *value;
{
  
  if ( identifier<0 || identifier>=optif_total ) return -1;

  if ( stream<0 || stream>MAX_TRANSMIT_STREAM ) return -2;

/* If the value is unset, return -3 */
  if(optif_info[identifier].bpm_testdata[stream] < 0 ) return -3;

/* only read the lowest 4 bits */
  *value = (optif_info[identifier].bpm_testdata[stream])&0xF;
  
  return 0;

}
/* end of optif_bpm_read_testdata */


/* optif_bpm_set_channelmask
 * This routine sets the channelmask on the BPM4B.
 * The input arguments are the board identifier, the stream number
 * and the mask to write. There is one register address per channel,
 * but this maps to a single register per BPM. Hence the bit pattern
 * is copied to the software structure for all streams on that BPM.
 * It might have been better for the third argument to have been an
 * on/off bit for this stream only - this can change on request.
 * return: -1 means invalid identifier 
 *         -2 means invalid stream number.
 */

int optif_bpm_set_channelmask (identifier, stream, mask)
int identifier;
short stream;
short mask;
{
  volatile unsigned short *ip;
  short i;

  if ( identifier<0 || identifier>=optif_total ) return -1;

  if ( stream<0 || stream>MAX_TRANSMIT_STREAM ) return -2;

/*set the register address */
  ip = (unsigned short *)
	  (optif_info[identifier].address + OPTIF_BPM_BASE 
	  + OPTIF_BPM_CHANNELMASK + (stream*0x10));

/* only set the lowest 4 bits */

  *ip = (mask&0xF);

/* Now load the same pattern into the structure for all
 * streams in use on this BPM (4 on 1st, 2 on 2nd)
 */

  for ( i=0; i<((stream<=3)?4:2); i++) {
        optif_info[identifier].bpm_channelmask[4*(stream/4)+i] = (mask&0xF);
  }

  return 0;

}
/* end of optif_bpm_set_channelmask */


/* optif_bpm_set_coarsedelay
 * This routine sets the coarse delay register on the BPM4B.
 * The input arguments are the board identifier, the stream number
 * and the delay to set. There is one register address per channel.
 * Only the lowest 5 bits are used. The value set is copied to the
 * internal structure.
 * return: -1 means invalid identifier 
 *         -2 means invalid stream number.
 */

int optif_bpm_set_coarsedelay (identifier, stream, delay)
int identifier;
short stream;
short delay;
{
  volatile unsigned short *ip;

  if ( identifier<0 || identifier>=optif_total ) return -1;

  if ( stream<0 || stream>MAX_TRANSMIT_STREAM ) return -2;

/*set the register address */
  ip = (unsigned short *)
	  (optif_info[identifier].address + OPTIF_BPM_BASE 
	  + OPTIF_BPM_COARSEDELAY + (stream*0x10));

/* only set the lowest 5 bits */

  *ip = (delay&0x1F);

/* Now load the same value into the structure */

  optif_info[identifier].bpm_coarsedelay[stream] = (delay&0x1F);

  return 0;

}
/* end of optif_bpm_set_coarsedelay */


/* optif_bpm_set_finedelay
 * This routine sets the fine delay register on the BPM4B.
 * The input arguments are the board identifier, the stream number
 * and the delay to set. There is one register address per channel.
 * Only the lowest 7 bits are used. The value set is copied to the
 * internal structure.
 * return: -1 means invalid identifier 
 *         -2 means invalid stream number.
 */

int optif_bpm_set_finedelay (identifier, stream, delay)
int identifier;
short stream;
short delay;
{
  volatile unsigned short *ip;

  if ( identifier<0 || identifier>=optif_total ) return -1;

  if ( stream<0 || stream>MAX_TRANSMIT_STREAM ) return -2;

/*set the register address */
  ip = (unsigned short *)
	  (optif_info[identifier].address + OPTIF_BPM_BASE 
	  + OPTIF_BPM_FINEDELAY + (stream*0x10));

/* only set the lowest 7 bits */

  *ip = (delay&0x7F);

/* Now load the same value into the structure */

  optif_info[identifier].bpm_finedelay[stream] = (delay&0x7F);

  return 0;

}
/* end of optif_bpm_set_finedelay */


/* optif_bpm_set_mkspneg
 * This routine sets the mark-space delay for the negative edge on
 * the BPM4B.
 * The input arguments are the board identifier, the stream number
 * and the delay to set. There is one register address per channel.
 * Only the lowest 6 bits are used. The value set is copied to the
 * internal structure.
 * return: -1 means invalid identifier 
 *         -2 means invalid stream number.
 */

int optif_bpm_set_mkspneg (identifier, stream, mksp)
int identifier;
short stream;
short mksp;
{
  volatile unsigned short *ip;

  if ( identifier<0 || identifier>=optif_total ) return -1;

  if ( stream<0 || stream>MAX_TRANSMIT_STREAM ) return -2;

/*set the register address */
  ip = (unsigned short *)
	  (optif_info[identifier].address + OPTIF_BPM_BASE 
	  + OPTIF_BPM_MKSPNEG + (stream*0x10));

/* only set the lowest 6 bits */

  *ip = (mksp&0x3F);

/* Now load the same value into the structure */

  optif_info[identifier].bpm_mkspneg[stream] = (mksp&0x3F);

  return 0;

}
/* end of optif_bpm_set_mkspneg */


/* optif_bpm_set_mksppos
 * This routine sets the mark-space delay for the positive edge
 * on the BPM4B.
 * The input arguments are the board identifier, the stream number
 * and the delay to set. There is one register address per channel.
 * Only the lowest 6 bits are used. The value set is copied to the
 * internal structure.
 * return: -1 means invalid identifier 
 *         -2 means invalid stream number.
 */

int optif_bpm_set_mksppos (identifier, stream, mksp)
int identifier;
short stream;
short mksp;
{
  volatile unsigned short *ip;

  if ( identifier<0 || identifier>=optif_total ) return -1;

  if ( stream<0 || stream>MAX_TRANSMIT_STREAM ) return -2;

/*set the register address */
  ip = (unsigned short *)
	  (optif_info[identifier].address + OPTIF_BPM_BASE 
	  + OPTIF_BPM_MKSPPOS + (stream*0x10));

/* only set the lowest 6 bits */

  *ip = (mksp&0x3F);

/* Now load the same value into the structure */

  optif_info[identifier].bpm_mksppos[stream] = (mksp&0x3F);
  
  return 0;

}
/* end of optif_bpm_set_mksppos */


/* optif_bpm_set_mode
 * This routine sets the mode bit in the syncreset register on
 * the BPM4B. The other bits are unchanged.
 * The input arguments are the board identifier, the stream number
 * and the delay to set. There is one register address per channel.
 * The value set is copied to the internal structure.
 * return: -1 means invalid identifier 
 *         -2 means invalid stream number.
 */

int optif_bpm_set_mode (identifier, stream, mode)
int identifier;
short stream;
short mode;
{
  volatile unsigned short *ip;

  if ( identifier<0 || identifier>=optif_total ) return -1;

  if ( stream<0 || stream>MAX_TRANSMIT_STREAM ) return -2;

/*set the register address */
  ip = (unsigned short *)
	  (optif_info[identifier].address + OPTIF_BPM_BASE 
	  + OPTIF_BPM_SYNCRESET + (stream*0x10));

/* Load the value into the correct bit in the structure. If the
 * syncreset register has not previously been written-to, we
 * should also set the fixed/random bit (arbitrarily I choose 0)
 * and ensure the syncreset bit is ON. 
 */
  if (optif_info[identifier].bpm_syncreset[stream] < 0) {
	  optif_info[identifier].bpm_syncreset[stream] = 0x1;
  }
  optif_info[identifier].bpm_syncreset[stream] |= ((mode&1)<<1);

/* then set the register */

  *ip = optif_info[identifier].bpm_syncreset[stream];

  return 0;

}
/* end of optif_bpm_set_mode */


/* optif_bpm_set_pulsewidth
 * This routine sets the pulse width register on the BPM4B.
 * The input arguments are the board identifier, the stream number
 * and the width to set. There is one register address per channel,
 * but this maps to a single register per BPM4B.
 * Only the lowest 2 bits are used. The value set is copied to the
 * internal structure.
 * return: -1 means invalid identifier 
 *         -2 means invalid stream number.
 */

int optif_bpm_set_pulsewidth (identifier, stream, width)
int identifier;
short stream;
short width;
{
  volatile unsigned short *ip;
  short i;

  if ( identifier<0 || identifier>=optif_total ) return -1;

  if ( stream<0 || stream>MAX_TRANSMIT_STREAM ) return -2;

/*set the register address */
  ip = (unsigned short *)
	  (optif_info[identifier].address + OPTIF_BPM_BASE 
	  + OPTIF_BPM_PULSEWIDTH + (stream*0x10));

/* only set the lowest 2 bits */

  *ip = (width&0x3);

 /* Now load the same pattern into the structure for all
 * streams in use on this BPM (4 on 1st, 2 on 2nd)
 */

  for ( i=0; i<((stream<=3)?4:2); i++) {
        optif_info[identifier].bpm_pulsewidth[4*(stream/4)+i] = (width&0x3);
  }

  return 0;

}
/* end of optif_bpm_set_pulsewidth */


/* optif_bpm_set_random
 * This routine sets the random/fixed bit in the syncreset register
 * on the BPM4B.
 * The input arguments are the board identifier, the stream number
 * and the setting. There is one register address per channel.
 * The value set is copied to the appropriate bit in the internal
 * structure.
 * return: -1 means invalid identifier 
 *         -2 means invalid stream number.
 */

int optif_bpm_set_random (identifier, stream, random)
int identifier;
short stream;
short random;
{
  volatile unsigned short *ip;

  if ( identifier<0 || identifier>=optif_total ) return -1;

  if ( stream<0 || stream>MAX_TRANSMIT_STREAM ) return -2;

/*set the register address */
  ip = (unsigned short *)
	  (optif_info[identifier].address + OPTIF_BPM_BASE 
	  + OPTIF_BPM_SYNCRESET + (stream*0x10));

 /* Load the value into the correct bit in the structure. If the
 * syncreset register has not previously been written-to, we
 * should also set the mode bit (arbitrarily I choose 0)
 * and ensure the syncreset bit is ON. 
 */
  if (optif_info[identifier].bpm_syncreset[stream] < 0) {
	  optif_info[identifier].bpm_syncreset[stream] = 0x1;
  }
  optif_info[identifier].bpm_syncreset[stream] |= ((random&0x1)<<2);

/* Now load the register */

  *ip = optif_info[identifier].bpm_syncreset[stream];

  return 0;

}
/* end of optif_bpm_set_random */


/* optif_bpm_set_testdata
 * This routine sets the test data register on the BPM4B.
 * The input arguments are the board identifier, the stream number
 * and the value to set. There is one register address per channel,
 * but only one register per BPM4B.
 * Only the lowest 4 bits are used. The value set is copied to the
 * internal structure.
 * return: -1 means invalid identifier 
 *         -2 means invalid stream number.
 */

int optif_bpm_set_testdata (identifier, stream, value)
int identifier;
short stream;
short value;
{
  volatile unsigned short *ip;
  short i;

  if ( identifier<0 || identifier>=optif_total ) return -1;

  if ( stream<0 || stream>MAX_TRANSMIT_STREAM ) return -2;

/*set the register address */
  ip = (unsigned short *)
	  (optif_info[identifier].address + OPTIF_BPM_BASE 
	  + OPTIF_BPM_TESTDATA + (stream*0x10));

/* only set the lowest 4 bits */

  *ip = (value&0xF);

/* Now load the same pattern into the structure for all
 * streams in use on this BPM (4 on 1st, 2 on 2nd)
 */

  for ( i=0; i<((stream<=3)?4:2); i++) {
        optif_info[identifier].bpm_testdata[4*(stream/4)+i] = (value&0xF);
  }

  return 0;

}
/* end of optif_bpm_set_testdata */


/* optif_bpm_syncreset
 * This routine issues a syncreset to a BPM4B channel.
 * The input arguments are the board identifier and the transmit
 * stream to reset. The reset is issued by writing 0 followed
 * by 1 into the appropriate bit of the syncreset register.
 * return: -1 means invalid identifier 
 *         -2 means invalid stream number.
 */

int optif_bpm_syncreset (identifier, stream)
int identifier;
short stream;
{
  int i;
  volatile unsigned short *ip;

  if ( identifier<0 || identifier>=optif_total ) return -1;

  if ( stream<0 || stream>MAX_TRANSMIT_STREAM ) return -2;

/*set the register address */
  ip = (unsigned short *)
	  (optif_info[identifier].address + OPTIF_BPM_BASE 
	  + OPTIF_BPM_SYNCRESET + (stream*0x10));

/* get the other bits from the structure and force bit
 * zero off, then on.
 */

  *ip = (optif_info[identifier].bpm_syncreset[stream] & ~0x1);
  *ip = (optif_info[identifier].bpm_syncreset[stream] | 0x1);

/* Now initialise the BPM part of the structure to -1, as the software
 * no longer knows what the settings are */
  for (i=0; i<=MAX_TRANSMIT_STREAM; i++) {
	  optif_info[identifier].bpm_syncreset[i] = -1;
	  optif_info[identifier].bpm_pulsewidth[i] = -1;
	  optif_info[identifier].bpm_coarsedelay[i] = -1;
	  optif_info[identifier].bpm_finedelay[i] = -1;
	  optif_info[identifier].bpm_mkspneg[i] = -1;
	  optif_info[identifier].bpm_mksppos[i] = -1;
	  optif_info[identifier].bpm_channelmask[i] = -1;
	  optif_info[identifier].bpm_testdata[i] = -1;
  }

  return 0;

}
/* end of optif_bpm_syncreset */

/* End of set of routines to access the BPM4B chips */


/* optif_exit.
 * Exit routine for OPTIF library. Should be called once only to tidy things
 * up. Unmaps the modules and deletes the malloced memory.
 */

int optif_exit()
{
  int i;
#ifdef RCC
  VME_ErrorCode_t status;
#else
  int status;
#endif
  
/* loop over the modules */
  for(i=0;i<optif_total;i++) {
#ifdef RAID
    status = RD13_VmeUnmap(optif_info[i].address);
#endif
#ifdef NIVXI
    status = UnMapVXIAddress(optif_info[i].window);
#endif
#ifdef QB_linux
    munmap((void *)(optif_info[i].address),WND_SIZE);
#endif
#ifdef RCC
    status = VME_MasterUnmap(optif_info[i].window);
#endif
  }  /*end of loop*/

#ifdef NIVXI
    CloseVXIlibrary();
#endif
#ifdef RCC
    status = VME_Close();
#endif

/* release memory */
  free((void *)optif_info);
  return 0;
}
/* end of optif_exit */


/* optif_init.
 * Initialise the OPTIF. Makes sure the module is in a standard state.
 * How much of the following is needed is open to investigation by
 * the real user.
 * Full setup of the BPM4B is done - this is to ensure the software knows
 * the state of all the (writeonly) registers.
 * The user should feel free to modify this routine to set OPTIF to an
 * appropriate initial state.
 * Input argument: OPTIF identifier.
 * Returns <0 on any failed call and passes on the return value.
 */

int optif_init(identifier)
int identifier;
{
  int status; 
  short stream;

/* Don't invert the clock */
  status = optif_set_clock_invert(identifier,0);
  if( status != 0) {
    printf("optif_init: optif_set_clock_invert returns %d\n",status);
    return status;
  }

/* Don't invert the command streams */
  status = optif_set_command_invert(identifier,0);
  if( status != 0) {
    printf("optif_init: optif_set_command_invert returns %d\n",status);
    return status;
  }

/* Don't invert the receive streams. 1 register for 4 streams */
  for( stream=0; stream<=MAX_RECEIVE_STREAM; stream+=4 ) {
    status = optif_set_receiver_invert(identifier,stream,1);
    if( status != 0) {
      printf("optif_init: optif_set_receiver_invert returns %d\n",status);
      return status;
    }
  }

/* Reset the BPM chips */
  status = optif_reset_bpm(identifier,-1);
  if( status != 0) {
    printf("optif_init: optif_reset_bpm returns %d\n",status);
    return status;
  }

/* Set threshold midway on all receive streams - reduces noise
   Threshold used to be initialised to zero, but Maurice and
   Tony have persuaded me that a higher level is more sensible!
   PWP 04.07.02 */
  for( stream=0 ; stream<=MAX_RECEIVE_STREAM ; stream++ ) {
    status = optif_set_threshold(identifier,stream,120);
    if( status != 0) {
      printf("optif_init: optif_set_threshold returns %d\n",status);
      return status;
    }
  }

/* Set the currents on all transmit streams to a nominal value.
 * MJG recommends a non-zero value.
 */
  for( stream=0 ; stream<=MAX_TRANSMIT_STREAM ; stream++ ) {
    status = optif_set_laser(identifier,stream,0x10);
    if( status != 0) {
      printf("optif_init: optif_set_laser returns %d\n",status);
      return status;
    }
  }
/* Set the various BPM registers. These are probably the values after
 * a reset, but the software needs to know what the registers are set
 * to, so set then anyway.
 */

/* Loop over the streams for the registers that exist per stream */

/* Set the mode and fixed/random bits in the syncreset register */
  for( stream=0; stream<=MAX_TRANSMIT_STREAM; stream++ ) {
	status = optif_bpm_set_mode(identifier, stream, 0);
	if( status != 0) {
      printf("optif_init: optif_bpm_set_mode returns %d\n",status);
      return status;
	}
	status = optif_bpm_set_random(identifier, stream, 0);
	if( status != 0) {
      printf("optif_init: optif_set_random returns %d\n",status);
      return status;
	}
/* Now set the delays */
	status = optif_bpm_set_coarsedelay(identifier, stream, 0);
	if( status != 0) {
      printf("optif_init: optif_bpm_set_coarsedelay returns %d\n",status);
      return status;
	}
   	status = optif_bpm_set_finedelay(identifier, stream, 0);
	if( status != 0) {
      printf("optif_init: optif_bpm_set_finedelay returns %d\n",status);
      return status;
	}
/* Next the mark-space registers */
	status = optif_bpm_set_mkspneg(identifier, stream, 14);  /* was 0 */
	if( status != 0) {
      printf("optif_init: optif_bpm_set_mkspneg returns %d\n",status);
      return status;
	}
   	status = optif_bpm_set_mksppos(identifier, stream, 12);  /* was 0 */
	if( status != 0) {
      printf("optif_init: optif_bpm_set_mksppos returns %d\n",status);
      return status;
	}
  }

/* Now do the registers that only exist once per BPM. The internal
 * structure keeps per stream versions of this information
 */
  for (stream=0; stream<=MAX_TRANSMIT_STREAM; stream+=4) {
/* start with the channelmask. Remember a 0 means the channel is ON
 * so initialise the streams to be all OFF */
    status = optif_bpm_set_channelmask(identifier, stream, 0xF);
  	  if( status != 0) {
        printf("optif_init: optif_bpm_set_channelmask returns %d\n",status);
        return status;
	  }
/* Now the pulsewidth */
    status = optif_bpm_set_pulsewidth(identifier, stream, 0);
	  if( status != 0) {
        printf("optif_init: optif_bpm_set_pulsewidth returns %d\n",status);
        return status;
	  }
/* Now the test data register */
    status = optif_bpm_set_testdata(identifier, stream, 0);
	  if( status != 0) {
        printf("optif_init: optif_bpm_set_testdata returns %d\n",status);
        return status;
	  }
  }
/* That should be everything nailed down */
  return 0;

}
/* end of optif_init */


/* optif_read_clock_invert
 * This routine reads the clock invert register.
 * The input arguments are the board identifier and
 * the vaue read is returned. The value read and the
 * contents of the structure are also compared, with the
 * latter corrected if different.
 * return: -1 means invalid identifier 
 *         -2 means hardware and software disagree.
 */

int optif_read_clock_invert (identifier, value)
int identifier;
short *value;
{
  volatile unsigned short *ip;

  if ( identifier<0 || identifier>=optif_total ) return -1;

/*set the register address */
  ip = (unsigned short *)
	  (optif_info[identifier].address + OPTIF_INVERT_CLOCK);

/*  bit 0 only is used */
  *value = (*ip)&0x1;

/* also check the structure and correct if necessary*/
  if( optif_info[identifier].invert_clock != *value ) {
          printf("optif_read_clock_invert: discrepancy ");
          printf("hardware %d, software %d\n",
		  *value, optif_info[identifier].invert_clock);
          optif_info[identifier].invert_clock = *value;
	    return -2;
  }

  return 0;

}
/* end of optif_read_clock_invert */


/* optif_read_command_invert.
 * This routine reads the command streams invert register.
 * The input arguments are the board identifier and
 * the value read is returned. The read value is compared
 * with the contents of the structure, with the latter 
 * corrected if different.
 * return: -1 means invalid identifier
 *         -2 means hardware and software disagree.
 */

int optif_read_command_invert (identifier, value)
int identifier;
short *value;
{
  volatile unsigned short *ip;

  if ( identifier<0 || identifier>=optif_total ) return -1;

/*set the register address */
  ip = (unsigned short *)
	  (optif_info[identifier].address + OPTIF_INVERT_COMMAND);

/*  bit 0 only is used */
  *value = (*ip)&0x1;

/* also check in the structure and correct if necessary*/
  if( optif_info[identifier].invert_command != *value ) {
          printf("optif_read_command_invert: discrepancy ");
	    printf("hardware %d, software %d\n",
		  *value, optif_info[identifier].invert_command);
          optif_info[identifier].invert_command = *value;
	    return -2;
  }

  return 0;

}
/* end of optif_read_command_invert */


/* optif_read_info
 * This routine reads the version information etc. registers,
 * and returns the information. The LS and MS bytes of the
 * manufacturer information are merged. Possibly this info
 * should be in the structure??
 */

int optif_read_info (identifier, firmware, manufacturer, version, type)
int identifier;
short *firmware;
short *manufacturer;
short *version;
short *type;
{
  volatile unsigned short *ip;
  short i;

/*set the register address */
  ip = (unsigned short *)
	  (optif_info[identifier].address + OPTIF_INFO);

/*  only get the lowest 8 bits in each case */
  
  *type = (*ip++)&0xFF;
  *version = (*ip++)&0xFF;
  i = (*ip++)&0xFF;  /* LS byte of manufacturer */
  *manufacturer = i + (((*ip++)&0xFF)<<8);
  *firmware = (*ip++)&0xFF;

  return 0;

}
/* end of optif_read_info */


/* optif_read_laser
 * This routine "reads" the laser setting for a stream.
 * The input arguments are the board identifier, stream number and
 * the value read is returned. Since the laser registers are 
 * writeonly, the contents of the structure are returned.
 * return: -1 means invalid identifier.
 *         -2 means invalid stream
 *         -3 means the structure still has the initialisation value.
 */

int optif_read_laser( identifier, stream, value)
int identifier;
short stream;
short *value;
{

  if ( identifier<0 || identifier>=optif_total ) return -1;

  if ( stream<0 || stream>MAX_TRANSMIT_STREAM ) return -2;


/*  bits 0-7 only are used */
  *value = (optif_info[identifier].laser[stream])&0xFF;

/* return -3 if the structure is unset */
  if (*value < 0) return -3;

  return 0;

}
/* end of optif_read_laser */

/* optif_read_receiver_invert
 * This routine reads the invert register for the receiver on a stream.
 * The input arguments are the board identifier, stream number and
 * the value read is returned. Also the read value and the structure
 * contents are compared and the latter updated if necessary.
 * 0 = don't invert, 1 = invert.
 * return: -1 means invalid identifier.
 *         -2 means invalid stream
 *         -3 means hardware and software don't agree.
 */

int optif_read_receiver_invert (identifier, stream, value)
int identifier;
short stream;
short *value;
{
  volatile unsigned short *ip;
  short i;
  int status;

  if ( identifier<0 || identifier>=optif_total ) return -1;

  if ( stream<0 || stream>MAX_RECEIVE_STREAM ) return -2;

/*set the register address */
  ip = (unsigned short *)
	  (optif_info[identifier].address + OPTIF_INVERT_RECEIVER + 2*(stream/4));

/*  bit 0 only is used */
  *value = (*ip)&1;

/* also check in the structure. Only one register per 4 streams is used, but the
 * software stores the setting for each stream separately. Hence we check 4 streams
 * below. This makes the code easier (I hope) to understand, and makes it easier to
 * change the code if one register/stream is ever used in the future.*/
  status = 0;
  for ( i=0 ; i<4 ; i++ ) {
	  if( optif_info[identifier].invert_receiver[i+(stream/4)*4] != *value ) {
		printf("optif_read_receiver_invert: discrepancy, stream %d ",i+(stream/4)*4);
		printf("hardware %d, software %d\n",
		    *value, optif_info[identifier].invert_receiver[i+(stream/4)*4]);
            optif_info[identifier].invert_receiver[i+(stream/4)*4] = *value;
		status = -3;
	  }
  }

  return status;

}
/* end of optif_read_receiver_invert */


/* optif_read_threshold
 * This routine "reads" the threshold setting for a stream.
 * The input arguments are the board identifier, stream number and
 * the value read is returned. Since the threshold registers are 
 * writeonly, the contents of the structure are returned.
 * return: -1 means invalid identifier.
 *         -2 means invalid stream
 *         -3 means the structure still has the initialisation value.
 */

int optif_read_threshold( identifier, stream, value)
int identifier;
short stream;
short *value;
{

  if ( identifier<0 || identifier>=optif_total ) return -1;

  if ( stream<0 || stream>MAX_RECEIVE_STREAM ) return -2;

/* return -3 if the structure is unset */
  if (optif_info[identifier].threshold[stream] < 0) return -3;

/*  bits 0-7 only are used */
  *value = (optif_info[identifier].threshold[stream])&0xFF;

  return 0;

}
/* end of optif_read_threshold */


/* optif_reset_bpm
 * This routine is used to reset one of the BPM4B chips on 
 * OPTIF. To reset requires writing 1 followed by 0 to the
 * reset address.
 * The input arguments are the board identifier and the chip
 * number (<0 means both chips).
 * return: -1 means invalid identifier
 *         -2 means invalid chip number (ie. 0 or 1 only). 
 */

int optif_reset_bpm (identifier, bpm)
int identifier;
short bpm;
{
  int i;
  volatile unsigned short *ip;

  if ( identifier<0 || identifier>=optif_total ) return -1;

  if ( bpm>1 ) return -2;

  ip = (unsigned short *)
	  (optif_info[identifier].address + OPTIF_RESET_BPM);

  if( bpm<0 ) {
/* reset both BPM chips */
	*ip = 1;
    *ip++ = 0;  /*move on to the 2nd BPM */
    *ip = 1;
	*ip = 0;
  }
  else {
/* do the specific BPM chip */
    *(ip+bpm) = 1;
	*(ip+bpm) = 0;
  }

/* Now initialise the BPM part of the structure to -1. This is to
 * indicate that the software doesn't KNOW what the settings currently
 * are.
 */
  for (i=0; i<=MAX_TRANSMIT_STREAM; i++) {
	  optif_info[identifier].bpm_syncreset[i] = -1;
	  optif_info[identifier].bpm_pulsewidth[i] = -1;
	  optif_info[identifier].bpm_coarsedelay[i] = -1;
	  optif_info[identifier].bpm_finedelay[i] = -1;
	  optif_info[identifier].bpm_mkspneg[i] = -1;
	  optif_info[identifier].bpm_mksppos[i] = -1;
	  optif_info[identifier].bpm_channelmask[i] = -1;
	  optif_info[identifier].bpm_testdata[i] = -1;
  }

  return 0;

}
/* end of optif_reset_bpm */


/* optif_set_clock_invert
 * This routine inverts (or not) the clock.
 * The input arguments are the board identifier and
 * the value to set.
 * 0 = don't invert, 1 = invert.
 * return: -1 means invalid identifier 
 */

int optif_set_clock_invert (identifier, value)
int identifier;
short value;
{
  volatile unsigned short *ip;

  if ( identifier<0 || identifier>=optif_total ) return -1;

/*set the register address */
  ip = (unsigned short *)
	  (optif_info[identifier].address + OPTIF_INVERT_CLOCK);

/*  bit 0 only is used */
  *ip = value&0x1;

/* also set in the structure */
  optif_info[identifier].invert_clock = value&0x1;

  return 0;

}
/* end of optif_set_clock_invert */


/* optif_set_command_invert.
 * This routine inverts (or not) the command streams.
 * The input arguments are the board identifier and
 * the value to set.
 * 0 = don't invert, 1 = invert.
 * return: -1 means invalid identifier 
 */

int optif_set_command_invert (identifier, value)
int identifier;
short value;
{
  volatile unsigned short *ip;

  if ( identifier<0 || identifier>=optif_total ) return -1;

/*set the register address */
  ip = (unsigned short *)
	  (optif_info[identifier].address + OPTIF_INVERT_COMMAND);

/*  bit 0 only is used */
  *ip = value&0x1;

/* also set in the structure */
  optif_info[identifier].invert_command = value&0x1;

  return 0;

}
/* end of optif_set_command_invert */


/* optif_set_laser
 * This routine sets the current for the laser on a transmit channel.
 * The input arguments are the board identifier, stream number and
 * the value to set.
 * return: -1 means invalid identifier
 *         -2 means invalid stream number. 
 */

int optif_set_laser (identifier, stream, value)
int identifier;
short stream;
short value;
{

  volatile unsigned short *ip;

  if ( identifier<0 || identifier>=optif_total ) return -1;

  if ( stream<0 || stream>MAX_TRANSMIT_STREAM ) return -2;

/*set the register address */
  ip = (unsigned short *)
	  (optif_info[identifier].address + OPTIF_LASER + 2*stream);

/* low 8 bits only are used */
  *ip = value&0xFF;

/* also set in the structure */
  optif_info[identifier].laser[stream] = value&0xFF;

  return 0;

}

/* end of optif_set_laser */


/* optif_set_receiver_invert
 * This routine inverts (or not) the receiver on a stream.
 * The input arguments are the board identifier, stream number and
 * the value to set.
 * 0 = don't invert, 1 = invert.
 * return: -1 means invalid identifier.
 *         -2 means invalid stream.
 */

int optif_set_receiver_invert (identifier, stream, value)
int identifier;
short stream;
short value;
{
  volatile unsigned short *ip;
  short i;

  if ( identifier<0 || identifier>=optif_total ) return -1;

  if ( stream<0 || stream>MAX_RECEIVE_STREAM ) return -2;

/*set the register address */
  ip = (unsigned short *)
	  (optif_info[identifier].address + OPTIF_INVERT_RECEIVER + 2*(stream/4));

/*  bit 0 only is used */
  *ip = value&0x1;

/* also set in the structure. Only one register per 4 streams is used, but the
 * software stores the setting for each stream separately. Hence we set 4 streams
 * below. This makes the code easier (I hope) to understand, and makes it easier to
 * change the code if one register/stream is ever used in the future.*/
  for ( i=0 ; i<4 ; i++ ) {
    optif_info[identifier].invert_receiver[i+(stream/4)*4] = value&0x1;
  }

  return 0;

}
/* end of optif_set_receiver_invert */


/* optif_set_threshold
 * This routine sets the threshold on a receive channel.
 * The input arguments are the board identifier, stream number and
 * the value to set.
 * return: -1 means invalid identifier
 *         -2 means invalid stream number. 
 */

int optif_set_threshold (identifier, stream, value)
int identifier;
short stream;
short value;
{
  volatile unsigned short *ip;

  if ( identifier<0 || identifier>=optif_total ) return -1;

  if ( stream<0 || stream>MAX_RECEIVE_STREAM ) return -2;

/*set the register address */
  ip = (unsigned short *)
	  (optif_info[identifier].address + OPTIF_THRESHOLD + 2*stream);

/* low 8 bits only are used */
  *ip = value&0xFF;

/* also set in the structure */
  optif_info[identifier].threshold[stream] = value&0xFF;

  return 0;

}
/* end of optif_set_threshold */


/* optif_status
 * Print out the status of an OPTIF. The input identifier identifies the OPTIF.
 * The information is taken mainly from the structure, as the OPTIF has
 * write-only registers for the most part. Where read/write registers exist, 
 * the information is read from the hardware, which forces an update to the
 * structure.
 * Return -1 if the identifier is illegal.
 */

int optif_status(identifier)
int identifier;
{
  int status;
  short stream;
  short value;
  short v1,v2,v3,v4;
 
  if ( identifier<0 || identifier>=optif_total ) return -1;

  printf("OPTIF #%d status\n\n",identifier);

/* First printout the identification registers */

  status = optif_read_info(identifier, &v1, &v2, &v3, &v4);

  printf("Hardware version %d, Type %d, Firmware version %d, Manufacturer %hx\n\n",
	v3, v4, v1, v2);  

/* print out the inverter settings */

/* first the clock */

  status = optif_read_clock_invert(identifier, &value);
  if( status!=0 ) {
/* report the error */
	  printf("Error from optif_read_clock_invert: %d\n",status);
  }
/* report the value - optif_read_clock_invert checks the hardware
 * and structure agree and updates the structure if not. 
 */
  printf("Polarities: clock: %hd",optif_info[identifier].invert_clock);

/* next the command */

  status = optif_read_command_invert(identifier, &value);
  if( status!=0 ) {
/* report the error */
	  printf("Error from optif_read_command_invert: %d\n",status);
  }
/* report the value - optif_read_command_invert checks the hardware
 * and structure agree and updates the structure if not. 
 */

  printf(" command: %hd",optif_info[identifier].invert_command);

/* finally get the receivers */

  for( stream=0; stream<11; stream+=4 ) {
       status = optif_read_receiver_invert(identifier, stream, &value);
       if( status!=0 ) {
/* report the error */
	    printf("Error from optif_read_receiver_invert: %d\n",status);
	   }
  }
/* report the value - optif_read_receiver_invert checks the hardware
 * and structure agree and updates the structure if not. 
 */

  printf(" Receiver streams (0-3,4-7,8-11): %hd %hd %hd\n\n",
    optif_info[identifier].invert_receiver[0],
    optif_info[identifier].invert_receiver[4],
    optif_info[identifier].invert_receiver[8]);

/* print the laser currents */

  printf("Laser currents:");
  for( stream=0 ; stream<=MAX_TRANSMIT_STREAM ; stream++ ) {
    printf(" %d",optif_info[identifier].laser[stream]);
  }
  printf("\n\n");

/* and the thresholds */

  printf("Receiver thresholds:");
  for( stream=0; stream<=MAX_RECEIVE_STREAM; stream++ ) {
    printf(" %d",optif_info[identifier].threshold[stream]);
  }
  printf("\n\n");

/* now BPM4B information */
  printf ("BPM settings:\n");
/* loop over streams */
  for( stream=0; stream<=MAX_TRANSMIT_STREAM; stream++ ) {
	status = optif_bpm_read_channelmask(identifier, stream, &value);
	if( status!=0 ) {
/* report the error */
	  printf("Error from optif_bpm_read_channelmask: %d\n",status);
	}
	if (stream==0) printf("Channelmask:");
	printf(" %d",optif_info[identifier].bpm_channelmask[stream]);
	if (stream==MAX_TRANSMIT_STREAM) printf("\n");
  }
/* */
  for( stream=0; stream<=MAX_TRANSMIT_STREAM; stream++ ) {
   	status = optif_bpm_read_coarsedelay(identifier, stream, &value);
	if( status!=0 ) {
/* report the error */
	  printf("Error from optif_bpm_read_coarsedelay: %d\n",status);
	}
	if (stream==0) printf("Coarse delay:");
	printf(" %d",optif_info[identifier].bpm_coarsedelay[stream]);
	if (stream==MAX_TRANSMIT_STREAM) printf("\n");
  }
/* */
  for( stream=0; stream<=MAX_TRANSMIT_STREAM; stream++ ) {
 	status = optif_bpm_read_finedelay(identifier, stream, &value);
	if( status!=0 ) {
/* report the error */
	  printf("Error from optif_bpm_read_finedelay: %d\n",status);
	}
	if (stream==0) printf("Fine delay:");
	printf(" %d",optif_info[identifier].bpm_finedelay[stream]);
	if (stream==MAX_TRANSMIT_STREAM) printf("\n");
  }
/* */
  for( stream=0; stream<=MAX_TRANSMIT_STREAM; stream++ ) {
	status = optif_bpm_read_mkspneg(identifier, stream, &value);
	if( status!=0 ) {
/* report the error */
	  printf("Error from optif_bpm_read_mkspneg: %d\n",status);
	}
	if (stream==0) printf("Mark/space, negative edge:");
	printf(" %d",optif_info[identifier].bpm_mkspneg[stream]);
	if (stream==MAX_TRANSMIT_STREAM) printf("\n");
  }
 /* */
  for( stream=0; stream<=MAX_TRANSMIT_STREAM; stream++ ) {
	status = optif_bpm_read_mksppos(identifier, stream, &value);
	if( status!=0 ) {
/* report the error */
	  printf("Error from optif_bpm_read_mksppos: %d\n",status);
	}
	if (stream==0) printf("Mark/space, positive edge:");
	printf(" %d",optif_info[identifier].bpm_mksppos[stream]);
	if (stream==MAX_TRANSMIT_STREAM) printf("\n");
  }
/* */
  for( stream=0; stream<=MAX_TRANSMIT_STREAM; stream++ ) {
	status = optif_bpm_read_pulsewidth(identifier, stream, &value);
	if( status!=0 ) {
/* report the error */
	  printf("Error from optif_bpm_read_pulsewidth: %d\n",status);
	}
	if (stream==0) printf("Pulse width:");
	printf(" %d",optif_info[identifier].bpm_pulsewidth[stream]);
	if (stream==MAX_TRANSMIT_STREAM) printf("\n");
  }
 /* for syncreset register, there are routines to access the mode
  * and fixed/random bits. However, it is probably sensible for this
  * status routine to report the combined register contents */
  for( stream=0; stream<=MAX_TRANSMIT_STREAM; stream++ ) {
	status = optif_bpm_read_mode(identifier, stream, &value);
	if( status!=0 ) {
/* report the error */
	  printf("Error from optif_bpm_read_mode: %d\n",status);
	}
	status = optif_bpm_read_random(identifier, stream, &value);
	if( status!=0 ) {
/* report the error */
	  printf("Error from optif_bpm_read_random: %d\n",status);
	}
	if (stream==0) printf("Syncreset register:");
	printf(" %d",optif_info[identifier].bpm_syncreset[stream]);
	if (stream==MAX_TRANSMIT_STREAM) printf("\n");
  }
/* */
  for( stream=0; stream<=MAX_TRANSMIT_STREAM; stream++ ) {
	status = optif_bpm_read_testdata(identifier, stream, &value);
	if( status!=0 ) {
/* report the error */
	  printf("Error from optif_bpm_read_testdata: %d\n",status);
	}
	if (stream==0) printf("Test data register:");
	printf(" %d",optif_info[identifier].bpm_testdata[stream]);
	if (stream==MAX_TRANSMIT_STREAM) printf("\n");
  }
/* I think that reports everything */
  return 0;

}
/* end of optif_status */


/* optif_vmemap
 * Map the OPTIF to the CPU address space. The input address is the true
 * VME address for UNIX and WNT implementations, mapped address for OS9 (this
 * is because the mapping in OS9 is CPU-specific and so it is not possible to
 * provide a generalised mapping within this routine). The mapping in UNIX/WNT
 * is NOT A32 - OPTIF is an A24 module.
 * For the SCTDAQ environment, this routine will probably need changing to cope
 * with the NI 4-window limit.
 * A failure in the mapping causes a non-zero return, with the status from the
 * mapping routine in the return value.
 */

int optif_vmemap(address, identifier)
unsigned long address;
int *identifier;
{
#if defined(NIVXI)
  short ret;
  UINT32 w,wstart,wend;
#elif defined(RCC)
  VME_ErrorCode_t status;
  VME_MasterMap_t master_map;
  u_int virtual_address;
  int handle;
#else
  int status;
#endif
  int i;
  int am;
  char *addr;

  am = 0;
  *identifier = optif_total++;		/*Simplest approach? */

/* Use malloc/realloc to get space for the structure */
  if(optif_total == 1) {
    optif_info = (struct optif_info_structure *) malloc(optif_info_size);
  }
  else {
    optif_info = (struct optif_info_structure *) realloc((void *)optif_info,optif_total*optif_info_size);
  }

/* Now initialise most of the structure to -1 (apart from the odd 0!) */
  for (i=0; i<=MAX_TRANSMIT_STREAM; i++) {
	  optif_info[*identifier].bpm_syncreset[i] = -1;
	  optif_info[*identifier].bpm_pulsewidth[i] = -1;
	  optif_info[*identifier].bpm_coarsedelay[i] = -1;
	  optif_info[*identifier].bpm_finedelay[i] = -1;
	  optif_info[*identifier].bpm_mkspneg[i] = -1;
	  optif_info[*identifier].bpm_mksppos[i] = -1;
	  optif_info[*identifier].bpm_channelmask[i] = -1;
	  optif_info[*identifier].bpm_testdata[i] = -1;
	  optif_info[*identifier].laser[i] = -1;
  }
  
  for (i=0; i<=MAX_RECEIVE_STREAM; i++) {
	  optif_info[*identifier].threshold[i] = -1;
  }

  optif_info[*identifier].address = 0;
#ifdef NIVXI
  optif_info[*identifier].window = 0;
#endif

/* Now do the mapping - CPU-dependent - allow a 0x10000 window */

#ifdef RAID
  am = 0x39;      /*A24D16 access*/
  status = RD13_VmeMap(address,am,0x10000,&addr,0);
  if(status != 0) {
     printf("VME mapping failed, status = %d\n",status);
     return status;
  }
  optif_info[*identifier].address = (unsigned long) addr;
#endif

#ifdef NIVXI
  if(optif_total == 1) {
    InitVXIlibrary();		/*only call once*/
    MapVXIAddressSize(0x80000);     /* allow 512k general purpose A24 window */
  }

  addr = (char *)MapVXIAddress(2,address,5000,&w,&ret);

  if(ret != 0) {
     printf("VME mapping failed, status = %d\n",ret);
     return (int)ret;
  }
  printf("optif_vmemap: MapVXIAddress ret %d win %d addr %8x for vme %8x\n", 
		ret, w, addr, address);
  ret = GetWindowRange(w, &wstart, &wend);
  printf("optif_vmemap: WindowRange start %lx end %lx\n", wstart, wend);
  optif_info[*identifier].address = (unsigned long) addr;
  optif_info[*identifier].window = (int)w;
#endif

#ifdef QB_linux
  if(optif_total == 1) {         /*only call once*/
    InitVME();
  }
  Initqb(&address ,WND_SIZE,VME_A24);
  optif_info[*identifier].address = address;
#endif

#ifdef RCC
  if(optif_total == 1) status = VME_Open();        /*Call first time only*/
/*Set up the structure for the mapping*/
  master_map.vmebus_address = address;
  master_map.window_size = 0x10000;
  master_map.address_modifier = VME_A24;
  master_map.options = 0;
  status = VME_MasterMap(&master_map, &handle);
  if(status != 0) {
    VME_ErrorPrint(status);
    return (int)status;
  }
  status = VME_MasterMapVirtualAddress(handle, &virtual_address);
  printf("optif_vmemap: VME address %x mapped to %x, handle %d\n",
    address, virtual_address,handle);
  optif_info[*identifier].address = (unsigned long) virtual_address;
  optif_info[*identifier].window = handle;
#endif

  return 0;
}
/* end of optif_vmemap */
