/*
 * temptress_lib.c. Version 0.4
 * Library of routines to handle TEMPTRESS 
 * version 0.1 PWP 10.12.00
 * version 0.2 JGK April 2001
 * Version 0.3 BJG/PWP 24.05.02 - additions for NIVXI under linux
 * Version 0.4 JCH 25.06.02 - support for ATLAS RCC library (for SBCs).
 */

/* includes and defines */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <time.h>

#ifdef NIVXI
#include <nivxi.h>
#endif

#ifdef OS9
#include <process.h>
#include <signal.h>
#include <types.h>
#endif

#ifdef RAID
#define CLOCKS_PER_SEC 100000
#endif

#ifdef RCC
#include <rcc_error.h>
#include <vme_rcc.h>
#endif

#include "temptress_lib.h"

/* Global variables */
short  temptress_total = 0;                    /* Total number of TEMPTRESS cards  */
struct temptress_info_structure *temptress_info;   /* structure to hold TEMPTRESS info */

/* Functions */

int temptress_exit() {
  /*
   * temptress_exit.
   * Exit routine for TEMPTRESS library. 
   * Should be called once only to tidy things up.
   * Unmaps the modules and deletes the malloced memory.
   */

  int i;
#ifdef RCC
  VME_ErrorCode_t status;
#else
  int status;
#endif

/* loop over the modules */
  for(i=0;i<temptress_total;i++) {
#ifdef RAID
    status = RD13_VmeUnmap(temptress_info[i].address);
#endif
#ifdef NIVXI
    status = UnMapVXIAddress(temptress_info[i].window);
#endif
#ifdef RCC
    status = VME_MasterUnmap(temptress_info[i].window);
#endif

  }  /*end of loop*/

#ifdef NIVXI
    CloseVXIlibrary();
#endif
#ifdef RCC
    status = VME_Close();
#endif

/* release memory */
  free((void *)temptress_info);
  return 0;
}


int temptress_print_status(short card){
  /*
   * Print temperature information for all channels
   */

  float temp[TEMPTRESS_MAX_CHANS];
  int chan,rcode; 

  rcode = temptress_read_all(card, temp);

  if(rcode ==-1){
    printf("temptress_status: card %d invalid\n",card);
    return -1;
  }

  for(chan=0;chan<TEMPTRESS_MAX_CHANS;chan++){
    printf("Chan %d Temperature %2.1f\n",chan,temp[chan]);
  }

  return rcode;
}


int temptress_read_all(short card, float* temp){
  /*
   * Read all channels
   * Return -1 if card invalid
   */

  int chan;
  volatile unsigned short *ip;
  unsigned short bits;

  for(chan=0;chan<TEMPTRESS_MAX_CHANS;chan++){
    temp[chan]=999;
  }

  if((card<0)||(card>=temptress_total)) return -1;

  for(chan=0;chan<TEMPTRESS_MAX_CHANS;chan++){
    ip = (short *)(temptress_info[card].address+chan*2);

    bits = *ip;
    if(bits == 0xCFFF){
      temp[chan] = 0.0;
    }else{
      temp[chan] = (float) ((float)bits/10.0);
    }
  }
  return 0;
}


int temptress_read_one(short card, short chan, float* temp) {
  /*
   * Read the requested channel
   * Return -1 if card invalid
   * Return -2 if chan invalid
   */

  volatile unsigned short *ip;
  unsigned short bits;

  *temp = 999;

  if((card<0)||(card>=temptress_total)) return -1;
  if((chan<0)||(chan>=TEMPTRESS_MAX_CHANS)) return -2;

  ip = (short *)(temptress_info[card].address+chan*2);

  bits = *ip;
  if(bits == 0xCFFF){
    *temp = 0.0;
  }else{
    *temp = (float) ((float)bits/10.0);
  }
  return 0;
}


int temptress_vmemap(unsigned long address, short *card) {
  /*
   * Map TEMPTRESS to the CPU address space. The input address is the true
   * VME address for UNIX and WNT implementations, mapped address for OS9 (this
   * is because the mapping in OS9 is CPU-specific and so it is not possible to
   * provide a generalised mapping within this routine). The mapping in UNIX/WNT
   * is A24.
   */

#if defined(NIVXI)
  short ret;
  int w;
  UINT32 wstart, wend;
#elif defined(RCC)
  VME_ErrorCode_t status;
  VME_MasterMap_t master_map;
  u_int virtual_address;
  int handle;
#else
  int status;
#endif
  int am;
  char *addr;

  am = 0x9;
  *card = temptress_total++;		/*Simplest approach? */

/* Use malloc/realloc to get space for the structure */
  if(temptress_total == 1)
    temptress_info = (struct temptress_info_structure *) malloc(temptress_info_size);
  else
    temptress_info = (struct temptress_info_structure *) realloc((void *)temptress_info,temptress_total*temptress_info_size);

  temptress_info[*card].address = address;
#ifdef NIVXI
  temptress_info[*card].window = 0;
#endif

/* Now do the mapping - CPU-dependent - allow a 0x1000 window */

#ifdef OS9
  status = _os_permit((void *)address, 0x1000,2,0);
  if(status != 0) {
     printf("VME mapping failed, status = %d\n",status);
     return status;
  }
#endif

#ifdef RAID
  am = 0x39;		/* A24 standard access */
  status = RD13_VmeMap(address,am,0x1000,&addr,0);
  if(status != 0) {
     printf("VME mapping failed, status = %d\n",status);
     return status;
  }
  temptress_info[*card].address = (unsigned long) addr;
#endif

#ifdef NIVXI  /* A24 */
  if(temptress_total == 1) {
    InitVXIlibrary();		/*only call once*/
    MapVXIAddressSize(0x1000); /* TEMPTRESS must live inside another window
                                   PWP 10.10.00 */
  }
  addr = (char *)MapVXIAddress(2,address,5000,&w,&ret);
  if(ret != 0) {
     printf("VME mapping failed, status = %d\n",ret);
     return (int)ret;
  }
  printf("temptress_vmemap: MapVXIAddress ret %d win %d addr %8x for vme %8x\n", 
		ret, w, addr, address);
  ret = GetWindowRange(w, &wstart, &wend);
  printf("temptress_vmemap: WindowRange start %lx end %lx\n", wstart, wend);
  temptress_info[*card].address = (unsigned long) addr;
  temptress_info[*card].window = w;
#endif

#ifdef RCC
  if(temptress_total == 1) status = VME_Open();		/*only call once*/
/* set up the map structure */
  master_map.vmebus_address = address;
  master_map.window_size = 0x1000;
  master_map.address_modifier = VME_A24;
  master_map.options = 0;
  status = VME_MasterMap(&master_map, &handle);
  if(status != VME_SUCCESS) {
    VME_ErrorPrint(status);
    return (int)status;
  }
  status = VME_MasterMapVirtualAddress(handle,&virtual_address);
  printf("temptress_vmemap: VME address %x mapped to %x, handle %d\n",
    address, virtual_address,handle);
  temptress_info[*card].address = (unsigned long) virtual_address;
  temptress_info[*card].window = handle;
#endif

  return 0;
}

