
/*
   stlib.c   SCT Multi-module Test System Library

   GFM May 1999

   Standard C Library implementing a set of functions which handle
   all calls to VME and file i/o.  These functions can be used to
   build higher-level scan programs.
*/

/* Last revised PWP 03.10.01 */

#include <stdio.h>
#include <stdlib.h>
#include <math.h>  /* added PWP 12.11.99 */
#ifdef RAID
#include <posix/time.h>
#endif

#ifdef __unix__
#include <unistd.h>
#endif

#ifdef WIN32
#include <windows.h>
#include <largeint.h>
#endif

#include "sct_hardware.h"

#include "../MUSTARD/mus_lib.h"
#include "../MUSTARD/mus_user.h" 
#include "../SLOG/slog_lib.h"
#include "../CLOAC/cloac_lib.h"
#include "../CLOAC/cloac_menu.h"
#include "../SCTLV/sctlv_lib.h"
#include "../scthv/scthv_lib.h"
#include "../BILED/biled_lib.h"
#include "../optif/optif_lib.h"
#include "../temptress/temptress_lib.h"

#include "abc.h"
#include "abc_config.h"
#include "sct.h"
#include "sct_scans.h"
#include "st_config.h"
#include "stout.h"

sysmap sct[SCT_MAX_MODULES];   /* The global system configuration database */

short  stream_2_mod[MUS_NUMBER][16];  /* Reverse sysmap information */
short  stream_2_link[MUS_NUMBER][16];

#define EVBUFSIZE 150000
#define EVMAX 1000

unsigned short evdata[EVBUFSIZE];
unsigned short *evpointer;


/* Burst results (a burst is defined as any acquisition interval between
   a Mustard enable and disable) */

short sct_histogram[SCT_MAX_MODULES][2][2048]; 
short sct_bc[SCT_MAX_MODULES][2][256];
short sct_l1[SCT_MAX_MODULES][2][16];

short sct_errs[SCT_MAX_MODULES][2];  /* How many error types do we want to log ??
                                        [module][link][chip][err_code] ???
                                        for now, [module][link], count total errors only */

int   mus_event_count[MUS_NUMBER];  /* 31 bits + overflow PWP 12.10.99 */
short mus_error_count[MUS_NUMBER];
short mus_control_errors[MUS_NUMBER];
short mus_timeout_errors[MUS_NUMBER];
short mus_used_streams[MUS_NUMBER];

int st_trigger_count;       /* Number of triggers sent this burst */
int scan_number = 0;          /* Sequence counter - equivalent to a count of 1-D scan number */
int burst_number = 0;

/*
 * Variables maintained by st_configure_variable, used to
 * track changes during scans - PWP 12.08.99
 */

short scan_pnt[4];     /* variables maintained by st_configure_variable */ 
short scan_typ[4];
float scan_var[4];

float scan_start[4];   /* variables which have to be maintained manually */
float scan_stop[4];    /* which are used to record scan definitions by   */
float scan_step[4];    /* stout.c - not used for now */
short scan_points[4];

short burst_max_timeouts = 5;   /* Maximum number of timeouts before burst terminates */
short old_ttype = -1;           /* keep record of last trigger type */

unsigned short slog_clk_enables[SLOG_NUMBER];    /* enable only used  clock streams  */
unsigned short slog_com_enables[SLOG_NUMBER];    /* enable only used command streams */
unsigned short slog_ncycles = 5;
unsigned short slog_do_loadtrig = 1;
unsigned short com_delay    = COM_DELAY;     /* delay between reset and calstrobe, used by SLOG only */
unsigned short pulse_delay  = PULSE_DELAY;   /* delay between pulse_input and L1, used by CLOAC and SLOG */
unsigned short trig_delay   = TRIG_DELAY;    /* delay between calstrobe and L1, used by CLOAC and SLOG */
unsigned short n_l1a        = NL1A;         /* number of multiple L1A triggers using SLOG */
unsigned short l1a_spacing  = L1A_SPACING;   /* interval between multiple L1A triggers using SLOG */
unsigned short sr_frequency = SR_FREQUENCY;  /* bits written to CLOAC SR_FREQUENCY register */
unsigned short tr_frequency = TR_FREQUENCY;  /* bits written to CLOAC SR_FREQUENCY register */

FILE* outf = NULL;
FILE* musf = NULL;

#include "stlib.h"

#ifdef WIN32
void usleep(int us){
  /*
   * An implementation of usleep for windows
   * This function uses the high performance counter from the windows 
   * API.  The performance of this counter may be expected to depend 
   * upon processor type and speed, however during tests with a 2GHz 
   * P4 system and with a 350MHz PII I got the same result: a maximum
   * latency of 5microS.  Hence this approach seems reasonable
   * for time intervals down to 10 (+5/-0) microS.
   * PWP 04.07.02
   */

  LARGE_INTEGER lFreq;
  LARGE_INTEGER lStart, lCount, lDiff;
  double delay,freq,target;

  QueryPerformanceCounter(&lStart);

  QueryPerformanceFrequency(&lFreq);
  freq = (double)lFreq.LowPart;

  target = ((double)us)/((double)1000000);

  delay=0;
  while(delay<target){
    QueryPerformanceCounter(&lCount);
    lDiff = LargeIntegerSubtract(lCount,lStart);
    delay = lDiff.LowPart/freq;
  }
  return;
}
#endif

int st_startup(){
  int rcode;

  printf("################################################\n");
  printf("#                                              #\n");
  printf("#   SCT System Startup                         #\n");
  printf("#                                              #\n");
  printf("#   Version GFM 990503                         #\n");
  printf("#                                              #\n");
  printf("################################################\n");
  printf("\n");

  rcode = st_vmemap();             /* Configure readout module addresses and do vmemaps       */
  rcode = st_configure_system();   /* Configure system channel mappings, enables, delays etc. */
  rcode = st_configure_dcs();      /* Set default values of DCS control values */
  rcode = st_configure_modules();  /* Configure module setings for all enabled modules        */
  rcode = st_init();               /* Execute initialisations for all enabled modules         */
                                   /* Note: any execution of one of these functions           */
                                   /* should be followed by the remaining ones                */

  return rcode;
}



int st_vmemap() {
  unsigned long address;
  int code,i;
  short id;
  int id_int; /* OPTIF library uses int - at least for time being - PWP 09.10.00 */
  char string[10];

  printf("################################################\n");
  printf("#                                              #\n");
  printf("#   Configuring READOUT  --- doing vmemaps     #\n");
  printf("#                                              #\n");
  printf("################################################\n");
  printf("\n");
  printf("Attempting to map all VME modules and set to idle\n");
  printf("\n");  

  /* Here we should map any readout modules. We just hard-code addresses
     and am-codes etc. since there aren't that many of them.  However,
     any readout modules implied to be present through detector modules
     being marked as 'present' in the sysmap should actually be present. */

  id = -1;
  
  /* Map A32 modules first. SLOG must always be the first of the A32 modules to be mapped.
   * A32 mapping is perhaps the most likely to fail on new VXI installations since the window 
   * is large (8MB).  Please do not change this!!  PWP 16.07.99 
   * On the contrary !  When VME mapping fails, it is natural to reorder mapping to determine 
   * whether the failure is device or order related.  As long as SLOG is the first A32 unit 
   * the order shouldn't be important.*/
   if (SLOG_NUMBER){
    printf("Attempting vmemap of SLOG  (Slow Command Generator)\n");
    address = SLOG_ADDRESS;
    for(i=0;i<SLOG_NUMBER;i++){
      code = slog_vmemap(address,&id);
      printf(" ... status %d id %x for address %x\n",code,id,address);
      if (id!=i) printf(" ... ERROR: SLOG %d identifier was not %d\n",i,id);
      if (code != 0) {
        printf("st_vmemap: FAILED with code %d for SLOG address %x; proceed ?\n", code, address);
        scanf("%s",string);
      }
    address += SLOG_SPACING;
    }
  }
  
  if (MUS_NUMBER){
    printf("st: Attempting vmemap of MUSTARD (Module Readout System)\n");
    address = MUS_ADDRESS;
    for(i=0;i<MUS_NUMBER;i++){
      code = mus_vmemap(address,&id);
      printf(" ... status %d id %x for address %x\n",code,id,address);
      if (id!=i) printf(" ... ERROR: MuSTARD %d identifier was not %d\n",i,id);
      if (code != 0) {
        printf("st_vmemap: FAILED with code %d for MuSTARD address %x; proceed ?\n", code, address);
        scanf("%s",string);
      }
    address += MUS_SPACING;
    }
  }
   /* Now map A24 modules.  The order is not strictly important - but I leave BiLED until last
   * since it perhaps least likely to be used.  Feel free to change this!!  PWP 16.06.99 */

  if (CLOAC_NUMBER){
    printf("st: Attempting vmemap of CLOAC (Master Clock and Trigger System)\n");
    address = CLOAC_ADDRESS;
    code = cc_vmemap(address);
    printf(" ... status %d for address %x\n",code,address);
    if (code != 0) {
      printf("st_vmemap: FAILED with code %d for CLOAC address %x; proceed ?\n", code, address);
      scanf("%s",string);
    }
  }
  
  if (SCTHV_NUMBER){
    printf("st: Attempting vmemap of SCTHV (VME based High Voltage Modules)\n");
    address = SCTHV_ADDRESS;
    for(i=0;i<SCTHV_NUMBER;i++){
      code = scthv_vmemap(address,&id);
      printf(" ... status %d id %x for address %x\n",code,id,address);
      if (id!=i) printf(" ... ERROR: SCTHV %d identifier was not %d\n",i,id);
      if (code != 0) {
        printf("st_vmemap: FAILED with code %d for SCTHV address %x; proceed ?\n", code, address);
        /* scanf("%s",string); */
        st_sleep(500);
      }
    address += SCTHV_SPACING;
    }
  }
  
  if (SCTLV_NUMBER){
    printf("Attempting vmemap of SCTLV (Low voltage power supply modules)\n");
    address = SCTLV_CRATE_ADDRESS;
    for(i=0;i<SCTLV_NUMBER;i++){
      code = sctlv_vmemap(address,&id);   /* Now map the individual modules PWP 17.01.01 */
      printf(" ... status %d for address %x\n",code,address);
      if (id!=i) printf(" ... ERROR: SCTLV %d identifier was not %d\n",i,id);
      if (code != 0) {
        printf("st_vmemap: FAILED with code %d for SCTLV address %x; proceed ?\n", code, address);
        /* scanf("%s",string); */
        st_sleep(500);
      }
    address += SCTLV_SPACING;
    }
  }

  if (BILED_NUMBER){
    printf("Attempting vmemap of BiLED (Biphase Mark encoder system)\n");
    address = BILED_ADDRESS;
    for(i=0;i<BILED_NUMBER;i++){
      code = biled_vmemap(address,&id);
      printf(" ... status %d id %x for address %x\n",code,id,address);
      if (id!=i) printf(" ... ERROR: BILED %d identifier was not %d\n",i,id);
      if (code != 0) {
        printf("st_vmemap: FAILED with code %d for BILED address %x; proceed ?\n", code, address);
        scanf("%s",string);
      }
      address += BILED_SPACING;
    }
  }

  if (OPTIF_NUMBER){
    printf("Attempting vmemap of OPTIF (optical transceiver system)\n");
    address = OPTIF_ADDRESS;
    for(i=0;i<OPTIF_NUMBER;i++){
      code = optif_vmemap(address,&id_int);
      printf(" ... status %d id %x for address %x\n",code,id_int,address);
      if (id_int!=i) printf(" ... ERROR: OPTIF %d identifier was not %d\n",i,id_int);
      if (code != 0) {
        printf("st_vmemap: FAILED with code %d for OPTIF address %x; proceed ?\n", code, address);
        scanf("%s",string);
      }
      address += OPTIF_SPACING;
    }
  }

  if (TEMPTRESS_NUMBER){
    printf("Attempting vmemap of TEMPTRESS (temperature measurement system)\n");
    address = TEMPTRESS_ADDRESS;
    for(i=0;i<TEMPTRESS_NUMBER;i++){
      code = temptress_vmemap(address,&id);
      printf(" ... status %d id %x for address %x\n",code,id,address);
      if (id!=i) printf(" ... ERROR: TEMPTRESS %d identifier was not %d\n",i,id);
      if (code != 0) {
        printf("st_vmemap: FAILED with code %d for TEMPTRESS address %x; proceed ?\n", code, address);
        scanf("%s",string);
      }
      address += TEMPTRESS_SPACING;
    }
  }
  
  evpointer = evdata;

  st_sleep(500); /* Wait 1000 mS */

  return code;
}

int st_configure_system(){

  printf("################################################\n");
  printf("#                                              #\n");
  printf("#   Configuring SYSTEM                         #\n");
  printf("#                                              #\n");
  printf("################################################\n");
  printf("\n");

  st_default_sysmap();

  st_show_sysmap();

  /* change any map settings from their default values */

  /* st_electro_sysmap(); */

  st_read_sysmap();
  printf("sysmap read OK\n");
  st_show_sysmap();

  st_sleep(500); /* Wait 500 mS */

  /* No hard code variations needed !!!
     NB. 1) MuSTARD stream delays will be read from system file.
            Values in module configuration files will be ignored.
         2) MuSTARD stream enables are not present in the system file.
            Settings will be inferred from the "link" variables 
            in the module configuration files.
     PWP 15.06.99 
  
     Note: link variables now used to implement redundant chip
     histogram readout: link0/1 == 2 => read both pages of that link 
     GFM 990722
  */

  return 0;
}

int st_configure_modules(){

  int j;

  printf("################################################\n");
  printf("#                                              #\n");
  printf("#   Configuring MODULES                        #\n");
  printf("#                                              #\n");
  printf("################################################\n");
  printf("\n");

  /* Review sysmap: for each 'present' detector, load its description
     file into the relevant module_cfg structure inside sysmap;
     if file read fails, mark module as absent.  If we wanted to,
     we could also review a table of successfully mapped readout
     modules, and if any required failed, we can mark their dependent
     detector modules absent too.  This establishes what the default
     working conditions for each module should be. */

  for (j=0; j<SCT_MAX_MODULES; j++) {
/*    printf("Entered module config read loop with j = %d\n",j);
      st_sleep(2000); */

    if(sct[j].present){
/*      printf("Entered module config read branch with j = %d\n",j);
        st_sleep(2000); */

      printf("st: configuring MODULE %d\n",j);

      if (st_read_module(j) == -1){
        printf("Marking module %d ABSENT due to error at file open\n");
        sct[j].active = 0;
        sct[j].present = 0;
        st_sleep(2000);
        printf("Continue ?\n");
        scanf("%d",&j);
      } else {
        st_show_module(j);
        if(sct[j].det.chipset==ABCDT){
          if(st_read_trims(j) == 0){
            st_sleep(500);
            st_show_trims(j);
          }
        }
        st_read_mask(j);
        st_read_rc(j);
        st_sleep(500);
      }
    }
  }

  st_show_sysmap();

  return 0;
}

 
int st_init(){
  short status;
  short i,j;

  printf("################################################\n");
  printf("#                                              #\n");
  printf("#   Initialising READOUT                       #\n");
  printf("#                                              #\n");
  printf("################################################\n");
  printf("\n");

  /* SCTLV cards are now initialised at map time PWP 22.02.01 */

  /* Here we prepare the readout modules for general running,
     eg, start the clocks etc. We should check all readout modules
     are really present and correct. */

  if(CLOAC_NUMBER>0){
    cc_init();                                   /* Starts clock; disables triggers */
    cc_load_delay((unsigned short)(trig_delay-CLOAC_CAL_OFFSET));  /* Load the system-wide common delay
                                                                      between calstrobe and L1A */
    cc_menu_srfreq(sr_frequency);
    cc_menu_trfreq(tr_frequency);
  }

  st_trigger_count = 0;

  for (i=0; i<MUS_NUMBER; i++) {
    mus_read_status(i,&status);
    mus_init(i);                   /* All input streams disabled */
    mus_set_readout_cntl(i,0,0);   /* Internal data transfer disabled */

    mus_error_count[i] = 0;
    mus_event_count[i] = 0;
    mus_control_errors[i] = 0;
    mus_timeout_errors[i] = 0;

    printf("Doing mus_clear_histogram()\n");
    for (j=0; j<12; j++)
      mus_clear_histogram(i,j);    /* Histogram cleared - all streams */
      
    mus_reset(i);                  /* Fifos and buffers emptied */
  }

  for (i=0; i<SLOG_NUMBER; i++) {
    slog_init(i);                  /* Stop all slow commands, enable all links */
    slog_clear(i);                 /* Clear memory */
    slog_show_ram(i, 0, 100);      /* Read back memory, and show stop bits everywhere */
    slog_show_status(i);    
  }

/* Even now that the vernier has gone, BiLED must still be
 * initialised manually.  PWP 01.10.99 */

/*  for (i=0; i<BILED_NUMBER; i++) biled_init(i); */

  for (i=0; i<OPTIF_NUMBER; i++) optif_init(i);

  printf("################################################\n");
  printf("#                                              #\n");
  printf("#   Initialising SYSTEM                        #\n");
  printf("#                                              #\n");
  printf("################################################\n");
  printf("\n");

  /* Here we review the sysmap and actively prepare the readout 
     module channels for the detector modules marked 'present' in the sysmap */

  st_execute_enables();

  for (i=0; i<MUS_NUMBER; i++) {
    mus_read_status(i,&status);
    mus_print_status(status);
  }

  /* Now is a good time to establish the reverse system mapping
     - must be done after st_execute enables */

  st_reverse_sysmap(); /* PWP 06.03.00 */


  /*
   * OPTO initialisation has been moved to the following new section
   * PWP 10.10.00
   */

  if((BILED_NUMBER>0)||(OPTIF_NUMBER>0)){
    printf("################################################\n");
    printf("#                                              #\n");
    printf("#   Initialising OPTO                          #\n");
    printf("#                                              #\n");
    printf("################################################\n");
    printf("\n");

    if(OPTIF_NUMBER>0){
      printf("Set OPTIF RX thresholds\n");
      st_optorx_init();
    }

    printf("Switch on VCSEL drive currents\n");
    st_optotx_on();

    st_sleep(2000);
  }

  printf("################################################\n");
  printf("#                                              #\n");
  printf("#   Initialising MODULES                       #\n");
  printf("#                                              #\n");
  printf("################################################\n");
  printf("\n");

  /* Here we physically write settings to modules and
   * to the DCS system etc. to bring the modules up to
   * the previously established default working conditions.
   */

  /* NB: Power cycling necessary to recover DORIC4 from 20MHz mode
   * PWP 07.07.99 */

  printf("Power OFF to all modules\n");
  st_lv_off();            /* Switch LV output off */
  printf("Initialising LV on all modules\n");
  st_lv_init();           /* Program LV supplies */
  printf("Power ON to all modules\n");
  st_lv_on();             /* Switch LV output on */
  st_sleep(100);          /* wait 100mS */

  printf("HARD RESET to all modules\n");
  st_hard_reset();        /* Hard reset to all modules */
  printf("EXECUTE CONFIG CONTROL BLOCKS to all modules\n"); 
  st_execute_configs(1,1);/* Send slow commands to configure modules */ 
  st_soft_reset();        /* Soft reset no longer in st_execute_configs,
                             must be done seperately (here) */

  st_scan_reset();   /*  reset variable change logs, as used by stout.c
                         to generate ASCII ntuple output */

  printf("################################################\n");
  printf("#                                              #\n");
  printf("#   End of System Startup : ready to go !      #\n");
  printf("#                                              #\n");
  printf("################################################\n");
  printf("\n");

  return 0;
}



int st_shutdown() {

  printf("################################################\n");
  printf("#                                              #\n");
  printf("#   Shutting down SCT System                   #\n");
  printf("#                                              #\n");
  printf("################################################\n");

  st_lv_off();
  st_hv_ramp(1);
  st_optotx_off();
/*
  if(BILED_NUMBER){
    biled_exit();
  }
//  if(CLOAC_NUMBER){
//    cloac_exit();
//  }
  if(MUS_NUMBER){
    mus_exit();
  }
  if(OPTIF_NUMBER){
    optif_exit();
  }
  if(SCTHV_NUMBER){
    scthv_exit();
  }
  if(SCTLV_NUMBER){
    sctlv_exit();
  }
  if(SLOG_NUMBER){
    slog_exit();
  }
  if(TEMPTRESS_NUMBER){
    temptress_exit();
  }
*/
  slog_exit();
  mus_exit();

  return 0;
}


int st_configure_enable(int det, short present, short active, short link0, short link1){

  /* 
     Modify sysmap on/off entries for detector module [d].  These are
     the parameters likely to change but not covered by st_configure_variable().

     Note: if links are enabled or disabled, the relevant readout channels 
     (Mustard, Slog, BiLED etc.) are appropriately changed by st_execute_enables.
  */ 

  if (det >= SCT_MAX_MODULES) return -1;

  if (present) sct[det].present = 1; else sct[det].present = 0;
  if (active) sct[det].active = 1; else sct[det].active = 0;
  if (link0) {
    sct[det].link0 = link0;
    sct[det].mus_enable0 = 1;
  } else {
    sct[det].link0 = 0;
    sct[det].mus_enable0 = 0;
  }
  if (link1) {
    sct[det].link1 = link1; 
    sct[det].mus_enable1 = 1;
  } else {
    sct[det].link1 = 0;
    sct[det].mus_enable1 = 0;
  }
  return 0;
}


int st_execute_enables(){
  short det;
  short ch, mus;
  short ndet = 0;
  short sid;
  unsigned short scl, sex, sin;
  
  for (sid=0; sid<SLOG_NUMBER; sid++){
        slog_clk_enables[sid] = 0xf000;        /* enable both NIM and both ECL channels as */
        slog_com_enables[sid] = 0xf000;        /* per firmware revision completed 22/07/99 */
  }

  for (mus=0; mus<MUS_NUMBER; mus++){  /* disable ALL MuSTARD channels */
    mus_used_streams[mus]=0;
    for(ch=0; ch<12; ch++) mus_set_stream_enable(mus, ch, 0);
  }
  
  for (det=0; det<SCT_MAX_MODULES; det++){  /* enable all USED MuSTARD channels */

    /* enable all used MuSTARD channels */

    if (sct[det].present && sct[det].link0 && (sct[det].suspend0==0) ){
      mus_used_streams[sct[det].mus_id]++;
      sct[det].mus_enable0 = 1;
      mus_set_stream_enable( sct[det].mus_id, sct[det].mus_ch0, sct[det].mus_enable0);
      mus_set_stream_delay( sct[det].mus_id, sct[det].mus_ch0, sct[det].mus_delay0);
    } else {
      sct[det].mus_enable0 = 0;
    }

    if (sct[det].present && sct[det].link1 && (sct[det].suspend1==0) )
    {
      mus_used_streams[sct[det].mus_id]++;
      sct[det].mus_enable1 = 1;
      mus_set_stream_enable( sct[det].mus_id, sct[det].mus_ch1, sct[det].mus_enable1);
      mus_set_stream_delay( sct[det].mus_id, sct[det].mus_ch1, sct[det].mus_delay1); 
    } else {
      sct[det].mus_enable1 = 0;
    }

    
    /* enable all used SLOG channels */
    
    /* OPTIF uses two clock signals, 0 and 1, and six command lines, 0-5.
     * Command streams 0-3 are encoded onto clock 0 and command 
     * streams 4-5 are encoded onto clock 1.  Hence clock 0 must be
     * enabled if _any_ of command streams 0-3 are in use, and
     * clock 1 must be enabled if either command stream 4 or 5 is in use.
     *
     * Now, one SLOG can drive two OPTIFs.  Let's assume that the
     * second one is wired to clocks 6,7 and command lines 6-11.
     * So if any of command streams 6-9 are enabled, we turn on
     * clock 6.  Similarly if command streams 10 or 11 are in use,
     * we must enable clock 7.
     *
     * Also, we leave the NIM command outputs enabled all the time. 
     *
     * Got all that?  PWP 10.10.00
     */

    if(sct[det].present){
      ndet ++;
      if (sct[det].det.select){ /* use redundancy */

        if ((sct[det].r_opto_tx_type==1)&&
            (sct[det].r_opto_tx_id>=0)&&(sct[det].r_opto_tx_id<OPTIF_NUMBER)){ /* OPTIF */

          if ((sct[det].r_slog_id>=0)&&(sct[det].r_slog_id<SLOG_NUMBER)){
            if(sct[det].r_slog_ch<4){         /* enable clock 0 */
              slog_clk_enables[sct[det].r_slog_id] |= 0x0001;
            } else if(sct[det].r_slog_ch<6){  /* enable clock 1 */
              slog_clk_enables[sct[det].r_slog_id] |= 0x0002;
            } else if(sct[det].r_slog_ch<10){ /* enable clock 6 */
              slog_clk_enables[sct[det].r_slog_id] |= 0x0040;
            } else if(sct[det].r_slog_ch<12){ /* enable clock 7 */
              slog_clk_enables[sct[det].r_slog_id] |= 0x0080;
            }
            slog_com_enables[sct[det].r_slog_id] |= 0x0001<<sct[det].r_slog_ch;
          }
        } else { /* assume electrical readout */
          if ((sct[det].r_slog_id>=0)&&(sct[det].r_slog_id<SLOG_NUMBER)){
            slog_clk_enables[sct[det].r_slog_id] |= 0x0001<<sct[det].r_slog_ch;
            slog_com_enables[sct[det].r_slog_id] |= 0x0001<<sct[det].r_slog_ch;
          }
        }
      } else { /* no redundancy */

        if ((sct[det].opto_tx_type==1)&&
            (sct[det].opto_tx_id>=0)&&(sct[det].opto_tx_id<OPTIF_NUMBER)){ /* OPTIF */

          if ((sct[det].slog_id>=0)&&(sct[det].slog_id<SLOG_NUMBER)){
            if(sct[det].slog_ch<4){         /* enable clock 0 */
              slog_clk_enables[sct[det].slog_id] |= 0x0001;
            } else if(sct[det].slog_ch<6){  /* enable clock 1 */
              slog_clk_enables[sct[det].slog_id] |= 0x0002;
            } else if(sct[det].slog_ch<10){ /* enable clock 6 */
              slog_clk_enables[sct[det].slog_id] |= 0x0040;
            } else if(sct[det].slog_ch<12){ /* enable clock 7 */
              slog_clk_enables[sct[det].slog_id] |= 0x0080;
            }
            slog_com_enables[sct[det].slog_id] |= 0x0001<<sct[det].slog_ch;
          }
        } else { /* assume electrical readout */
          if ((sct[det].slog_id>=0)&&(sct[det].slog_id<SLOG_NUMBER)){
            slog_clk_enables[sct[det].slog_id] |= 0x0001<<sct[det].slog_ch;
            slog_com_enables[sct[det].slog_id] |= 0x0001<<sct[det].slog_ch;
          }
        }
      }
    }
  }


  if (ndet>0) printf("st_execute_enables: %d modules are present in the sysmap\n",ndet);
  else        printf("st_execute_enables: WARNING - NO MODULES IN THE SYSMAP!!\n");

  for (sid=0; sid<SLOG_NUMBER; sid++) {
    slog_set_clock(sid, (unsigned short)(slog_clk_enables[sid]&0x3fff));
      /* ECL clocks enabled by bits 12,13 
       * for clock_enable, bits 14, 15 have no function */
    if(CLOAC_NUMBER>0) slog_set_external(sid, (unsigned short)(slog_com_enables[sid]&0x7fff)); /* was 0x5fff PWP 03.10.00 */
      /* keep data bit 13 (lower NIM) for pretrigger pulse -
         this is actually enabled by bit 15 */
    slog_set_internal(sid, 0);
    slog_get_clock(sid,&scl);
    slog_get_external(sid,&sex);
    slog_get_internal(sid,&sin);
    printf("st_execute_enables: SLOG %d clock %4x external %4x internal %4x\n",
      sid, scl, sex, sin);
  }

  return 0;
}


int st_coerce_variable(short typ, float var, float *newvar, short is_step) {
  
  /*
   * returns "newvar" .. nearest possible dac setting to "var"
   * We must add a half step to avoid rounding errors !!
   * PWP 07.07.99
   *
   * Should not coerce stepsize to lowest possible setting
   * in case of SCTLV parameters - scanning from 3.5 to 4.5
   * volts Vdd in 1.0 volt steps isn't much use!!
   * PWP bugfix 11.03.00
   */

  float v;

  switch(typ) {

    case ST_ELAPSED:
      if (var < 0) v = 0.0;
      else if (var > ST_MAX_SCANPOINTS) v = ST_MAX_SCANPOINTS;
      else v = ((int)((var+0.5)/(float)1.0)) * (float)1.0;   /* Integer */
      break;

    /* ABCD chip parameters: perform both range checking */

    case ST_VTHR:
      if (var < 0) v = 0.0;
      else if (var > 637.5) v = 637.5;
      else v = ((int)((var+1.25)/(float)2.5)) * (float)2.5;   /* 2.5mV / DAC bit */
      break;

    case ST_VCAL:
      if (var < 0) v = 0.0;
      else if (var > 159.375) v = 159.375;
      else v = ((int)((var+0.312)/(float)0.625)) * (float)0.625;   /* 0.625mV / DAC bit */
      break;

    case ST_STROBE_DELAY:
      if (var < 0) v = 0.0;
      else if (var > 63.0) v = 63.0;
      else v = ((int)((var+0.5)/(float)1.0)) * (float)1.0;   /* Set in DAC bits */
      break;

    case ST_PREAMP:
      if (var < 0) v = 0.0;
      else if (var >285.2) v = (float)285.2;
      else v = ((int)((var+4.6)/(float)9.2)) * (float)9.2;  /* 9.2uA / DAC bit */ 
      break;

    case ST_SHAPER:
      if (var < 0) v = 0.0;
      else if (var > 37.2) v = (float)37.2;
      else v = ((int)((var+0.6)/(float)1.2)) * (float)1.2;   /* 1.2uA / DAC bit */
      break;

    case ST_TRIM:
      if (var < 0) v = 0.0;
      else if (var > 15.0) v = (float) 15.0;
      else v = ((int)((var+0.5)/(float)1.0)) * (float)1.0;
      break;

    case ST_MASK:
      if (var < 0) v = 0.0;
      else if (var > 3.0) v = (float) 3.0;
      else v = ((int)((var+0.5)/(float)1.0)) * (float)1.0;
      break;

    case ST_ROLE: /* PWP 08.12.00 */
      if (var < 0) v = 0.0;
      else if (var > 5.0) v = (float) 5.0;
      else v = ((int)((var+0.5)/(float)1.0)) * (float)1.0;
      break;

    case ST_NMASK: /* PWP 10.05.01 */
      if (var < 0) v = 0.0;
      else if (var > 1535.0) v = (float) 1535.0;
      else v = ((int)((var+0.5)/(float)1.0)) * (float)1.0;
      break;

    case ST_CAL_MODE:
    case ST_COMPRESSION:
    case ST_TRIM_RANGE:
      if (var < 0) v = 0.0;
      else if (var > 3.0) v = (float) 3.0;
      else v = ((int)((var+0.5)/(float)1.0)) * (float)1.0;
      break;

    case ST_EDGE_DETECT:
    case ST_SEND_MASK:
    case ST_ACCUMULATE:
    case ST_DTM:
    case ST_ACTIVE:  /* PWP 08.11.01 */
      if (var > 0.5) v = (float)1.0;
      else v = (float)0.0;
      break;

    case ST_BYPASS:
      if (var < 0) v = 0.0;
      else if (var > 14.0) v = (float) 14.0;
      else v = ((int)((var+0.5)/(float)1.0)) * (float)1.0;
      break;

   case ST_TOKEN:
      if (var < 0) v = 0.0;
      else if (var > 61.0) v = (float) 61.0;
      else v = ((int)((var+0.5)/(float)1.0)) * (float)1.0;
      break;

    /* Dervied ABCD chip parameters:
     * perform no range checking or coersion for the time being 
     * other than to check that value is positive.
     * Actual value used will be chipset dependent */

    case ST_QTHR:
      if (var < -1.0) v = (float)-1.0;
      else v = var;
      break;

    case ST_QCAL:
      if (var < 0.0) v = (float)0.0;
      else v = var;
      break;


    /* MuSTARD parameters: perform range checking */

    case ST_STREAM_DELAY:
    case ST_STREAM_DELAY0:    
    case ST_STREAM_DELAY1:  
      if (var < 0) v = 0.0;
      else if (var > 24.0) v = 24.0;   /* PWP bugfix 12.08.99 */
      else v = ((int)((var+0.5)/(float)1.0)) * (float)1.0;
      break;
   
    case ST_VDET:
      if (var < 0) v = 0.0;
      else if (var > 500.0) v = 500.0;
      else v = ((int)((var+(V_STEP/2))/(float)V_STEP)) * (float)V_STEP;
      break;

    /* SCTLV parameters
     * Section revised to permit operation with a mix of
     * SCTLV2 and SCTLV3 cards. In general the step sizes used 
     * by SCTLV3 are equal to, or are integer multiples of, 
     * those used by SCTLV2.  Hence the coersion ranges and 
     * step sizes used for lines available from either card 
     * are mostly those of SCTLV3.
     * Exception: VPIN has a step size of 40mV for SCTLV3 
     * and 50mV for LV3: in this case the step size is simply 
     * coerced to the nearest 10mV.
     * PWP 04.07.01
     * 
     * Minimum possible values for Vcc and Vdd are non zero.
     * No range checking is performed in these cases such that 
     * scan_step, which may be less than the minimum possible 
     * V setting, can still be coerced to a possible step size.
     * Any attempt to set an impossible value will be correctly
     * coerced by sctlv_lib.
     * PWP 07.07.99 */

    case ST_VCC: /* use larger step sizes and limits of SCTLV3 */
      if(is_step)
      v = (float)((double)var / vi3_scale[Vcc])*(float)vi3_scale[Vcc];
      else if(var < vi3_min[Vcc]) v = vi3_min[Vcc];
      else if(var > vi3_max[Vcc]) v = vi3_max[Vcc];
      else v = (float)((double)var / vi3_scale[Vcc])*(float)vi3_scale[Vcc];
      break;

    case ST_VI1: /* SCTLV2 */
      if(is_step)
      v =  (float)((double)var / vi2_scale[Vi1])*(float)vi2_scale[Vi1];
      else if(var < vi2_min[Vi1]) v = vi2_min[Vi1];
      else if(var > vi2_max[Vi1]) v = vi2_max[Vi1];
      else v = (float)((double)var / vi2_scale[Vi1])*(float)vi2_scale[Vi1];
      break;

    case ST_VDD: /* use larger step sizes and limits of SCTLV3 */
      if(is_step) v = (float)((double)var / vi3_scale[Vdd])*(float)vi3_scale[Vdd];
      else if(var < vi3_min[Vdd]) v = vi3_min[Vdd];
      else if(var > vi3_max[Vdd]) v = vi3_max[Vdd];
      else v = (float)((double)var / vi3_scale[Vdd])*(float)vi3_scale[Vdd];
      break;

    case ST_VPIN:
      /* limits are the same for SCTLV2 and SCTLV3, but the step sizes 
       * are 40mV and 50mV respectively.  For now, check range limits
       * but coerce only to nearest 10mV.
       * PWP 04.07.01
       */
      if(is_step) v = (float)((double)var / 10)*(float)10;
      else if(var < vi3_min[Vpin]) v = vi3_min[Vpin];
      else if(var > vi3_max[Vpin]) v = vi3_max[Vpin];
      else v = (float)((double)var / 10)*(float)10;
      break;

    case ST_VLED: /* same limits and step for all SCTLV */
      if(is_step) v = (float)((double)var / vi3_scale[Vld0])*(float)vi3_scale[Vld0];
      else if(var < vi3_min[Vld0]) v = vi3_min[Vld0];
      else if(var > vi3_max[Vld0]) v = vi3_max[Vld0];
      else v = (float)((double)var / vi3_scale[Vld0])*(float)vi3_scale[Vld0];
      break;
      
    case ST_VLED0: /* same limits and step for all SCTLV */
      if(is_step) v = (float)((double)var / vi3_scale[Vld0])*(float)vi3_scale[Vld0];
      else if(var < vi3_min[Vld0]) v = vi3_min[Vld0];
      else if(var > vi3_max[Vld0]) v = vi3_max[Vld0];
      else v = (float)((double)var / vi3_scale[Vld0])*(float)vi3_scale[Vld0];
      break;

    case ST_VLED1: /* same limits and step for all SCTLV */
      if(is_step) v = (float)((double)var / vi3_scale[Vld1])*(float)vi3_scale[Vld1];
      else if(var < vi3_min[Vld1]) v = vi3_min[Vld1];
      else if(var > vi3_max[Vld1]) v = vi3_max[Vld1];
      else v = (float)((double)var / vi3_scale[Vld1])*(float)vi3_scale[Vld1];
      break; 

    case ST_SELECT: /* added PWP 01.10.00 */
      if (var > 0.5) v = (float)1.0;
      else v = (float)0.0;
      break;

    case ST_RAMP:   /* added PWP 12.12.00 */
      if(var < 1) v = 1.0;
      else if(var > 4.0) v = 4.0;
      else v = ((int)((var+0.5)/(float)1.0)) * (float)1.0;
      break;

    case ST_IDET:   /* added PWP 12.07.01 */
      if(var < 0) v = 0.0;
      else if(var > 5000.0) v = 5000.0;
      else v = var;
      break;

    case ST_ICC: /* use larger step sizes and limits of SCTLV3 */
      if(is_step)
      v = (float)((double)var / vi3_scale[Icc])*(float)vi3_scale[Icc];
      else if(var < vi3_min[Icc]) v = vi3_min[Icc];
      else if(var > vi3_max[Icc]) v = vi3_max[Icc];
      else v = (float)((double)var / vi3_scale[Icc])*(float)vi3_scale[Icc];
      break;

    case ST_IVI1: /* SCTLV2 */
      if(is_step)
      v =  (float)((double)var / vi2_scale[Ivi1])*(float)vi2_scale[Ivi1];
      else if(var < vi2_min[Ivi1]) v = vi2_min[Ivi1];
      else if(var > vi2_max[Ivi1]) v = vi2_max[Ivi1];
      else v = (float)((double)var / vi2_scale[Ivi1])*(float)vi2_scale[Ivi1];
      break;

    case ST_IDD: /* use larger step sizes and limits of SCTLV3 */
      if(is_step) v = (float)((double)var / vi3_scale[Idd])*(float)vi3_scale[Idd];
      else if(var < vi3_min[Idd]) v = vi3_min[Idd];
      else if(var > vi3_max[Idd]) v = vi3_max[Idd];
      else v = (float)((double)var / vi3_scale[Idd])*(float)vi3_scale[Idd];
      break;

    case ST_IPIN:
      v = var;
      /* Note:  Although Ipin can be monitored, no limit can be set. */
      break;

    case ST_ILED: /* SCTLV2 */
      if(is_step) v = (float)((double)var / vi2_scale[Ild0])*(float)vi2_scale[Ild0];
      else if(var < vi2_min[Ild0]) v = vi2_min[Ild0];
      else if(var > vi2_max[Ild0]) v = vi2_max[Ild0];
      else v = (float)((double)var / vi2_scale[Ild0])*(float)vi2_scale[Ild0];
      break;
      
    case ST_ILED0: /* SCTLV2 */
      if(is_step) v = (float)((double)var / vi2_scale[Ild0])*(float)vi2_scale[Ild0];
      else if(var < vi2_min[Ild0]) v = vi2_min[Ild0];
      else if(var > vi2_max[Ild0]) v = vi2_max[Ild0];
      else v = (float)((double)var / vi2_scale[Ild0])*(float)vi2_scale[Ild0];
      break;

    case ST_ILED1: /* SCTLV2 */
      if(is_step) v = (float)((double)var / vi2_scale[Ild1])*(float)vi2_scale[Ild1];
      else if(var < vi2_min[Ild1]) v = vi2_min[Ild1];
      else if(var > vi2_max[Ild1]) v = vi2_max[Ild1];
      else v = (float)((double)var / vi2_scale[Ild1])*(float)vi2_scale[Ild1];
      break;

	/* SLOG and CLOAC */

    case ST_CLOCK_DUTY:
      if(var < 0) v = 0.0;
      else if(var > 15.0) v = 15.0;
      else v = ((int)((var+0.5)/(float)1.0)) * (float)1.0;
      break; 
    
    case ST_CLOCK_PHASE:
      if(var < 0) v = 0.0;
      else if(var > 31.0) v = 31.0;
      else v = ((int)((var+0.5)/(float)1.0)) * (float)1.0;
      break; 
    
    case ST_COM_DELAY:  /* delay between soft reset and calstrobe, SLOG only */
      if(var < 0) v = 0.0;
      /* To allow use of this feature to inject external noise pulses into 
       * modules, and to scan over long delays, the upper limit has been raised 
       * PWP 10.01.00 */
      else if(var > 1400.0) v = 1400.0;  /* cb buffer is 1600 words long */
      else v = ((int)((var+0.5)/(float)1.0)) * (float)1.0;
      break;

    case ST_PULSE_DELAY:  /* delay between pulse_input and calstrobe */
    case ST_TRIG_DELAY:   /* delay between calstrobe and L1 */
      if(var < 0) v = 0.0;
      else if(var > 143.0) v = 143.0;
      else v = ((int)((var+0.5)/(float)1.0)) * (float)1.0;
      break; 
    
    case ST_SR_FREQUENCY: /* added PWP 27.06.00 */
    case ST_TR_FREQUENCY: /* added PWP 27.06.00 */
      if(var < 0) v = 0.0;
      else if(var > 31.0) v = 31.0;
      else v = ((int)((var+0.5)/(float)1.0)) * (float)1.0;
      break; 

    case ST_NL1A: /* added PWP 22.07.00 */
      if(var < 0) v = 0.0;
      else if(var > 8.0) v = 8.0;
      else v = ((int)((var+0.5)/(float)1.0)) * (float)1.0;
      break;

    case ST_L1A_SPACING: /* added PWP 22.07.00 */
      if(var < 0) v = 0.0;
      else if(var > 1024.0) v = 1024.0;
      else v = ((int)((var+0.5)/(float)1.0)) * (float)1.0;
      break; 
      
    /* OPTO - added by PWP, 10.10.00 */

    case ST_OPTO_RX_THR:
    case ST_OPTO_RX_THR0:
    case ST_OPTO_RX_THR1:
    case ST_OPTO_TX_DAC:   /* all 8 bit DACs */
      if(var < 0) v = 0.0;
      else if(var > 255.0) v = 255.0;
      else v = ((int)((var+0.5)/(float)1.0)) * (float)1.0;
      break;

    case ST_OPTO_TX_MKSPNEG:
    case ST_OPTO_TX_MKSPPOS:   /* 6 bit */
      if(var < 0) v = 0.0;
      else if(var > 63.0) v = 63.0;
      else v = ((int)((var+0.5)/(float)1.0)) * (float)1.0;
      break;
    
    case ST_OPTO_TX_CDELAY:   /* 5 bit */
      if(var < 0) v = 0.0;
      else if(var > 31.0) v = 31.0;
      else v = ((int)((var+0.5)/(float)1.0)) * (float)1.0;
      break;
    
    case ST_OPTO_TX_FDELAY:   /* 7 bit */
      if(var < 0) v = 0.0;
      else if(var > 127.0) v = 127.0;
      else v = ((int)((var+0.5)/(float)1.0)) * (float)1.0;
      break;

    /* FORWARD SUPPORT CARD settings - PWP 15.05.01 */

    case ST_FWD_SC_MODE:
      if(var < 0) v = 0.0;
      else if(var > 4) v = 4;
      else v = ((int)((var+0.5)/(float)1.0)) * (float)1.0;
      break; 

    case ST_BPM_DRIVE:
      if(var < 0) v = 0.0;
      else if(var > 3) v = 3;
      else v = ((int)((var+0.5)/(float)1.0)) * (float)1.0;
      break; 

    case ST_VDC_DAC:
    case ST_VDC_DAC0:
    case ST_VDC_DAC1:
      if(var < 0) v = 0.0;
      else if(var > 1023.0) v = 1023.0;
      else v = ((int)((var+0.5)/(float)1.0)) * (float)1.0;
      break; 


    default:      /* variable not to be scanned */
      v = 0;
      break;
  }

  *newvar = v;
  if (v<var) {
    printf("st_coerce_variable: parameter %hd  reduced  from %3.2f to %3.2f\n",typ,var,v);
    return(2);
  }
  else if (v>var) {
    printf("st_coerce_variable: parameter %hd increased from %3.2f to %3.2f\n",typ,var,v);
    return(3);
  }
  else if  (v==var) {
    printf("st_coerce_variable: parameter %hd unchanged from %3.2f\n",typ,v);
    return(1);
  }
  return(0);
}  

void st_configure_variable(short mid, short typ, float var) {
  
  /*
     Configure one variable of type [typ] to the value [var] for module [mid].
     If [mid] == -1, do all present and active modules.

     Usage:
        st_coerce_variable()    // performs validity and range checks

        one or more calls to st_configure_variable() for one or more
          modules, or for all active modules, or for one or more variable types.

        st_execute_configs()    // rebuilds and executes config command blocks
                                // for all *present* modules 
  */

  float v, vv;
  short lo_mod,hi_mod,i,j,k;
  short mask = (short)0xffff;
  unsigned short conf[] = {1, 3, 5, 7, 11, 13, 15, 21, 23, 27, 29, 31, 43, 45, 47, 53, 55, 59, 61, 63};

  if (mid >= SCT_MAX_MODULES){
    printf("st_configure_variable: module %d outside limit of %d modules\n", mid, SCT_MAX_MODULES);
    return;
  }

  printf("st_configure_variable: module %hd typ %hd value %3.2f\n", mid, typ, var);

  if (mid < 0) {   /*  many module case  */
    lo_mod = 0;
    hi_mod = SCT_MAX_MODULES;
  } else {         /* single module case */
    if ((sct[mid].present) == 0) return;
    lo_mod = mid;
    hi_mod = mid+1;
  }

  /* printf("lo_mod = %hd hi_mod = %hd\n", lo_mod, hi_mod);*/

  for (i=lo_mod; i<hi_mod; i++) {

    /* update variable if present AND (active or single module) */

    if ((sct[i].present) && ((sct[i].active)||((lo_mod+1)==hi_mod))){

      switch(typ) {

        case ST_VTHR:
          for (j=0; j<12; j++)
            if (sct[i].det.chips[j].active) sct[i].det.chips[j].vthr = var;
          break;

        case ST_VCAL:
          for (j=0; j<12; j++)
            if (sct[i].det.chips[j].active) sct[i].det.chips[j].vcal = var;
          break;

        case ST_STROBE_DELAY:
          for (j=0; j<12; j++)
            if (sct[i].det.chips[j].active) sct[i].det.chips[j].delay = var;
          break;

        case ST_PREAMP:
          for (j=0; j<12; j++)
            if (sct[i].det.chips[j].active) sct[i].det.chips[j].febias = var;
          break;

        case ST_SHAPER:
          for (j=0; j<12; j++)
            if (sct[i].det.chips[j].active) sct[i].det.chips[j].feshaper = var;
          break;

        case ST_TRIM:
          if (sct[i].det.chipset==ABCDT){
            for (j=0; j<12; j++){
              if (sct[i].det.chips[j].active){
                for (k=0; k<128; k++) sct[i].det.chips[j].trim[k] = (short)var;
              }
            }
            sct[i].det.updatetrims = 1;
          }
          break;

        case ST_MASK:
          /* Revised PWP 01.11.00
           * This is for global, simple mask register settings.
           * To mask an individual channel, use st_mask_chan in st_config.c
           */

          if(var==0.0) mask = (short)0x0000;
          if(var==1.0) mask = (short)0x5555;
          if(var==2.0) mask = (short)0xaaaa;
          if(var==3.0) mask = (short)0xffff;

          for (j=0; j<12;j++)
            if (sct[i].det.chips[j].active) {
              for (k=0; k<8; k++) sct[i].det.chips[j].mask[k] = mask;
            }
          break;

        case ST_ROLE: /* PWP 12.12.00 */
          for (j=0; j<12; j++)
            if (sct[i].det.chips[j].active) sct[i].det.chips[j].role = (short)var;
          break;

        case ST_NMASK:
          /* PWP 10.05.01
           * Mask first "var" channels of a module
           * may be useful togther with SEND_MASK during diagnostics
           * since the event size is fixed.
           */

          for (j=0;j<12;j++){         /* unmask all channels */
            for (k=0; k<8; k++) sct[i].det.chips[j].mask[k] = (short) 0xffff;
          }
          for (j=0;j<(short)var;j++){ /* mask (var) channels */
            st_mask_chan(i,j);
          }
          break;

        case ST_CAL_MODE:
          for (j=0; j<12; j++)
            if (sct[i].det.chips[j].active) sct[i].det.chips[j].cal_mode = (short)var;
          break;

        case ST_COMPRESSION:
          for (j=0; j<12; j++)
            if (sct[i].det.chips[j].active) sct[i].det.chips[j].compression = (short)var;
          break;

        case ST_TRIM_RANGE:
          for (j=0; j<12; j++)
            if (sct[i].det.chips[j].active) sct[i].det.chips[j].trim_range = (short)var;
          break;

        case ST_EDGE_DETECT:
          for (j=0; j<12; j++)
            if (sct[i].det.chips[j].active) sct[i].det.chips[j].edge_detect = (short)var;
          break;

        case ST_SEND_MASK:
          for (j=0; j<12; j++)
            if (sct[i].det.chips[j].active) sct[i].det.chips[j].mask_readout = (short)var;
          break;

        case ST_ACCUMULATE:
          for (j=0; j<12; j++)
            if (sct[i].det.chips[j].active) sct[i].det.chips[j].accumulate = (short)var;
          break;

        case ST_DTM:
          sct[i].det.dtm=(short)var;
          break;

        case ST_BYPASS:
          sct[mid].det.bypass = (short)var;
          for (j=0; j<12; j++){
            sct[i].det.chips[j].role=MISSING;
          }
          switch((int)var){
            case 0:
              sct[i].det.chips[0].role=MASTER;
              sct[i].det.chips[1].role=END;
              sct[i].det.chips[6].role=MASTER;
              sct[i].det.chips[7].role=END;
              sct[i].link0 = 1;
              sct[i].link1 = 1;
              sct[i].mus_enable0 = 1;
              sct[i].mus_enable1 = 1;
              break;
            case 1:
              sct[i].det.chips[0].role=MASTER;
              sct[i].det.chips[1].role=SLAVE;
              sct[i].det.chips[2].role=END;
              sct[i].det.chips[6].role=MASTER;
              sct[i].det.chips[7].role=SLAVE;
              sct[i].det.chips[8].role=END;
              sct[i].link0 = 1;
              sct[i].link1 = 1;
              sct[i].mus_enable0 = 1;
              sct[i].mus_enable1 = 1;
              break;
            case 2:
              sct[i].det.chips[0].role=MASTER;
              sct[i].det.chips[1].role=SLAVE;
              sct[i].det.chips[2].role=SLAVE;
              sct[i].det.chips[3].role=END;
              sct[i].det.chips[6].role=MASTER;
              sct[i].det.chips[7].role=SLAVE;
              sct[i].det.chips[8].role=SLAVE;
              sct[i].det.chips[9].role=END;
              sct[i].link0 = 1;
              sct[i].link1 = 1;
              sct[i].mus_enable0 = 1;
              sct[i].mus_enable1 = 1;
              break;
            case 3:
              sct[i].det.chips[0].role=MASTER;
              sct[i].det.chips[1].role=SLAVE;
              sct[i].det.chips[2].role=SLAVE;
              sct[i].det.chips[3].role=SLAVE;
              sct[i].det.chips[4].role=END;
              sct[i].det.chips[6].role=MASTER;
              sct[i].det.chips[7].role=SLAVE;
              sct[i].det.chips[8].role=SLAVE;
              sct[i].det.chips[9].role=SLAVE;
              sct[i].det.chips[10].role=END;
              sct[i].link0 = 1;
              sct[i].link1 = 1;
              sct[i].mus_enable0 = 1;
              sct[i].mus_enable1 = 1;
              break;
            /* case 4 is the default configuration */
            case 5:
              sct[i].det.chips[0].role=MASTER;
              sct[i].det.chips[1].role=DEAD;
              sct[i].det.chips[2].role=END;
              sct[i].det.chips[6].role=MASTER;
              sct[i].det.chips[7].role=DEAD;
              sct[i].det.chips[8].role=END;
              sct[i].link0 = 1;
              sct[i].link1 = 1;
              sct[i].mus_enable0 = 1;
              sct[i].mus_enable1 = 1;
              break;
            case 6:
              sct[i].det.chips[0].role=MASTER;
              sct[i].det.chips[1].role=SLAVE;
              sct[i].det.chips[2].role=DEAD;
              sct[i].det.chips[3].role=END;
              sct[i].det.chips[6].role=MASTER;
              sct[i].det.chips[7].role=SLAVE;
              sct[i].det.chips[8].role=DEAD;
              sct[i].det.chips[9].role=END;
              sct[i].link0 = 1;
              sct[i].link1 = 1;
              sct[i].mus_enable0 = 1;
              sct[i].mus_enable1 = 1;
              break;
            case 7:
              sct[i].det.chips[0].role=MASTER;
              sct[i].det.chips[1].role=SLAVE;
              sct[i].det.chips[2].role=SLAVE;
              sct[i].det.chips[3].role=DEAD;
              sct[i].det.chips[4].role=END;
              sct[i].det.chips[6].role=MASTER;
              sct[i].det.chips[7].role=SLAVE;
              sct[i].det.chips[8].role=SLAVE;
              sct[i].det.chips[9].role=DEAD;
              sct[i].det.chips[10].role=END;
              sct[i].link0 = 1;
              sct[i].link1 = 1;
              sct[i].mus_enable0 = 1;
              sct[i].mus_enable1 = 1;
              break;
            case 8:
              sct[i].det.chips[0].role=MASTER;
              sct[i].det.chips[1].role=SLAVE;
              sct[i].det.chips[2].role=SLAVE;
              sct[i].det.chips[3].role=SLAVE;
              sct[i].det.chips[4].role=DEAD;
              sct[i].det.chips[5].role=END;
              sct[i].det.chips[6].role=MASTER;
              sct[i].det.chips[7].role=SLAVE;
              sct[i].det.chips[8].role=SLAVE;
              sct[i].det.chips[9].role=SLAVE;
              sct[i].det.chips[10].role=DEAD;
              sct[i].det.chips[11].role=END;
              sct[i].link0 = 1;
              sct[i].link1 = 1;
              sct[i].mus_enable0 = 1;
              sct[i].mus_enable1 = 1;
              break;
            case 9:
              sct[i].det.chips[0].role=MASTER;
              sct[i].det.chips[1].role=SLAVE;
              sct[i].det.chips[2].role=SLAVE;
              sct[i].det.chips[3].role=SLAVE;
              sct[i].det.chips[4].role=SLAVE;
              sct[i].det.chips[5].role=SLAVE;
              sct[i].det.chips[6].role=DEAD;
              sct[i].det.chips[7].role=SLAVE;
              sct[i].det.chips[8].role=SLAVE;
              sct[i].det.chips[9].role=SLAVE;
              sct[i].det.chips[10].role=SLAVE;
              sct[i].det.chips[11].role=END;
              sct[i].link0 = 2;
              sct[i].link1 = 0;
              sct[i].mus_enable0 = 1;
              sct[i].mus_enable1 = 0;
              break;
            case 10:
              sct[i].det.chips[0].role=DEAD;
              sct[i].det.chips[1].role=SLAVE;
              sct[i].det.chips[2].role=SLAVE;
              sct[i].det.chips[3].role=SLAVE;
              sct[i].det.chips[4].role=SLAVE;
              sct[i].det.chips[5].role=END;
              sct[i].det.chips[6].role=MASTER;
              sct[i].det.chips[7].role=SLAVE;
              sct[i].det.chips[8].role=SLAVE;
              sct[i].det.chips[9].role=SLAVE;
              sct[i].det.chips[10].role=SLAVE;
              sct[i].det.chips[11].role=SLAVE;
              sct[i].link0 = 0;
              sct[i].link1 = 2;
              sct[i].mus_enable0 = 0;
              sct[i].mus_enable1 = 1;
              break;
            case 11:
              sct[i].det.chips[0].role=MASTER;
              sct[i].det.chips[1].role=SLAVE;
              sct[i].det.chips[2].role=SLAVE;
              sct[i].det.chips[3].role=SLAVE;
              sct[i].det.chips[4].role=SLAVE;
              sct[i].det.chips[5].role=SLAVE;
              sct[i].det.chips[6].role=SLAVE;
              sct[i].det.chips[7].role=SLAVE;
              sct[i].det.chips[8].role=SLAVE;
              sct[i].det.chips[9].role=SLAVE;
              sct[i].det.chips[10].role=SLAVE;
              sct[i].det.chips[11].role=END;
              sct[i].link0 = 2;
              sct[i].link1 = 0;
              sct[i].mus_enable0 = 1;
              sct[i].mus_enable1 = 0;
              break;
            case 12:
              sct[i].det.chips[0].role=SLAVE;
              sct[i].det.chips[1].role=SLAVE;
              sct[i].det.chips[2].role=SLAVE;
              sct[i].det.chips[3].role=SLAVE;
              sct[i].det.chips[4].role=SLAVE;
              sct[i].det.chips[5].role=END;
              sct[i].det.chips[6].role=MASTER;
              sct[i].det.chips[7].role=SLAVE;
              sct[i].det.chips[8].role=SLAVE;
              sct[i].det.chips[9].role=SLAVE;
              sct[i].det.chips[10].role=SLAVE;
              sct[i].det.chips[11].role=SLAVE;
              sct[i].link0 = 0;
              sct[i].link1 = 2;
              sct[i].mus_enable0 = 0;
              sct[i].mus_enable1 = 1;
              break;
            case 13:
              sct[i].det.chips[0].role=MASTER;
              sct[i].det.chips[1].role=SLAVE;
              sct[i].det.chips[2].role=SLAVE;
              sct[i].det.chips[3].role=SLAVE;
              sct[i].det.chips[4].role=SLAVE;
              sct[i].det.chips[5].role=DEAD;
              sct[i].det.chips[6].role=SLAVE;
              sct[i].det.chips[7].role=SLAVE;
              sct[i].det.chips[8].role=SLAVE;
              sct[i].det.chips[9].role=SLAVE;
              sct[i].det.chips[10].role=SLAVE;
              sct[i].det.chips[11].role=END;
              sct[i].link0 = 2;
              sct[i].link1 = 0;
              sct[i].mus_enable0 = 1;
              sct[i].mus_enable1 = 0;
              break;
            case 14:
              sct[i].det.chips[0].role=SLAVE;
              sct[i].det.chips[1].role=SLAVE;
              sct[i].det.chips[2].role=SLAVE;
              sct[i].det.chips[3].role=SLAVE;
              sct[i].det.chips[4].role=SLAVE;
              sct[i].det.chips[5].role=END;
              sct[i].det.chips[6].role=MASTER;
              sct[i].det.chips[7].role=SLAVE;
              sct[i].det.chips[8].role=SLAVE;
              sct[i].det.chips[9].role=SLAVE;
              sct[i].det.chips[10].role=SLAVE;
              sct[i].det.chips[11].role=DEAD;
              sct[i].link0 = 0;
              sct[i].link1 = 2;
              sct[i].mus_enable0 = 0;
              sct[i].mus_enable1 = 1;
              break;
            case 4:
            default:
              sct[i].det.chips[0].role=MASTER;
              sct[i].det.chips[1].role=SLAVE;
              sct[i].det.chips[2].role=SLAVE;
              sct[i].det.chips[3].role=SLAVE;
              sct[i].det.chips[4].role=SLAVE;
              sct[i].det.chips[5].role=END;
              sct[i].det.chips[6].role=MASTER;
              sct[i].det.chips[7].role=SLAVE;
              sct[i].det.chips[8].role=SLAVE;
              sct[i].det.chips[9].role=SLAVE;
              sct[i].det.chips[10].role=SLAVE;
              sct[i].det.chips[11].role=END;
              sct[i].link0 = 1;
              sct[i].link1 = 1;
              sct[i].mus_enable0 = 1;
              sct[i].mus_enable1 = 1;
              break;
          }
          break; /* end case ST_BYPASS */

        case ST_TOKEN:
	  /*
	   * This variable iterates over a set of token/data passing
	   * schemes to determine the functionallity of each token link.
	   * The sequence sets all the combinations to read-out
	   * n chips by first testing the bypass token (chip n-2 to n) 
	   * for all combinations of (n-2) chips. Secondly trying the 
	   * direct token (n-1 to n) for all combinations of (n-1) chips.
	   * For n<=6 (configurations 0-19), both link0 and link1 are 
	   * tested simultaneously. When n>6 one link is read out at the
	   * time, alternating even numbers link0 and odd numbers link1.

	   * Configurations 20-35 reads out S9 (S1 for odd numbers) via
	   * the 8 possible combinations of reading out chip E5 (E13).

	   * Configurations 36-61 reads chip M8 (M1) via the bypass token 
	   * for all combinations of accessing S4 (S12). Then via the direct
	   * token for all combinations of accessing E5 (E13). 

	   * The logical inconsistecy in the order between cases 20-35 and
	   * 36-61 is done to make the possible read-out schemes for the 
	   * barrel hybrid a consecutive series.

	   * LE 3/3-2001*/

          sct[mid].det.token = (short)var;
	  if(var<20) {
	    sct[i].link0 = 1;
	    sct[i].link1 = 1;
	    sct[i].mus_enable0 = 1;
	    sct[i].mus_enable1 = 1;
	    st_configure_token(i, (conf[(int)var] + (conf[(int)var] << 6)), 1, 1);
	  } else if(var<36) {
	    if(!((int)var%2)) {
	      sct[i].link0 = 2;
	      sct[i].link1 = 0;
	      sct[i].mus_enable0 = 1;
	      sct[i].mus_enable1 = 0;
	      st_configure_token(i, (conf[(int)(var-20)/2+12] + (0x1 << 7)), 2, 0);
	    } else {
	      sct[i].link0 = 0;
	      sct[i].link1 = 2;
	      sct[i].mus_enable0 = 0;
	      sct[i].mus_enable1 = 1;
	      st_configure_token(i, (0x2 + (conf[((int)var-21)/2+12] << 6)), 0, 2);
	    }
	  } else { /* only possible for forward modules */
	    if(!((int)var%2)) {
	      sct[i].link0 = 2;
	      sct[i].link1 = 0;
	      sct[i].mus_enable0 = 1;
	      sct[i].mus_enable1 = 0;
	      st_configure_token(i, (conf[(int)((var-36)/2+7)] + (1 << 6)), 2, 0);
	    } else {
	      sct[i].link0 = 0;
	      sct[i].link1 = 2;
	      sct[i].mus_enable0 = 0;
	      sct[i].mus_enable1 = 1;
	      st_configure_token(i, (0x1 + (conf[(int)((var-37)/2+7)] << 6)), 0, 2);
	    }
	  }
	  break;

        case ST_QTHR:
          for (j=0; j<12; j++){
            if (sct[i].det.chips[j].active){
              switch(sct[i].det.chips[j].rc_function){
                case 1: /* polynomial */
                  v = sct[i].det.chips[j].rc_par[0] + 
                      var * (sct[i].det.chips[j].rc_par[1] + 
                      (var * sct[i].det.chips[j].rc_par[2]) );
                  break; 
                case 2: /* grillo */
                  if((var==0)|| (sct[i].det.chips[j].rc_par[1]==0))
                    vv = 0;
                  else
                    vv = sct[i].det.chips[j].rc_par[2] / 
                        (sct[i].det.chips[j].rc_par[1] *var);
                  v  = sct[i].det.chips[j].rc_par[0] + 
                       sct[i].det.chips[j].rc_par[2] / (float) sqrt(1+(vv*vv)); 
                  break; 
                case 3: /* exp */
                  if(sct[i].det.chips[j].rc_par[1]==0)
                    v = 0;
                  else
                    v = sct[i].det.chips[j].rc_par[2] + 
                        sct[i].det.chips[j].rc_par[0]/(1+ (float)exp(-var/sct[i].det.chips[j].rc_par[1]));
                  break;  
                case 4: /* linear fit */
                  v = sct[i].det.chips[j].rc_par[0] + 
                      (var * sct[i].det.chips[j].rc_par[1]);
                  break; 
                default:
                  v = 0;
                  break;
              }
              vv = ((int)((v+1.25)/(float)2.5)) * (float)2.5;   /* 2.5mV / DAC bit */
              sct[i].det.chips[j].vthr = vv;
            }
          }
          break;

        case ST_QCAL:
          for (j=0; j<12; j++){
            if (sct[i].det.chips[j].active){

              if(sct[i].det.chipset == ABC_CAFE){
                if (var > 9.9609375) v = 9.9609375;   /* 0.0390625fC / DAC bit */
                else v = ((int)((var+0.01953125)/(float)0.0390625)) * (float)0.625;

              } else { /* ABCD */
                /*
                 * calculate the actual step size taking into
                 * account the capacitor correction factor:
                 * step size = (c_factor * 0.0625) fC / DAC bit 
                 */
                vv = (float) 0.0625 * sct[i].det.chips[j].c_factor;

                if (var > (255*vv)) v = 15.9375; /* maximum vcal setting */
                else v = ((int)((var+(vv/2))/vv)) * (float) 0.625;
              }

              sct[i].det.chips[j].vcal = v;

            } /* end if(active) */
          } /* end for(chip) */
          break;

        case ST_STREAM_DELAY:
          sct[i].mus_delay0 = (short)var;
          sct[i].mus_delay1 = (short)var;
	
          if (MUS_NUMBER > 0) {
            mus_set_stream_delay(sct[i].mus_id, sct[i].mus_ch0, sct[i].mus_delay0);
            mus_set_stream_delay(sct[i].mus_id, sct[i].mus_ch1, sct[i].mus_delay1);
          }
          break;

        case ST_STREAM_DELAY0:
          sct[i].mus_delay0 = (short)var;		

          if (MUS_NUMBER > 0) {
             mus_set_stream_delay(sct[i].mus_id, sct[i].mus_ch0, sct[i].mus_delay0);
          }    
          break;

        case ST_STREAM_DELAY1:

          sct[i].mus_delay1 = (short)var;
  		
          if (MUS_NUMBER > 0) {
            mus_set_stream_delay(sct[i].mus_id, sct[i].mus_ch1, sct[i].mus_delay1);
          }
          break;

      
        case ST_VDET:
          sct[i].det.vdet_set = var;
          /*
           * DO NOT SEND COMMANDS TO SCTHV MODULES HERE!!
           * PWP 12.07.01
           */
          break;

        case ST_VCC:
          sct[i].det.vcc_set = var;
          if((sct[i].lv_ch>=0)&&(sct[i].lv_ch<sctlv_n_channels()))
            sctlv_set_vcc(sct[i].lv_ch, sct[i].det.vcc_set, sct[i].det.icc_lim);
          break;

        case ST_VI1:
          sct[i].det.vi1_set = var;
          if((sct[i].lv_ch>=0)&&(sct[i].lv_ch<sctlv_n_channels()))
            sctlv_set_vvi1(sct[i].lv_ch, sct[i].det.vi1_set, sct[i].det.ivi1_lim);
          break;

        case ST_VDD:
          sct[i].det.vdd_set = var;
          if((sct[i].lv_ch>=0)&&(sct[i].lv_ch<sctlv_n_channels()))
            sctlv_set_vdd(sct[i].lv_ch, sct[i].det.vdd_set, sct[i].det.idd_lim);
          break;

        case ST_VPIN:
          /* Note:  Vpin is programmable only for 
             SCTLV versions >= 5, PWP 15.12.99 */
          sct[i].det.vpin_set = var;
          if((sct[i].lv_ch>=0)&&(sct[i].lv_ch<sctlv_n_channels()))
            sctlv_set_vpin(sct[i].lv_ch, sct[i].det.vpin_set); 
          break;

        case ST_VLED:
          sct[i].det.vled0_set = var;
          sct[i].det.vled1_set = var;
          if((sct[i].lv_ch>=0)&&(sct[i].lv_ch<sctlv_n_channels())){
            sctlv_set_vled0(sct[i].lv_ch, sct[i].det.vled0_set, sct[i].det.iled0_lim);
            sctlv_set_vled1(sct[i].lv_ch, sct[i].det.vled1_set, sct[i].det.iled1_lim);
          }
          break;
      
        case ST_VLED0:
          sct[i].det.vled0_set = var;
          if((sct[i].lv_ch>=0)&&(sct[i].lv_ch<sctlv_n_channels()))
            sctlv_set_vled0(sct[i].lv_ch, sct[i].det.vled0_set, sct[i].det.iled0_lim);
          break;
  
        case ST_VLED1:
          sct[i].det.vled1_set = var;
          if((sct[i].lv_ch>=0)&&(sct[i].lv_ch<sctlv_n_channels()))
            sctlv_set_vled1(sct[i].lv_ch, sct[i].det.vled1_set, sct[i].det.iled1_lim);
          break; 
   
        case ST_SELECT: /* added PWP 01.10.00 */
          sct[i].det.select = (short)var;
          if((sct[i].lv_ch>=0)&&(sct[i].lv_ch<sctlv_n_channels()))
            sctlv_set_select(sct[i].lv_ch, sct[i].det.select);
          break; 

        case ST_RAMP: /* added PWP 12.12.00 */
          sct[i].det.ramp = (short)var;
          break; 

        case ST_IDET:
          sct[i].det.idet_lim = var;
          break;

        case ST_ICC:
          sct[i].det.icc_lim = var;
          if((sct[i].lv_ch>=0)&&(sct[i].lv_ch<sctlv_n_channels()))
            sctlv_set_vcc(sct[i].lv_ch, sct[i].det.vcc_set, sct[i].det.icc_lim);
          break;

        case ST_IVI1:
          sct[i].det.ivi1_lim = var;
          if((sct[i].lv_ch>=0)&&(sct[i].lv_ch<sctlv_n_channels()))
            sctlv_set_vvi1(sct[i].lv_ch, sct[i].det.vi1_set, sct[i].det.ivi1_lim);
          break;

        case ST_IDD:
          sct[i].det.idd_lim = var;
          if((sct[i].lv_ch>=0)&&(sct[i].lv_ch<sctlv_n_channels()))
            sctlv_set_vdd(sct[i].lv_ch, sct[i].det.vdd_set, sct[i].det.idd_lim);
          break;

        case ST_IPIN:
          /* Note:  Vpin is programmable only for 
             SCTLV versions >= 5, PWP 15.12.99 */
          sct[i].det.ipin_lim = var;
          if((sct[i].lv_ch>=0)&&(sct[i].lv_ch<sctlv_n_channels()))
            sctlv_set_vpin(sct[i].lv_ch, sct[i].det.vpin_set); 
          break;

        case ST_ILED:
          sct[i].det.iled0_lim = var;
          sct[i].det.iled1_lim = var;
          if((sct[i].lv_ch>=0)&&(sct[i].lv_ch<sctlv_n_channels())){
            sctlv_set_vled0(sct[i].lv_ch, sct[i].det.vled0_set, sct[i].det.iled0_lim);
            sctlv_set_vled1(sct[i].lv_ch, sct[i].det.vled1_set, sct[i].det.iled1_lim);
          }
          break;
      
        case ST_ILED0:
          sct[i].det.iled0_lim = var;
          if((sct[i].lv_ch>=0)&&(sct[i].lv_ch<sctlv_n_channels()))
            sctlv_set_vled0(sct[i].lv_ch, sct[i].det.vled0_set, sct[i].det.iled0_lim);
          break;
  
        case ST_ILED1:
          sct[i].det.iled1_lim = var;
          if((sct[i].lv_ch>=0)&&(sct[i].lv_ch<sctlv_n_channels()))
            sctlv_set_vled1(sct[i].lv_ch, sct[i].det.vled1_set, sct[i].det.iled1_lim);
          break;

 
           /* The following cases are timing parameters which can only be set for
           all modules together (eg, trigger delay) or for all modules
           connected to a single slog (eg, clock duty cycle).  They are
           included here to allow a common scanning mechanism even though this
           means they will be repeatedly called for each active module.  Care
           must be taken that the called functions are efficient, for example,
           returning immediately if the requested value is already set. */
  
        case ST_CLOCK_DUTY:
          if (sct[i].det.select == 0) slog_set_clock_duty(sct[i].slog_id, (short)var);
          else                        slog_set_clock_duty(sct[i].r_slog_id, (short)var);
          break;

        case ST_CLOCK_PHASE:
          if (sct[i].det.select == 0) slog_set_clock_phase(sct[i].slog_id, (short)var);
          else                        slog_set_clock_phase(sct[i].r_slog_id, (short)var);
          break;

        case ST_COM_DELAY:
          com_delay = (unsigned short)var;
          break;

        case ST_PULSE_DELAY:
          pulse_delay = (unsigned short)var;
          break;
        
        case ST_TRIG_DELAY:
          trig_delay = (unsigned short)var;
          if(CLOAC_NUMBER>0) cc_load_delay((unsigned short)(trig_delay-CLOAC_CAL_OFFSET));
          break;

        case ST_SR_FREQUENCY:  /* added PWP 27.06.00 */
          sr_frequency = (unsigned short)var;
          if(CLOAC_NUMBER>0) cc_menu_srfreq((unsigned short)sr_frequency);
          break;

        case ST_TR_FREQUENCY:  /* added PWP 27.06.00 */
          tr_frequency = (unsigned short)var;
          if(CLOAC_NUMBER>0) cc_menu_trfreq((unsigned short)tr_frequency);
          break;

        case ST_NL1A:        /* added PWP 22.07.00 */
          n_l1a = (unsigned short) var;
          break;

        case ST_L1A_SPACING: /* added PWP 22.07.00 */
          l1a_spacing = (unsigned short) var;
          break; 

        /* OPTO - added by PWP 10.10.00*/

        case ST_OPTO_RX_THR:
          sct[i].opto_rx_thr0 = (short)var;
          sct[i].opto_rx_thr1 = (short)var;
          if((sct[i].opto_rx_id>=0)&&(sct[i].opto_rx_id<OPTIF_NUMBER)){
            optif_set_threshold(sct[i].opto_rx_id, sct[i].opto_rx_ch0, sct[i].opto_rx_thr0);
            optif_set_threshold(sct[i].opto_rx_id, sct[i].opto_rx_ch1, sct[i].opto_rx_thr1);
          }
          break;

        case ST_OPTO_RX_THR0:
          sct[i].opto_rx_thr0 = (short)var;
          if((sct[i].opto_rx_id>=0)&&(sct[i].opto_rx_id<OPTIF_NUMBER)){
            optif_set_threshold(sct[i].opto_rx_id, sct[i].opto_rx_ch0, sct[i].opto_rx_thr0);
          }
          break;

        case ST_OPTO_RX_THR1:
          sct[i].opto_rx_thr1 = (short)var;
          if((sct[i].opto_rx_id>=0)&&(sct[i].opto_rx_id<OPTIF_NUMBER)){
            optif_set_threshold(sct[i].opto_rx_id, sct[i].opto_rx_ch1, sct[i].opto_rx_thr1);
          }
          break;

        case ST_OPTO_TX_DAC:
          sct[i].opto_tx_iset = (short)var;
          if((sct[i].opto_tx_type==0)&&(sct[i].opto_tx_id>=0)&&(sct[i].opto_tx_id<BILED_NUMBER)){
            biled_set_dac(sct[i].opto_tx_id,(short)(sct[i].opto_tx_ch+8),sct[i].opto_tx_iset);            
          }
          if((sct[i].opto_tx_type==1)&&(sct[i].opto_tx_id>=0)&&(sct[i].opto_tx_id<OPTIF_NUMBER)){
            optif_set_laser(sct[i].opto_tx_id, sct[i].opto_tx_ch, sct[i].opto_tx_iset);
          }
          break;

        case ST_OPTO_TX_MKSPNEG:
          if((sct[i].opto_tx_id>=0)&&(sct[i].opto_tx_id<OPTIF_NUMBER)){
            optif_bpm_set_mkspneg(sct[i].opto_tx_id, sct[i].opto_tx_ch, (short)var);
          }
          break;

        case ST_OPTO_TX_MKSPPOS:
          if((sct[i].opto_tx_id>=0)&&(sct[i].opto_tx_id<OPTIF_NUMBER)){
            optif_bpm_set_mksppos(sct[i].opto_tx_id, sct[i].opto_tx_ch, (short)var);
          }
          break;

        case ST_OPTO_TX_CDELAY:
          if((sct[i].opto_tx_id>=0)&&(sct[i].opto_tx_id<OPTIF_NUMBER)){
            optif_bpm_set_coarsedelay(sct[i].opto_tx_id, sct[i].opto_tx_ch, (short)var);
          }
          break;

        case ST_OPTO_TX_FDELAY:
          if((sct[i].opto_tx_id>=0)&&(sct[i].opto_tx_id<OPTIF_NUMBER)){
            optif_bpm_set_finedelay(sct[i].opto_tx_id, sct[i].opto_tx_ch, (short)var);
          }
          break;

        /* FORWARD SUPPORT CARD - PWP 15.05.01 */

        case ST_FWD_SC_MODE:
          sct[i].det.sc_mode = (short)var;
          break; 

        case ST_BPM_DRIVE:
          sct[i].det.bpm_drive = (short)var;
          break; 

        case ST_VDC_DAC:
          sct[i].det.vdc_dac0 = (short)var;
          sct[i].det.vdc_dac1 = (short)var;
          break;

        case ST_VDC_DAC0:
          sct[i].det.vdc_dac0 = (short)var;
          break;

        case ST_VDC_DAC1:
          sct[i].det.vdc_dac1 = (short)var;
          break;

        default:      /* variable not to be scanned */
          break;
       } /* end switch */
    } /* end if */
  } /* end for */

  /* if the bypass configuration or SELECT line has been changed, 
   * we must enable the relevant clock, command and data streams
   * PWP 02.11.00 
   */

  if((typ==ST_BYPASS)||(typ==ST_SELECT)||(typ==ST_TOKEN)){
    st_execute_enables();
  }

  /* 
   * New section PWP 16.07.99 revised 25.08.99- 
   * maintain a record of the last * FOUR * variables * which have been 
   * changed and the values assigned to them.  Four variables is
   * consistent with making gain measurements (i.e. a series of threshold
   * scans at different q) as a function of two other variables.  No action
   * is taken in the case of ST_CAL_MODE, or if the variable has been 
   * changed for only one module (and hence is probably not part of a scan).
   */
  
  if((typ != ST_CAL_MODE) && (hi_mod>lo_mod)){
    /* revised setting of one of the three scan parameters */
    if(typ == scan_typ[0]){
      scan_pnt[0] ++;      /* increment scan point counter */
      scan_var[0] = var;
    }
    else if(typ == scan_typ[1]){
      scan_pnt[0] = 0;     /* reset counter for variable 1 */
      scan_pnt[1] ++;
      scan_var[1] = var;
    }
    else if(typ == scan_typ[2]){
      scan_pnt[0] = 0;     /* reset counter for variables 1 + 2 */
      scan_pnt[1] = 0;
      scan_pnt[2] ++;
      scan_var[2] = var;
    }
    else if(typ == scan_typ[3]){
      scan_pnt[0] = 0;     /* reset counter for variables 1, 2 and 3 */
      scan_pnt[1] = 0;
      scan_pnt[2] = 0;
      scan_pnt[3] ++;
      scan_var[3] = var;
    }
    else
    {      /* move along, right down the bus please !! */
      scan_typ[3] = scan_typ[2];
      scan_typ[2] = scan_typ[1];
      scan_typ[1] = scan_typ[0];
      scan_typ[0] = typ;
      scan_var[3] = scan_var[2];
      scan_var[2] = scan_var[1];
      scan_var[1] = scan_var[0];
      scan_var[0] = var;
      scan_pnt[3] = 1;
      scan_pnt[2] = 1;
      scan_pnt[1] = 1;
      scan_pnt[0] = 1;
    }

    slog_do_loadtrig = 1; /* PWP bugfix 15.11.00 */
  }
}  
void st_configure_token(int mid, unsigned token, int link0, int link1) {
  /*
   * Configures chips as present or missing according to 
   * the bit-pattern given in 'token'. Sets Master, slave 
   * and End chips appropriately. LE 23/2-2001.
   */

  int j;

  if(link0>0) {
    sct[mid].mus_enable0 = 1; 
  } else {
    sct[mid].mus_enable0 = 0;
  } 
  if(link1>0) { 
    sct[mid].mus_enable1 = 1; 
  } else { 
    sct[mid].mus_enable1 = 0; 
  }
  sct[mid].link0 = link0;
  sct[mid].link1 = link1;
  


  for(j=0; j<12; j++){
    
    if( (token >> j) & 0x1 ) {

      if(j==0) {
	if(link0) {
	  if((token&0x3F)==1) {
	    sct[mid].det.chips[j].role=LONELY;
	  } else {
	    sct[mid].det.chips[j].role=MASTER;
	  }
	} else {
	  sct[mid].det.chips[j].role=END;
	}

      } else if(j==6) {
	if (link1) { 
	  if( (token&0x0FC0)==64) {
	    sct[mid].det.chips[j].role=LONELY;
	  } else {
	    sct[mid].det.chips[j].role=MASTER;
	  }
	} else {
	  sct[mid].det.chips[j].role=END;
	}

      } else if(j<6) {
	if( (link0<2) && (pow(2,(j+1))>(token&0x3F)) ) {
	  sct[mid].det.chips[j].role=END;	  
	} else {
	  sct[mid].det.chips[j].role=SLAVE;
	}

      } else {
	if( (link1<2) && (pow(2,(j+1))>token) ) {
	  sct[mid].det.chips[j].role=END;	  
	} else {
	  sct[mid].det.chips[j].role=SLAVE;
	}
      }

    } else {
      sct[mid].det.chips[j].role=DEAD;
    }
  }
}


void st_configure_trim(short mid, short cid, short chan, short var) {
  
  /*
     Configure the trim value of channel chan of chip cid of 
     module mid to value var

     PWP 01.11.00
  */

  short v2;

  if ((mid<0)||(mid >= SCT_MAX_MODULES)){
    printf("st_configure_trim: mod %d chip %d chan %d  - invalid module number\n",
            mid,cid,chan);
    return;
  }

  if ((cid<0)||(cid >= 12)){
    printf("st_configure_trim: mod %d chip %d chan %d  - invalid chip number\n",
            mid,cid,chan);
    return;
  }

  if ((chan<0)||(chan >= 128)){
    printf("st_configure_trim: mod %d chip %d chan %d  - invalid channel number\n",
            mid,cid,chan);
    return;
  }


  v2=var;

  if (var<0){
    printf("st_configure_trim: mod %d chip %d chan %d var %d - invalid trim, using 0\n",
            mid,cid,chan,var);
    v2=0;
  }

  if (var>15){
    printf("st_configure_trim: mod %d chip %d chan %d var %d - invalid trim, using 15\n",
            mid,cid,chan,var);
    v2=15;
  }

  if ((sct[mid].present) && (sct[mid].active) && (sct[mid].det.chips[cid].active)){

    sct[mid].det.chips[cid].trim[chan]=v2;
    sct[mid].det.updatetrims = 1;          /* PWP bugfix 15.11.00 */
  }

}


void st_configure_chip_variable(short mid, short cid, short typ, float var) {
  
  /*
     Configure one variable of type [typ] to the value [var] for chip [cid] 
     of module [mid].  This allows variables such as the strobe delay to 
     be set automatically on a chip by chip basis.

     cut and paste from st_configure_variable, PWP 25.10.00
  */

  float v, vv;
  short i,j,k;
  short mask = (short)0xffff;

  if ((mid<0)||(mid >= SCT_MAX_MODULES)){
    printf("st_configure_chip_variable: module %d outside limit of %d modules\n", mid, SCT_MAX_MODULES);
    return;
  }

  if ((cid<0)||(cid >= 12)){
    printf("st_configure_chip_variable: chip %d outside range of 0 to 11", cid);
    return;
  }

  i = mid; /*    much easier than rewriting the     */
  j = cid; /* contents of the main switch construct */

  printf("st_configure_chip_variable: module %hd chip %hd typ %hd value %3.2f\n", i, j, typ, var);


  /* do not update anything unless module is present and active */

  if((sct[i].present) && (sct[i].active)){


    if(typ==ST_ACTIVE){

     /* ST_ACTIVE may be changed whether chip is active or not */
     sct[i].det.chips[j].active = (short)var;

    }else if(sct[i].det.chips[j].active){

     /* update other variables only if chip is already active */

     switch(typ) {

      case ST_VTHR:
        sct[i].det.chips[j].vthr = var;
        break;

      case ST_VCAL:
        sct[i].det.chips[j].vcal = var;
        break;

      case ST_STROBE_DELAY:
        sct[i].det.chips[j].delay = var;
        break;

      case ST_PREAMP:
        sct[i].det.chips[j].febias = var;
        break;

      case ST_SHAPER:
        sct[i].det.chips[j].feshaper = var;
        break;

      case ST_TRIM:
        if (sct[i].det.chipset==ABCDT){
          for (k=0; k<128; k++) sct[i].det.chips[j].trim[k] = (short)var;
          sct[i].det.updatetrims = 1;
        }
        break;

      case ST_MASK:
        /* Revised PWP 01.11.00
         * This is for global, simple mask register settings.
         * To mask an individual channel, use st_mask_chan in st_config.c
         */

        if(var==0.0) mask = (short)0x0000;
        if(var==1.0) mask = (short)0x5555;
        if(var==2.0) mask = (short)0xaaaa;
        if(var==3.0) mask = (short)0xffff;

        for (k=0; k<8; k++) sct[i].det.chips[j].mask[k] = mask;
        break;

      case ST_ROLE: /* PWP 08.12.00 */
        sct[i].det.chips[j].role = (short)var;
        break;

      case ST_CAL_MODE:
        sct[i].det.chips[j].cal_mode = (short)var;
        break;

      case ST_COMPRESSION:
        sct[i].det.chips[j].compression = (short)var;
        break;

      case ST_TRIM_RANGE:
        sct[i].det.chips[j].trim_range = (short)var;
        break;

      case ST_EDGE_DETECT:
        sct[i].det.chips[j].edge_detect = (short)var;
        break;

      case ST_SEND_MASK:
        sct[i].det.chips[j].mask_readout = (short)var;
        break;

      case ST_ACCUMULATE:
        sct[i].det.chips[j].accumulate = (short)var;
        break;


      case ST_QTHR:
        
        switch(sct[i].det.chips[j].rc_function){
          case 1: /* polynomial */
            v = sct[i].det.chips[j].rc_par[0] + 
                var * (sct[i].det.chips[j].rc_par[1] + 
                (var * sct[i].det.chips[j].rc_par[2]) );
            break; 
          case 2: /* grillo */
            if((var==0)|| (sct[i].det.chips[j].rc_par[1]==0))
              vv = 0;
            else
              vv = sct[i].det.chips[j].rc_par[2] / 
                  (sct[i].det.chips[j].rc_par[1] *var);
            v  = sct[i].det.chips[j].rc_par[0] + 
                 sct[i].det.chips[j].rc_par[2] / (float) sqrt(1+(vv*vv)); 
            break; 
          case 3: /* exp */
            if(sct[i].det.chips[j].rc_par[1]==0)
              v = 0;
            else
              v = sct[i].det.chips[j].rc_par[2] + 
                  sct[i].det.chips[j].rc_par[0]/(1+ (float)exp(-var/sct[i].det.chips[j].rc_par[1]));
            break;  
          case 4: /* linear fit */
            v = sct[i].det.chips[j].rc_par[0] + 
                (var * sct[i].det.chips[j].rc_par[1]);
            break; 
          default:
            v = 0;
            break;
        }
        vv = ((int)((v+1.25)/(float)2.5)) * (float)2.5;   /* 2.5mV / DAC bit */
        sct[i].det.chips[j].vthr = vv;
        break;

      case ST_QCAL:
        if(sct[i].det.chipset == ABC_CAFE){
          if (var > 9.9609375) v = 9.9609375;   /* 0.0390625fC / DAC bit */
          else v = ((int)((var+0.01953125)/(float)0.0390625)) * (float)0.625;

        } else {              /* ABCD */
          if (var > 15.9375) v = 15.9375;      /* 0.0625fC / DAC bit */
          else v = ((int)((var+0.0312)/(float)0.0625)) * (float)0.625;
        } 
        sct[i].det.chips[j].vcal = v;
        break;


      default:      /* not a chip variable */
        break;

     } /* end switch */
    } /* end if(chip active) */
  } /* end if(module present and active) */

}

  
int st_execute_configs(short dump, short full_config) {

  int j,k;
  short opto_tx_type, opto_tx_id, opto_tx_ch;
  short slog_ch, slog_id;
  unsigned short scl, scc, sex, sin;
  unsigned short cb[1000];
  unsigned int cb_size;
  unsigned short cbr[32000];
  unsigned int cbr_size = 32000; 
  unsigned short cb_mask = 0x3fff;
  unsigned short slog_runstate;

  /* Initialise multiple modules, one at a time, masking control
     streams at the BiLED board and Slog.   To do: make more efficient
     by downloading multiple module configurations in parallel.
     Assuming a single SLOG, 6 (or 12) ELECTRO modules may be 
     initialised at once.  In the OPTO case,  only one command 
     stream can be sent to each BiLED board - hence the optimum
     would be to loop, configuring two modules at a time. */

  for (j=0; j<SLOG_NUMBER; j++){
    slog_disable_external((short)j);
  }

  for (j=0; j<SCT_MAX_MODULES; j++) {
    if (sct[j].present) {

      if (dump>4) st_show_module(j);

      if (sct[j].det.select){
        opto_tx_type = sct[j].r_opto_tx_type;
        opto_tx_id   = sct[j].r_opto_tx_id;
        opto_tx_ch   = sct[j].r_opto_tx_ch;
        slog_ch = sct[j].r_slog_ch;
        slog_id = sct[j].r_slog_id;
      } else {
        opto_tx_type = sct[j].opto_tx_type;
        opto_tx_id   = sct[j].opto_tx_id;
        opto_tx_ch   = sct[j].opto_tx_ch;
        slog_ch = sct[j].slog_ch;
        slog_id = sct[j].slog_id;
      }

      /* Configure enables such that only one module receives the commands */

      /* disable all OPTO and ELECTRO config streams */

      for (k=0; k<BILED_NUMBER; k++) biled_channel_disable(0,k);
      for (k=0; k<SLOG_NUMBER;  k++) slog_disable_internal((short)k);

      if((opto_tx_type==0)&&(opto_tx_id>=0)&&(opto_tx_id<BILED_NUMBER)){
        
        /* OPTO configuration using BiLED :
         *
         * disable all LVDS outputs; sysmap must correctly associate slog
         * logical bitstream with correct ECL output.
         * enable both ECL channels (bits 12,13 of enable register)
         * and both NIM channels (bits 14,15).  Only one channel of
         * one BiLED board is enabled at any one time... PWP 19.07.99
         */

        biled_channel_enable (opto_tx_ch, opto_tx_id); /* enable correct BiLED ch */
        slog_set_internal(slog_id,0x7000);             /* enable both ECL channels
                                                          and NIM channel 12*/

      } else {                                     /* ELECTRO configuration */
        /*
         * enable the correct LVDS output, and disable all others and all ECL outputs
         * set SLOG internal enables to route command to NIM channel 12 only
         * - applies to SLOGs with current firmware.  PWP 03.10.00
         */

        slog_set_internal(slog_id,(unsigned short)((0x0001<<(unsigned short)slog_ch) | 0x4000));
      } 
 
      if (dump){
        printf("st_execute_configs: mod %d sel %d using SLOG id %d ch %d OPTO TX type %d id %d ch %d\n", 
          j, sct[j].det.select, slog_id, slog_ch, opto_tx_type, opto_tx_id, opto_tx_ch);   
      }

      /* Build, load and execute CBs */

      /* PWP change 10.01.99
       * wish to use NIM output from ch 13 as pretrigger signal 
       * when using SLOG control blocks - now route command stream
       * to ch 12 only */

      cb_mask = (0x0001 << slog_ch) | 0x1000; /* was 0x3000; */

      /* The "| 0x3000" feeds data to channels 12,13 (NIM).
       * Note that although the NIM channels are now enabled by 
       * bits 14,15 their data still comes from bits 12,13 */

      abc_configure(&sct[j].det,cb,&cb_size,full_config);
      cb_clear(cbr,cbr_size);
      cb_rotate(cb,cb_size,cbr,&cbr_size,cb_mask);

      if (SLOG_NUMBER) {
        if (dump){
          slog_get_clock(slog_id,&scl);
          slog_get_clk_cntl(slog_id,&scc);
          slog_get_external(slog_id,&sex);
          slog_get_internal(slog_id,&sin);
          printf("st_execute_configs: SLOG %d enables were clock %4x external %4x internal %4x clk_cntl %4x\n",
            slog_id, scl, sex, sin, scc);
        }

        if((sct[j].det.chipset==ABCDT)&&(sct[j].det.updatetrims==1)){

        /* Load and execute trimdac control blocks while using execution time to
           build next control block. */

          for(k=0; k<12; k++){                       /*   loop over chips  */
            if(sct[j].det.chips[k].role!=MISSING){   /* if chip is present */

              if(dump) printf("st_execute_configs: initialising trimdacs for module %d chip %dx\n",j,k);

              slog_load(slog_id, cbr, cbr_size);     /*  load and execute  */
              slog_set_cycles(slog_id,slog_ncycles); /*       previous     */
              slog_start(slog_id);                   /*    control block   */

              abc_trims(&sct[j].det,k, cb,&cb_size); /*  generate next CB  */ 
              cb_clear(cbr,cbr_size);
              cb_rotate(cb,cb_size,cbr,&cbr_size,cb_mask);

              slog_get_runstate(slog_id, &slog_runstate);
              while (slog_runstate&0x01) slog_get_runstate(slog_id, &slog_runstate);
            }
          } 
          sct[j].det.updatetrims = 0;
        
        } else {
          if(dump) printf("st_execute_configs: module: %d chipset: %d  update_flag: %d: skipping TRIMS\n",
                     j, sct[j].det.chipset, sct[j].det.updatetrims);
        }
        slog_load(slog_id, cbr, cbr_size);           /*     load and       */
        slog_set_cycles(slog_id,slog_ncycles);
        slog_start(slog_id);                         /*  execute final CB  */
        slog_get_runstate(slog_id, &slog_runstate);
        while (slog_runstate&0x01)  slog_get_runstate(slog_id, &slog_runstate);

      } else {
        /* Try CLOAC instead */
        printf("st_execute_configs: No SLOGs available: trying CLOAC broadcast slowcom\n"); 
        if(CLOAC_NUMBER>0) cc_slowcommand(cb, cb_size);
      }
    }
  }  

  if(BILED_NUMBER) { /* Re-enable all used Biled channels */
    for (j=0; j<SCT_MAX_MODULES; j++) {
      if(sct[j].present) {
        if (sct[j].det.select){
          opto_tx_type = sct[j].r_opto_tx_type;
          opto_tx_id   = sct[j].r_opto_tx_id;
          opto_tx_ch   = sct[j].r_opto_tx_ch;
        } else {
          opto_tx_type = sct[j].opto_tx_type;
          opto_tx_id   = sct[j].opto_tx_id;
          opto_tx_ch   = sct[j].opto_tx_ch;
        }
        if((opto_tx_type==0)&&(opto_tx_id>=0)&&(opto_tx_id<BILED_NUMBER)){
          biled_channel_enable (opto_tx_ch, opto_tx_id);
        }
      }
    }
  }

  /* Re-enable all connected Slog channels to pass transparently external commands and triggers */ 
  for (j=0; j<SLOG_NUMBER; j++) {
    slog_disable_internal((short)j);
    if(CLOAC_NUMBER>0) slog_set_external((short)j,(unsigned short)(slog_com_enables[j]&0x7fff));
  }
    
  old_ttype = -1; /* force reload of slog trigger if appropriate */

  return 0;
}


int st_vburst(int tmax, short ttype) {

  /* 
     Single Cloac VBURST of [tmax] type [ttype] triggers.
     Mustards in Decoding and Sample-all
     Mustards tested until [tmax] events reached,
     subject to timeouts.
  */

  printf("st_vburst:  NOT YET IMPLEMENTED\n");
  return -99;
}

int st_sburst(int interval, short ttype, int *count) {

  /*
     Mustard free run for [interval] milliseconds
     Cloac trigger type [ttype], issuing internal triggers
        if enabled
     Return number [count] of events read
  */

  printf("st_sburst:  NOT YET IMPLEMENTED\n");
  return -99;
}

int st_iburst(int irequest, short ttype, int *icount) {

  /*
     Mustard free run accumulating at least [irequest] events
        or time-outs
     Cloac trigger type [ttype], issuing internal triggers
        if enabled
     Return number [icount] of valid events read
  */

  printf("st_iburst:  NOT YET IMPLEMENTED\n");
  return -99;
}

int st_tburst(int tmax, short ttype,
              short cfgfreq, short srfreq, short bcfreq, short dump) {
  /* 
     Burst of [tmax] individual Cloac VME triggers of type [ttype]:
       0 = isolated L1A for noise studies; 3 = Calstrobe+L1A; 
       10 = simulated empty event & 11 = simulated event with single hit, for loop-back use.
       20 = pulse input register (for digital tests)
     Mustards placed into Decode and Sample-All modes
     Mustard events tested and cleared,
       if((dump)&&(error)) event downloaded and dumped to screen
       if(musf)    all events written to ascii file
  */
    
  short mid = 0;    /* ### To do: loop over all active Mustards */
  int i,j,m,t,strip;

  short run = 1;
  short config_frequency, next_config;
  short soft_reset_frequency, next_soft_reset;
  short bc_reset_frequency, next_bc_reset;

  int   evcount =0; 
  short evlength, everror, status;
  short prev_everror = 0;
  int   total_timeout_errors = 0;

  short rcode;

  short *hpointer;
  short hlength;

  short dcnt;
  short mus_dump = 0;

  if (dump <= 0) dcnt = 0;
  else if (dump > 24) dcnt = 24;
  else dcnt = dump;

/*  if (musf!=NULL) mus_dump=1; */

  for (m=0; m<SCT_MAX_MODULES; m++) {   /* loop over all modules */
    if (sct[m].present) {
      /* Clearing internal histograms here for security - not strictly necessary */
      mus_clear_histogram(sct[m].mus_id,sct[m].mus_ch0);
      mus_clear_histogram(sct[m].mus_id,sct[m].mus_ch1);
    }
  }

  for (mid=0; mid<MUS_NUMBER; mid++) {  /* loop over all MuSTARDs */
    mus_timeout_errors[mid] = 0;
    mus_control_errors[mid] = 0;

     if ((rcode = mus_set_readout_modes(mid,0,1)) != 0) { /* SAMPLE ALL */
      printf("tburst: ERROR for Mustard %d: mus_set_readout_modes returns %d\n", mid, rcode);
      return -2;
    }
   
    if ((rcode = mus_set_readout_cntl(mid,1,1)) != 0) {
      printf("tburst: ERROR for Mustard %d: mus_set_readout_cntl returns %d\n", mid, rcode);
      return -3;
    }
  }

  st_soft_reset();  /* Always do a global synchronous soft reset at start of burst */

  if (cfgfreq > 0) config_frequency = cfgfreq;
  else             config_frequency = -1;
  next_config = config_frequency;

  if (srfreq > 0) soft_reset_frequency = srfreq;
  else            soft_reset_frequency = -1;

  next_soft_reset = soft_reset_frequency;

  if (bcfreq > 0) bc_reset_frequency = bcfreq;
  else            bc_reset_frequency = -1;

  next_bc_reset = bc_reset_frequency;

  if(outf != NULL) stout_sob(ST_TBURST,0,tmax,ttype,cfgfreq,srfreq,bcfreq);

  st_sleep(10);   /* Wait 10ms to ensure Mustard ready to receive first event */

  t = 0;
  while (run) {     /* Enter event loop */
	  next_config--;
    next_soft_reset--;
    next_bc_reset--;
    if (!next_config) {
      st_execute_configs(0,1);
      next_config = config_frequency;
      st_sleep(1);
      if (dump) printf("tburst: done EXECUTE CONFIGS\n");
    }

    if (!next_soft_reset) {
      st_soft_reset();
      next_soft_reset = soft_reset_frequency;
      st_sleep(1);
      if (dump) printf("tburst: done SOFT RESET\n");
    }

    if (!next_bc_reset) {
      st_bc_reset();
      next_bc_reset = bc_reset_frequency;
      st_sleep(1);
      if (dump) printf("tburst: done BC RESET\n");
    }

    if (dump > 7) printf("tburst: sending trigger %d  of %d  type %d\n", t, tmax, ttype);
	
    st_single(ttype);

    for (mid=0; mid<MUS_NUMBER; mid++) {  /* loop over all MuSTARDs */
      if(mus_used_streams[mid]>0){
        everror = prev_everror;

        evpointer = evdata;
        rcode = mus_test_event(mid,&evlength,&evpointer,&evcount,&everror,mus_dump);

        /*
        if (everror != prev_everror) {
          prev_everror = everror;
          if (dump) printf("tburst: DECODING ERROR for Mustard %d rcode %d: num stream decoding errors changed %d\n", mid,rcode,everror);
          mus_dump_event(mid,evdata);
        }
        */

        prev_everror = everror;

     //   if(musf!=NULL){
     //     /* dump event to ascii file */
     //     mus_dump_event_to_file(musf,evdata);
     //   }

        if (rcode != 0) {
          if (rcode == -5) {
            if (dump > 5) printf("tburst: CHIP ERROR or PROTOCOL VIOLATION detected by Mustard %d\n", mid);
            /* Take diagnostic action 
            mus_dump_event(mid,evdata); */
          } else if (rcode == -4) {
            if (dump > 5) printf("tburst: TIMEOUT for Mustard %d\n", mid);
            mus_timeout_errors[mid]++;
            total_timeout_errors++;
            evlength = 0;  
            evcount = 0;
            everror = 0;
            /* next_soft_reset = 0;  not yet */
          } else {
            if (dump > 5) printf("tburst: CONTROL ERROR for Mustard %d: mus_test_event returns %d\n", mid, rcode);
            mus_control_errors[mid]++;
            /* next_soft_reset = 0;  not yet */
          }
          mus_read_status(mid,&status);
          mus_print_status(status);
        }
        /*
         * Do not update error or event 
         * counters after TIMEOUT
         */
        if(rcode != -4) {
          mus_event_count[mid] = evcount;
          mus_error_count[mid] = everror;
        }
      } /* end if(mus_used_streams>0) */ 
    }
    t++;
    if (t >= tmax) run = 0;
    if (total_timeout_errors > burst_max_timeouts) run = 0;
  }

  for (mid=0; mid<MUS_NUMBER; mid++) mus_set_readout_cntl(mid,0,0);  


  st_trigger_count = t;


  for (m=0; m<SCT_MAX_MODULES; m++) {
    if (sct[m].present) {
      if (sct[m].link0) {
     
        printf("*****************************************************\n");
        printf("* Doing normal read of chips 0-5 through link 0     *\n");
        printf("*****************************************************\n");

        /* Do normal readout of first 6 chips */
        hpointer = sct_histogram[m][0];
        mus_read_histogram(sct[m].mus_id, 0, sct[m].mus_ch0,&hlength,&hpointer); 

        if (dcnt) {
          printf("tburst: reading hists for module %d: mustard %d stream %d  length %d\n",
            m, sct[m].mus_id, sct[m].mus_ch0,hlength);
          for (i=0;i<dcnt;i++){
            strip=i*32;
            printf("%2d 0 %3d *",m,strip);
            for (j=0;j<32;j++) printf(" %d",sct_histogram[m][0][strip+j]);
            printf("\n");
          }
        }

        /* Now check if readout redundancy in use and if so read second 6 chips  GFM 990723*/
        if ((sct[m].link0 == 2) && (sct[m].link1 == 0)) {

          printf("*****************************************************\n");
          printf("* Doing redundant read of chips 8-13 through link 0 *\n");
          printf("*****************************************************\n");

          hpointer = sct_histogram[m][1];
          mus_read_histogram(sct[m].mus_id, 1, sct[m].mus_ch0, &hlength, &hpointer);

          if (dcnt) {
            printf("tburst: reading hists for module %d: mustard %d stream %d  length %d  ALT CHIPS 8-13\n",
              m, sct[m].mus_id, sct[m].mus_ch0,hlength);
            for (i=0;i<dcnt;i++){
              strip=i*32;
              printf("%2d 0 %3d *",m,strip);
              for (j=0;j<32;j++) printf(" %d",sct_histogram[m][1][strip+j]);
              printf("\n");
            }
          }
        }       
        
        mus_clear_histogram(sct[m].mus_id,sct[m].mus_ch0);
      }
      if (sct[m].link1){

        /* Do normal readout of second 6 chips */
        hpointer = sct_histogram[m][1];
        mus_read_histogram(sct[m].mus_id, 1, sct[m].mus_ch1,&hlength,&hpointer);

        if (dcnt) {
          printf("tburst: reading hists for module %d: mustard %d stream %d  length %d\n",
            m, sct[m].mus_id, sct[m].mus_ch1,hlength); 
          for (i=0;i<dcnt;i++) {
            strip=i*32;
            printf("%2d 1 %3d *",m,strip);
            for (j=0;j<32;j++) printf(" %d",sct_histogram[m][1][strip+j]);
            printf("\n");
          }
        }

        /* Now check if readout redundancy in use and if so read first 6 chips */
        if ((sct[m].link1 == 2) && (sct[m].link0 == 0)) {
          hpointer = sct_histogram[m][0];
          mus_read_histogram(sct[m].mus_id, 0, sct[m].mus_ch1, &hlength, &hpointer);

          if (dcnt) {
            printf("tburst: reading hists for module %d: mustard %d stream %d length %d  ALT CHIPS 0-5\n",
              m, sct[m].mus_id, sct[m].mus_ch1,hlength); 
            for (i=0;i<dcnt;i++) {
              strip=i*32;
              printf("%2d 1 %3d *",m,strip);
              for (j=0;j<32;j++) printf(" %d",sct_histogram[m][0][strip+j]);
              printf("\n");
            }
          }
        }       
        
        mus_clear_histogram(sct[m].mus_id,sct[m].mus_ch1);
      }
    }
  }

  for (mid=0; mid<MUS_NUMBER; mid++) mus_reset(mid); 

  if(outf != NULL) stout_eob();
  burst_number++; 

  if ( evcount+1 < tmax ) 
    return -1;
  else 
    return 0;
}


/* These commands are primarily CLOAC actuated events which we
   may wish to initiate from a higher level application */


int st_run(short ttype, short tfreq) {
  short i;

  if(CLOAC_NUMBER>0){
    if ((ttype == 1)||(ttype == 2)) cc_menu_enable(ttype);
    else return -1;  /* internal repeating fixed or random frequency triggers */

    cc_menu_trfreq(tfreq);
  } else {
    st_single(0);  /* l1A only */

    for (i=0; i<SLOG_NUMBER; i++){
      slog_set_cycles(i,0xffff);  /* bit 15 set - infinite cycles for SLOG B */
      slog_set_internal(i,slog_com_enables[i]);
      slog_start(i);
    }
  }
  return 0;
}

int st_stop() {
  short i;

  if(CLOAC_NUMBER>0){
    cc_menu_disable(1);
    cc_menu_disable(2);
  } else {
    for (i=0; i<SLOG_NUMBER; i++){
      slog_stop(i);
    }
  }

  return 0;
}

int st_single(short ttype) {

  /* Set flag to load trigger to SLOG if new trigger type
   * is different from the old one.  This is much neater,
   * and hopefully more robust, than the old way of doing this.
   * PWP 03.10.01
   */
  unsigned short r,s;

  if(old_ttype!=ttype){
    old_ttype=ttype;
    slog_do_loadtrig = 1;
  }

  switch(ttype){   /* Note: trigger types 0-6 follow CLOAC command definitions */
   
    case 0: {  /* Single L1A */
               if(CLOAC_NUMBER>0){
                 cc_load_command(0);
                 cc_load_command(VSINGLETRIG);
                 cc_load_command(0);
               } else {
                 st_slog_single(0,0,1,com_delay,0,0);
               }
               break;
            }
    case 1: {  /* Single soft reset */
               if(CLOAC_NUMBER>0){
                 cc_load_command(0);
                 cc_load_command(VSRST);
                 cc_load_command(0);
               } else {
                 st_slog_single(2,0,0,com_delay,0,0);
               }
               break;
            }
    case 2: {  /* Single bc reset */
               if(CLOAC_NUMBER>0){
                 cc_load_command(0);
                 cc_load_command(VBCRST);
                 cc_load_command(0);
               } else {
                 st_slog_single(3,0,0,com_delay,0,0);
               }
               break;
            }
    case 3: {  /* Single cal strobe followed by L1A */
               if(CLOAC_NUMBER>0){
                 cc_load_command(0);
                 cc_load_command(VCAL);
                 cc_load_command(0);
               } else {
                 st_slog_single(5,0,1,trig_delay,0,0);
               }
               break;
            }
    case 5: {  /* Single over-all reset */
               if(CLOAC_NUMBER>0){
                 cc_load_command(0);
                 cc_load_command(VOARST);
                 cc_load_command(0);
               }
               break;
            }
    case 10:{  /* Single simulated empty ABC/D event */
               st_echo_event(-1);
               break;
            }
    case 11:{  /* Single simulated ABC/D event with one hit in strip 0 */
               st_echo_event(0);
               break;
            }
    case 12:{  /* Single simulated ABC/D event with one hit in a random strip */
               st_echo_event(rand());
               break;
            }
    case 13:{  /* Single simulated ABC/D event with one hit in an incrementing strip */
               st_echo_event(st_trigger_count);
               break;
            }
    case 20:{  /* Single pulse_input_register followed by L1A */
               if(CLOAC_NUMBER>0){
                 cc_load_pulse(); /* Load pulse_input_reg command and set delay */
                 cc_load_delay((unsigned short)(pulse_delay-CLOAC_CAL_OFFSET));
                 cc_load_command(0);
                 cc_load_command(VCAL);
                 cc_load_command(0);
                 cc_load_cal();   /* Restore default cal command and set delay */
                 cc_load_delay((unsigned short)(trig_delay-CLOAC_CAL_OFFSET));
               } else {
                 st_slog_single(4,0,1,pulse_delay,0,0);
               }
               break;
            }
    case 25:{  /* Single external trigger L1A and internal busy ON */
               /*         (Contributed by Pavel Reznicek)          */
               if(CLOAC_NUMBER>0){
                 cc_load_command(0);
                 cc_load_command(VBURST);
                 cc_load_string(1);
                 cc_load_command(VBURST+LDVBURST);
                 cc_read_enable(&r);
                 cc_load_enable((unsigned short)(r|(1<<ENEXTTRIG)|(1<<ENINTBUSY)));
                 do cc_read_state(&s); while(!(s&(1<<7)));
                 cc_load_enable(r);
                 cc_load_command(0);
               } else {
                 /* do nothing - can't get external triggers with no CLOAC */
               }
               break;
            }
    case 30:{ /* single L1A using SLOG */
              st_slog_single(0,0,1,com_delay,0,0);
              break;
            }
    case 31:{ /* single soft reset and L1A using SLOG */
              st_slog_single(2,0,1,com_delay,0,0);
              break;
            }
    case 32:{ /* single BC   reset and L1A using SLOG */
              st_slog_single(3,0,1,com_delay,0,0);
              break;
            }
    case 33:{ /* single soft reset, calstrobe and L1A using SLOG */
              st_slog_single(2,5,1,com_delay,trig_delay,0);
              break;
            }
    case 34:{ /* single BC   reset, calstrobe and L1A using SLOG */
              st_slog_single(3,5,1,com_delay,trig_delay,0);
              break;
            }
    case 35:{ /* single soft reset, pulse input register and L1A using SLOG */
              st_slog_single(2,4,1,com_delay,pulse_delay,0);
              break;
            }
    case 36:{ /* single BC   reset, pulse input register and L1A using SLOG */
              st_slog_single(3,4,1,com_delay,pulse_delay,0);
              break;
            }
    case 37:{ /* single cal and L1A using SLOG */
              st_slog_single(5,0,1,trig_delay,0,0);
              break;
            }
    case 38:{ /* single pulse and L1A using SLOG */
              st_slog_single(4,0,1,pulse_delay,0,0);
              break;
            }
    case 39:{ /* user defined SLOG sequence */
              st_slog_user("slog.txt",0); /* gives 1 cycle PWP 20.08.02 */
              break;
            }
    case 40:{ /* multiple L1A using SLOG */
              st_slog_single(0,0,8,trig_delay,0,0); /* changed from com_delay. PWP 16.03.00 */
              break;
            }
    case 41:{ /* single soft reset and multiple L1A using SLOG */
              st_slog_single(2,0,8,trig_delay,0,0); /* changed from com_delay. PWP 16.03.00 */
              break;
            }
    case 42:{ /* single BC   reset and multiple L1A using SLOG */
              st_slog_single(3,0,8,trig_delay,0,0);
              break;
            }
    case 43:{ /* single soft reset, calstrobe and multiple L1A using SLOG */
              st_slog_single(2,5,8,com_delay,trig_delay,0);
              break;
            }
    case 44:{ /* single BC   reset, calstrobe and multiple L1A using SLOG */
              st_slog_single(3,5,8,com_delay,trig_delay,0);
              break;
            }
    case 45:{ /* single soft reset, calstrobe and multiple L1A using SLOG */
              st_slog_single(2,4,8,com_delay,trig_delay,0);
              break;
            }
    case 46:{ /* single BC   reset, pulse input register and L1A using SLOG */
              st_slog_single(3,4,8,com_delay,pulse_delay,0);
              break;
            }
    case 47:{ /* single cal and multiple L1A using SLOG */
              st_slog_single(5,0,8,trig_delay,0,0);
              break;
            }
    case 48:{ /* single pulse and multiple L1A using SLOG */
              st_slog_single(4,0,8,pulse_delay,0,0);
              break;
            }
    case 50:{ /* soft reset, pulse input and multiple L1A using SLOG */
              st_slog_single(2,4,n_l1a,com_delay,pulse_delay,l1a_spacing);
              break;
            }
    case 51:{ /* soft reset, cal and multiple L1A using SLOG */
              st_slog_single(2,5,n_l1a,com_delay,trig_delay,l1a_spacing);
              break;
            }
    case 52:{ /* soft reset and multiple L1A using SLOG */
              st_slog_single(2,0,n_l1a,com_delay,0,l1a_spacing);
              break;
            }
    case 53:{ /* multiple L1A using SLOG */
              st_slog_single(0,0,n_l1a,trig_delay,0,l1a_spacing); /* no SR */
              break;
            }
    case 60:{ /* double cal and single L1A using SLOG */
              st_slog_single(5,5,1,trig_delay,trig_delay,0);
              break;
            }
    case 98: { /* Send enable data  */
              st_slog_single(7,0,1,0,0,0);
              break;
            }
    case 99: { /* Send enable data  */
              st_slog_single(6,0,1,0,0,0);
              break;
            }
    default:   return -1;
  }

  return 0;
}

void st_echo_event(int seed){
  /* Use CLOAC slowcommand to send a single event with a single hit
     for use with a loop-back test.  If seed < 0, send valid but empty event, 
     otherwise derive l1, bc, chip and address from seed */

  short l1a,bc,chip,strip,bins;
  unsigned short cb[4];

  if (seed < 0) chip = seed;
  else {
    l1a = seed & 0x000f; 
    bc = seed & 0x00ff;
    chip = (seed % 768) / 128;
    strip = seed & 0x007f;
    bins = 0x7;
  }

  abc_echo_event( l1a, bc, chip, strip, bins, cb);
    
  if(CLOAC_NUMBER>0) cc_slowcommand(cb, 4);   /* Send simulated event data */
}
  
void st_enable_datataking(){
  unsigned short cb[4];
  
  abc_global_enabledata(cb);
  
  if(CLOAC_NUMBER>0) cc_slowcommand(cb, 4);   /* Send EnableDataTaking to all modules */
                             /* Note: this interim solution ignores module config DTM */
}

void st_bc_reset(){
  st_single(2);
}

void st_soft_reset(){
  st_single(1);
}

void st_hard_reset(){
  int j;

  if (sctlv_n_channels>0){
    printf("Sending HARD RESET to all modules!!\n");
    for (j=0; j<SCT_MAX_MODULES; j++) {
      if((sct[j].present)&&(sct[j].lv_ch<sctlv_n_channels())){
        sctlv_set_reset(sct[j].lv_ch, 1); 
      }
    }
    st_sleep(100); /* wait 100mS */
    for (j=0; j<SCT_MAX_MODULES; j++){
      if((sct[j].present)&&(sct[j].lv_ch<sctlv_n_channels())){
        sctlv_set_reset(sct[j].lv_ch, 0);
      }
    }
    st_sleep(100); /* wait 100mS */
  } else printf("st_hard_reset: no SCTLV channels!!!\n");
}

   
void st_sleep(clock_t ms){
  /*
   * st_sleep now calls usleep, a version of this having been
   * provided for use under windows.  Units here are still mS,
   * however where a shorter delay is needed usleep is now 
   * called directly.  PWP 04.07.02
   */

   usleep(ms*1000);
}


/* These functions are getters for external code, eg, root */

short st_give_nmods(short *list){
  short m;
  short count = 0;

  for (m=0; m<SCT_MAX_MODULES; m++) 
    if (sct[m].present) {
       count++;
       list[m] = 1;
    } else {
       list[m] = 0;
    }

  return count;
}


short st_give_modname(short mid, char *name){

  if (mid >= SCT_MAX_MODULES){
    sprintf (name,"invalid");
    return -1;
  }
  if (!sct[mid].present){
    sprintf (name,"missing");
    return -2;
  }
  sprintf(name,"%s", sct[mid].detname);
  return 0;
}

short st_give_devtype(short mid, char *name){

  if (mid >= SCT_MAX_MODULES){
    sprintf (name,"invalid");
    return -1;
  }
  if (!sct[mid].present){
    sprintf (name,"missing");
    return -2;
  }
  sprintf(name,"%s", sct[mid].devtype);
  return 0;
}


short st_give_data(short module, short index, short *hist0, short *hist1){
  int i, start, step;

  if ((index<0)||(index>3)){
    start = 0;
    step = 1;
  } else {
    start = index;
    step = 4;
  }

  if (sct[module].present) {
    /* We copy values rather than pass a pointer for ROOT/CINT compatibility */
    if (sct[module].link0) {
      for (i=start; i<768; i+=step) hist0[i] = sct_histogram[module][0][i];
      if ((sct[module].link0 == 2) && (sct[module].link1 == 0))
        for (i=start; i<768; i+=step) hist1[i] = sct_histogram[module][1][i];
    }
    if (sct[module].link1) {
      for (i=start; i<768; i+=step) hist1[i] = sct_histogram[module][1][i];
      if ((sct[module].link0 == 0) && (sct[module].link1 == 2))
        for (i=start; i<768; i+=step) hist0[i] = sct_histogram[module][0][i];
    }
    return 0;
  } else {
    return -1;
  }
}

short st_give_counters(short mus_id, int *tsent, int *evcnt, short *ercnt, short *tocnt, short *xecnt){

  if (mus_id >= MUS_NUMBER) return -1;

  printf("st_give_counters: mus_id %d  tsent %d  evcnt %d  strm_ercnt %d  timeouts %d  cntrl_ercnt %d\n",
      mus_id, st_trigger_count, mus_event_count[mus_id], mus_error_count[mus_id], 
      mus_timeout_errors[mus_id], mus_control_errors[mus_id]);

  *tsent = st_trigger_count;
  *evcnt = mus_event_count[mus_id]+1;  /* GFM 991019  This now returns 1 even in case Mustard has not yet been read */
  *ercnt = mus_error_count[mus_id];
  *tocnt = mus_timeout_errors[mus_id];
  *xecnt = mus_control_errors[mus_id];

  return 0;
}

int st_rawburst(int tmax, short ttype,
                short cfgfreq, short srfreq, short bcfreq, short dump){
  /* Place MuSTARDs in RAW and SAMPLE ALL modes, 
   * compile histogram of first 768 bits of RAW data
   * - cf. digital 'scope in summing mode... */
    
  short mid = 0;    /* Should now loop over all MuSTARDs - PWP 29.02.00 */
  int i,j,m,t;
  int bit = 0;
  int hlength = 768;
  int old_stream, stream;
  
  short run = 1;

  short config_frequency, next_config;
  short soft_reset_frequency, next_soft_reset;
  short bc_reset_frequency, next_bc_reset;

  int   evcount = 0;
  short everror = 0;
  short evlength, status;
  short prev_everror = 0;
  int   total_timeout_errors = 0;

  short rcode;

  short dbuffer[0x4000]; /* 16k d16 words max event size */
  short *dpointer;
  short dlength;

  

  for (m=0; m<SCT_MAX_MODULES; m++) {   /* loop over all modules */
    if (sct[m].present) {
      /* Clearing internal histograms here for security - not strictly necessary */
      mus_clear_histogram(sct[m].mus_id,sct[m].mus_ch0);
      mus_clear_histogram(sct[m].mus_id,sct[m].mus_ch1);
    }
  }

  for (mid=0; mid<MUS_NUMBER; mid++) {  /* loop over all MuSTARDs */
    mus_timeout_errors[mid] = 0;
    mus_control_errors[mid] = 0;

    if ((rcode = mus_set_readout_modes(mid,1,1)) != 0) { /* RAW, SAMPLE ALL */
      printf("rawburst: ERROR for Mustard %d: mus_set_readout_modes returns %d\n", mid, rcode);
      return -2;
    }
  
    if ((rcode = mus_set_readout_cntl(mid,1,1)) != 0) {
      printf("rawburst: ERROR for Mustard %d: mus_set_readout_cntl returns %d\n", mid, rcode);
      return -3;
    }
  }
  
  /* Assume st_reverse_sysmap has been called hence the relationship 
   * between streams and modules/links is already known PWP 29.02.00 */

  for (m=0; m<SCT_MAX_MODULES; m++){
    if(sct[m].mus_id<MUS_NUMBER){
      for(i=0; i<hlength; i++){    /* clear software histogram */
        sct_histogram[m][0][i] = 0;
        sct_histogram[m][1][i] = 0;
      }
    }
  }

  
  if (cfgfreq > 0) config_frequency = cfgfreq;
  else             config_frequency = -1;
  next_config = config_frequency;

  if (srfreq > 0)  soft_reset_frequency = srfreq;
  else             soft_reset_frequency = -1;
  next_soft_reset = soft_reset_frequency;

  if (bcfreq > 0) bc_reset_frequency = bcfreq;
  else             bc_reset_frequency = -1;
  next_bc_reset = bc_reset_frequency;

  if(outf != NULL) stout_sob(ST_RAWBURST,0,tmax,ttype,cfgfreq,srfreq,bcfreq);

  st_sleep(10);   /* Wait 10ms to ensure Mustard ready to receive first event */

  t = 0;
  while (run) {     /* Enter event loop */

    next_config--;
    next_soft_reset--;
    next_bc_reset--;
    
    if (!next_config) {
      st_execute_configs(0,1);
      next_config = config_frequency;
      st_sleep(1);
      if (dump) printf("rawburst: done EXECUTE CONFIGS\n");
    }
    
    if (!next_soft_reset) {
      st_soft_reset();
      next_soft_reset = soft_reset_frequency;
      st_sleep(1);
      if (dump) printf("rawburst: done SOFT RESET\n");
    }
    
    if (!next_bc_reset) {
      st_bc_reset();
      next_bc_reset = bc_reset_frequency;
      st_sleep(1);
      if (dump) printf("rawburst: done BC RESET\n");
    }
    
    if (dump > 7) printf("rawburst: sending trigger %d  of %d  type %d\n", t, tmax, ttype);
    
    st_single(ttype);

  /*  st_sleep(1); */
    
    for (mid=0; mid<MUS_NUMBER; mid++) {  /* loop over all MuSTARDs */
      if(mus_used_streams[mid]>0){
        dpointer = dbuffer;
        rcode = mus_read_event(mid, &dlength, &dpointer);
    
        if (rcode != 0) {
          if (rcode == -4) {
            printf("rawburst: TIMEOUT for Mustard %d\n", mid);
            mus_timeout_errors[mid]++;
            total_timeout_errors++;
            evlength = 0;  
            evcount = 0;
            everror = 0;
          } else {
            printf("rawburst: CONTROL ERROR for Mustard %d: mus_read_event returns %d\n", mid, rcode);
            mus_control_errors[mid]++;
          }
          mus_read_status(mid,&status);
          mus_print_status(status);
        } else {
          evlength = (dbuffer[0] & 0x3fff);
          everror  = (dbuffer[1] & 0xffff);   /* limited relevance in RAW mode */
          evcount  = ( ((dbuffer[3]<<16)&0xffff0000) + (dbuffer[2] & 0xffff) );

          /* Now histogram the raw data */

          old_stream = -1;
          for (i=4; i<(evlength); i++){       /* loop over words */
            stream = (dbuffer[i]&0x1e00)>>9;

            if (old_stream != stream){        /* reset bit count */
              bit = 0;
              old_stream = stream;
            }
            for (j=0; j<8; j++){              /* loop over bit */
              if ((dbuffer[i]&(0x0080>>j)) && (bit<hlength)){
                sct_histogram[stream_2_mod[mid][stream]][stream_2_link[mid][stream]][bit] ++;
              }
              bit++;
            }
          }
        }
        /*
         * Do not update error or event 
         * counters after TIMEOUT
         */
        if(rcode != -4) {
          mus_event_count[mid] = evcount;
          mus_error_count[mid] = everror;
        }
      }
    }

    t++;
    if (t >= tmax) run = 0;
    if (total_timeout_errors > burst_max_timeouts) run = 0;
  }

  st_trigger_count = t;

  for (mid=0; mid<MUS_NUMBER; mid++) {  /* loop over all MuSTARDs */
    mus_set_readout_cntl(mid,0,0);  
    mus_reset(mid);  
  }

  for (m=0; m<SCT_MAX_MODULES; m++) {
    if (sct[m].present) {
      if (sct[m].link0) {
        if (dump) {
          printf("rawburst: histogram for module %d: mustard %d stream %d  length %d\n",
            m, sct[m].mus_id, sct[m].mus_ch0,hlength);
          for (i=0;i<8;i++){
            for (j=(i*10);j<((i+1)*10);j++){
              printf(" %4d",sct_histogram[m][0][j]);
            }
            printf("\n");
          }
        }
      }
      if (sct[m].link1){
        if (dump) {
          printf("rawburst: histogram for module %d: mustard %d stream %d  length %d\n",
            m, sct[m].mus_id, sct[m].mus_ch1,hlength);      
          for (i=0;i<8;i++){
            for (j=(i*10);j<((i+1)*10);j++){
              printf(" %4d",sct_histogram[m][0][j]);
            }
            printf("\n");
          }
        }
      }
    }
  }

  if(outf != NULL) stout_eob();
  burst_number++; 

  if ( evcount+1 < tmax ) 
    return -1;
  else 
    return 0;
}

int st_compare_sendid(short n_ev, short dump){
  /* 
   Calculate the data expected from each module in send_id mode
	 subtract the expected result from a RAW data histogram
	 Count + report errors!!!
  */
	
  int i,m;
  int word, bit;
  int db_size;
  int n_errs;
  int t_errs = 0;        /* May have _lots_ of bit errors!! */
  short link;

  unsigned short db[24]; /* enough bits for 12 chips in send_ID mode */

  for (m=0; m<SCT_MAX_MODULES; m++){  /* loop over modules */
    if(sct[m].present){
      for (link=0;link<2; link++){    /* loop over links */
       
        if ( ((link==0)&&(sct[m].link0>0)) || ((link==1)&&(sct[m].link1>0)) ){
                                  /* modify tests from "==1" to ">0" for redundant readout  GFM 990723 */
          n_errs = 0;
          abc_sendid(&sct[m].det, link, db, &db_size);
          
          for(i=0; i<db_size; i++){     /* loop over expected data */
            word = i / 16;
            bit  = i % 16;

            if (0x8000&(db[word]<<bit)) { /* 1 expected */              
              sct_histogram[m][link][i] -= n_ev;
            }
          }
          for (i=6; i<18; i++) {     /* mask counter bits */
            sct_histogram[m][link][i] = 0;
          }
          for(i=0; i<db_size; i++){     /* loop over expected data */
            n_errs += abs (sct_histogram[m][link][i]);   /* tally errors */
          }
          sct_errs[m][link] = n_errs;
          t_errs += n_errs;
          printf("Module %s link %d bit error tally = %ld\n",sct[m].detname,link,n_errs);
        } /* end if link enabled*/
      }
    }
  }
  /* mus_error_count[0] += (short) t_errs; ### mus_error_count is intended to store
                                            the most recent Mustard error counter value
                                            which is a short: To use it to accumulate
                                            bit errors in this way would require a much
                                            larger type */  

  return 0;
}

int st_scan_definition(int dimension, float start, float stop, float step, short points){

  if((dimension<0)||(dimension>3)){
    printf("st_scan_definition called for invalid dimension %d\n",dimension);
    return -1;
  }

  scan_start[dimension] = start;
  scan_stop[dimension] = stop;
  scan_step[dimension] = step;
  scan_points[dimension] = points;

  return 0;
}

void st_scan_increment(){
  scan_number++;
}

void st_scan_reset(){
  int i;

  for(i=0;i<4;i++){
    scan_pnt[i] = 0;
    scan_typ[i] = 0;
    scan_var[i] = 0;

    scan_start[i]  = -1;
    scan_stop[i]   = -1;
    scan_step[i]   = -1;
    scan_points[i] = -1;
  }
}

int st_optorx_init(){
  int i, status;
  int j=0;

  if(OPTIF_NUMBER==0){
    printf("st_optorx_init: OPTIF_NUMBER = 0!!!\n");
    return -1;
  }

  for(i=0;i<SCT_MAX_MODULES;i++){

    if((sct[i].opto_rx_id>=0)&&(sct[i].opto_rx_id<OPTIF_NUMBER)){ /* range check */

      printf("st_optorx_init: module %d (%s) OptIF id %d RX ch %d thr => %d RX ch %d thr => %d\n",
           i, sct[i].detname, sct[i].opto_tx_id, sct[i].opto_tx_ch, sct[i].opto_tx_iset);

      if((sct[i].opto_rx_ch0>=0)&&(sct[i].opto_rx_ch0<12)){ /* ch range check */
        status = optif_set_threshold(sct[i].opto_rx_id,sct[i].opto_rx_ch0,sct[i].opto_rx_thr0);
        j++;

      } else {
        printf("st_optorx_init: module %d (%s) ERROR, OptIF id %d RX ch0 = %d out of range (0-11)\n",
          i, sct[i].detname, sct[i].opto_rx_id, sct[i].opto_rx_ch0);
      }

      if((sct[i].opto_rx_ch1>=0)&&(sct[i].opto_rx_ch1<12)){ /* ch range check */
        status = optif_set_threshold(sct[i].opto_rx_id,sct[i].opto_rx_ch1,sct[i].opto_rx_thr1);
        j++;

      } else {
        printf("st_optorx_init: module %d (%s) ERROR, OptIF id %d RX ch1 = %d out of range (0-11)\n",
          i, sct[i].detname, sct[i].opto_rx_id, sct[i].opto_rx_ch1);
      }
    }       
  }

  if(j) printf("st_optorx_init: %d thresholds have been set\n",j);
  else  printf("st_optorx_init: no OPTO RX modules found\n");

  return 0;
}

int st_optotx_on(){
  int i, status;
  int j=0;
  short mask;

  if((BILED_NUMBER==0)&&(OPTIF_NUMBER==0)){
    printf("st_optotx_on: BILED_NUMBER = 0 and OPTIF_NUMBER = 0!!!\n");
    return -1;
  }

  for(i=0;i<SCT_MAX_MODULES;i++){
    if(sct[i].present){
      if (sct[i].det.select==0){
        /*
         * VcSELs are only turned on for modules where the primary
         * clock and command option is in use.  For clock and command
         * to reach a module which is using the redundancy option,
         * when a "real" module is not present in the sysmap at the
         * neighbour location, a "dummy" module should be added to
         * the sysmap such that the VcSEL is enabled.
         * PWP 10.10.00 
         */

        if(sct[i].opto_tx_type==0){        /* BiLED */

          if((sct[i].opto_tx_id>=0)&&(sct[i].opto_tx_id<BILED_NUMBER)){ /* id range check */

            printf("st_optotx_on: module %d (%s) BiLED id %d ch %d TX DAC => %d bits\n",
              i, sct[i].detname, sct[i].opto_tx_id, sct[i].opto_tx_ch, sct[i].opto_tx_iset);

            if((sct[i].opto_tx_ch>=1)&&(sct[i].opto_tx_ch<5)){          /* ch range check */

              /* Turn ON VcSEL and enable command transmission */
          
              biled_set_dac(sct[i].opto_tx_id,(short)(sct[i].opto_tx_ch+8),sct[i].opto_tx_iset);
              biled_channel_enable(sct[i].opto_tx_ch,sct[i].opto_tx_id);
              j++;

            } else {
              printf("st_optotx_on: module %d (%s) ERROR, BiLED channel %d out of range (1-4)\n",
                i, sct[i].detname, sct[i].opto_tx_ch);
            }
          } else {
              printf("st_optotx_on: module %d (%s) ERROR, BiLED id %d out of range (0-5)\n",
                i, sct[i].detname, sct[i].opto_tx_id);
          }

        } else if(sct[i].opto_tx_type==1){ /* OptIF */

          if((sct[i].opto_tx_id>=0)&&(sct[i].opto_tx_id<OPTIF_NUMBER)){ /* range check */

            printf("st_optotx_on: module %d (%s) OptIF id %d ch %d TX DAC => %d bits\n",
              i, sct[i].detname, sct[i].opto_tx_id, sct[i].opto_tx_ch, sct[i].opto_tx_iset);

            if((sct[i].opto_tx_ch>=0)&&(sct[i].opto_tx_ch<6)){          /* ch range check */

              /* Turn ON VcSEL*/

              status = optif_set_laser(sct[i].opto_tx_id,sct[i].opto_tx_ch,sct[i].opto_tx_iset);

              if( status != 0) {
                printf("st_optotx_on: ERROR, optif_set_laser returns %d\n",status);
              } else {
                j++;  /* VcSEL has been turned on */

                /* Enable command transmission: 0 means enabled */

                status = optif_bpm_read_channelmask(sct[i].opto_tx_id, sct[i].opto_tx_ch, &mask);
                printf("            mask was 0x%x\n",mask);
                mask ^= (0x0001<<(sct[i].opto_tx_ch%4));
                status = optif_bpm_set_channelmask(sct[i].opto_tx_id, sct[i].opto_tx_ch, mask);

                printf("st_opto_tx_on: OPTIF id %d ch id mask is now 0x%x\n",
                  sct[i].opto_tx_id,sct[i].opto_tx_ch,mask);
              }
            } else {
              printf("st_optotx_on: module %d (%s) ERROR, OptIF channel %d out of range (0-5)\n",
                i, sct[i].detname, sct[i].opto_tx_ch);
            }
          } else {
              printf("st_optotx_on: module %d (%s) ERROR, OptIF id %d out of range (0-5)\n",
                i, sct[i].detname, sct[i].opto_tx_id);
          }
        }
      } /* end if(select==0) */
    } /* end if(present) */
  }

  if(j) printf("st_optotx_on: CAUTION!! %d VcSELs have been switched on\n",j);
  else  printf("st_optotx_on: no OPTO modules found\n");

  return 0;
}

int st_optotx_off(){
  int i,j,status;

  if((BILED_NUMBER==0)&&(OPTIF_NUMBER==0)){
    printf("st_optotx_off: BILED_NUMBER = 0 and OPTIF_NUMBER = 0!!!\n");
    return -1;
  }

  for(i=0;i<BILED_NUMBER;i++){
    printf("st_optotx_off: switch off all VCSELS of BiLED board %d\n",i);
    for(j=9;j<13;j++) biled_set_dac((short)i,(short)j,0); /* DACs 9 to 12 */
  }

  for(i=0;i<OPTIF_NUMBER;i++){
    printf("st_optotx_off: turning down all VCSELS of OPTIF board %d\n",i);

    /* Q. Why down and not off?
     * A. "MJG recommends non zero default current setting"
     * N.B. The setting of 0x10 seems to be safely below the lasing threshold
     * PWP 09.10.00
     */

    for(j=0;j<6;j++) {
      status = optif_set_laser(i,(short)j,0x10);
      if( status != 0) {
        printf("st_optotx_off: optif_set_laser returns %d\n",status);
      }
    }
  }

  return 0;
}

int st_seu_slog_load(short seu_invert){

  int j;
  unsigned short cb[2000];
  unsigned int cb_size = 2000;
  unsigned short cbr[32000];      /* Main Buffer */
  unsigned int cbr_size = 32000;

  unsigned short cbr1[32000];     /* First Part */
  unsigned int cbr1_size = 32000;
  unsigned short cbr2[32000];     /* Second Part */
  unsigned int cbr2_size = 32000;

  unsigned short cb_mask = 0x3fff;
  short    sid, slog_ch, slog_id;

  /* Build and load the following sequence to be executed immediately
   * after start of spill during SEU tests:
   *   Write Mask Register (all chips, data = all zeros)
   *   Soft Reset
   *   DTM
   *   Soft Reset
   *   L1A        => mask reference event
   *   Delay to accomodate maximum length of event (guess 480*16 BCOs)
   *   Write configuration register - all command streams executed in 
   *     parallel, all chips serially within the command stream.
   *     (Purpose during SEUTest - set accumulator bit and go to 
   *     SendID mode)
   * PWP 10.04.00
   */

  for (sid=0; sid<SLOG_NUMBER; sid++){

    cb_clear(cb,cb_size);
    abc_seu_setup(cb,&cb_size,seu_invert);
    cb_clear(cbr1,cbr1_size);
    cb_rotate(cb,cb_size,cbr1,&cbr1_size,(unsigned short)(slog_com_enables[sid]&0x5fff));
    
    cb_clear(cbr2,cbr2_size);
    cbr2_size=0;

    for (j=0; j<SCT_MAX_MODULES; j++) {
      if (sct[j].present) {

        if (sct[j].det.select){
          slog_ch = sct[j].r_slog_ch;
          slog_id = sct[j].r_slog_id;
        } else {
          slog_ch = sct[j].slog_ch;
          slog_id = sct[j].slog_id;
        }

        if (slog_id == sid){

           /* Select Appropriate mask bit */
          cb_mask = (0x0001 << slog_ch);
          cb_clear(cb,cb_size);
          abc_configure(&sct[j].det,cb,&cb_size,0);

          cbr_size=32000;
          cb_clear(cbr,cbr_size);
          cb_rotate(cb,cb_size,cbr,&cbr_size,cb_mask);

          if(cbr_size>30000) cbr_size = 30000;
          if(cbr_size>cbr2_size) cbr2_size = cbr_size;
          
          cb_merge(cbr2,cbr,cbr2,cbr2_size);
        }
      }
    }
    cbr_size=32000;
    cb_clear(cbr,cbr_size);
    cbr_size=0;
    cb_cat(cbr,&cbr_size,cbr1,cbr1_size);
    cb_cat(cbr,&cbr_size,cbr2,cbr2_size);

    slog_disable_external(sid);      
    slog_load(sid, cbr, cbr_size);
    slog_set_cycles(sid,0);  /* gives one cycle */
  }

  if(SLOG_NUMBER==0) {
    printf("st_seu_slog_load: SLOG_NUMBER = 0!!\n");
    return -1;
  }

  old_ttype=-1;

  return 0;
}


int st_seu_slog_run(){
  short i;
  unsigned short slog_runstate;

  if (SLOG_NUMBER) {

    for (i=0; i<SLOG_NUMBER; i++){
      slog_set_internal(i,slog_com_enables[i]);
      slog_start(i);
    }

    for (i=0; i<SLOG_NUMBER; i++) {
      slog_get_runstate(i, &slog_runstate);
      while (slog_runstate&0x01) slog_get_runstate(i, &slog_runstate);
       slog_disable_internal(i);
       if(CLOAC_NUMBER>0) slog_set_external(i,(unsigned short)(slog_com_enables[i]&0x5fff));
    }
	
    return 0;
  } else {
    printf("st_seu_slog_run: SLOG_NUMBER = 0!!\n");
    return -1;
  }
}

int st_slog_single(int com1, int com2, int ntrigs, unsigned short delay1, unsigned short delay2, unsigned short delay3){
  short i;
  unsigned short cb[100];
  unsigned int cb_size = 100;
  unsigned short cbr[1600];
  unsigned short cbr1[1600];
  unsigned short cbr2[1600];
  unsigned int cbr_size = 1600; 
  unsigned short slog_runstate;

  if (SLOG_NUMBER) {

    if(slog_do_loadtrig){
      cb_clear(cb,cb_size);
      abc_command(cb,&cb_size,com1,com2,ntrigs,delay1,delay2,delay3);

      /* Generate pretrigger cb */
      cb_clear(cbr1,cbr_size);
      cbr1[32] = 0x2000;

      for (i=0; i<SLOG_NUMBER; i++) {
        /* Generate main CB - masking slog ch 13 (NIM) (enable bit 15) for use as pretrigger */
        cb_clear(cbr2,cbr_size);
        cb_rotate(cb,cb_size,cbr2,&cbr_size,(unsigned short)(slog_com_enables[i]&0x5fff));
        cb_merge(cbr1,cbr2,cbr,cbr_size);
        slog_disable_external(i);      
        slog_load(i, cbr, cbr_size);
      }
      slog_do_loadtrig = 0;
    }

    for (i=0; i<SLOG_NUMBER; i++){
      slog_set_cycles(i,0);  /* gives one cycle */
                             /* must be done every time, bugfix 16.04.02 */
      slog_set_internal(i,slog_com_enables[i]);
      slog_start(i);
    }

    for (i=0; i<SLOG_NUMBER; i++) {
      slog_get_runstate(i, &slog_runstate);
      while (slog_runstate&0x01) slog_get_runstate(i, &slog_runstate);
      slog_disable_internal(i);
      if(CLOAC_NUMBER>0) slog_set_external(i,(unsigned short)(slog_com_enables[i]&0x7fff));
    }
	
    return 0;
  } else {
    printf("st_slog_single: SLOG_NUMBER = 0!!\n");
    return -1;
  }
}

int st_slog_user(char* fname, unsigned short ncycles){
  short i;
  unsigned short slog_runstate;

  /* split into three subroutines PWP 13.08.02 */

  if (SLOG_NUMBER) {

    st_slog_load(fname);
    st_slog_start(1, ncycles);

    for (i=0; i<SLOG_NUMBER; i++) {
      slog_get_runstate(i, &slog_runstate);
      while (slog_runstate&0x01) slog_get_runstate(i, &slog_runstate);
    }

    st_slog_stop();

    return 0;
  } else {
    printf("st_slog_user: SLOG_NUMBER = 0!!\n");
    return -1;
  }
}


int st_slog_load(char* fname){
  char fname2[128];
  short i;

  if (SLOG_NUMBER) {
    if(slog_do_loadtrig){
      sprintf(fname2,"%s%s",ST_CONFIG_PATH,fname);

      for (i=0; i<SLOG_NUMBER; i++) {
        slog_fload(i, fname2);
      }
      slog_do_loadtrig = 0;
    }
    return 0;
  } else {
    printf("st_slog_load: SLOG_NUMBER = 0!!\n");
    return -1;
  }
}


int st_slog_start(short set_cycles,unsigned short ncycles){
  short i;

  if (SLOG_NUMBER) {
    for (i=0; i<SLOG_NUMBER; i++){  /* set enables */
      slog_disable_external(i);
      slog_set_internal(i,slog_com_enables[i]);
      if(set_cycles) slog_set_cycles(i,ncycles);
    }
    for (i=0; i<SLOG_NUMBER; i++){  /* start */
      slog_start(i);
    }
    return 0;
  } else {
    printf("st_slog_start: SLOG_NUMBER = 0!!\n");
    return -1;
  }
}  

 
int st_slog_stop(){
  short i;

  if (SLOG_NUMBER) {
    for (i=0; i<SLOG_NUMBER; i++){ 
      slog_stop(i);
    }
    for (i=0; i<SLOG_NUMBER; i++){ 
      slog_disable_internal(i);
      if(CLOAC_NUMBER>0) slog_set_external(i,(unsigned short)(slog_com_enables[i]&0x7fff));
    }
    return 0;
  } else {
    printf("st_slog_stop: SLOG_NUMBER = 0!!\n");
    return -1;
  }
}  
 
int st_nseburst(int tmin, int tmax, short target, short ttype, short dump) {
  /* 
   * Take data  using DECODE and SAMPLE_ALL modes until either
   *   >50% of all present and active channels have occupancy > target
   * OR
   *   n_trigs > tmax
   * PWP 27.10.99
   */
    
  short mid = 0;    /* Should now loop over all MuSTARDs - PWP 29.02.00 */
  int i,j,m,t;

  short loop = 0;
  short run = 1;

  int evcount = 0; 
  short evlength, everror, status;
  short prev_everror = 0;
  short dump_next=0;
  int   total_timeout_errors = 0;
  short rcode;

  short *hpointer;
  short hlength;

  int   n_trigs=0;
  short n_chans[SCT_MAX_MODULES];
  short n_chans_over_target[SCT_MAX_MODULES];
  short n_overflows[SCT_MAX_MODULES];

  /*
   * For each module, we need to establish how many chips are present
   * and active.  (Assume that all channels are equal)
   */

  for (m=0; m<SCT_MAX_MODULES; m++){
    n_chans[m]=0;
    if ((sct[m].present)&&(sct[m].active)){
      for(i=0;i<12;i++){
        if(sct[m].det.chips[i].active) n_chans[m] +=128;
      }
    }
  }


  for (mid=0; mid<MUS_NUMBER; mid++) {  /* loop over all MuSTARDs */
    mus_timeout_errors[mid] = 0;
    mus_control_errors[mid] = 0;
    mus_event_count[mid] = 0;
    mus_error_count[mid] = 0;
  }
   
  if(outf != NULL) stout_sob(ST_TBURST,0,tmax,ttype,0,0,0);

  while(run){ 

    /* Always do a global synchronous soft reset at start of burst */

    st_soft_reset();

    /* Clear MuSTARD histograms on all used streams */

    for (mid=0; mid<MUS_NUMBER; mid++) mus_reset(mid);
    
    for (m=0; m<SCT_MAX_MODULES; m++) {   /* loop over all modules */
      if (sct[m].present) {
        mus_clear_histogram(sct[m].mus_id,sct[m].mus_ch0);
        mus_clear_histogram(sct[m].mus_id,sct[m].mus_ch1);
      }
    }

    /* Put MuSTARDs into correct modes */

    for (mid=0; mid<MUS_NUMBER; mid++){
      if ((rcode = mus_set_readout_modes(mid,0,1)) != 0) { /* SAMPLE ALL */
        printf("nseburst: ERROR for Mustard %d: mus_set_readout_modes returns %d\n", mid, rcode);
        return -2;
	  }
      if ((rcode = mus_set_readout_cntl(mid,1,1)) != 0) {
        printf("rawburst: ERROR for Mustard %d: mus_set_readout_cntl returns %d\n", mid, rcode);
        return -3;
	  }
    }

    /* take tmin events */
    t = 0;
    loop ++;
    st_sleep(10);   /* Wait 10ms to ensure Mustard ready to receive first event */

    if(dump) printf("nseburst: starting loop %d\n",loop); 

    while ((t<tmin)&&(total_timeout_errors< burst_max_timeouts)) {

      if (dump > 7) printf("nseburst: sending trigger %d  of %d  type %d\n", t, tmax, ttype);
	
      st_single(ttype);
      /* st_sleep(1); */

      for (mid=0; mid<MUS_NUMBER; mid++) {
        if(mus_used_streams[mid]>0){
          everror = prev_everror;

          evpointer = evdata;
          rcode = mus_test_event(mid,&evlength,&evpointer,&evcount,&everror,dump_next);

          prev_everror = everror;

          if (rcode != 0) {
            if (rcode == -5) {
              if (dump > 5) printf("nseburst: CHIP ERROR or PROTOCOL VIOLATION detected by Mustard %d\n", mid);
              /* Take diagnostic action   */
              mus_dump_event(mid,evdata);
              dump_next=2; 
            } else if (rcode == -4) {
              if (dump > 5) printf("nseburst: TIMEOUT for Mustard %d\n", mid);
              mus_timeout_errors[mid]++;
              total_timeout_errors++;
              evlength = 0;  
              evcount = 0;
              everror = 0;
              /* next_soft_reset = 0;  not yet */
            } else {
              if (dump > 5) printf("nseburst: CONTROL ERROR for Mustard %d: mus_test_event returns %d\n", mid, rcode);
              mus_control_errors[mid]++;
              /* next_soft_reset = 0;  not yet */
            }
            mus_read_status(mid,&status);
            mus_print_status(status);
          } else if(dump_next){
            mus_dump_event(mid,evdata);
            mus_read_status(mid,&status);
            mus_print_status(status);
            dump_next--;
          }
        }
      } /* end for (mid=0; */
      t++;
    }

    run =0;
    n_trigs+=t;
    for (mid=0; mid<MUS_NUMBER; mid++) {
      mus_event_count[mid] += evcount;
      mus_error_count[mid] += everror;
    }
    if(dump) printf("nseburst: loop %d completed, total %ld triggers so far\n",loop, n_trigs); 


    /* Read MuSTARD histograms */

    st_sleep(10);   /* Wait 10ms to ensure Mustard ready to be read out */

    for (mid=0; mid<MUS_NUMBER; mid++) mus_set_readout_cntl(mid,0,0);  

    for (m=0; m<SCT_MAX_MODULES; m++) {
      if (sct[m].present) {

        if (sct[m].link0) {
  
          /* Do normal readout of first 6 chips */
          hpointer = sct_histogram[m][0];
          if(loop==1) mus_read_histogram(sct[m].mus_id, 0, sct[m].mus_ch0,&hlength,&hpointer);
          else        mus_add_histogram(sct[m].mus_id, 0, sct[m].mus_ch0,&hlength,&hpointer);
  
          /* If readout redundancy is in use, read the remaining 6 chips */

          if ((sct[m].link0 == 2) && (sct[m].link1 == 0)) {
            hpointer = sct_histogram[m][1];
            if(loop==1) mus_read_histogram(sct[m].mus_id, 1, sct[m].mus_ch0, &hlength, &hpointer);
            else        mus_add_histogram(sct[m].mus_id, 1, sct[m].mus_ch0, &hlength, &hpointer);
          }       
        }

        if (sct[m].link1){
  
          /* Do normal readout of second 6 chips */
          hpointer = sct_histogram[m][1];
          if(loop==1) mus_read_histogram(sct[m].mus_id, 1, sct[m].mus_ch1,&hlength,&hpointer);
          else        mus_add_histogram(sct[m].mus_id, 1, sct[m].mus_ch1,&hlength,&hpointer);
  
          /* If readout redundancy is in use, read the remaining 6 chips */

          if ((sct[m].link1 == 2) && (sct[m].link0 == 0)) {
            hpointer = sct_histogram[m][0];
            if(loop==1) mus_read_histogram(sct[m].mus_id, 0, sct[m].mus_ch1, &hlength, &hpointer);
            else        mus_add_histogram(sct[m].mus_id, 0, sct[m].mus_ch1, &hlength, &hpointer);
          }       
        }


        /* Count how many channels have exceeded the target occupancy value */

        n_chans_over_target[m]=0;
        n_overflows[m]=0;

        for(j=0;j<768;j++){
          if(sct_histogram[m][0][j] >= target){
            n_chans_over_target[m]++;
          }else if(sct_histogram[m][0][j]< 0){
            n_overflows[m]++; 
            n_chans_over_target[m]++;
          }

          if(sct_histogram[m][1][j] >= target){
            n_chans_over_target[m]++;
          }else if(sct_histogram[m][1][j]< 0){
            n_overflows[m]++; 
            n_chans_over_target[m]++;
          }
        }

        if(dump){
          printf("module %d: %d chans over target inc. %d overflows\n",
                  m,n_chans_over_target[m],n_overflows[m]);    
        }

        if(n_chans_over_target[m] < (n_chans[m]/2)) run = 1; 

        if(total_timeout_errors>=burst_max_timeouts)  run = 0;  /* PWP bugfix 04.11.99 */

      } /* end if (sct[m].present) */
    } /* end for m<SCT_MAX_MODULES */

    if(n_trigs>=tmax) run =0;

  } /* end while (run) */

  /* Reset MuSTARD and clear histograms */  

  for (mid=0; mid<MUS_NUMBER; mid++) mus_reset(mid);
 
  for (m=0; m<SCT_MAX_MODULES; m++){
    if (sct[m].present){
      if (sct[m].link0) mus_clear_histogram(sct[m].mus_id,sct[m].mus_ch0);
      if (sct[m].link1) mus_clear_histogram(sct[m].mus_id,sct[m].mus_ch1);
    }
  }
  
  if(outf != NULL) stout_eob();
  burst_number++; 
  st_trigger_count = n_trigs;
  
  if (n_trigs > evcount+1) return -1;
  if (n_trigs >= tmax)     return -4;
  else return 0;
}

void st_set_slog_loadtrig(short val){

  slog_do_loadtrig=1;
}

void st_set_role(short mid, short cid, short role){

   if (mid>SCT_MAX_MODULES) return;
   if (cid>11) return;

   sct[mid].det.chips[cid].role=role;

} 

short st_used_streams(short mid){

  if((mid<0)||(mid>=MUS_NUMBER)) return 0;

  return mus_used_streams[mid];
}


void st_hv_off(){
  int rcode;
  short card, chan;
  short nch=0;

  /* turn off ALL HV channels */

  if (SCTHV_NUMBER){
    printf("\nst_hv_off:\n");

    for (card=0; card<SCTHV_NUMBER; card++){
      for (chan=0; chan<4; chan++){
        rcode=scthv_off(card, chan);
        st_sleep(100);

        if(rcode==0) nch++;
        else {
          scthv_print_error(rcode,card,chan);
        }
      }
    }
    printf("st_hv_off:  %ld channels stopped \n",nch);
  } else printf("st_hv_off: SCTHV_NUMBER = 0!!!\n");
}


void st_hv_on(){
  int rcode;
  short j;
  short nch=0;

  /* turn on USED HV channels */

  if (SCTHV_NUMBER){
    printf("\nst_hv_on:\n");

    for (j=0; j<SCT_MAX_MODULES; j++){
      if (sct[j].present&&(sct[j].hv_id>=0)&&(sct[j].hv_ch>=0)){
        printf("MODULE %d (%s) HV channel started\n",j,sct[j].detname);
        rcode=scthv_on(sct[j].hv_id,sct[j].hv_ch);
        st_sleep(100);

        if(rcode==0) nch++;
        else {
          scthv_print_error(rcode,sct[j].hv_id,sct[j].hv_ch);
        }
      }
    }
    printf("st_hv_on:  %ld channels started \n",nch);
  } else printf("st_hv_on: SCTHV_NUMBER = 0!!!\n");
}


void st_hv_ramp(short to_zero){
  int rcode;
  int loop;
  short j;

  short ch_started;
  short ch_ramping;
  short ch_hung;
  short ch_tripped;
  short ch_complete;

  /* request voltage ramp from USED HV channels */

  if (SCTHV_NUMBER){
    if(to_zero){
      printf("\nst_hv_ramp: RAMPING DOWN before SWITCHING OFF\n");
    }else{
      printf("\nst_hv_ramp: RAMPING TO CONFIGURED LEVEL\n");

    }
    /* Turn ON HV channels */

    ch_started=0;
    for (j=0; j<SCT_MAX_MODULES; j++){
      if (sct[j].present&&(sct[j].hv_id>=0)&&(sct[j].hv_ch>=0)){
        if(scthv_ison(sct[j].hv_id,sct[j].hv_ch)==0){ /* channel OFF */
          rcode=scthv_on(sct[j].hv_id,sct[j].hv_ch);
          st_sleep(100);

          if(rcode){
            scthv_print_error(rcode,sct[j].hv_id,sct[j].hv_ch);
          }else{
            ch_started++;
          }
        }
      }
    }

    if(ch_started){
      printf("st_hv_ramp: waiting two seconds to give the newly started channel processors time to settle\n");
      st_sleep(2000);
    }


    /* send ramp commands */
    printf("st_hv_ramp: sending ramp commands\n");


    ch_started=0;
    for (j=0; j<SCT_MAX_MODULES; j++){
      if (sct[j].present&&(sct[j].hv_id>=0)&&(sct[j].hv_ch>=0)){
        if(to_zero){
          rcode=scthv_request(sct[j].hv_id,sct[j].hv_ch,(float)0.0,sct[j].det.idet_lim,sct[j].det.ramp);
        }else{
          rcode=scthv_request(sct[j].hv_id,sct[j].hv_ch,sct[j].det.vdet_set,sct[j].det.idet_lim,sct[j].det.ramp);
        }
        st_sleep(100);

        if(rcode){
          scthv_print_error(rcode,sct[j].hv_id,sct[j].hv_ch);
        }else{
          ch_started++;
        }
      }
    }


    /*
     * Now loop until all channels have successfully reached the target
     * (or else tripped!!)
     */

    loop=0;
    ch_ramping=0;
    ch_hung=0;
    ch_tripped=0;
    ch_complete=0;

    while(((ch_complete+ch_tripped)<ch_started)&&(loop<SCTHV_MAX_LOOPS)){
      ch_ramping=0;
      ch_hung=0;
      ch_tripped=0;
      ch_complete=0;

      st_sleep(2000);

      for (j=0; j<SCT_MAX_MODULES; j++){
        if (sct[j].present&&(sct[j].hv_id>=0)&&(sct[j].hv_ch>=0)){
          rcode = scthv_print_status(sct[j].hv_id,sct[j].hv_ch);
          st_sleep(100);

          switch(rcode){
            case 4:  /* ramping */
              ch_ramping++;
              break;

            case 3:  /* hung_up - can sometimes be cured by sending request again */
              ch_hung++;
              st_sleep(1000);
              rcode=scthv_repeat_request(sct[j].hv_id,sct[j].hv_ch);
              st_sleep(100);
              break;

            case 2:  /* over voltage trip */
            case 1:  /* over current trip */
              ch_tripped++;
              break;

            case 0:  /* ramping complete */
              if(loop==0){  /* was the command missed? */
                st_sleep(1000);
                if(to_zero){
                  rcode=scthv_request(sct[j].hv_id,sct[j].hv_ch,(float)0.0,sct[j].det.idet_lim,sct[j].det.ramp);
                }else{
                  rcode=scthv_request(sct[j].hv_id,sct[j].hv_ch,sct[j].det.vdet_set,sct[j].det.idet_lim,sct[j].det.ramp);
                }
                st_sleep(100);
              }
              ch_complete++;
              break;

            default:
              scthv_print_error(rcode,sct[j].hv_id,sct[j].hv_ch);
              break;
          }

        }
      }
      printf("st_hv_ramp:  loop %d, %d started, %d ramping, %d hung, %d tripped, %d complete\n\n",
              loop, ch_started, ch_ramping, ch_hung, ch_tripped, ch_complete);
      loop++;
    }
    if(loop>=SCTHV_MAX_LOOPS){
      printf("st_hv_ramp: TIMEOUT waiting for ramp to complete\n");
    }

    if(to_zero){
      /* switch all channels off now */
      st_sleep(2000);
      st_hv_off();
    }
  } else printf("st_hv_ramp: SCTHV_NUMBER = 0!!!\n");
}


void st_hv_status(){
  short j;

  /* print status of USED HV channels */

  if (SCTHV_NUMBER){
    printf("\nst_hv_status:\n");

    for (j=0; j<SCT_MAX_MODULES; j++){
      if (sct[j].present&&(sct[j].hv_id>=0)&&(sct[j].hv_ch>=0)){
        printf("MODULE %d (%s) HV status enquiry:\n",
          j,sct[j].detname);
        scthv_print_status(sct[j].hv_id,sct[j].hv_ch);
        st_sleep(100);
      }
    }
  } else printf("st_hv_status: SCTHV_NUMBER = 0!!!\n");
}


int st_hv_trips(short recover){
  int rcode;
  short j;

  short ch_tripped =0;
  short ch_stopped =0;

  /* Identify tripped HV channels, switch them off, then ramp up again */

  if (SCTHV_NUMBER){

    /*
     * Loop to identify any channels which have tripped
     * and switch them off
     */

    printf("st_hv_trips: Identification of tripped channels\n");

    for (j=0; j<SCT_MAX_MODULES; j++){
      if (sct[j].present&&(sct[j].hv_id>=0)&&(sct[j].hv_ch>=0)){
        rcode = scthv_print_status(sct[j].hv_id,sct[j].hv_ch);
        st_sleep(100);

        switch(rcode){
          case 2:  /* over voltage trip */
          case 1:  /* over current trip */

            scthv_print_error(rcode,sct[j].hv_id,sct[j].hv_ch);
            st_sleep(1000);
            ch_tripped++;

            if(recover){
              /* we must switch off tripped channel before we can ramp it up */

              rcode=scthv_off(sct[j].hv_id,sct[j].hv_ch);
              st_sleep(1000);
              ch_stopped++;
            }
            break;

          default:
            break;
        }
      }
    }

    printf("st_hv_trips: %d tripped channels were found\n",ch_tripped);

    if(ch_stopped){
      printf("             %d tripped channels have been stopped\n",ch_stopped);
      printf("             waiting two seconds...\n");
      st_sleep(2000);

      printf("             attempting to recover these channels...\n");
      st_hv_ramp(0);
    }

  } else printf("st_hv_trips: SCTHV_NUMBER = 0!!!\n");

  return ch_tripped;
}


void st_lv_init(){
  short j;

  if (sctlv_n_channels>0){
    for (j=0; j<SCT_MAX_MODULES; j++){
      if((sct[j].present)&&(sct[j].lv_ch<sctlv_n_channels())){
        sctlv_check_HW_on(sct[j].lv_ch);

        printf("Setting low voltages for Module %d\n", j);
      
        sctlv_set_vcc(sct[j].lv_ch, sct[j].det.vcc_set, sct[j].det.icc_lim);  
        sctlv_set_vvi1(sct[j].lv_ch, sct[j].det.vi1_set, sct[j].det.ivi1_lim);
        sctlv_set_vdd(sct[j].lv_ch, sct[j].det.vdd_set, sct[j].det.idd_lim);
        sctlv_set_vled0(sct[j].lv_ch, sct[j].det.vled0_set, sct[j].det.iled0_lim);
        sctlv_set_vled1(sct[j].lv_ch, sct[j].det.vled1_set, sct[j].det.iled1_lim);
        sctlv_set_select(sct[j].lv_ch, sct[j].det.select);
        sctlv_print_status(sct[j].lv_ch);
      
        printf("Finished setting low voltages for Module %d\n\n", j);
      }
    }
  } else printf("st_init_lv: no SCTLV channels!!!\n");
}

void st_lv_1off(short id){
  /* turn off single LV channel */

  if((sct[id].present)&&(sct[id].lv_ch<sctlv_n_channels())){
    printf("\nst_lv_1off:");
    printf("\nModule %d (%s)\n", id, sct[id].detname);
    sctlv_check_HW_on(sct[id].lv_ch);
    sctlv_set_power(sct[id].lv_ch, 0);
    st_sleep(100); /* wait 100mS */
    sctlv_brief_status(sct[id].lv_ch);
  }
}


void st_lv_off(){
  short j;

  /* turn off ALL LV channels */

  if (sctlv_n_channels()>0){
    for (j=0; j<sctlv_n_channels(); j++){
      /* Power off ALL channels, in case of sysmap changes */
      sctlv_check_HW_on(j);
      sctlv_set_power(j, 0);
    }

    st_sleep(100); /* wait 100mS */

    printf("\nst_lv_off:");
    for (j=0; j<SCT_MAX_MODULES; j++){
      if((sct[j].present)&&(sct[j].lv_ch<sctlv_n_channels())){
        printf("\nModule %d (%s)\n", j, sct[j].detname);
        sctlv_brief_status(sct[j].lv_ch);
      }
    }
  } else printf("st_lv_off: no SCTLV channels!!!\n");
}


void st_lv_1on(short id){
  /* rewritten to use ramping PWP 04.07.02 */
  float v;

  /* turn on one LV channel */

  if((sct[id].present)&&(sct[id].lv_ch<sctlv_n_channels())){
    printf("st_lv_1on:\n");
    printf("Module %d (%s)\n", id, sct[id].detname);

    if(BLUE_HANDBAG){
      printf("Ramping low voltages\n");

      for(v=(float)2.5; v<(float)4.0; v+=(float)0.2){
        printf("  setting %3.2fV (or limit)\n",v);
        if((sct[id].present)&&(sct[id].lv_ch<sctlv_n_channels())){
          if(v<sct[id].det.vdd_set){
            sctlv_set_vdd(sct[id].lv_ch, v, sct[id].det.idd_lim);
          }
          if(v<sct[id].det.vcc_set){
            sctlv_set_vcc(sct[id].lv_ch, v, sct[id].det.icc_lim);
          }
          sctlv_check_pwron(sct[id].lv_ch);
        }
        st_sleep(10); /* wait 10mS */
      }
    } /* end if (BLUE_HANDBAG) */

    printf("Requesting desired voltages\n");

    sctlv_set_vdd(sct[id].lv_ch, sct[id].det.vdd_set, sct[id].det.idd_lim);
    sctlv_set_vcc(sct[id].lv_ch, sct[id].det.vcc_set, sct[id].det.icc_lim);
    sctlv_check_pwron(sct[id].lv_ch);
    /* reload ABCD3T trims after power cycle */
    if(sct[id].det.chipset==ABCDT) sct[id].det.updatetrims = 1;

    st_sleep(10); /* wait 10mS */
    sctlv_brief_status(sct[id].lv_ch);
  }
}

void st_lv_on(){
  /* rewritten to use ramping PWP 04.07.02 */
  float v;
  short j;

  /* turn on USED LV channels */

  if (sctlv_n_channels()>0){
    printf("st_lv_on:\n");

    if(BLUE_HANDBAG){
      printf("Ramping low voltages\n");

      for(v=(float)2.5; v<(float)4.0; v+=(float)0.2){
        printf("  setting %3.2fV (or limit)\n",v);
        for(j=0; j<SCT_MAX_MODULES; j++){
          if((sct[j].present)&&(sct[j].lv_ch<sctlv_n_channels())){
            if(v<sct[j].det.vdd_set){
              sctlv_set_vdd(sct[j].lv_ch, v, sct[j].det.idd_lim);
            }
            if(v<sct[j].det.vcc_set){
              sctlv_set_vcc(sct[j].lv_ch, v, sct[j].det.icc_lim);
            }
            sctlv_check_pwron(sct[j].lv_ch);
          }
        }
        st_sleep(10); /* wait 10mS */
      }
    } /* end if (BLUE_HANDBAG) */

    printf("Requesting desired voltages\n");

    for (j=0; j<SCT_MAX_MODULES; j++){
      if((sct[j].present)&&(sct[j].lv_ch<sctlv_n_channels())){
        sctlv_set_vdd(sct[j].lv_ch, sct[j].det.vdd_set, sct[j].det.idd_lim);
        sctlv_set_vcc(sct[j].lv_ch, sct[j].det.vcc_set, sct[j].det.icc_lim);
        sctlv_check_pwron(sct[j].lv_ch);
        /* reload ABCD3T trims after power cycle */
        if(sct[j].det.chipset==ABCDT) sct[j].det.updatetrims = 1;
      }
    }

    st_sleep(10); /* wait 10mS */
    st_lv_status();

  } else printf("st_lv_on: no SCTLV channels!!!\n");
}


void st_lv_status(){
  short j;

  /* print status of USED LV channels */

  if (sctlv_n_channels()>0){
    printf("\nst_lv_status:\n");

    for (j=0; j<SCT_MAX_MODULES; j++){
      if((sct[j].present)&&(sct[j].lv_ch<sctlv_n_channels())){
        printf("\nModule %d (%s)\n", j, sct[j].detname);
        sctlv_brief_status(sct[j].lv_ch);
      }
    }
  } else printf("st_lv_status: no SCTLV channels!!!\n");
}

int st_lv_trips(short recover){
  /* rewritten to call st_lv_1on - PWP 04.07.02 */
  short j;
  int ntrips = 0;

  if (sctlv_n_channels()>0){
    printf("\nst_lv_trips\n:");
    for (j=0; j<SCT_MAX_MODULES; j++){
      if((sct[j].present)&&(sct[j].lv_ch<sctlv_n_channels())){
        if (sctlv_check_trip(sct[j].lv_ch)){
          ntrips++;
          printf("\nModule %d (%s) has tripped\n", j, sct[j].detname);
          if(recover){    
            sctlv_set_power(sct[j].lv_ch, 0);
            st_sleep(100); /* wait 100mS */

            st_lv_1on(j);
          }
        }
      }
    }
    printf("\n  %d trip(s) found\n",ntrips);
  } else printf("st_lv_trips: no SCTLV channels!!!\n");

  return ntrips;
}

int st_configure_dcs(){
  int id;

  /* for now, just clear the trip state */

  for(id=0;id<SCT_MAX_MODULES;id++){
    sct[id].det.suspended = 0;
    sct[id].det.hv_trip = 0;
    sct[id].det.lv_trip = 0;
    sct[id].det.mt_trip = 0;
  }

  return 0;
}

void st_dcs_mon(short id){

  if(SCTHV_NUMBER>0) st_hv_mon(id);
  if(SCTLV_NUMBER>0) st_lv_mon(id);
  if(TEMPTRESS_NUMBER>0) st_t_mon(id);

  return;
}

int st_hv_mon(short id) {
  int rcode = 0;

  if(sct[id].present&&(sct[id].hv_id<SCTHV_NUMBER)){

    rcode = scthv_monitor(sct[id].hv_id,sct[id].hv_ch,
                         &sct[id].det.vdet,
                         &sct[id].det.idet,
                         &sct[id].det.hv_status);

    if((rcode==3)||(rcode==5)||(rcode==-8)){ /* stuck or error */
      scthv_print_error(rcode,sct[id].hv_id,sct[id].hv_ch);

      /* may recover if we repeat the last request */
      rcode=scthv_repeat_request(sct[id].hv_id,sct[id].hv_ch);
      st_sleep(1000);

      /* now send the monitor command again */
      rcode = scthv_monitor(sct[id].hv_id,sct[id].hv_ch,
                           &sct[id].det.vdet,
                           &sct[id].det.idet,
                           &sct[id].det.hv_status);
    }

    if((rcode==1)||(rcode==2)){
      if(sct[id].det.hv_trip==0){
        sct[id].det.hv_trip = 1;
      }
    } else {
      sct[id].det.hv_trip = 0;
    }
  }
  return rcode;
}

int st_hv_chan(short id) {

  return sct[id].hv_ch;
}
 
int st_lv_mon(short id){
  char version;
  float var[LV_NMON];
  int rcode = 0;

  if(sct[id].present&&(sct[id].lv_ch<sctlv_n_channels())){

    version = 0;
    sctlv_get_version(sct[id].lv_ch,&version);

    rcode = sctlv_monitor(sct[id].lv_ch,&sct[id].det.lv_status,
                          &sct[id].det.lv_temp, var);

    if(version>=6){  /* SCTLV3 */

      sct[id].det.vccout= var[VccOut];
      sct[id].det.vcc   = var[Vcc];
      sct[id].det.vi1   = (float)-1;   /* gone!! */
      sct[id].det.vddout= var[VddOut];
      sct[id].det.vdd   = var[Vdd];
      sct[id].det.vpin  = var[Vpin];
      sct[id].det.vled0 = var[Vld0];
      sct[id].det.vled1 = var[Vld1];

      sct[id].det.icc   = var[Icc];
      sct[id].det.ivi1  = (float)-1;   /* gone!! */
      sct[id].det.idd   = var[Idd];
      sct[id].det.ipin  = var[Ipin];
      sct[id].det.iled0 = (float)-1;   /* unavailable */
      sct[id].det.iled1 = (float)-1;   /* unavailable */

      sct[id].det.t1    = var[T0];
      sct[id].det.t2    = var[T1];

    }else{  /* SCTLV1 or SCTLV2 */

      sct[id].det.vccout= (float)-1;   /* unavailable */
      sct[id].det.vcc   = var[Vcc];
      sct[id].det.vi1   = var[Vi1];
      sct[id].det.vddout= (float)-1;   /* unavailable */
      sct[id].det.vdd   = var[Vdd];
      sct[id].det.vpin  = var[Vpin];
      sct[id].det.vled0 = var[Vld0];
      sct[id].det.vled1 = var[Vld1];

      sct[id].det.icc   = var[Icc];
      sct[id].det.ivi1  = var[Ivi1];
      sct[id].det.idd   = var[Idd];
      sct[id].det.ipin  = (float)-1;
      sct[id].det.iled0 = var[Ild0];
      sct[id].det.iled1 = var[Ild1];

      sct[id].det.t1    = (float)-999; /* unavailable */
      sct[id].det.t2    = (float)-999; /* unavailable */

    }
    
    if(rcode>0){
      if(sct[id].det.lv_trip==0){
        sct[id].det.lv_trip = 1;
      }
    } else {
      sct[id].det.lv_trip = 0;
    }
  }
  return rcode;
}

int st_lv_recover(short id){
  /* rewritten to call st_lv_1on - PWP 04.07.02 */
  int rcode = 0;

  if((sct[id].present)&&(sct[id].lv_ch<sctlv_n_channels())){
    rcode = sctlv_check_trip(sct[id].lv_ch);

    if(rcode==2){ /* timeout, card has crashed */
      if(sctlv_board_restart(id)){
        /* restart failed */
        rcode = -1;
      }
    }

    if(rcode>0){
      sctlv_set_power(sct[id].lv_ch, 0);
      st_sleep(100); /* wait 100mS */

      st_lv_1on(id);
    }
  }
  return rcode;
}


int st_t_mon(short id) {
  /*
   * if SCTLV revsion 6 or later (SCTLV3) is in use, we assume 
   * that temperature information has already come from SCTLV,
   * otherwise try TEMPTRESS.
   * PWP 09.02.01
   * revised PWP 04.07.01
   */

  char version;

  if(sct[id].present){
    
    version = 0;
    sctlv_get_version(sct[id].lv_ch,&version); /* protected against out of range */
    
    if(version<6){ /* not SCTLV3, hence no temp monitor */
      if(sct[id].lv_id<(TEMPTRESS_NUMBER)*16){
        temptress_read_one((short)(sct[id].lv_id/16),(short)((sct[id].lv_ch%16)*2),&sct[id].det.t1);
        temptress_read_one((short)(sct[id].lv_id/16),(short)(((sct[id].lv_ch%16)*2)+1),& sct[id].det.t2);
      }
    }
  }
  return 0;
}

void st_clear_suspense(){
  /*
   * Clear all suspended links
   * PWP 11.04.01
   */

  int mid;

  for(mid=0;mid<SCT_MAX_MODULES;mid++){
    sct[mid].suspend0 = 0;
    sct[mid].suspend1 = 0;
  }

  st_execute_enables();
}

void st_clear_1suspense(short mid){
  /*
   * Clear suspense for a single module
   * PWP 17.04.01
   */

  if (mid >= SCT_MAX_MODULES){
    printf("st_clear_suspense: module %d outside limit of %d modules\n", mid, SCT_MAX_MODULES);
    return;
  }

  sct[mid].suspend0 = 0;
  sct[mid].suspend1 = 0;

  st_execute_enables();
}

void st_suspend_link(short mid, short lid){
  /*
   * suspend a stream indexed by MODULE and LINK numbers 
   *  - remove it from the data aquisition as may be required if 
   *    the error rate is high or the link has died
   * PWP 11.04.01
   */

  if (mid >= SCT_MAX_MODULES){
    printf("st_suspend_link: module %d outside limit of %d modules\n", mid, SCT_MAX_MODULES);
    return;
  }

  switch(lid){
    case 0:
      sct[mid].suspend0 = 1;
      break;
    case 1:
      sct[mid].suspend1 = 1;
      break;
    default:
      printf("st_suspend_link: link %d not allowed\n",lid);
      break;
  }

  st_execute_enables();
}

void st_suspend_module(short mid){
  /*
   * suspend a module indexed by MODULE number 
   *  - remove it from the data aquisition as may be required if 
   *    the error rate is high or the link has died
   * PWP 11.04.01
   */

  if (mid >= SCT_MAX_MODULES){
    printf("st_suspend_link: module %d outside limit of %d modules\n", mid, SCT_MAX_MODULES);
    return;
  }

  sct[mid].suspend0 = 1;
  sct[mid].suspend1 = 1;

  st_execute_enables();
}

void st_suspend_stream(short mid, short sid){
  /*
   * suspend a stream indexed by MuSTARD and STREAM numbers 
   *  - remove it from the data aquisition as may be required if 
   *    the error rate is high or the link has died
   * PWP 11.04.01
   */

  short module, link;

  if (mid >= MUS_NUMBER){
    printf("st_suspend_stream: MuSTARD %d outside limit of %d MuSTARDs\n", mid, MUS_NUMBER);
    return;
  }

  if (sid >= 12){
    printf("st_suspend_stream: stream %d outside limit of 12 streams\n", mid);
    return;
  }

  module = st_find_module(mid,sid);
  link   = st_find_link(mid,sid);

  st_suspend_link(module,link);
}

void st_retrieve_variable(short mid, short cid, short chan, short typ, float* var)  {
  /*
   * Retrieve just about any piece of information from the sysmap.
   * This version by PWP 12.12.00
   * BJG 30.04.02 - If module not present return 0
   */

  short short_var;

  if (mid >= SCT_MAX_MODULES){
    printf("st_retrieve_variable: module %d outside limit of %d modules\n", mid, SCT_MAX_MODULES);
    return;
  }

  if (cid >= 12){
    printf("st_retrieve_variable: chip %d outside limit of %d chips\n", mid, SCT_MAX_MODULES);
    return;
  }

  if (cid<0){
    cid=0;
  }

  /* return variable if module is present */

  if (sct[mid].present){

    switch(typ) {

      case ST_VTHR:
        *var = sct[mid].det.chips[cid].vthr;
        break;

      case ST_VCAL:
        *var = sct[mid].det.chips[cid].vcal;
        break;

      case ST_STROBE_DELAY:
        *var = sct[mid].det.chips[cid].delay;
        break;

      case ST_PREAMP:
        *var = sct[mid].det.chips[cid].febias;
        break;

      case ST_SHAPER:
        *var = sct[mid].det.chips[cid].feshaper;
        break;

      case ST_TRIM:
        if(chan>127) *var = 0;
        else         *var = (float)sct[mid].det.chips[cid].trim[chan];
        break;

      case ST_MASK:
        if(chan>7) *var = 0;
        else       *var = (float)sct[mid].det.chips[cid].mask[chan];
        break;

      case ST_ROLE:
        *var = (float)sct[mid].det.chips[cid].role;
        break;

      case ST_CAL_MODE:
        *var = (float)sct[mid].det.chips[cid].cal_mode;
        break;

      case ST_COMPRESSION:
        *var = (float)sct[mid].det.chips[cid].compression;
        break;

      case ST_TRIM_RANGE:
        *var = (float)sct[mid].det.chips[cid].trim_range;
        break;

      case ST_EDGE_DETECT:
        *var = (float)sct[mid].det.chips[cid].edge_detect;
        break;

      case ST_SEND_MASK:
        *var = (float)sct[mid].det.chips[cid].mask_readout;
        break;

      case ST_ACCUMULATE:
        *var = (float)sct[mid].det.chips[cid].accumulate;
        break;

      case ST_DTM:
        *var = (float)sct[mid].det.dtm;
        break;

      case ST_BYPASS:
        *var = (float) sct[mid].det.bypass;
        break;

      case ST_ACTIVE: /* PWP 08.11.01 */
        if(sct[mid].active){
          *var = (float) sct[mid].det.chips[cid].active;
        }else{
          *var = (float) 0.0;
        }
        break;

      case ST_TOKEN:
        *var = (float) sct[mid].det.token;
        break;

      case ST_FWD_SC_MODE:
        *var = (float) sct[mid].det.sc_mode;
        break; 

      case ST_BPM_DRIVE:
        *var = (float) sct[mid].det.bpm_drive;
        break;

      case ST_VDC_DAC:
      case ST_VDC_DAC0:
        *var = (float) sct[mid].det.vdc_dac0;
        break;

      case ST_VDC_DAC1:
        *var = (float) sct[mid].det.vdc_dac1;
        break;

      case ST_QTHR:
        switch(sct[mid].det.chips[cid].rc_function){
          case 1: /* polynomial */
            *var = 0;
            break; 
          case 2: /* grillo */
            *var = 0;
            break; 
          case 3: /* exp */
            *var = 0;
            break;  
          case 4: /* linear fit */
            if(sct[mid].det.chips[cid].rc_par[1]==0){
              *var = 0;
            } else {
              *var = (sct[mid].det.chips[cid].vthr - sct[mid].det.chips[cid].rc_par[0]) / 
                     sct[mid].det.chips[cid].rc_par[1];
            }
            break; 
          default:
            *var = 0;
            break;
          }
        break;

      case ST_QCAL:
        *var = (float)10.0 * sct[mid].det.chips[cid].vcal;
        break;

      /*
       * Off Detector Tuning:
       * MuSTARD, CLOAC and SLOG
       */

      case ST_STREAM_DELAY:
      case ST_STREAM_DELAY0:
        *var = (float)sct[mid].mus_delay0;		
        break;

      case ST_STREAM_DELAY1:
        *var = (float)sct[mid].mus_delay1;
        break;

      case ST_CLOCK_DUTY:
        if (sct[mid].det.select == 0) slog_get_clock_duty(sct[mid].slog_id, &short_var);
        else                        slog_get_clock_duty(sct[mid].r_slog_id, &short_var);
        *var = (float)short_var;
        break;

      case ST_CLOCK_PHASE:
        if (sct[mid].det.select == 0) slog_get_clock_phase(sct[mid].slog_id, &short_var);
        else                        slog_get_clock_phase(sct[mid].r_slog_id, &short_var);
        *var = (float)short_var;
        break;

      case ST_COM_DELAY:
        *var = (float)com_delay;
        break;

      case ST_PULSE_DELAY:
        *var = (float)pulse_delay;
        break;
        
      case ST_TRIG_DELAY:
        *var = (float)trig_delay;
        break;

      case ST_SR_FREQUENCY:
        *var = (float)sr_frequency;
        break;

      case ST_TR_FREQUENCY:
        *var = (float)tr_frequency;
        break;

      case ST_NL1A:
        *var = (float)n_l1a;
        break;

      case ST_L1A_SPACING:
        *var = (float)l1a_spacing;
        break; 

      /*
       * Off Detector: OPTO
       */

      case ST_OPTO_RX_THR:
      case ST_OPTO_RX_THR0:
        *var = (float)sct[mid].opto_rx_thr0;
        break;

      case ST_OPTO_RX_THR1:
        *var = (float)sct[mid].opto_rx_thr1;
        break;

      case ST_OPTO_TX_DAC:
        *var = (float)sct[mid].opto_tx_iset;
        break;

      case ST_OPTO_TX_MKSPNEG:
        if((sct[mid].opto_tx_id>=0)&&(sct[mid].opto_tx_id<OPTIF_NUMBER)){
          optif_bpm_read_mkspneg(sct[mid].opto_tx_id, sct[mid].opto_tx_ch, &short_var);
        }
        *var = (float)short_var;
        break;

      case ST_OPTO_TX_MKSPPOS:
        if((sct[mid].opto_tx_id>=0)&&(sct[mid].opto_tx_id<OPTIF_NUMBER)){
          optif_bpm_read_mksppos(sct[mid].opto_tx_id, sct[mid].opto_tx_ch, &short_var);
        }
        *var = (float)short_var;
        break;

      case ST_OPTO_TX_CDELAY:
        if((sct[mid].opto_tx_id>=0)&&(sct[mid].opto_tx_id<OPTIF_NUMBER)){
          optif_bpm_read_coarsedelay(sct[mid].opto_tx_id, sct[mid].opto_tx_ch, &short_var);
        }
        *var = (float)short_var;
        break;

      case ST_OPTO_TX_FDELAY:
        if((sct[mid].opto_tx_id>=0)&&(sct[mid].opto_tx_id<OPTIF_NUMBER)){
          optif_bpm_read_finedelay(sct[mid].opto_tx_id, sct[mid].opto_tx_ch, &short_var);
        }
        *var = (float)short_var;
        break;

      /*
       * DCS requested settings
       */

      case ST_VDET:
        *var = sct[mid].det.vdet_set;
        break;

      case ST_VCC:
        *var = sct[mid].det.vcc_set;
        break;

      case ST_VI1:
        *var = sct[mid].det.vi1_set;
        break;

      case ST_VDD:
        *var = sct[mid].det.vdd_set;
        break;

      case ST_VPIN:
        *var = sct[mid].det.vpin_set;
        break;

      case ST_VLED:
      case ST_VLED0:
        *var = sct[mid].det.vled0_set;
        break;
  
      case ST_VLED1:
        *var = sct[mid].det.vled1_set;
        break; 
   
      case ST_SELECT:
        *var = sct[mid].det.select;
        break; 

      case ST_RAMP:
        *var = sct[mid].det.ramp;
        break; 

      /*
       * DCS requested limits
       */

      case ST_IDET:
        *var = sct[mid].det.idet_lim;
        break;

      case ST_ICC:
        *var = sct[mid].det.icc_lim;
        break;

      case ST_IVI1:
        *var = sct[mid].det.ivi1_lim;
        break;

      case ST_IDD:
        *var = sct[mid].det.idd_lim;
        break;

      case ST_IPIN:
        *var = sct[mid].det.ipin_lim;
        break;

      case ST_ILED:
      case ST_ILED0:
        *var = sct[mid].det.iled0_lim;
        break;
  
      case ST_ILED1:
        *var = sct[mid].det.iled1_lim;
        break; 

      case ST_T1_LIMIT:
        *var = sct[mid].det.t1_lim;
        break; 

      case ST_T2_LIMIT:
        *var = sct[mid].det.t2_lim;
        break; 

      /*
       * DCS control parameters
       */

      case ST_SUSPENDED:
        *var = sct[mid].det.suspended;
        break;      
        
      case ST_HVTRIP:
        *var = (float)sct[mid].det.hv_trip;
        break; 

      case ST_LVTRIP:
        *var = (float)sct[mid].det.lv_trip;
        break; 

      case ST_MTTRIP:
        *var = (float)sct[mid].det.mt_trip;
        break; 
      

      /*
       * DCS monitored values
       */

      case ST_MON_VDET:
        *var = sct[mid].det.vdet;
        break;

      case ST_MON_VCC:
        *var = sct[mid].det.vcc;
        break;

      case ST_MON_VI1:
        *var = sct[mid].det.vi1;
        break;

      case ST_MON_VDD:
        *var = sct[mid].det.vdd;
        break;

      case ST_MON_VPIN:
        *var = sct[mid].det.vpin;
        break;

      case ST_MON_VLED:
      case ST_MON_VLED0:
        *var = sct[mid].det.vled0;
        break;
  
      case ST_MON_VLED1:
        *var = sct[mid].det.vled1;
        break; 
   
      case ST_MON_IDET:
        *var = sct[mid].det.idet;
        break;

      case ST_MON_ICC:
        *var = sct[mid].det.icc;
        break;

      case ST_MON_IVI1:
        *var = sct[mid].det.ivi1;
        break;

      case ST_MON_IDD:
        *var = sct[mid].det.idd;
        break;

      case ST_MON_IPIN:
        *var = sct[mid].det.ipin;
        break;

      case ST_MON_ILED:
      case ST_MON_ILED0:
        *var = sct[mid].det.iled0;
        break;
  
      case ST_MON_ILED1:
        *var = sct[mid].det.iled1;
        break; 

      case ST_MON_T1:
        *var = sct[mid].det.t1;
        break; 

      case ST_MON_T2:
        *var = sct[mid].det.t2;
        break; 

      case ST_HV_STATUS:
        *var = (float)sct[mid].det.hv_status;
        break; 

      case ST_LV_STATUS:
        *var = (float)sct[mid].det.lv_status;
        break;

      case ST_LV_TEMP:
        *var = (float)sct[mid].det.lv_temp;
        break; 

      default:
        break;
     } /* end switch */
  } else { 
    *var = 0;
  } /* end if present*/
}

int st_execute_sc_configs(short dump, short hard_reset) {
  /*
   * Initialise multiple forward support cards, one at a time, 
   * masking control streams at Slog.
   * PWP 15.05.01
   */

  int j,k;
  short slog_ch, slog_id;
  unsigned short scl, scc, sex, sin;
  unsigned short cb[30];
  unsigned int cb_size = 30;
  unsigned short cbr[480];
  unsigned int cbr_size = 480; 
  unsigned short cb_mask = 0x3fff;
  unsigned short slog_runstate;


  if(hard_reset){  /* Assert HARD RESET */
    for (j=0; j<SCT_MAX_MODULES; j++) {
      if (sct[j].present) {
        if((sct[j].lv_ch>=0)&&(sct[j].lv_ch<sctlv_n_channels())){
          sctlv_set_reset(sct[j].lv_ch, 1);
        }
      }
    st_sleep(100); /* wait for RESET to be asserted */
    }
  }

  for (j=0; j<SLOG_NUMBER; j++){
    slog_disable_external((short)j);
  }

  for (j=0; j<SCT_MAX_MODULES; j++) {
    if (sct[j].present) {

      /* here we send the commands AGAINST the setting of the
       * select line, so the modules do not see them */

      if (sct[j].det.select){
        slog_ch = sct[j].slog_ch;
        slog_id = sct[j].slog_id;
      } else {
        slog_ch = sct[j].r_slog_ch;
        slog_id = sct[j].r_slog_id;
      }

      /* Configure enables such that only one support card receives the commands */

      for (k=0; k<SLOG_NUMBER;  k++) slog_disable_internal((short)k);

      slog_set_internal(slog_id,(unsigned short)((0x0001<<(unsigned short)slog_ch) | 0x4000));
 
      if (dump){
        printf("st_execute_sc_configs: mod %d sel %d using SLOG id %d ch %d\n", 
          j, sct[j].det.select, slog_id, slog_ch);
      }

      /* Build, load and execute CBs */

      cb_mask = (0x0001 << slog_ch) | 0x1000;
      sc_configure(&sct[j].det,cb,&cb_size);
      cb_clear(cbr,cbr_size);
      cb_rotate(cb,cb_size,cbr,&cbr_size,cb_mask);

      if (SLOG_NUMBER) {
        if (dump){
          slog_get_clock(slog_id,&scl);
          slog_get_clk_cntl(slog_id,&scc);
          slog_get_external(slog_id,&sex);
          slog_get_internal(slog_id,&sin);
          printf("st_execute_sc_configs: SLOG %d enables were clock %4x external %4x internal %4x clk_cntl %4x\n",
            slog_id, scl, sex, sin, scc);
        }

        slog_load(slog_id, cbr, cbr_size);           /*   load and   */
        slog_set_cycles(slog_id,0);
        slog_start(slog_id);                         /*  execute CB  */
        slog_get_runstate(slog_id, &slog_runstate);
        while (slog_runstate&0x01)  slog_get_runstate(slog_id, &slog_runstate);
      }
    }
  }  

  /* Re-enable all connected Slog channels to pass transparently external commands and triggers */ 
  for (j=0; j<SLOG_NUMBER; j++) {
    slog_disable_internal((short)j);
    if(CLOAC_NUMBER>0) slog_set_external((short)j,(unsigned short)(slog_com_enables[j]&0x7fff));
  }
    

  if(hard_reset){  /* Clear HARD RESET */
    for (j=0; j<SCT_MAX_MODULES; j++) {
      if (sct[j].present) {
        if((sct[j].lv_ch>=0)&&(sct[j].lv_ch<sctlv_n_channels())){
          sctlv_set_reset(sct[j].lv_ch, 0);
        }
      }
    }
  }

  return 0;
}

void st_clock_off(){
  short sid;

  for (sid=0; sid<SLOG_NUMBER; sid++) {
    slog_set_clock(sid, (unsigned short)0);
  }
}

void st_clock_on(){
  short sid;

  for (sid=0; sid<SLOG_NUMBER; sid++) {
    slog_set_clock(sid, (unsigned short)(slog_clk_enables[sid]&0x3fff));
  }
}

